/*
  $Id: ApplicationContext.java,v 1.6 2004/03/11 22:10:24 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.application;

import java.io.File;
import java.util.Iterator;
import java.util.ArrayList;

/**
 *  \brief Manage properties for an application
 *  A property is an option passed to the program being launched <BR>
 *  It can be passed in several ways :
 *  <ul>
 *    <li>In a classic java property file </li>
 *    <li>On the command line </li>
 *  </ul>
 *  <BR>
 *  Here we provide several facilities :
 *  <ul>
 *    <li>Declare the list of properties available (including property type, property default value, property description)</li>
 *    <li>Load the property file</li>
 *    <li>Use the properties</li>
 *  </ul>
 */

public interface ApplicationContext
{
  /**
   * \brief Declare a new property
   *
   * \param key the property key
   * \param type the property type
   * \param help a brief description of the property meaning
   * \param defaultValue the default value of the property
   * \throw PropertyTypeMismatchException if type of defaultValue is not of type type.
   */
  void declareProperty(String key, PropertyKindEnum type, String help, String defaultValue) throws PropertyException;

  /**
   * \brief remove a property from context
   *
   * \parap key the property to remove
   */
  void removeProperty(String key);

  /**
   * \brief Set a property to a value
   *
   * \param key a defined property
   * \param value the value that will be associated to property key
   * \throw PropertyException can be either PropertyNotFoundException if property key has not been defined, or PropertyTypeMismatchException if value is not of the type associated with property key
   */
  void setProperty(String key, String value) throws PropertyException;

  /**
   *  \brief Load a java property file
   *
   *  \param file the property file to load
   *  \throw java.io.IOException if file doesn't exists
   *  \throw PropertyException if a property is defined in the file but was not declared or value os of the wrong type
   */
  void loadPropertiesFromFile(File file) throws java.io.IOException, PropertyException;

  /**
   *  \brief Load properties from environment
   *  If a property is "dotted" (aka it uses the dotted notation), all dot are converted to underscores<BR>
   *  e.g : property.file will be converted to property_file then will be searched in environment
   *
   *  \throw PropertyException if a property is defined in the file but was not declared or value os of the wrong type
   */
  void loadPropertiesFromEnv() throws PropertyException;

  /**
   * \brief Iterate through all properties
   *
   * \return Iterator<ApplicationProperty>
   */
  Iterator iterateProperties();

  /**
    * \brief Iterate through all properties by alphabetical order
    *
    * \return Iterator<ApplicationProperty>
    */
  Iterator iteratePropertiesByName();


  /**
   * \brief get a property of type string
   *
   * \param key the name of the property to get
   * \return the value of the property in string format
   * \throw PropertyException if property doesn't exists or of the wrong format
   */
  String getPropertyString(String key) throws PropertyException;

  /**
   * \brief get a property of type int
   *
   * \param key the name of the property to get
   * \return the value of the property in int format
   * \throw PropertyException if property doesn't exists or of the wrong format
   */
  int getPropertyInt(String key) throws PropertyException;

  /**
   * \brief get a property of type File
   *
   * \param key the name of the property to get
   * \return the value of the property in path format
   * \throw PropertyException if property doesn't exists or of the wrong format
   */
  File getPropertyPath(String key) throws PropertyException;

  /**
   * \brief get a property of type ArrayList<File>
   *
   * \param key the name of the property to get
   * \return the value of the property as a list of File
   * \throw PropertyException if property doesn't exists or of the wrong format
   */
  ArrayList getPropertyPathList(String key) throws PropertyException;

  /**
   * \brief get a property of type boolean
   *
   * \param key the name of the property to get
   * \return the value of the property in boolean format
   * \throw PropertyException if property doesn't exists or of the wrong format
   */
  boolean getPropertyBoolean(String key) throws PropertyException;

  /**
   * \brief Return the version associated with this program
   *
   * \return a string describing the version of this program
   */
  String  getVersion();

  /**
   * \brief terminate program with message
   *
   * \param message the message to display before terminating
   */
  void terminate(String message);
}
