/*
  $Id: ApplicationProperty.java,v 1.5 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.application;

import java.util.Iterator;
import java.util.ArrayList;
import java.io.File;

/**
 * \brief Representation of a property suitable for this application
 * It contains info about type of properties, default value, help ...
 */

public interface ApplicationProperty
{
  /**
   * \brief Return the name of this property
   *
   * \return name of property
   */
  String           getName();

  /**
   * \brief Return type of property
   *
   * \return type of property
   */
  PropertyKindEnum getType();

  /**
   * \brief Return help message associated with property
   *
   * \return help message
   */
  String           getHelp();

  /**
   * \brief Return the string representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.StringValue
   *
   * \return string value if property is a String, undefined value otherwise
   */
  String           getStringValue();

  /**
   * \brief Return the integer representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.IntValue
   *
   * \return integer value if property is an int, undefined value otherwise
   */
  int              getIntValue();

  /**
   * \brief Return the boolean representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.BooleanValue
   *
   * \return boolean value if property is a boolean, undefined value otherwise
   */
  boolean          getBooleanValue();

  /**
   * \brief Return the path representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.PathValue
   *
   * \return path value if property is a Path, undefined value otherwise
   */
  File             getPathValue();

  /**
   * \brief Return the list of path representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.PathListValue
   *
   * \return list of path value if property is a path list, undefined value otherwise (ArrayList<File>)
   */
  ArrayList           getPathListValue();

  /**
   * \brief Return an iterator on the path list representation of this property
   * No check for its type is performed, so one must ensure first that
   * getType() == PropertyTypeEnum.PathListValue
   *
   * \return an iterator on a Path list value if property is a path list, undefined value otherwise (Iterator<File>)
   */
  Iterator         iteratePathListValue();

  /**
   * \brief Return true if property was not entered
   *
   * \return true if property is empty, false if there is a value
   */
  boolean          isEmpty();

  /**
   * \brief load a value into property
   *
   * \param value value to load into property
   * \throw PropertyException if value cannot fit into property because of type checking
   */
  void load(String value) throws PropertyException;
}