/*
  $Id: SymbolEnvironmentDisplayer.java,v 1.5 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.application;

import net.nosica.compiler.application.Environment;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.symbol.Accessible;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.Argument;
import net.nosica.symbol.ResultType;
import net.nosica.symbol.Field;
import net.nosica.symbol.AliasTable;
import net.nosica.symbol.SymbolException;
import net.nosica.util.IndentedPrintWriter;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.Map;

/**
 * \brief a class to display the contents of SymbolEnvironment
 * Display all contents of SymbolEnvironment. For degug purpose
 */

public class SymbolEnvironmentDisplayer
{
  private IndentedPrintWriter writer;
  private Environment env = Environment.instance();

  /**
   * \brief creates a new displayer
   *
   * \param writer an IndentedPrintWriter
   */
  public SymbolEnvironmentDisplayer(IndentedPrintWriter writer) { this.writer = writer; }

  /**
   * \brief Display all contents of SymbolEnvironment
   * the writer is used as output channel
   */
  public void display()
  {
    Iterator it = env.getSE().iterateTypeName();

    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();

      try
      {
        GenericType type = env.getSE().getGenericType(typeName);
        display(type);
        AliasTable aliases = type.getAliasTable();
        // Aliases
        writer.println("ALIASES --");
        display(aliases);
      }
      catch (SymbolException e)
      {
        throw new Error("Internal Error : " + e.getMessage());
      }
    }
  }

  public void display(GenericType type)
  {
    // TypeName, kind of type, property of type
    display((Accessible)type);
    writer.print(type.getKind().toString() + " " + type.getName() + "[");
    if (type.isNative()) writer.print("N");
    if (type.isPrimitive()) writer.print("P");
    if (type.isAbstract()) writer.print("A");
    if (type.isFinal()) writer.print("F");
    if (type.isInner()) writer.print("I");
    writer.println("]");

    // Extends
    writer.println("EXTENDS --");
    displayTypeNameSet(type.iterateExtends());

    // Implements
    writer.println("IMPLEMENTS --");
    displayTypeNameSet(type.iterateImplements());

    // fields
    writer.println("FIELDS --");
    writer.addIndent();
    Iterator fieldIt = type.iterateFields();
    while (fieldIt.hasNext())
    {
      Field field = (Field)fieldIt.next();
      display(field);
    }
    writer.decIndent();

    // destructor
    if (type.hasDestructor())
    {
      writer.println("DESTRUCTOR--");
      writer.addIndent();
      display(type.getDestructor());
      writer.decIndent();
    }

    // cast operators
    writer.println("CAST OPERATORS--");
    writer.addIndent();
    Iterator castIt = type.getCastOperators().values().iterator();
    while (castIt.hasNext())
    {
      GenericMethod method = (GenericMethod)castIt.next();
      display(method);
    }
    writer.decIndent();

    // overloadables
    writer.println("OVERLOADABLES--");
    writer.addIndent();
    Iterator methodsIt = type.getOverloadables().values().iterator();
    while (methodsIt.hasNext())
    {
      Iterator methodIt = ((ArrayList)methodsIt.next()).iterator();
      while (methodIt.hasNext())
      {
        GenericMethod method = (GenericMethod)methodIt.next();
        display(method);
      }
    }
    writer.decIndent();
  }

  private void display(Accessible accessible)
  {
    if (accessible.isPublic())
      writer.print("public ");
    else
    if (accessible.isProtected())
      writer.print("protected ");
    else
    if (accessible.isPrivate())
      writer.print("private ");
  }

  private void display(AliasTable aliases)
  {
    if (aliases.hasEnclosing())
      display(aliases.getEnclosing());

    writer.addIndent();
    Iterator aliasesIt = aliases.iterateAliases();
    while (aliasesIt.hasNext())
    {
      Map.Entry entry = (Map.Entry)aliasesIt.next();
      TypeName alias = (TypeName)entry.getKey();
      TypeName aliased = (TypeName)entry.getValue();
      writer.println(alias + " -> " + aliased);
    }
    writer.decIndent();
  }

  public void display(Field field)
  {
    display((Accessible)field);
	if (field.isMutable()) writer.print("M"); else writer.print("X");
	if (field.isStatic()) writer.print("S"); else writer.print("X");
	if (field.isTransient()) writer.print("T"); else writer.print("X");
    writer.print((field.isConst() ? " const " : " ") + field.getTypeName());
    writer.print(" " + field.getName());
    writer.println();
  }

  public void display(GenericMethod method)
  {
    display((Accessible)method);

    if (method.getKind() == MethodKindEnum.castOperator)
    {
      writer.print("cast " + method.getResultType() + "() const");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.constructor)
    {
      writer.print("constructor(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.destructor)
    {
      writer.println("destructor()");
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.infix)
    {
      display(method.getResultType());
      writer.print(" infix(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.postfix)
    {
      display(method.getResultType());
      writer.print(" postfix(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.prefix)
    {
      display(method.getResultType());
      writer.print(" prefix " + method.getName() + "(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.property)
    {
      display(method.getResultType());
      writer.print(" property " + method.getName() + "(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.propertyArray)
    {
      display(method.getResultType());
      writer.print(" property []" + (method.getName() != null ? method.getName() : "") + "(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.method)
    {
      display(method.getResultType());
      writer.print(" " + method.getName() + "(");
      displayArgs(method.getArguments());
      writer.print(")");
      displayThrowList(method.getThrowList());
      writer.println();
      writer.flush();
    }
    else
      throw new Error("Internal error : unrecognized method kind");
  }

  private void display(ResultType resultType)
  {
    if (resultType == null)
      writer.print("sub ");
    else
    {
      if (resultType.isConst())
        writer.print("const ");
      writer.print(resultType.getTypeName());
    }
  }

  private void display(Argument arg)
  {
    if (arg.isConst())
      writer.print("const ");
    writer.print(arg.getTypeName() + " " + arg.getName());
  }

  private void displayArgs(ArrayList args)
  {
    Iterator it = args.iterator();
    while (it.hasNext())
    {
      display((Argument)it.next());
      if (it.hasNext())
        writer.print(", ");
    }
  }

  private void displayThrowList(Set throwList)
  {
    Iterator it = throwList.iterator();
    while (it.hasNext())
    {
      writer.print(((TypeName)it.next()).toString());
      if (it.hasNext())
        writer.print(", ");
    }
  }

  private void displayTypeNameSet(Iterator it)
  {
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      writer.print(typeName);
      if (it.hasNext())
        writer.print(", ");
    }

    writer.println();
    writer.flush();
  }
}