/*
  $Id: SymbolEnvironmentImpl.java,v 1.6 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.application;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponent;
import net.nosica.parser.node.util.generic.GenericParameter;

import net.nosica.symbol.GenericType;
import net.nosica.symbol.SymbolException;

import net.nosica.util.Helper;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;

/**
 * \brief Stores GenericType and AliasTable
 * As we need to store all different implementation of generic types to handle
 * partial generic implementation, we store GenericTypes in an unusual way <BR>
 * This is a Map<TypeName, Map<TypeName, GenericType> ><BR>
 * To store a GenericType, here are the steps :
 * <ul>
 *   <li> create a non generic name from the generic name of the type
 *   <li> if the map contains the non generic name
 *   <ul>
 *     <li> get the underlying map
 *   </ul>
 *   <li> else
 *   <ul>
 *     <li> create a new Map
 *     <li> add it to the global Map
 *   </ul>
 *   <li>Now that we know the underlying map exists, get the underlying Map
 *   <li>simply add the GenericType to it, using it's generic name as a key
 *   <li>report error if type exists
 * </ul>
 * To get a GenericType :
 * <ul>
 *   <li>if the underlyng map exists, access it and retrieve the GenericType
 *   <li>else report an error
 * </ul>
 * As we see, we use a Map<NonGenericName, Map<GenericName, GenericType> ><BR>
 * This way, if we have a class Test<A, B> that has a partial implementation, we can retrieve<BR>
 * all the implementations of Test at once and decide which one to use
 */

public class SymbolEnvironmentImpl implements SymbolEnvironment
{
  private static SymbolEnvironmentImpl theInstance = new SymbolEnvironmentImpl();
  private SymbolEnvironmentImpl() {}
  public static SymbolEnvironmentImpl instance() { return theInstance; }
  public static void instanceReset() { theInstance = new SymbolEnvironmentImpl(); }

  private Map types   = new HashMap();            // Map<TypeName, Map<TypeName, GenericType> >
  private Map fullGenericTypes = new HashMap();   // Map<TypeName, GenericType>

  public void addGenericType(GenericType type) throws SymbolException
  {
    TypeName nonGenericName = Helper.createRadixTypeName(type.getName());

    // access underlying map by non generic name key
    Map typeRadixMap;
    if (types.containsKey(nonGenericName))
    {
      // map exist : get it
      typeRadixMap = (Map)types.get(nonGenericName);

      // does the key already exists ?
      if (typeRadixMap.containsKey(type.getName()))
        throw new SymbolException("Type [" + type.getName() + "] already exists");

      // if model is already present in repository, check type against it !
      if (fullGenericTypes.containsKey(nonGenericName))
        checkGenericityCompatibility(type, (GenericType)fullGenericTypes.get(nonGenericName));
    }
    else
    {
      // new non generic key : create the map
      typeRadixMap = new HashMap();
      // and add it
      types.put(nonGenericName, typeRadixMap);
    }

    // is this THE model ?
    if (type.hasGenericParameter() && !type.isPartial())
    {
      // this is THE model !
      if (fullGenericTypes.containsKey(nonGenericName))
        throw new SymbolException("Type [" + type.getName() + "] already exists");

      fullGenericTypes.put(nonGenericName, type);

      // verify all models previously stored
      if (types.containsKey(nonGenericName))
      {
        Iterator it = ((Map)types.get(nonGenericName)).values().iterator();
        while (it.hasNext())
        {
          GenericType partialType = (GenericType)it.next();

          // don't check against itself
          if (partialType != type)
            checkGenericityCompatibility(partialType, type);
        }
      }
    }

    // finally add resolved type
    typeRadixMap.put(type.getName(), type);
  }

  private void checkGenericityCompatibility(GenericType partialType, GenericType model) throws SymbolException
  {
    Iterator i = model.getName().iterateComponents();
    Iterator j = partialType.getName().iterateComponents();

    boolean error = false;
loop :
    // iterate components
    while (i.hasNext() && j.hasNext())
    {
      TypeNameComponent component1 = (TypeNameComponent)i.next();
      TypeNameComponent component2 = (TypeNameComponent)j.next();

      if (!component1.getPackageElement().equals(component2.getPackageElement()))
      {
        error = true;
        break loop;
      }

      if (component1.hasGenericExtension() != component2.hasGenericExtension())
      {
        error = true;
        break loop;
      }

      if (component1.hasGenericExtension())
      {
        // iterate generic extension
        Iterator k = component1.iterateGenericExtension();
        Iterator l = component2.iterateGenericExtension();
        while (k.hasNext() && l.hasNext())
        {
          TypeName genericParm1 = (TypeName)k.next();
          TypeName genericParm2 = (TypeName)l.next();

          GenericParameter modelGenericParm = model.getGenericParameter(genericParm1);

          if (partialType.isGenericParameter(genericParm2))
          {
            // this is a generic parameter : it has to implements same things ...
            GenericParameter partialGenericParm = partialType.getGenericParameter(genericParm2);

            if (!partialGenericParm.getImplements().equals(modelGenericParm.getImplements()))
            {
              error = true;
              break loop;
            }
          }
          else
          {
            // this is a partial generic parameter
            if (!doesImplements(genericParm2, modelGenericParm.getImplements()))
            {
              error = true;
              break loop;
            }
          }
        }

        if (k.hasNext() || l.hasNext())
        {
          error = true;
          break loop;
        }
      }
    }

    // @TODO@ : check interfaces

    // if two generic parameters are not compatible, or there is not the same number of
    // generic parameters
    if (error)
      throw new SymbolException("Partial type [" + partialType.getName() + "] is not compatible with type [" + model.getName() + "]");
  }

  public GenericType getGenericType(TypeName typeName) throws SymbolException
  {
    TypeName nonGenericName = Helper.createRadixTypeName(typeName);

    if (types.containsKey(nonGenericName))
    {
      Map radixMap = (Map)types.get(nonGenericName);
      if (radixMap.containsKey(typeName))
        return (GenericType)radixMap.get(typeName);
    }

    throw new SymbolException("Unknown type [" + typeName + "]");
  }

  public GenericType getBestSuitableGenericType(TypeName typeName, Map genericAssociation) throws SymbolException
  {
    try
    {
      // try to get a fully resolved generic type (just in case ...)
      return getGenericType(typeName);
    }
    catch (SymbolException e)
    {
      // seems we will have to resolve which partial generic instantiation
      // will be the best
    }

    // retrieve all implementations
    Map suitableTypes = getGenericTypes(typeName);
    int maxScore = -1;
    ArrayList bestOnes = null;
    ArrayList genericAssociations = null;

    // iterate through them
    Iterator it = suitableTypes.values().iterator();
    while (it.hasNext())
    {
      int score = 0;
      GenericType genericType = (GenericType)it.next();
      Map genericAssoc = new HashMap();

      TypeName genericName = genericType.getName();

      ArrayList partialComponentList = new ArrayList();
      Iterator i = genericName.iterateComponents();
      Iterator j = typeName.iterateComponents();
      boolean matching = true;

      // iterate through component of typeName
thisGenericType:
      while (i.hasNext() && j.hasNext())
      {
        TypeNameComponent component1 = (TypeNameComponent)i.next();
        TypeNameComponent component2 = (TypeNameComponent)j.next();

        partialComponentList.add(component1);

        // wrong package element ?
        if (!component1.getPackageElement().equals(component2.getPackageElement()))
        {
          matching = false;
          break thisGenericType;
        }

        // wrong genericity information ?
        if (component1.hasGenericExtension() != component2.hasGenericExtension())
        {
          matching = false;
          break thisGenericType;
        }

        if (component1.hasGenericExtension())
        {
          TypeName partialTypeName = new TypeNameImpl(partialComponentList);
          GenericType maybeEnclosing = getGenericType(partialTypeName);

          Iterator k = component1.iterateGenericExtension();
          Iterator l = component2.iterateGenericExtension();

          // iterate through generic extension
          while (k.hasNext() && l.hasNext())
          {
            TypeName genericParm1 = (TypeName)k.next();
            TypeName genericParm2 = (TypeName)l.next();

            if (!genericParm1.equals(genericParm2))
            {
              if (maybeEnclosing.isGenericParameter(genericParm1))
              {
                // genericParm1 must be a full generic parameter (such as class Toto<T>)
                // class must have a description of this parameter
                // it is possible that partialComponentList refer to an enclosing type.
                // In this case we have to ask to this enclosing type what is this parameter
                GenericParameter genericParm = maybeEnclosing.getGenericParameter(genericParm1);
                if (!doesImplements(genericParm2, genericParm.getImplements()))
                {
                  matching = false;
                  break thisGenericType;
                }

                genericAssoc.put(genericParm1, genericParm2);

                score += 1;
              }
              else
              {
                matching = false;
                break thisGenericType;
              }
            }
            else
              score += 2;
          }

          // requested type name has more generic parameter than current generic type name ?
          if (l.hasNext())
            break thisGenericType;

          // iterate through remaining generic extension of generic types for default name
          while (k.hasNext())
          {
            TypeName genericParm1 = (TypeName)k.next();

            GenericParameter genericParm = maybeEnclosing.getGenericParameter(genericParm1);
            if (!genericParm.hasDefaultType())
            {
              matching = false;
              break thisGenericType;
            }

            genericAssoc.put(genericParm.getName(), genericParm.getDefaultType());
          }
        }
      }

      if (matching)
      {
        // all check successfull : add this type !
        if (score == maxScore)
        {
          bestOnes.add(genericType);
          genericAssociations.add(genericAssoc);
        }

        if (score > maxScore)
        {
          maxScore = score;

          bestOnes = new ArrayList();
          bestOnes.add(genericType);
          genericAssociations = new ArrayList();
          genericAssociations.add(genericAssoc);
        }
      }
    }

    // report errors
    if (bestOnes == null)
    {
      // no match !
      throw new SymbolException("Could not resolve symbol [" + typeName + "] : no match");
    }

    if (bestOnes.size() != 1)
    {
      StringBuffer sb = new StringBuffer();

      it = bestOnes.iterator();
      while (it.hasNext())
      {
        GenericType type = (GenericType)it.next();
        sb.append(type.getName().toString());
        if (it.hasNext())
          sb.append(", ");
      }

      throw new SymbolException("Could not resolve symbol [" + typeName + "] : several matches : " + sb.toString());
    }

    // transfer result of association
    it = ((Map)genericAssociations.get(0)).entrySet().iterator();
    while (it.hasNext())
    {
      Map.Entry entry = (Map.Entry)it.next();
      genericAssociation.put(entry.getKey(), entry.getValue());
    }

    // return result
    return (GenericType)bestOnes.get(0);
  }

  private boolean doesImplements(TypeName typeName, TypeName interfaceName)
  {
    // @TODO@ : implement this
    return true;
  }

  public Map getGenericTypes(TypeName typeName) throws SymbolException
  {
    TypeName nonGenericName = Helper.createRadixTypeName(typeName);

    if (types.containsKey(nonGenericName))
      return (Map)types.get(nonGenericName);

    throw new SymbolException("Unknown type [" + typeName + "]");
  }

  public Iterator iterateTypeName()
  {
    return new SymbolEnvironmentIterator();
  }

  public boolean genericTypeExists(TypeName typeName)
  {
    TypeName nonGenericName = Helper.createRadixTypeName(typeName);

    return types.containsKey(nonGenericName);
  }

  private class SymbolEnvironmentIterator implements Iterator
  {
    Iterator firstLevel;
    Iterator secondLevel;

    public SymbolEnvironmentIterator()
    {
      firstLevel = types.values().iterator();
      if (firstLevel.hasNext())
      {
        Map radixMap = (Map)firstLevel.next();
        secondLevel = radixMap.keySet().iterator();
      }
      else
        secondLevel = null;
    }

    public boolean hasNext()
    {
      if (secondLevel == null)
        return false;

      if (!secondLevel.hasNext())
      {
        if (firstLevel.hasNext())
        {
          Map radixMap = (Map)firstLevel.next();
          secondLevel = radixMap.keySet().iterator();
        }
        else
          return false;
      }

      return true;
    }

    public Object next() { return secondLevel.next(); }
    public void remove() { throw new UnsupportedOperationException(); }
  }
}