/*
  $Id: Main.java,v 1.25 2004/03/11 22:10:25 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler;

import net.nosica.compiler.application.Environment;
import net.nosica.compiler.application.MessageServiceImpl;

import net.nosica.application.ApplicationContext;
import net.nosica.application.ApplicationProperty;
import net.nosica.application.PropertyKindEnum;
import net.nosica.application.PropertyException;
import net.nosica.application.NosicaPropertySetup;

import java.io.File;
import java.util.Iterator;

/**
 * \brief Main class
 * Declares valid properties, version information, handle command line parsing, as well
 * as environment parsing
 */

public class Main
{
  private String systemPropertyFile = Config.SYSTEM_PROPERTY_FILE;
  private String userPropertyFile     = "";

  public static void main(String[] args)
  {
    try
    {
      Main main = new Main(args);
      main.run();
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      System.exit(2);
    }
  }

  /**
   * \brief Constructs a new instance ready to compile
   *
   * \param args arguments of command line
   */
  public Main(String[] args) throws PropertyException, java.io.IOException
  {
    ApplicationContext context = Environment.instance().getAC();

    context.declareProperty("property.file",  PropertyKindEnum.PathValue,     "Path to property file", "nosicacc.properties");
    context.loadPropertiesFromEnv();
    context.removeProperty("property.file");

    NosicaPropertySetup.setup(context);

    commandLineAnalysis(context, args);

    try
    {
      File systemProperties = new File(systemPropertyFile);
      if (systemProperties.exists())
      {
        System.out.println("A system property file has been found, loading from [" + systemProperties.getAbsolutePath() + "]");
        context.loadPropertiesFromFile(systemProperties);
      }
      else
      {
        Environment.instance().terminate("Could not load system property file from " + systemPropertyFile);
      }

      File userProperties = new File(userPropertyFile);
      if (userProperties.exists())
      {
        System.out.println("A user property file has been found, loading from [" + userProperties.getAbsolutePath() + "]");
        context.loadPropertiesFromFile(userProperties);
      }
    }
    catch (PropertyException e)
    {
      Environment.instance().terminate(e.getMessage());
    }
  }

  private void headline()
  {
    System.out.println("The Nosica compiler -  Copyright (c) 2001-2002 David Jobet, Eric Nicolas");
    System.out.println("Nosica is distributed under the GNU General Public License");
    System.out.println("version " + Config.VERSION);
  }

  private void help(ApplicationContext context)
  {
    headline();
    System.out.println("Usage: nosicacc [options] (fileName)");
    System.out.println("Options :");
    System.out.println("  -h               : help");
    System.out.println("  -s file          : uses this system property file instead of " + Config.SYSTEM_PROPERTY_FILE);
    System.out.println("  -p file          : loads this property file");
    System.out.println("  -property value  : sets a property for the compilation session");
    System.out.println("  --property       : sets a boolean property to true");
    System.out.println("Available properties : ");

    Iterator it = context.iteratePropertiesByName();
    while (it.hasNext())
    {
      ApplicationProperty prop = (ApplicationProperty)it.next();
      String name = prop.getName();
      System.out.println(name + padWithSpace(15 - name.length()) + " - " + prop.getHelp());
    }

    Environment.instance().terminate("");
  }

  private String padWithSpace(int nbSpace)
  {
    StringBuffer result = new StringBuffer();
    for (int i = 0; i < nbSpace; ++i)
      result.append(" ");
    return result.toString();
  }

  private void commandLineAnalysis(ApplicationContext context, String[] args) throws PropertyException
  {
    // OK, extract all parameters passed to the compiler
    // that is all parameters beginning with a "-"
    int i = 0;
    boolean foundSource = false;

    while (i < args.length)
    {
      if (args[i].length() != 0)
      {
        if (args[i].charAt(0) == '-')
        {
          if (args[i].charAt(1) == '-')
          { // activating a boolean property
            try
            {
              context.setProperty(args[i].substring(2), "true");
            }
            catch (PropertyException e)
            {
              Environment.instance().terminate("option " + args[i] + " does not exist");
            }
          }
          else
          if (args[i].equals("-h"))
          { // help
            help(context);
          }
          else
          if (args[i].equals("-s"))
          {
                // specify system property file
                if (i == args.length - 1)
                {
                  Environment.instance().terminate("-s must be followed by the system property file name");
                }

                systemPropertyFile = args[i + 1];

                ++i;
          }
          else
          if (args[i].equals("-p"))
          {
            // specify property file
            if (i == args.length - 1)
            {
              Environment.instance().terminate("-p must be followed by the property file name");
            }

            userPropertyFile = args[i + 1];

            ++i;
          }
          else
          {
            if (i == args.length-1)
            {
              Environment.instance().terminate("-option must be followed by its value");
            }

            try
            {
              context.setProperty(args[i].substring(1), args[i+1]);
            }
            catch (PropertyException e)
            {
              Environment.instance().terminate("option -" + args[i].substring(1) + " does not exist");
            }

            ++i;
          }
        }
        else  // no "-" found : the file name
        {
          foundSource = true;
          context.setProperty("source", args[i]);
        }
      }

      ++i;
    }

    if (!foundSource)
    {
      Environment.instance().terminate("No Main class found on command line");
    }
  }

  public void run()
  {
    Environment.instance().setMS(MessageServiceImpl.instance());
    Environment.instance().getCP().compil();
  }
}
