/*
  $Id: GlobalAnalysisDriverImpl.java,v 1.27 2004/03/11 22:10:25 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.analysis;

import net.nosica.framework.VisitorException;

import net.nosica.application.MessageLevel;
import net.nosica.application.PropertyException;
import net.nosica.application.NativeTypes;

import net.nosica.compiler.application.ExceptionEnvironmentDisplayer;

import net.nosica.compiler.application.Environment;

import net.nosica.compiler.intermediate.Function;
import net.nosica.compiler.intermediate.Record;
import net.nosica.compiler.intermediate.Argument;

import net.nosica.symbol.SymbolException;
import net.nosica.symbol.MethodKindEnum;

import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.Type;

import net.nosica.parser.node.util.var.TypeName;

import net.nosica.util.IndentedPrintWriter;

import java.util.ArrayList;
import java.util.Stack;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

public class GlobalAnalysisDriverImpl implements GlobalAnalysisDriver, PossibleTypeNameAnalysisCallBack
{
  private Environment env = Environment.instance();
  private Stack functionToAnalyse = new Stack();  // Stack<Function>
  private Set functionAnalysed = new HashSet();   // Set<Function>
  private Set exceptionFunctionAnalysed = new HashSet();   // Set<Function>
  private Map callerRepository = new HashMap();   // Map<Function, Set<Function> >
  private Set destructors = new HashSet();        // Set<Function>
  private boolean analyseDestructor = false;

  private boolean debugRecordPTN = false;
  private String  debugRecordPTNType = null;

  private boolean debugPTN = false;
  private String  debugPTNType = null;

  private boolean debugEE = false;
  private String debugEEType = null;

  private Function activeFunction = null;

  public GlobalAnalysisDriverImpl()
  {
    try
    {
      debugRecordPTN = env.getAC().getPropertyBoolean("debug.recordptn");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugRecordPTNType = env.getAC().getPropertyString("debug.recordptn.type");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugPTN = env.getAC().getPropertyBoolean("debug.ptn");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugPTNType = env.getAC().getPropertyString("debug.ptn.type");
    }
    catch (PropertyException e)
    {
    }

  try
  {
    debugEE = env.getAC().getPropertyBoolean("debug.exception");
  }
  catch (PropertyException e)
  {
  }

  try
  {
    debugEEType = env.getAC().getPropertyString("debug.exception.type");
  }
  catch (PropertyException e)
  {
  }
  }

  public void globalAnalysis(Type mainType)
  {
    // first stage : analyse static constructors
    bootstrapStaticConstructors();
    analyse();

    // second stage : analyse normal methods
    bootstrapMainType(mainType);
    analyse();

    // third stage : analyse destructors
    bootstrapDestructors();
    analyse();

    // Fourth stage : analyse static destructors
    bootstrapStaticDestructors();
    analyse();

    // now clean up unused methods (methods that could not be reached by following path of calls)
    cleanUpUnusedMethods();

    debug();
  }

  private void bootstrapStaticConstructors()
  {
    try
    {
      Iterator it = env.getIRRE().iterateRecords();
      while (it.hasNext())
      {
        Record r = (Record)it.next();
        Type type = env.getTE().getType(r.getName());

        if (type.getName().isArray())
          continue;

        if (type.getName().equals(NativeTypes.get_net_nosica_lang_Object()))
          continue;

        if (type.isInterface())
          continue;

        addFunctionToAnalyse(env.getIRFE().getFunction(type.getInitializer()));
      }
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
  }

  private void bootstrapMainType(Type mainType)
  {
    // search for main method and bootstrap only this method ...
    try
    {
      Method mainMethod = env.getMR().findMainMethod(mainType.getName());
      if (mainMethod == null)
      {
        env.terminate("Internal error : Could not find main method in main type");
      }

      Function mainFunction = env.getIRFE().getFunction(mainMethod);
      // also add IR type for argument of main method
      Argument arg = (Argument)mainFunction.iterateArgs().next();
      PossibleTypeName ptn = env.getPTNE().getPTN(arg);
      PossibleTypeName realPtn = PossibleTypeNameImpl.createScalar(
        NativeTypes.get_net_nosica_lang_string_array()
      );
      ptn.merge(realPtn);

      addFunctionToAnalyse(mainFunction);

      // bootstrap also constructor of string (needed by main)
      Method stringConstructor = env.getMR().getDefaultConstructor(
        NativeTypes.get_net_nosica_lang_string(),
        NativeTypes.get_net_nosica_lang_string(),
        new ArrayList()
      );

      Function stringConstructorFunction = env.getIRFE().getFunction(stringConstructor);

      addFunctionToAnalyse(stringConstructorFunction);
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
  }

  private void bootstrapDestructors()
  {
    analyseDestructor = true;
    Iterator it = destructors.iterator();
    while (it.hasNext())
    {
      Function f = (Function)it.next();
      addFunctionToAnalyse(f);
    }
  }

  private void bootstrapStaticDestructors()
  {
    try
    {
      Iterator it = env.getIRRE().iterateRecords();
      while (it.hasNext())
      {
        Record r = (Record)it.next();
        Type type = env.getTE().getType(r.getName());

        if (type.getName().isArray())
          continue;

        if (type.getName().equals(NativeTypes.get_net_nosica_lang_Object()))
          continue;

        if (type.isInterface())
          continue;

        addFunctionToAnalyse(env.getIRFE().getFunction(type.getDeinitializer()));
      }
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
  }

  private void analyse()
  {
    while (!functionToAnalyse.isEmpty())
    {
      Function f = (Function)functionToAnalyse.pop();

      env.getMS().report(MessageLevel.DBG, "Analysis001", f.getMethod().toString(), null);

      if (f.getStatement() == null)
      {
        functionAnalysed.add(f);
        continue;
      }

      Set exceptions = env.getEE().getExceptions(f);
      PossibleTypeNameAnalysisVisitor visitor =
        new PossibleTypeNameAnalysisVisitor(
          exceptions,
          this
        );

      try
      {
        activeFunction = f;
        f.getStatement().accept(visitor);
        functionAnalysed.add(f);

        if (f.getMethod().getKind() == MethodKindEnum.destructor)
        {
          // a destructor has no right to launch an exception.
          // even if it is a RuntimeException
          // the user has to catch them !
          if (!exceptions.isEmpty())
          {
            env.getMS().report(MessageLevel.ERR, "Exception001", f.getMethod().toString(), null);
          }
        }
      }
      catch (VisitorException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }
  }

  private void cleanUpUnusedMethods()
  {
    Set functionToCleanUp = new HashSet();

    Iterator it = env.getIRFE().iterateFunctions();
    while (it.hasNext())
    {
      Function f = (Function)it.next();

      // don't remove net_nosica_lang_string_array as we need it to pass String[]
      // to main (@TODO@ : find a better way to do it)
      if (
          !functionAnalysed.contains(f) &&
          !f.getMethod().getType().equals(NativeTypes.get_net_nosica_lang_string_array())
         )
      {
        functionToCleanUp.add(f);
      }
    }

    it = functionToCleanUp.iterator();
    while (it.hasNext())
    {
      Function f = (Function)it.next();
      env.getMS().report(MessageLevel.DBG, "Analysis002", f.getMethod().toString(), null);
      env.getIRFE().removeFunction(f);
    }
  }

  public void analyse(Function function) throws VisitorException
  {
    addFunctionToAnalyse(function);
  }

  public void analyseCallers() throws VisitorException
  {
    Set s = (Set)callerRepository.get(activeFunction);
    if (s == null)
      return;

    Iterator it = s.iterator();
    while (it.hasNext())
    {
      Function f = (Function)it.next();
      addFunctionToAnalyse(f);
    }

    // remove active functions from list of methods to analyse
    functionToAnalyse.remove(activeFunction);
  }

  public void addCaller(Function f) throws VisitorException
  {
    Set s = (Set)callerRepository.get(f);

    if (s == null)
    {
      s = new HashSet();
      callerRepository.put(f, s);
    }

    s.add(activeFunction);

    if (!functionAnalysed.contains(f))
      addFunctionToAnalyse(f);
  }

  public void analyseActive() throws VisitorException
  {
    addFunctionToAnalyse(activeFunction);
  }

  public void analyseType(TypeName typeName) throws VisitorException
  {
    Type type = env.getTE().getType(typeName);

    // overloadables
    Iterator it = type.iterateOverloadables();
    while (it.hasNext())
    {
      ArrayList methods = (ArrayList)it.next();
      Iterator j = methods.iterator();
      while (j.hasNext())
      {
        Method m = (Method)j.next();
        Function f = env.getIRFE().getFunction(m);

        if (functionToAnalyse.contains(f) || functionAnalysed.contains(f))
        {
          addFunctionToAnalyse(f);
        }
      }
    }

    // cast operators
    it = type.iterateCastOperators();
    while (it.hasNext())
    {
      Method m = (Method)it.next();
      Function f = env.getIRFE().getFunction(m);

      if (functionToAnalyse.contains(f) || functionAnalysed.contains(f))
      {
        addFunctionToAnalyse(f);
      }
    }

    Function f = env.getIRFE().getFunction(type.getDestructor());
    addFunctionToAnalyse(f);
  }

  public boolean mergeFunctionResult(PossibleTypeName ptn)
  {
    PossibleTypeName fPTN = env.getPTNE().getPTN(activeFunction);

    // a method that returns a primitive type has only one type in it.
    if (fPTN.isPrimitive() && fPTN.isEmpty())
    {
      PossibleTypeName primitivePTN = PossibleTypeNameImpl.createScalar(fPTN.getTypeName());
      fPTN.merge(primitivePTN);

      return false;
    }

    return fPTN.merge(ptn);
  }

  public void addFieldArraySourceToActive(Set fieldArraySource)
  {
    Set fFieldArraySource = env.getPTNE().getFieldArraySource(activeFunction);
    fFieldArraySource.addAll(fieldArraySource);
  }

  public boolean alreadyAnalysed(Function f)
  {
    return exceptionFunctionAnalysed.contains(f);
  }

  private void addFunctionToAnalyse(Function f)
  {
    if ((f.getMethod().getKind() == MethodKindEnum.destructor) && !analyseDestructor)
    {
      destructors.add(f);
      return;
    }

    if (!functionToAnalyse.contains(f))
      functionToAnalyse.add(f);
  }

  private void debug()
  {
    if (debugRecordPTN)
    {
      try
      {
        IndentedPrintWriter printer = new IndentedPrintWriter(System.out);
        TypeNameEnvironmentDisplayer displayer = new TypeNameEnvironmentDisplayer(printer);

        displayer.displayRecords();
      }
      catch (VisitorException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }
    else
    if (debugRecordPTNType != null)
    {
      Iterator it = env.getIRRE().iterateRecords();
      while (it.hasNext())
      {
        Record r = (Record)it.next();

        if (r.getName().toString().equals(debugRecordPTNType))
        {
          try
          {
            IndentedPrintWriter printer = new IndentedPrintWriter(System.out);
            TypeNameEnvironmentDisplayer displayer = new TypeNameEnvironmentDisplayer(printer);

            displayer.display(r);
          }
          catch (VisitorException e)
          {
            env.terminate("Internal error : " + e.getMessage());
          }
        }
      }
    }

    if (debugPTN)
    {
      try
      {
        IndentedPrintWriter printer = new IndentedPrintWriter(System.out);
        TypeNameEnvironmentDisplayer displayer = new TypeNameEnvironmentDisplayer(printer);

        displayer.displayFunctions();
      }
      catch (VisitorException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }
    else
    if (debugPTNType != null)
    {
      Iterator it = env.getIRFE().iterateFunctions();
      while (it.hasNext())
      {
        Function f = (Function)it.next();

        if (f.getMethod().getType().toString().equals(debugPTNType))
        {
          TypeNameEnvironmentDisplayer displayer = new TypeNameEnvironmentDisplayer(
            new IndentedPrintWriter(System.out)
          );

          try
          {
            displayer.display(f);
          }
          catch (VisitorException e)
          {
            env.terminate("Internal error : " + e.getMessage());
          }
        }
      }
    }

    if (debugEE)
    {
      IndentedPrintWriter printer = new IndentedPrintWriter(System.out);
      ExceptionEnvironmentDisplayer displayer = new ExceptionEnvironmentDisplayer(printer);

      displayer.display();
    }
    else
    if (debugEEType != null)
    {
      IndentedPrintWriter printer = new IndentedPrintWriter(System.out);
      ExceptionEnvironmentDisplayer displayer = new ExceptionEnvironmentDisplayer(printer);

      displayer.displayType(debugEEType);
    }
  }
}
