/*
  $Id: PossibleTypeNameAnalysisVisitor.java,v 1.12 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.analysis;

import net.nosica.application.NativeTypes;
import net.nosica.compiler.application.Environment;
import net.nosica.compiler.application.PTNEnvironment;

import net.nosica.util.Pair;

import net.nosica.framework.VisitorException;

import net.nosica.parser.node.util.var.TypeName;

import net.nosica.symbol.MethodKindEnum;

import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.MethodHelper;
import net.nosica.compiler.typechecker.Type;

import net.nosica.compiler.intermediate.IRVisitor;
import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.Call;
import net.nosica.compiler.intermediate.MemberAccess;
import net.nosica.compiler.intermediate.Literal;
import net.nosica.compiler.intermediate.Sequence;
import net.nosica.compiler.intermediate.StatementExpression;
import net.nosica.compiler.intermediate.Variable;
import net.nosica.compiler.intermediate.Move;
import net.nosica.compiler.intermediate.ExtendedSequence;
import net.nosica.compiler.intermediate.Jump;
import net.nosica.compiler.intermediate.LabelName;
import net.nosica.compiler.intermediate.ConditionalJump;
import net.nosica.compiler.intermediate.Cast;
import net.nosica.compiler.intermediate.InstanceOf;
import net.nosica.compiler.intermediate.RawAllocation;
import net.nosica.compiler.intermediate.Init;
import net.nosica.compiler.intermediate.Function;
import net.nosica.compiler.intermediate.Field;
import net.nosica.compiler.intermediate.Statement;
import net.nosica.compiler.intermediate.StatementList;
import net.nosica.compiler.intermediate.ArrayGet;
import net.nosica.compiler.intermediate.ArraySet;
import net.nosica.compiler.intermediate.ConstraintError;
import net.nosica.compiler.intermediate.TryRegister;
import net.nosica.compiler.intermediate.TryUnregister;
import net.nosica.compiler.intermediate.Throw;
import net.nosica.compiler.intermediate.VariableScopeLeave;
import net.nosica.compiler.intermediate.Return;

import java.util.Iterator;
import java.util.Set;
import java.util.Map;
import java.util.Stack;

/**
 * \brief a visitor that analyses all possible Run Time Types that variable can take
 */
public class PossibleTypeNameAnalysisVisitor implements IRVisitor
{
  private Environment env = Environment.instance();
  private PTNEnvironment PTNEnv = env.getPTNE();
  private PossibleTypeNameAnalysisCallBack callBack;

  // for exception analysis
  private Set exceptions;               // list of exception thrown by this function
  private Stack tryStack = new Stack(); // Stack<TryRegister>

  public PossibleTypeNameAnalysisVisitor(Set exceptions, PossibleTypeNameAnalysisCallBack callBack)
  {
    this.callBack = callBack;
    this.exceptions = exceptions;
  }

  public void visit(Call call) throws VisitorException
  {
    // analyse each arguments (because of ArrayGet)
    Iterator it = call.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }

    if (call.hasSubscript())
    {
      // analyse subscript (because of ArrayGet)
      call.getSubscript().accept(this);
    }

    // a ref comparison does not modify anything
    if (MethodHelper.isRefComparison(call.getMethod()))
    {
      // these operators cannot be overloaded
      // add a resolved call
      Map resolvedMethod = PTNEnv.getResolvedMethod(call);
      resolvedMethod.put(
        NativeTypes.get_net_nosica_lang_Object(),
        env.getIRFE().getFunction(call.getMethod())
      );

      return;
    }

    if (
        call.hasSubscript() &&
        call.getMethod().getKind() != MethodKindEnum.constructor &&
        call.getMethod().getKind() != MethodKindEnum.destructor
       )
    {
      PossibleTypeName subscriptPTN = PTNEnv.getPTN(call.getSubscript());
      Map resolvedMethod = PTNEnv.getResolvedMethod(call);

      // iterate over all possible type of subscript to extract all callable polymorphic methods
      it = subscriptPTN.iteratePossibleType();
      while (it.hasNext())
      {
        PossibleTypeName ptn = (PossibleTypeName)it.next();

        if (ptn.getTypeName().equals(NativeTypes.get_net_nosica_lang_None()))
          continue;

        Method calledMethod = env.getMR().findMethod(ptn.getTypeName(), call.getMethod());
        Function f = env.getIRFE().getFunction(calledMethod);

        // complete resolved method info
        if (!resolvedMethod.containsKey(ptn.getTypeName()))
        {
          resolvedMethod.put(ptn.getTypeName(), f);
        }

        callBack.addCaller(f);
        mergeArguments(f, call);
        mergeResult(f, call);

        // this info
        Variable fThis = f.getThisInfo();
        PossibleTypeName thisPTN = PTNEnv.getPTN(fThis);

        // let's say we have an interface I with implementations I1, I2, I3, I4 and we have ptn=I1
        // subscript PTN may contain all derived implementation I1, I2, I3, I4
        // but the thisPTN must only contain derived types from I1
        PossibleTypeName derivedPTN = PossibleTypeNameImpl.filterDerivedTypes(thisPTN.getTypeName(), subscriptPTN);

        if (thisPTN.merge(derivedPTN))
        {
          // reanalyse f
          callBack.analyse(f);
        }

        // exception
        analyseException(f);
      }
    }
    else
    {
      // static method
      Function f = env.getIRFE().getFunction(call.getMethod());
      callBack.addCaller(f);
      mergeArguments(f, call);
      mergeResult(f, call);

      analyseException(f);
    }
  }

  private void mergeArguments(Function f, Call call) throws VisitorException
  {
    boolean fNeedReanalyse = false;
    Iterator it = call.iterateArgs();
    Iterator j = f.iterateArgs();

    while (it.hasNext() && j.hasNext())
    {
      Expression callArg = (Expression)it.next();
      Variable fArg = (Variable)j.next();

      PossibleTypeName callArgPTN = PTNEnv.getPTN(callArg);
      PossibleTypeName fArgPTN = PTNEnv.getPTN(fArg);

      if (fArgPTN.merge(callArgPTN))
      {
        fNeedReanalyse = true;
      }

      Set callArgFAS = PTNEnv.getFieldArraySource(callArg);
      Set fArgFAS = PTNEnv.getFieldArraySource(fArg);

      fArgFAS.addAll(callArgFAS);
    }

    if (it.hasNext() || j.hasNext())
    {
      // argument number does not match !
      System.err.println("call = " + display(call.getMethod()));
      System.err.println("f = " + display(f.getMethod()));
      env.terminate("Internal error : cannot merge arguments, number are not matching");
    }

    if (fNeedReanalyse)
      callBack.analyse(f);
  }

  private void mergeResult(Function f, Call call)
  {
    // merge the result of f to get the result of the call
    if (!call.getMethod().isSub())
    {
      PossibleTypeName fPTN = PTNEnv.getPTN(f);

      if (fPTN.isPrimitive() && fPTN.isEmpty())
      {
        PossibleTypeName primitivePTN = PossibleTypeNameImpl.createScalar(fPTN.getTypeName());
        fPTN.merge(primitivePTN);
      }

      PossibleTypeName callPTN = PTNEnv.getPTN(call);
      callPTN.merge(fPTN);

      // add also field array source to call if any
      if (PTNEnv.hasFieldArraySource(f))
      {
        Set fFieldArraySource = PTNEnv.getFieldArraySource(f);
        Set callFieldArraySource = PTNEnv.getFieldArraySource(call);

        callFieldArraySource.addAll(fFieldArraySource);
      }
    }
  }

  private String display(Method m)
  {
    StringBuffer sb = new StringBuffer();

    if (m.isStatic())
      sb.append("static ");

    if (m.isSub())
      sb.append("sub ");
    else
    {
      if (m.getResultType().isConst())
        sb.append("const ");
      sb.append(m.getResultType().getTypeName() + " ");
    }

    sb.append(m.getType() + ".");

    if (m.getModel().getKind() == MethodKindEnum.property)
      sb.append("property ");
    else
    if (m.getModel().getKind() == MethodKindEnum.propertyArray)
      sb.append("property[]");
    else
    if (m.getModel().getKind() == MethodKindEnum.infix)
      sb.append("infix ");
    else
    if (m.getModel().getKind() == MethodKindEnum.postfix)
      sb.append("postfix ");
    else
    if (m.getModel().getKind() == MethodKindEnum.prefix)
      sb.append("prefix ");
    else
    if (m.getModel().getKind() == MethodKindEnum.castOperator)
      sb.append("cast ");

    if (m.getName() != null)
      sb.append(m.getName());

    sb.append("(");
    Iterator it = m.iterateArguments();
    while (it.hasNext())
    {
      net.nosica.compiler.typechecker.Argument arg = (net.nosica.compiler.typechecker.Argument)it.next();

      if (arg.isConst())
        sb.append("const ");
      sb.append(arg.getTypeName() + " " + arg.getName());

      if (it.hasNext())
        sb.append(", ");
    }

    sb.append(")");

    return sb.toString();
  }

  public void visit(MemberAccess memberAccess) throws VisitorException
  {
    if (memberAccess.hasSubscript())
    {
      // because of ArrayGet
      memberAccess.getSubscript().accept(this);

      if (memberAccess.isSuper() || memberAccess.isThis())
      {
        // something like
        // aVariable.super.XXXX
        // in this case, we have to merge the PTN of super as if it was the PTN of aVariable
        PossibleTypeName memberPTN = PTNEnv.getPTN(memberAccess);
        PossibleTypeName subscriptPTN = PTNEnv.getPTN(memberAccess.getSubscript());

        memberPTN.merge(subscriptPTN);
      }
    }

    PossibleTypeName memberPTN = PTNEnv.getPTN(memberAccess);
    PossibleTypeName fieldPTN = PTNEnv.getPTN(memberAccess.getField());

    memberPTN.merge(fieldPTN);
  }

  public void visit(Literal literal) throws VisitorException
  {
    // set the ptn of literal
    PossibleTypeName ptn = PTNEnv.getPTN(literal);
    if (ptn.isEmpty())
    {
      PossibleTypeName tempPTN = PossibleTypeNameImpl.createScalar(
        literal.getTypeName()
      );

      ptn.merge(tempPTN);
    }
  }

  public void visit(Sequence sequence) throws VisitorException
  {
    throw new UnsupportedOperationException();
  }

  public void visit(StatementExpression statementExpression) throws VisitorException
  {
    statementExpression.getExpression().accept(this);
  }

  public void visit(Variable variable) throws VisitorException
  {
    // represents an access or a set : does not modify variable
    PossibleTypeName ptn = PTNEnv.getPTN(variable);
    if (ptn.isPrimitive() && ptn.isEmpty())
    {
      PossibleTypeName primitivePTN = PossibleTypeNameImpl.createScalar(ptn.getTypeName());
      ptn.merge(primitivePTN);
    }
  }

  public void visit(Move move) throws VisitorException
  {
    move.getVariable().accept(this);
    move.getRhs().accept(this);

    PossibleTypeName lhsPTN = PTNEnv.getPTN(move.getVariable());
    PossibleTypeName rhsPTN = PTNEnv.getPTN(move.getRhs());

    if (lhsPTN.merge(rhsPTN))
    {
      // set function to be reanalysed
      callBack.analyseActive();

      onMemberAccessChanged(move.getVariable(), move.getRhs(), rhsPTN);
    }
  }

  public void visit(ExtendedSequence exSeq) throws VisitorException
  {
    throw new UnsupportedOperationException();
  }

  public void visit(Jump jump) throws VisitorException
  {
    // nothing to do
  }

  public void visit(LabelName labelName) throws VisitorException
  {
    // nothing to do
  }

  public void visit(ConditionalJump conditionalJump) throws VisitorException
  {
    conditionalJump.getCondition().accept(this);
  }

  public void visit(Cast castExp) throws VisitorException
  {
    castExp.getExpression().accept(this);

    // before returning, set the ptn of cast exp
    PossibleTypeName ptn = PTNEnv.getPTN(castExp);
    PossibleTypeName expPTN = PTNEnv.getPTN(castExp.getExpression());
    ptn.merge(expPTN);
  }

  public void visit(InstanceOf instanceOfExp) throws VisitorException
  {
    instanceOfExp.getExpression().accept(this);

    // before returning, set the ptn of instanceof
    PossibleTypeName ptn = PTNEnv.getPTN(instanceOfExp);
    if (ptn.isEmpty())
    {
      PossibleTypeName tempPTN = PossibleTypeNameImpl.createScalar(
        NativeTypes.get_net_nosica_lang_bool()
      );

      ptn.merge(tempPTN);
    }
  }

  public void visit(RawAllocation rawAllocation) throws VisitorException
  {
    // set the ptn of rawAllocation
    PossibleTypeName ptn = PTNEnv.getPTN(rawAllocation);
    if (ptn.isEmpty())
    {
      PossibleTypeName tempPTN = PossibleTypeNameImpl.createScalar(
        rawAllocation.getTypeName()
      );

      ptn.merge(tempPTN);
    }
  }

  private void analyseDestructor(PossibleTypeName ptn) throws VisitorException
  {
    Iterator it = ptn.iteratePossibleType();
    while (it.hasNext())
    {
      PossibleTypeName aPTN = (PossibleTypeName)it.next();

      Type type = env.getTE().getType(aPTN.getTypeName());
      Function  destructor = env.getIRFE().getFunction(type.getDestructor());
      Variable fThis = destructor.getThisInfo();
      PossibleTypeName fPTN = PTNEnv.getPTN(fThis);

      if (aPTN.getTypeName().isArray())
      {
        callBack.addCaller(destructor);
        analyseDestructor(aPTN.getArrayElementPossibleTypes());
      }
      else
      {
        if (fPTN.merge(aPTN))
        {
          callBack.addCaller(destructor);
          callBack.analyse(destructor);
        }
      }
    }
  }

  public void visit(Init init) throws VisitorException
  {
    visit((Move)init);  // @TODO@ init should be removed, it is of no interest !
  }

  public void visit(StatementList statementList) throws VisitorException
  {
    Iterator it = statementList.iterateStatementList();

    while (it.hasNext())
    {
      Statement s = (Statement)it.next();
      s.accept(this);
    }
  }

  public void visit(ArraySet arraySet) throws VisitorException
  {
    // analyse subscript because it can be an ArrayGet
    arraySet.getSubscript().accept(this);
    // analyse arguments
    Iterator it = arraySet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
    // analyse rhs
    arraySet.getRhs().accept(this);

    PossibleTypeName lhsPTN = PTNEnv.getPTN(arraySet.getSubscript());
    PossibleTypeName rhsPTN = PTNEnv.getPTN(arraySet.getRhs());

    PossibleTypeName tempPtn = PossibleTypeNameImpl.createScalar(arraySet.getSubscript().getTypeName());
    tempPtn.getArrayElementPossibleTypes().merge(rhsPTN);

    if (lhsPTN.merge(tempPtn))
    {
      // set function to be reanalysed
      callBack.analyseActive();

      onMemberAccessChanged(arraySet.getSubscript(), arraySet.getRhs(), tempPtn);

      onFieldArraySourceChange(arraySet.getSubscript(), lhsPTN);
    }
  }

  public void visit(ArrayGet arrayGet) throws VisitorException
  {
    // because subscript can be an ArrayGet
    arrayGet.getSubscript().accept(this);
    // analyse arguments
    Iterator it = arrayGet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }

    PossibleTypeName arrayGetPTN = PTNEnv.getPTN(arrayGet);
    PossibleTypeName underlyingPTN = PTNEnv.getPTN(arrayGet.getSubscript());

    PossibleTypeName arrayElement = null;

    it = underlyingPTN.iteratePossibleType();
    while (it.hasNext())
    {
      PossibleTypeName PTN = (PossibleTypeName)it.next();
      if (PTN.getTypeName().isArray())
      {
        arrayElement = PTN;
        break;
      }
    }

    if (arrayElement == null)
      return;

    arrayGetPTN.merge(
      arrayElement.getArrayElementPossibleTypes()
    );
  }

  public void visit(ConstraintError constraintError) throws VisitorException
  {
    // nothing to do
  }

  public void visit(TryRegister tryRegister) throws VisitorException
  {
    tryStack.push(tryRegister);

    if (tryRegister.hasFinallyStatement())
    {
      tryRegister.getFinallyStatement().accept(this);
    }
  }

  public void visit(TryUnregister tryUnregister) throws VisitorException
  {
    tryStack.pop();
  }

  public void visit(Throw throwStatement) throws VisitorException
  {
    throwStatement.getExpression().accept(this);
    analyseException(throwStatement.getExpression().getTypeName());
  }

  public void visit(VariableScopeLeave varScopeLeave) throws VisitorException
  {
    varScopeLeave.getVariable().accept(this);

    // a VariableScopeLeave call the underlying destructor
    PossibleTypeName ptn = PTNEnv.getPTN(varScopeLeave.getVariable());
    analyseDestructor(ptn);
  }

  public void visit(Return returnStatement) throws VisitorException
  {
    returnStatement.getResult().accept(this);
    PossibleTypeName ptn = PTNEnv.getPTN(returnStatement.getResult());

    if (callBack.mergeFunctionResult(ptn))
    {
      callBack.analyseCallers();

      onReturnChanged(returnStatement.getResult());
    }
  }

  private void onMemberAccessChanged(Expression lhs, Expression rhs, PossibleTypeName rhsPTN) throws VisitorException
  {
    // left instance is a member, reanalyse all methods of type
    if (lhs instanceof MemberAccess)
    {
      MemberAccess member = (MemberAccess)lhs;
      PossibleTypeName memberPTN = PTNEnv.getPTN(member.getField());

      if (memberPTN.merge(rhsPTN))
        callBack.analyseType(member.getField().getEnclosing());
    }

    // right instance is a member
    if (rhs instanceof MemberAccess)
    {
      MemberAccess member = (MemberAccess)rhs;

      // and an array, let's modify lhs list of source field array
      if (member.getField().getTypeName().isArray())
      {
        Set fieldArraySource = PTNEnv.getFieldArraySource(lhs);

        fieldArraySource.add(member.getField());
      }
    }

    // now copy field array source from rhs to lhs if any
    if (PTNEnv.hasFieldArraySource(rhs))
    {
      Set rhsFieldArraySource = PTNEnv.getFieldArraySource(rhs);
      Set lhsFieldArraySource = PTNEnv.getFieldArraySource(lhs);

      lhsFieldArraySource.addAll(rhsFieldArraySource);
    }
  }

  private void onReturnChanged(Expression exp) throws VisitorException
  {
    if (PTNEnv.hasFieldArraySource(exp))
    {
      Set fieldArraySource = PTNEnv.getFieldArraySource(exp);
      callBack.addFieldArraySourceToActive(fieldArraySource);
    }
  }

  private void onFieldArraySourceChange(Expression subscript, PossibleTypeName ptn) throws VisitorException
  {
    Set fieldArraySource = PTNEnv.getFieldArraySource(subscript);

    Iterator it = fieldArraySource.iterator();
    while (it.hasNext())
    {
      Field field = (Field)it.next();
      PossibleTypeName fieldPTN = PTNEnv.getPTN(field);

      if (fieldPTN.merge(ptn))
      {
        callBack.analyseType(field.getEnclosing());
      }
    }
  }

  private void analyseException(Function f) throws VisitorException
  {
    Set functionExceptions = env.getEE().getExceptions(f);
    Iterator j = functionExceptions.iterator();

    while (j.hasNext())
    {
      TypeName typeName = (TypeName)j.next();
      analyseException(typeName);
    }
  }

  private void analyseException(TypeName typeName) throws VisitorException
  {
    Iterator it = tryStack.iterator();

    while (it.hasNext())
    {
      TryRegister tryRegister = (TryRegister)it.next();

      Map exceptionMap = tryRegister.getExceptionMap();
      Iterator j = exceptionMap.entrySet().iterator();
      while (j.hasNext())
      {
        Map.Entry entry = (Map.Entry)j.next();
        TypeName exceptionType = (TypeName)entry.getKey();

        if (
            env.getTE().doesInherits(typeName, exceptionType) ||
            typeName.equals(exceptionType)
        )
        {
          Pair labelVariable = (Pair)entry.getValue();
          Variable var = (Variable)labelVariable.second;

          PossibleTypeName varPTN = PTNEnv.getPTN(var);
          PossibleTypeName ptn = PossibleTypeNameImpl.createScalar(typeName);

          // as the catch site are located after
          varPTN.merge(ptn);

          return;
        }
      }
    }

    // exception is launched through this method, add it to the list of thrown exception
    exceptions.add(typeName);

    // notify analysis driver that we need to re-analyse callers of this method
    callBack.analyseCallers();
  }
}
