/*
  $Id: Environment.java,v 1.21 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
    along wtith Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.application.*;
import net.nosica.compiler.typechecker.MethodResolver;
import net.nosica.compiler.typechecker.MethodResolverImpl;
import net.nosica.compiler.typechecker.Type;
import net.nosica.compiler.production.CodeProduction;
import net.nosica.compiler.production.c.CCodeProductionImpl;

/**
 * \brief A kind of super Factory for whole nosica compiler
 * Used to get all implementations. A client should not try to directly instantiates
 * a particular type but get it through this class
 */

public class Environment
{
  private static Environment theInstance = new Environment();
  private Environment() {}
  public static synchronized Environment instance() { return theInstance; }

  private MessageService messageServiceImpl = null;

  public boolean debug = false;

  /**
   * \brief get implementation of the ErrorRepository
   *
   * \return an ErrorRepository instance
   */
  public void setMS(MessageService ms) { messageServiceImpl = ms; }


  /**
   * \brief get implementation of the ErrorRepository
   *
   * \return an ErrorRepository instance
   */
  public MessageService getMS() { return messageServiceImpl; }

  /**
   * \brief get implementation of the ApplicationContext
   *
   * \return an ApplicationContext instance
   */
  public ApplicationContext getAC() { return ApplicationContextImpl.instance(); }

  /**
   * \brief get implementation of the CompilerProcessor
   *
   * \return a CompilerProcessor instance
   */
  public CompilerProcessor getCP() { return SerialCompilerProcessorImpl.instance(); }

  /**
   * \brief get implementation of the SymbolEnvironment
   *
   * \return an SymbolEnvironment instance
   */
  public SymbolEnvironment getSE() { return SymbolEnvironmentImpl.instance(); }

  /**
   * \brief get implementation of the FileLoader
   *
   * \return a FileLoader instance
   */
  public FileLoader getFL() { return FileLoaderImpl.instance(); }

  /**
   * \brief get implementation of the TypeEnvironment
   *
   * \return a TypeEnvironment instance
   */
  public TypeEnvironment getTE() { return TypeEnvironmentImpl.instance(); }

  /**
   * \brief get implementation of the TypeLoader
   *
   * \return a TypeLoader instance
   */
  public TypeLoader getTL() { return TypeLoaderImpl.instance(); }

  /**
   * \brief get implementation of the TypeNameResolver
   *
   * \return a TypeNameResolver instance
   */
  public TypeNameResolver getTR() { return TypeNameResolverImpl.instance(); }

  /**
   * \brief get implementation of the GenericitySolver
   *
   * \return a GenericitySolver instance
   */
  public GenericitySolver getGS() { return GenericitySolverImpl.instance(); }

  /**
   * \brief get implementation of MethodResolver
   *
   * \return a MethodResolver instance
   */
  public MethodResolver getMR() { return MethodResolverImpl.instance(); }

  /**
   * \brief get implementation of IRRecordEnvironment
   */
  public IRRecordEnvironment getIRRE() { return IRRecordEnvironmentImpl.instance(); }

  /**
   * \brief get implementation of IRFunctionEnvironment
   */
  public IRFunctionEnvironment getIRFE() { return IRFunctionEnvironmentImpl.instance(); }

  /**
   * \brief get implementation of NameGenerator
   */
  public NameGenerator getNG() { return NameGeneratorImpl.instance(); }

  /**
   * \brief get implementation of PTNEnvironment
   */
  public PTNEnvironment getPTNE() { return PTNEnvironmentImpl.instance();}

  /**
   * \brief get implementation of ExceptionEnvironment
   */
  public ExceptionEnvironment getEE() { return ExceptionEnvironmentImpl.instance();}

  /**
   * \brief create a new implementation of CodeProduction
   */
  public CodeProduction createCP(Type mainType)
  {
    try
    {
      String code = getAC().getPropertyString("code.language");

      if (code.equals("c"))
        return new CCodeProductionImpl(mainType);

      getMS().report(MessageLevel.NFO, "Property code.language is set to an unknown language, reverting to default");
    }
    catch (PropertyException e)
    {
      getMS().report(MessageLevel.NFO, "Property code.language is not set, reverting to default");
    }

    return new CCodeProductionImpl(mainType);
  }

  /**
   * \brief terminates program with an error message
   *
   * \param message message that will be output using System.out before termination
   */
  public void terminate(String message)
  {
    getAC().terminate(message);
  }

  /**
   * \brief check that an assertion is true, otherwise, call terminate(message)
   */
  public static void check(boolean condition, String message)
  {
    if (!condition)
    {
      theInstance.terminate("Check failed : " + message);
    }
  }

  /**
   * \brief reset all environements in order to start a  new compilation
   *
   * Warning: Do not reset ApplicationContext : the properties remain valid from
   * one compilation to the next.
   */
  public void reset()
  {
    MessageServiceImpl.instanceReset();
    SerialCompilerProcessorImpl.instanceReset();
    SymbolEnvironmentImpl.instanceReset();
    FileLoaderImpl.instanceReset();
    TypeEnvironmentImpl.instanceReset();
    TypeLoaderImpl.instanceReset();
    TypeNameResolverImpl.instanceReset();
    GenericitySolverImpl.instanceReset();
    MethodResolverImpl.instanceReset();
    IRRecordEnvironmentImpl.instanceReset();
    IRFunctionEnvironmentImpl.instanceReset();
    NameGeneratorImpl.instanceReset();
  }

  static public class CompilationAbortException extends RuntimeException
  {
    public CompilationAbortException() { super("Compilation Abort Exception"); }
  }
}
