/*
  $Id: IRTranslator.java,v 1.24 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.parser.node.util.var.TypeName;

import net.nosica.compiler.typechecker.MethodCallContext;
import net.nosica.compiler.typechecker.FieldContext;
import net.nosica.compiler.typechecker.Variable;
import net.nosica.compiler.typechecker.Label;

import net.nosica.compiler.intermediate.MemberAccess;
import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.Statement;

import java.util.ArrayList;
import java.util.Map;

/**
 * \brief translate type checked expression into Intermediate Representation(IR)
 * This module hides the manipulation of IR node from the type checker
 */
public interface IRTranslator
{
  /**
   * \brief translate a variable access
   *
   * \param variable the variable to translate
   * \return translated variable
   */
  net.nosica.compiler.intermediate.Variable translateVariable(Variable variable);

  /**
   * \brief translate a reference assignment
   *
   * \param left must be a variable
   * \param right an expression
   *
   * \return a ref assignment
   */
  Expression translateReferenceAssignment(Expression left, Expression right);

  /**
   * \brief translate a method call
   *
   * \param mcc the method call context
   * \param args ArrayList<Expression> the arguments. Instance on which call has to be performed must be first element
   *
   * \return a method call
   */
  Expression translateMethodCall(MethodCallContext mcc, ArrayList args);

  /**
   * \brief translate an assignment
   *
   * \param mcc null for a reference one, a method call context for a primitive one
   * \param tLeft left expression
   * \param tRight right expression
   * \param isInit if this is a reference init
   *
   * \return translated assignment
   */
  Expression translateAssignment(
    MethodCallContext mcc,
    Expression tLeft,
    Expression tRight,
    boolean isInit
  );

  /**
   * \brief translate a primitive variable construction
   *
   * \param mcc the constructor to call
   * \param args the arguments of the constructor
   *
   * \return a primitive construction
   */
  Expression translatePrimitiveConstruction(MethodCallContext mcc, ArrayList args);

  /**
   * \brief translate a primitive variable construction
   *
   * \param var the variable to call constructor on
   * \param mcc the constructor to call
   * \param args the arguments of the constructor
   *
   * \return a primitive construction
   */
  Expression translatePrimitiveConstruction(Expression var, MethodCallContext mcc, ArrayList args);

  /**
   * \brief translate allocation of an array
   *
   * \param constructors ArrayList<MethodCallContext> list of constructors to call to construct array
   * \param args ArrayList<ArrayList<Expression> > list of arguments to apply to constructors
   *
   * \return allocation
   */
  Expression translateArrayAllocation(ArrayList constructors, ArrayList args);

  /**
   * \brief translate allocation of a reference variable
   *
   * \param mcc the constructor to call on allocation
   * \param args the arguments to apply to constructor
   *
   * \return allocation
   */
  Expression translateAllocation(MethodCallContext mcc, ArrayList args);

  /**
   * \brief translate a cast expression
   *
   * \param castTypeName the type name of the cast expression
   * \param exp the expression to cast
   *
   * \return cast expression
   */
  Expression translateCastExpression(TypeName castTypeName, Expression exp);

  /**
   * \brief translate an instanceof expression
   *
   * \param instanceOfTypeName the type name of the instanceof expression
   * \param exp the expression to test against
   *
   * \return instanceof expression
   */
  Expression translateInstanceOfExpression(TypeName instanceOfTypeName, Expression exp);

  /**
   * \brief translate a literal
   *
   * \param typeName typeName of the literal
   * \param str literal string
   *
   * \return translated literal
   */
  Expression translateLiteral(TypeName typeName, String str);

  /**
   * \brief translate this
   *
   * \return translated this expression
   */
  net.nosica.compiler.intermediate.Variable translateThis();

  /**
   * \brief transalte a super access
   *
   * \return translated super access
   */
  MemberAccess translateSuper();

  /**
   * \brief translate a member access
   *
   * \param exp the subscript of the member access
   * \param fieldContext the field context
   *
   * \return a member access
   */
  MemberAccess translateMemberAccess(Expression exp, FieldContext fieldContext);

  /**
   * \brief translate an extended sequence
   *
   * \param s the side effect
   * \param e the expression with side effect
   *
   * \return an extended expression containing both the statement and the expression
   */
  Expression translateExtendedSequence(Statement s, Expression e);

  /**
   * \brief translate a sequence of two statements
   *
   * \param s1 the first statement
   * \param s2 the second statement
   *
   * \return a statement consisting of the sequence of the two statements
   */
  Statement translateSequence(Statement s1, Statement s2);

  /**
   * \brief translate a sequence of a statement and an expression
   *
   * \param s the statement
   * \param e the expression
   *
   * \return a sequence of s and e
   */
  Statement translateSequence(Statement s, Expression e);

  /**
   * \brief translate the destruction of variables
   *
   * \param statement the statement to write to
   * \param variables ArrayList<Variable> the variables to destroy
   *
   * \return the completed statement
   */
  Statement translateDeadVariableDestruction(Statement statement, ArrayList variables);

  /**
   * \brief translate field initialization
   *
   * \param statement the statement to write to
   * \param variables ArrayList<Field> the field to initialize
   *
   * \return completed statement
   */
  Statement translateFieldInitialization(Statement statement, ArrayList variables);

  /**
   * \brief translate a label
   *
   * \param label the label to translate
   *
   * \return translated label
   */
  net.nosica.compiler.intermediate.LabelName translateLabel(Label label);

  /**
   * \brief translate a jump to a label
   *
   * \param label the label to jump to
   *
   * \return translated jump
   */
  net.nosica.compiler.intermediate.Jump translateJump(Label label);

  /**
   * \brief translate a conditional jump
   *
   * \param cond the condition
   * \param trueLabel the label to jump to if condition is true
   * \param falseLabel the label to jump to if condition is false
   *
   * \return translated conditional jump
   */
  Statement translateConditionalJump(net.nosica.compiler.intermediate.Variable cond, Label trueLabel, Label falseLabel);

  /**
   * \brief translate a return that yields a result
   *
   * \param exp the expression to return
   *
   * \return the translated return statement
   */
  Statement translateReturn(Expression exp);

  /**
   * \brief translate initialization of a variable
   *
   * \param var the variable to initialize
   * \param exp the expression to assign to variable
   *
   * \return translated statement
   */
  Statement translateInit(net.nosica.compiler.intermediate.Variable var, Expression exp);

  /**
   * \brief translate initialization of a variable as an expression
   *
   * \param var the variable to initialize
   * \param exp the expression to assign to variable
   *
   * \return translated expression
   */
  Expression translateInitAsExpression(net.nosica.compiler.intermediate.Variable var, Expression exp);

  /**
   * \brief translate a try site register
   *
   * \param exceptionMap Map<TypeName, Pair<Label, Variable> > description of which exceptions are caught, and the correspondant catch block label and exception variables
   * \param finallyStatement the finally statement if any
   *
   * \return a translated try site register
   */
  Statement translateTryRegister(Map exceptionMap, Statement finallyStatement);

  /**
   * \brief translate the end of a try site to let the runtime unregister it
   *
   * \return translated tryUnregister
   */
  Statement translateTryUnregister();

  /**
   * \brief translate a thrown expression
   *
   * \param exp the expression to throw
   * \return translated statement
   */
  Statement translateThrow(Expression exp);
}