/*
  $Id: IRTranslatorImpl.java,v 1.43 2004/05/06 19:47:00 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.parser.node.util.var.TypeName;

import net.nosica.application.NativeTypes;
import net.nosica.application.MessageLevel;
import net.nosica.compiler.typechecker.*;
import net.nosica.compiler.typechecker.Type;
import net.nosica.compiler.typechecker.MethodCallContext;
import net.nosica.compiler.typechecker.MethodCallContextImpl;
import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.FieldContext;
import net.nosica.compiler.typechecker.FieldContextImpl;
import net.nosica.compiler.typechecker.Label;
import net.nosica.compiler.typechecker.LabelImpl;
import net.nosica.compiler.typechecker.TypeCheckerException;

import net.nosica.symbol.SymbolException;
import net.nosica.symbol.MethodKindEnum;

import net.nosica.compiler.intermediate.Record;
import net.nosica.compiler.intermediate.Field;
import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.Statement;
import net.nosica.compiler.intermediate.LiteralImpl;
import net.nosica.compiler.intermediate.Variable;
import net.nosica.compiler.intermediate.VariableImpl;
import net.nosica.compiler.intermediate.Call;
import net.nosica.compiler.intermediate.CallImpl;
import net.nosica.compiler.intermediate.MemberAccess;
import net.nosica.compiler.intermediate.MemberAccessImpl;
import net.nosica.compiler.intermediate.SequenceImpl;
import net.nosica.compiler.intermediate.StatementExpressionImpl;
import net.nosica.compiler.intermediate.MoveImpl;
import net.nosica.compiler.intermediate.ExtendedSequenceImpl;
import net.nosica.compiler.intermediate.JumpImpl;
import net.nosica.compiler.intermediate.LabelNameImpl;
import net.nosica.compiler.intermediate.ConditionalJumpImpl;
import net.nosica.compiler.intermediate.InstanceOfImpl;
import net.nosica.compiler.intermediate.CastImpl;
import net.nosica.compiler.intermediate.RawAllocationImpl;
import net.nosica.compiler.intermediate.InitImpl;
import net.nosica.compiler.intermediate.ArrayGetImpl;
import net.nosica.compiler.intermediate.ArraySetImpl;
import net.nosica.compiler.intermediate.ArgumentImpl;
import net.nosica.compiler.intermediate.TryRegisterImpl;
import net.nosica.compiler.intermediate.TryUnregisterImpl;
import net.nosica.compiler.intermediate.ThrowImpl;
import net.nosica.compiler.intermediate.VariableScopeLeaveImpl;
import net.nosica.compiler.intermediate.LabelName;
import net.nosica.compiler.intermediate.Jump;
import net.nosica.compiler.intermediate.ReturnImpl;

import net.nosica.compiler.application.Environment;
import net.nosica.util.Pair;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

/**
 * \brief translate high level parser node to low level IR nodes
 *
 * Some nodes are inserted, removed or transformed by ExSeqRemoverVisitor<br>
 *
 * Nodes removed later by ExSeqRemoverVisitor :
 * <ul>
 *   <li>ExtendedSequence
 *   <li>Sequence
 * </ul>
 * Nodes inserted later by ExSeqRemoverVisitor :
 * <ul>
 *   <li>StatementList
 *   <li>CheckNull
 *   <li>ConstraintError
 * </ul>
 * Nodes modified later by ExSeqRemoverVisitor :
 * <ul>
 *   <li>ArrayGet
 *   <li>ArraySet
 * </ul>
 */
public class IRTranslatorImpl implements IRTranslator
{
  private Environment env = Environment.instance();
  private Map variableCache = new HashMap(); // Map<typechecker.Variable, intermediate.Variable>
  private VariableEnvironment varEnv;
  private Variable thisVariable = null;
  private Method activeMethod;

  private static int staticUid = 0;
  private String seedString;
  private int uid = 0;

  public IRTranslatorImpl(Method activeMethod, VariableEnvironment varEnv)
  {
    this.activeMethod = activeMethod;
    this.seedString = "IRT_" + staticUid++ + "_";

    if (!activeMethod.isStatic())
      thisVariable = new VariableImpl(activeMethod.getType(), "this");

    this.varEnv = varEnv;
  }

  public Expression translateReferenceAssignment(Expression left, Expression right)
  {
    if (left instanceof Variable)
    {
      Variable vLeft = (Variable)left;

      // assign temporary to left variable
      Statement statement = new MoveImpl(
        vLeft,
        right
      );

      // and return left variable as result
      return new ExtendedSequenceImpl(
        statement,
        vLeft
      );
    }
    else
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck036", left);
      throw new TypeCheckerErrorExpressionException();
    }
  }

  public Expression translateMethodCall(MethodCallContext mcc, ArrayList args)
  {
    if (mcc.getMethod().getKind() == MethodKindEnum.propertyArray && mcc.getMethod().getType().isArray())
    {
      // array access
      return translateArrayAccess(mcc, args);
    }

    // postfix method
    if (mcc.getMethod().getKind() == MethodKindEnum.postfix)
    {
      if (args.size() != 2)
      {
        env.terminate("Internal error : a postfix operator has 2 arguments");
      }

      // swap arguments
      Object temp = args.get(0);
      args.set(0, args.get(1));
      args.set(1, temp);
    }

    Expression subscript = (Expression)args.get(0);
    args.remove(0);

    // translate expression like super.super.method()
    if (!mcc.getMethod().isStatic())
    {
      Iterator it = mcc.iterateSubscript();
      while (it.hasNext())
      {
        Pair pair = (Pair)it.next();

        if (((Boolean)pair.first).booleanValue())
          subscript = translateEnclosing(subscript);
        else
          subscript = translateSuper(subscript);
      }
    }

    // translate primitive cast on arguments
    Iterator mccArgsIt = mcc.iterateArgumentCasts();
    Iterator argsIt    = args.iterator();
    ArrayList newArgs = new ArrayList(); // ArrayList<Expression>

    while (mccArgsIt.hasNext() && argsIt.hasNext())
    {
      MethodCallContext argMcc = (MethodCallContext)mccArgsIt.next();
      Expression        argExp = (Expression)argsIt.next();

      if (argMcc == null || argMcc.getMethod() == null)
      {
        // no primitive cast on this one : continue
        newArgs.add(argExp);

        continue;
      }

      // apply method on argument
      ArrayList castArgs = new ArrayList();
      castArgs.add(argExp);

      argExp = translateMethodCall(
        argMcc,
        castArgs
      );

      newArgs.add(argExp);
    }

    if (mccArgsIt.hasNext() || argsIt.hasNext())
    {
      env.terminate("Internal error : number of arguments do not match");
    }

    // set old args to point to new args (the one with cast operators set)
    args = newArgs;

    Statement statement = null;

    // if method is sub
    if (mcc.getMethod().isSub())
    {
      // just call method
      statement = translateSequence(
        statement,
        new StatementExpressionImpl(
          new CallImpl(
            subscript,
            mcc.getMethod(),
            args
          )
        )
      );

      return new ExtendedSequenceImpl(
        statement,
        translateLiteral(
          NativeTypes.get_net_nosica_lang_None(),
          "null"
        )
      );
    }
    else
    {
      Call call = new CallImpl(subscript, mcc.getMethod(), args);
/*
      Type type = null;

      try
      {
        type = env.getTE().getType(mcc.getMethod().getResultType().getTypeName());
      }
      catch (SymbolException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }

      if (type.isPrimitive())
      {
        // means we have to add a temporary and add it to variable environment for further destruction
        // example :
        // Console.out << f() << "\n";
        // string f() { return ""; }
        // if we don't create a temp for the primitive temp variable returned by f(),
        // then the memory storage for the string will never be destroyed (net.nosica.lang.char[])

        // create variable that will hold the result
        net.nosica.compiler.typechecker.Variable tempVar = new net.nosica.compiler.typechecker.VariableImpl(
          false,
          type.getName(),
          env.getNG().generateTemporaryName(),
          true
        );

        Variable var = translateVariable(tempVar);

        // and add it to VariableEnvironment for further destruction
        try
        {
          varEnv.add(tempVar);
        }
        catch (TypeCheckerException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }

        // and add a INIT(VAR, CALL) statement
        statement = translateSequence(
          statement,
          translateInit(var, call)
        );

        return new ExtendedSequenceImpl(
          statement,
          var
        );
      }
*/
      if (statement == null)
        return call;

      return new ExtendedSequenceImpl(
        statement,
        call
      );
    }
  }

  public Expression translateAssignment(
                MethodCallContext mcc,
                Expression tLeft,
                Expression tRight,
                boolean isInit
              )
  {
    // translation
    if (mcc == null)
    {
      // assign operator
      if (isInit)
      {
        if (tLeft instanceof Variable)
        {
          return translateInitAsExpression(
            (net.nosica.compiler.intermediate.Variable)tLeft,
            tRight
          );
        }
        else
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck036", tLeft);
          throw new TypeCheckerErrorExpressionException();
        }
      }
      else
        return translateReferenceAssignment(tLeft, tRight);
    }
    else
    {
      // primitive assignment
      ArrayList args = new ArrayList();
      args.add(tLeft); args.add(tRight);

      return translateMethodCall(mcc, args);
    }
  }

  public Expression translateArrayAccess(MethodCallContext mcc, ArrayList args)
  {
    Expression subscript = (Expression)args.get(0);
    args.remove(0);

    Statement statement = null;

    if (!mcc.getMethod().isSub())
    {
      // property get
      statement = checkArrayBounds(statement, args);

      if (statement == null)
      {
        return new ArrayGetImpl(
          subscript,
          args
        );
      }
      else
      {
        return new ExtendedSequenceImpl(
          statement,
          new ArrayGetImpl(
            subscript,
            args
          )
        );
      }
    }
    else
    {
      // property set
      Expression rhs = (Expression)args.get(args.size() - 1);
      args.remove(args.size() - 1);

      statement = translateSequence(
        statement,
        new ArraySetImpl(
          subscript,
          rhs,
          args
        )
      );

      statement = checkArrayBounds(statement, args);

      return new ExtendedSequenceImpl(
        statement,
        translateLiteral(                   // in fact, we do not return anything
          NativeTypes.get_net_nosica_lang_None(),   // but, as we need an expression, we return a literal
          "null"                            // that is not used, thus will be discarded later
        )
      );
    }
  }

  private Statement checkArrayBounds(Statement statement, ArrayList args)
  {
    // @TODO@
    return statement;
  }

  private Expression translateEnclosing(Expression subscript)
  {
    try
    {
      Record record = env.getIRRE().getRecord(subscript.getTypeName());
      Field field = record.getField("___enclosing");

      return new MemberAccessImpl(
        subscript,
        field
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
      return null;
    }
  }

  private Expression translateSuper(Expression subscript)
  {
    try
    {
      Record record = env.getIRRE().getRecord(subscript.getTypeName());
      Field field = record.getField("super");

      return new MemberAccessImpl(
        subscript,
        field
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
      return null;
    }
  }

  private Label generateLabel()
  {
    return new LabelImpl(seedString + uid++);
  }

  public Expression translateExtendedSequence(Statement s, Expression e)
  {
    if (s == null)
      return e;
    return new ExtendedSequenceImpl(s, e);
  }

  public Statement translateSequence(Statement s1, Statement s2)
  {
    if (s1 == null)
      return s2;
    if (s2 == null)
      return s1;

    return new SequenceImpl(
      s1, s2
    );
  }

  public Statement translateSequence(Statement s, Expression e)
  {
    if (s == null)
      return new StatementExpressionImpl(e);
    if (e == null)
      return s;

    return new SequenceImpl(
      s,
      new StatementExpressionImpl(
        e
      )
    );
  }

  public Statement translateDeadVariableDestruction(Statement statement, ArrayList variables)
  {
    Iterator it = variables.iterator();
    while (it.hasNext())
    {
      Object o = it.next();

      Variable irVar = null;

      if (o instanceof net.nosica.compiler.typechecker.Variable)
      {
        net.nosica.compiler.typechecker.Variable var =
          (net.nosica.compiler.typechecker.Variable)o;

        irVar = translateVariable(var);
      }
      else
      if (o instanceof net.nosica.compiler.typechecker.Field)
      {
        net.nosica.compiler.typechecker.Field field =
          (net.nosica.compiler.typechecker.Field)o;

        FieldContext fieldContext = new FieldContextImpl(
          field
        );

        irVar = translateMemberAccess(translateThis(), fieldContext);
      }

      if (isPrimitive(irVar.getTypeName()))
      {
        try
        {
          ArrayList args = new ArrayList();
          args.add(irVar);

          statement = translateSequence(
            statement,
            new StatementExpressionImpl(
              translateMethodCall(
                new MethodCallContextImpl(
                  null,
                  env.getTE().getType(irVar.getTypeName()).getDestructor()
                ),
                args
              )
            )
          );
        }
        catch (SymbolException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }
      }

      // and variable is at end of scope
      statement = translateSequence(
        statement,
        new VariableScopeLeaveImpl(
          irVar
        )
      );
    }

    return statement;
  }

  public Statement translateFieldInitialization(Statement statement, ArrayList variables)
  {
    Iterator it = variables.iterator();
    while (it.hasNext())
    {
      net.nosica.compiler.typechecker.Field field =
        (net.nosica.compiler.typechecker.Field)it.next();

      MemberAccess exp = translateMemberAccess(
        translateThis(),
        new FieldContextImpl(
          field
        )
      );

      ArrayList args = new ArrayList();
      args.add(exp);

      if (isPrimitive(exp.getTypeName()))
      {
        // primitive variable : call default constructor if it exists
        ArrayList possibleMethods = new ArrayList();

        try
        {
          Method defaultConstructor = env.getMR().getDefaultConstructor(
            exp.getTypeName(),
            activeMethod.getType(),
            possibleMethods
          );

          if (defaultConstructor != null)
          {
            statement = translateSequence(
              statement,
              new StatementExpressionImpl(
                translateMethodCall(
                  new MethodCallContextImpl(
                    null,
                    defaultConstructor,
                    new ArrayList()
                  ),
                  args
                )
              )
            );
          }
        }
        catch (SymbolException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }
      }
      else
      {
        // reference variable : INIT it to null
        statement = translateSequence(
          statement,
          new InitImpl(
            exp,
            translateLiteral(
              NativeTypes.get_net_nosica_lang_None(),
              "null"
            )
          )
        );
      }
    }

    return statement;
  }

  public Variable translateVariable(net.nosica.compiler.typechecker.Variable variable)
  {
    Variable IRVar = null;

    if (variableCache.containsKey(variable))
      IRVar = (Variable)variableCache.get(variable);
    else
    {
      if (variable.isArgument())
        IRVar = new ArgumentImpl(
          variable.getTypeName(),
          variable.isConst(),
          variable.getName()
        );
      else
        IRVar = new VariableImpl(
          variable.getTypeName(),
          variable.getName()
        );

      variableCache.put(variable, IRVar);
    }

    return IRVar;
  }

  public Expression translateCastExpression(TypeName castTypeName, Expression exp)
  {
    return new CastImpl(exp, castTypeName);
  }

  public Expression translateInstanceOfExpression(TypeName instanceOfTypeName, Expression exp)
  {
    return new InstanceOfImpl(exp, instanceOfTypeName);
  }

  public Expression translateLiteral(TypeName typeName, String str)
  {
    return new LiteralImpl(
      typeName,
      str
    );
  }

  public Variable translateThis()
  {
    return thisVariable;
  }

  public MemberAccess translateSuper()
  {
    try
    {
      Record record = env.getIRRE().getRecord(activeMethod.getType());
      Field field = record.getField("super");

      return new MemberAccessImpl(
        translateThis(),
        field
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
      return null;
    }
  }

  public MemberAccess translateMemberAccess(Expression exp, FieldContext fieldContext)
  {
    Iterator i = fieldContext.iterateSubscript();
    while (i.hasNext())
    {
      Pair element = (Pair)i.next();
      boolean  elementIsEnclosing = ((Boolean)element.first).booleanValue();

      if (elementIsEnclosing)
        exp = translateEnclosing(exp);
      else
        exp = translateSuper(exp);
    }

    net.nosica.compiler.typechecker.Field  field  = fieldContext.getField();
    try
    {
      Record record = env.getIRRE().getRecord(field.getType());

      Field IRField = record.getField(field.getName());

      return new MemberAccessImpl(
        exp,
        IRField
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
      return null;
    }
  }

  public Expression translatePrimitiveConstruction(MethodCallContext mcc, ArrayList args)
  {
    // create variable that will hold the result
    net.nosica.compiler.typechecker.Variable tempVar = new net.nosica.compiler.typechecker.VariableImpl(
      false,
      mcc.getMethod().getType(),
      env.getNG().generateTemporaryName(),
      false
    );

    Variable var = translateVariable(tempVar);

    return translatePrimitiveConstruction(var, mcc, args);
  }

  public Expression translatePrimitiveConstruction(Expression var, MethodCallContext mcc, ArrayList args)
  {
    args.add(0, var);
    return new ExtendedSequenceImpl(
      new StatementExpressionImpl(
        translateMethodCall(mcc, args)
      ),
      var
    );
  }

  public Expression translateArrayAllocation(ArrayList constructors, ArrayList args)
  {
    MethodCallContext mcc = (MethodCallContext)constructors.remove(0);
    ArrayList mArgs = (ArrayList)args.remove(0);

    // create variable that will hold the result
    net.nosica.compiler.typechecker.Variable tempVar = new net.nosica.compiler.typechecker.VariableImpl(
      false,
      mcc.getMethod().getType(),
      env.getNG().generateTemporaryName(),
      false
    );

    Variable var = translateVariable(tempVar);

    // and add it to VariableEnvironment for further destruction
    try
    {
      varEnv.add(tempVar);
    }
    catch (TypeCheckerException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }

    // var = RAW ALLOCATION
    Statement s = new InitImpl(
      var,
      new RawAllocationImpl(
        var.getTypeName()
      )
    );

    ArrayList constructionArgs = new ArrayList();
    constructionArgs.add(var);
    constructionArgs.addAll(mArgs);

    // VAR.constructor(dimensions)
    s = translateSequence(
      s,
      translateMethodCall(
        mcc,
        constructionArgs
      )
    );

    // create one counter for each dimension.
    ArrayList loopCounters = new ArrayList();
    Iterator it = mArgs.iterator();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      loopCounters.add(
        new VariableImpl(
          exp.getTypeName(),
          env.getNG().generateTemporaryName()
        )
      );
    }

    s = recTranslateArrayLoops(s, constructors, args, var, loopCounters, mArgs, 0);

    return new ExtendedSequenceImpl(
      s,
      var
    );
  }

  /**
   * \param s statement to complete
   * \param constructors ArrayList<MethodCallContext>
   * \param args ArrayList<ArrayList<Expression> >
   * \param instance the array instance to work on
   * \param loopCounters ArrayList<Variable> variable to use to construct loop
   * \param loopLimits ArrayList<Expression> limit of each loop
   * \param activeLoop loop being generated
   */
  private Statement recTranslateArrayLoops(
    Statement s,
    ArrayList constructors,
    ArrayList args,
    Variable instance,
    ArrayList loopCounters,
    ArrayList loopLimits,
    int activeLoop
  )
  {
    Variable loopCounter = (Variable)loopCounters.get(activeLoop);
    Expression loopLimitExp = (Expression)loopLimits.get(activeLoop);

    Method assignOp = env.getMR().getOperatorInstanceAssign(loopCounter.getTypeName());

    // transfer loopLimit into a variable
    Variable loopLimit = new VariableImpl(loopLimitExp.getTypeName(), env.getNG().generateTemporaryName());
    ArrayList loopLimitInitArgs = new ArrayList();
    loopLimitInitArgs.add(loopLimit);
    loopLimitInitArgs.add(loopLimitExp);

    // loopLimit = loopLimitExp
    s = translateSequence(
      s,
      translateMethodCall(
        new MethodCallContextImpl(
          null,
          assignOp
        ),
        loopLimitInitArgs
      )
    );

    // initialise loop counter to 0
    ArrayList loopCounterInitArgs = new ArrayList();
    loopCounterInitArgs.add(loopCounter);
    loopCounterInitArgs.add(
      translateLiteral(
        loopCounter.getTypeName(),
        "0"
      )
    );

    // LoopCounter = 0
    s = translateSequence(
      s,
      translateMethodCall(
        new MethodCallContextImpl(
          null,
          assignOp
        ),
        loopCounterInitArgs
      )
    );

    // generate jump to condition
    Label condLabel = generateLabel();

    // GOTO COND
    s = translateSequence(
      s,
      translateJump(
        condLabel
      )
    );

    // insert a label to begin of block
    Label beginLabel = generateLabel();

    // BEGIN :
    s = translateSequence(
      s,
      translateLabel(
        beginLabel
      )
    );

    // recursive call to generate either
    // - inner loop
    // - initialisation of array element
    if ((activeLoop + 1) < loopCounters.size())
    {
      // inner loop
      s = recTranslateArrayLoops(
        s,
        constructors,
        args,
        instance,
        loopCounters,
        loopLimits,
        activeLoop + 1
      );
    }
    else
    {
      // construct array element
      try
      {
        Type arrayElementType = env.getTE().getType(instance.getTypeName().getArrayElementType());

        Expression arrayElement = null;

        if (arrayElementType.isPrimitive())
        {
          // call default constructor passed in parameter
          if (constructors.size() != 1)
          {
            env.terminate("Internal error : while constructing array initialiser, list of constructors should be 1");
          }

          // create a temporary
          arrayElement = new VariableImpl(arrayElementType.getName(), env.getNG().generateTemporaryName());

          MethodCallContext mcc = (MethodCallContext)constructors.get(0);
          ArrayList arrayElementConstructorArgs = new ArrayList();
          arrayElementConstructorArgs.add(arrayElement);

          // temporary.constructor()
          s = translateSequence(
            s,
            translateMethodCall(
              mcc,
              arrayElementConstructorArgs
            )
          );
        }
        else
        {
        // reference type
          if (arrayElementType.getName().isArray() && constructors.size() > 0)
          {
            // recursive call for array element types (if any)
            arrayElement = translateArrayAllocation(
              constructors,
              args
            );
          }
          else
          {
            // initialisation to null
            arrayElement = translateLiteral(
              NativeTypes.get_net_nosica_lang_None(),
              "null"
            );
          }
        }

        // ARRAYSET
        s = translateSequence(
          s,
          new ArraySetImpl(
          instance,
          arrayElement,
          loopCounters
          )
        );
      }
      catch (SymbolException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }

    // increment of loop counter
    Method incMethod = env.getMR().getOperatorPlusPlus(loopCounter.getTypeName());
    ArrayList incArgs = new ArrayList();
    incArgs.add(loopCounter);

    // ++loopCounter
    s = translateSequence(
      s,
      translateMethodCall(
        new MethodCallContextImpl(
          null,
          incMethod
        ),
        incArgs
      )
    );

    // COND :
    s = translateSequence(
      s,
      translateLabel(
        condLabel
      )
    );

    // translate comparison
    Method ltOperator = env.getMR().getOperatorLt(loopCounter.getTypeName());
    ArrayList ltArgs = new ArrayList();
    ltArgs.add(loopCounter);
    ltArgs.add(loopLimit);

    Variable testResult = new VariableImpl(NativeTypes.get_net_nosica_lang_bool(), env.getNG().generateTemporaryName());
    Method boolAssign = env.getMR().getOperatorInstanceAssign(testResult.getTypeName());
    ArrayList boolAssignArgs = new ArrayList();
    boolAssignArgs.add(testResult);
    boolAssignArgs.add(
      translateMethodCall(
        new MethodCallContextImpl(
          null,
          ltOperator
        ),
        ltArgs
      )
    );

    // testResult = loopCounter < loopLimit
    s = translateSequence(
      s,
      translateMethodCall(
        new MethodCallContextImpl(
          null,
          boolAssign
        ),
        boolAssignArgs
      )
    );

    // CONDJUMP(testResult, begin, end)
    Label endLabel = generateLabel();

    s = translateSequence(
      s,
      translateConditionalJump(
        testResult,
        beginLabel,
        endLabel
      )
    );

    // END :
    s = translateSequence(
      s,
      translateLabel(endLabel)
    );

    // destroy temporaries
    s = translateSequence(
      s,
      new VariableScopeLeaveImpl(testResult)
    );

    s = translateSequence(
      s,
      new VariableScopeLeaveImpl(loopLimit)
    );

    s = translateSequence(
      s,
      new VariableScopeLeaveImpl(loopCounter)
    );

    return s;
  }

  public Expression translateAllocation(MethodCallContext mcc, ArrayList args)
  {
    // create variable that will hold the result
    net.nosica.compiler.typechecker.Variable tempVar = new net.nosica.compiler.typechecker.VariableImpl(
      false,
      mcc.getMethod().getType(),
      env.getNG().generateTemporaryName(),
      false
    );

    Variable var = translateVariable(tempVar);

    // and add it to VariableEnvironment for further destruction
    try
    {
      varEnv.add(tempVar);
    }
    catch (TypeCheckerException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }

    Expression result = new ExtendedSequenceImpl(
      new InitImpl(
        var,
        new RawAllocationImpl(
          var.getTypeName()
        )
      ),
      var
    );

    args.add(0, result);
    return new ExtendedSequenceImpl(
      new StatementExpressionImpl(
        translateMethodCall(mcc, args)
      ),
      var
    );
  }

  public LabelName translateLabel(net.nosica.compiler.typechecker.Label label)
  {
    return new LabelNameImpl(label);
  }

  public Jump translateJump(net.nosica.compiler.typechecker.Label label)
  {
    return new JumpImpl(label);
  }

  public Statement translateConditionalJump(Variable cond, Label trueLabel, Label falseLabel)
  {
    return new ConditionalJumpImpl(
      cond,
      trueLabel,
      falseLabel
    );
  }

  public Statement translateReturn(Expression exp)
  {
    return new ReturnImpl(exp);
  }

  public Statement translateInit(Variable var, Expression exp)
  {
    return new InitImpl(var, exp);
  }

  public Expression translateInitAsExpression(Variable var, Expression exp)
  {
    // assign temporary to left variable
    Statement statement = new InitImpl(var, exp);

    // and return left variable as result
    return new ExtendedSequenceImpl(
      statement,
      var
    );
  }

  public Statement translateTryRegister(Map exceptionMap, Statement finallyStatement)
  {
    return new TryRegisterImpl(
      exceptionMap,
      finallyStatement
    );
  }

  public Statement translateTryUnregister()
  {
    return new TryUnregisterImpl();
  }

  public Statement translateThrow(Expression exp)
  {
    return new ThrowImpl(exp);
  }

  private boolean isPrimitive(TypeName typeName)
  {
    try
    {
      Type type = env.getTE().getType(typeName);
      return type.isPrimitive();
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
      return false;
    }
  }
}
