/*
  $Id: MessageServiceImpl.java,v 1.12 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.application.*;
import net.nosica.framework.Localizable;

import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.File;
import java.util.StringTokenizer;
import java.util.Map;
import java.util.HashMap;

public class MessageServiceImpl implements MessageService
{
  private static MessageServiceImpl theInstance = new MessageServiceImpl();
  synchronized public static MessageServiceImpl instance() { return theInstance; }
  synchronized public static void instanceReset() { theInstance = new MessageServiceImpl(); }
  private MessageServiceImpl() { initialize(); }

  private int nbError = 0;
  private int nbErrorMax;
  private int nbWarning = 0;
  private Map messages = new HashMap();  // Map<String, Message>
  private Filter filter = new SimpleFilter();

  public void report(MessageLevel level, String msgId, String[] params, Localizable localizable)
  {
    if (level == MessageLevel.DBG && !filter.messageOK(msgId))
      return;

    if (messages.containsKey(msgId))
      reportError(level, (Message)messages.get(msgId), params, localizable);
    else
      reportError(level, msgId, params, localizable);
  }

  public void report(MessageLevel level, String msgId, String param, Localizable localizable)
  {
    report(level, msgId, new String[]{param}, localizable);
  }

  public void report(MessageLevel level, String msgId, Localizable localizable)
  {
    report(level, msgId, new String[]{}, localizable);
  }

  public void report(MessageLevel level, String msgId)
  {
    report(level, msgId, new String[]{}, null);
  }

  public int getNbError()   { return nbError; }
  public int getNbWarning() { return nbWarning; }

  public void attachFilter(Filter filter)
  {
    this.filter = filter;
  }

  private void reportError(MessageLevel level, Message message, String[] params, Localizable localizable)
  {
    StringBuffer result = new StringBuffer();

    // display type of error
    if (level == MessageLevel.NFO)
    {
      result.append("NFO:");
    }
    else if (level == MessageLevel.DBG)
    {
      result.append("DBG:");
    }
    else if (level == MessageLevel.WRN)
    {
      result.append("WRN:");
      ++nbWarning;
    }
    else if (level == MessageLevel.ERR)
    {
      result.append("ERR:");
      ++nbError;
    }

    result.append(message.id).append("\n");

    // display localization of error
    if (localizable != null)
      result.append("-- ").append(localizable.localization()).append("\n");

    // display message + parameters
    result.append("-- ");
    int pos = 0;
    int oldPos = 0;
    while ((pos = message.message.indexOf('%', pos)) != -1)
    {
      String part = message.message.substring(oldPos, pos);
      result.append(part);

      StringBuffer strNumber = new StringBuffer();
      while (pos + 1 < message.message.length())
      {
        ++pos;
        char ch = message.message.charAt(pos);
        if (Character.isDigit(ch))
          strNumber.append(ch);
        else
          break;
      }

      int number = Integer.parseInt(strNumber.toString());

      if (number < params.length)
        result.append(params[number]);
      else
        result.append("MISSING PARAMETER");

      oldPos = pos;
    }

    result.append(message.message.substring(oldPos));

    // take action
    if (level == MessageLevel.ERR)
      System.err.println(result.toString());
    else
      System.out.println(result.toString());

    if (nbError >= nbErrorMax)
      Environment.instance().terminate("-- Too many errors (" + nbError + "), correct above errors");
  }

  private void reportError(MessageLevel level, String msgId, String[] params, Localizable localizable)
  {
    StringBuffer result = new StringBuffer();

    result.append("UNKNOWN ERROR [").append(msgId).append("] (message not found in message file)\n");
    if (localizable != null)
      result.append(localizable.localization()).append("\n");
    if (params.length > 0)
      result.append("PARAMETERS :");
    boolean needComa = false;
    for (int i = 0; i < params.length; ++i)
    {
      if (needComa)
        result.append(", ");
      result.append("[").append(params[i]).append("]");
      needComa = true;
    }

    System.out.println(result.toString());
  }

  private void initialize()
  {
    try
    {
      nbErrorMax = Environment.instance().getAC().getPropertyInt("error.max");

      nbError    = 0;
      nbWarning  = 0;

      File file  = ApplicationContextImpl.instance().getPropertyPath("error.messages");
      if (!file.exists())
        throw new java.io.IOException();

      LineNumberReader fileReader  =
        new LineNumberReader(
          new InputStreamReader(
            new FileInputStream(file)));

      String id;
      String message;

      while(fileReader.ready())
      {
        String line = fileReader.readLine();
        if (line == null) break;
        if (line.length() == 0) continue;
        if (line.charAt(0) == '#') continue;
        //
        StringTokenizer st = new StringTokenizer(line, ";");
        if (st.hasMoreTokens()) id      = st.nextToken().trim();
        else throw new Exception();
        if (st.hasMoreTokens()) message = st.nextToken().trim();
        else throw new Exception();
        //
        messages.put(id, new Message(id, message));
      }
    }
    catch (PropertyException e)
    {
      Environment.instance().terminate("FATAL: Property error.messages must be defined");
    }
    catch (java.io.IOException e)
    {
      Environment.instance().terminate("FATAL: can't load messages file");
    }
    catch (Exception e)
    {
      Environment.instance().terminate("FATAL: invalid format in message file");
    }
  }

  static private class Message
  {
    public String id;
    public String message;

    public Message(String id, String message)
    {
      this.id = id;
      this.message = message;
    }
  }
}
