/*
  $Id: SerialCompilerProcessorImpl.java,v 1.52 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponentImpl;

import net.nosica.symbol.GenericType;
import net.nosica.symbol.AliasTable;
import net.nosica.symbol.SymbolException;

import net.nosica.application.NativeTypes;
import net.nosica.application.PropertyException;
import net.nosica.application.MessageLevel;
import net.nosica.compiler.typechecker.*;
import net.nosica.compiler.typechecker.TypeCheckVisitor;
import net.nosica.compiler.typechecker.StatementVisitor;
import net.nosica.compiler.typechecker.TypeCheckerException;
import net.nosica.compiler.typechecker.Type;
import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.Field;
import net.nosica.compiler.typechecker.Variable;
import net.nosica.compiler.typechecker.Label;
import net.nosica.compiler.typechecker.Argument;
import net.nosica.compiler.typechecker.ExceptionEnvironment;
import net.nosica.compiler.typechecker.ExceptionEnvironmentImpl;

import net.nosica.compiler.intermediate.IRVisitorRewriter;
import net.nosica.compiler.intermediate.ExSeqRemoverVisitor;
//import net.nosica.compiler.intermediate.BasicBlockVisitor;
import net.nosica.compiler.intermediate.Record;
import net.nosica.compiler.intermediate.Statement;
import net.nosica.compiler.intermediate.Function;
import net.nosica.compiler.intermediate.FunctionImpl;

import net.nosica.compiler.analysis.GlobalAnalysisDriver;
import net.nosica.compiler.analysis.GlobalAnalysisDriverImpl;

import net.nosica.compiler.production.CodeProduction;

import net.nosica.framework.VisitorException;
import net.nosica.framework.Localizable;

import net.nosica.util.IndentedPrintWriter;

import java.util.Stack;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * \brief implementation of a CompilerProcessor as a serial one.
 * That is, do not use thread, parallel processing and all this kind of stuff.
 * These things will be implemented later in a ParallelCompilerProcessorImpl :-)
 */
public class SerialCompilerProcessorImpl implements CompilerProcessor
{
  private static SerialCompilerProcessorImpl theInstance = new SerialCompilerProcessorImpl();
  public static SerialCompilerProcessorImpl instance() { return theInstance; }
  public static void instanceReset() { theInstance = new SerialCompilerProcessorImpl(); }

  private Environment env = Environment.instance();
  private boolean debugRecord;
  private String debugRecordType = null;
  private boolean debugIRMethods;
  private String debugIRMethodsOfType = null;

  private Type mainType;

  private Stack typesToCompil = new Stack();      // Stack<Type>
  private Set typesCompiled = new HashSet();      // Set<TypeName>

  private int nbMethod = 0;

  private SerialCompilerProcessorImpl()
  {
    try
    {
      debugRecord = env.getAC().getPropertyBoolean("debug.record");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugRecordType = env.getAC().getPropertyString("debug.record.type");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugIRMethods = env.getAC().getPropertyBoolean("debug.ir");
    }
    catch (PropertyException e)
    {
    }

    try
    {
      debugIRMethodsOfType = env.getAC().getPropertyString("debug.ir.type");
    }
    catch (PropertyException e)
    {
    }
  }

  private void bootstrap() throws CompilationAbortException
  {
    try
    {
      // get name of main class of program to compil
      String className = env.getAC().getPropertyString("source");

      if (className.length() == 0)
      {
        env.getMS().report(MessageLevel.NFO, "Waiting for a class name");
        return;
      }

      // sanity check for typeName : must not be generic nor array
      int lt  = className.indexOf('<');
      int gt  = className.indexOf('>');
      int arrayBegin = className.indexOf('[');
      int arrayEnd = className.indexOf(']');

      if (lt != -1 || gt != -1)
        env.terminate("Fatal - Cannot take a generic TypeName as the main class to compil");
      if (arrayBegin != -1 || arrayEnd != -1)
        env.terminate("Fatal - Cannot take an array TypeName as the main class to compil");

      // conversion
      TypeName typeName = string2TypeName(className);

      // because we know this type is not generic and TypeLoader can mess up because there
      // is no generic type loaded at all for now
      try
      {
        env.getTL().loadSymbols(typeName);
      }
      catch (SymbolException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      GenericType type = null;
      try
      {
        type = env.getSE().getGenericType(typeName);
      }
      catch (SymbolException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      GenericitySolver genericitySolver = env.getGS();
      Type instantiatedType = null;

      try
      {
        instantiatedType = genericitySolver.instantiateType(type, new java.util.HashMap(), typeName);
      }
      catch (GenericityException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      try
      {
        env.getTE().addType(instantiatedType);
      }
      catch (SymbolException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      // load type only when type is referenced by TypeEnvironment
      // this avoids recursion
      try
      {
        env.getTL().loadDepends(instantiatedType);
      }
      catch (GenericityException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }
      catch (SymbolException e)
      {
        // could not load type typeName
        System.out.println(e.getMessage());
        throw new CompilationAbortException();
      }

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      // verify interface if class
      if (instantiatedType.isClass() && !instantiatedType.isAbstract())
        env.getTL().verifyInterface(instantiatedType);

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      // verify some other details
      env.getTL().verifyType(instantiatedType);

      if (env.getMS().getNbError() != 0)
      {
        return;
      }

      addUsedType(instantiatedType.getName());

      this.mainType = instantiatedType;

      // verify mainType has a main method
      verifyMainMethod(mainType);
    }
    catch (PropertyException  e)
    {
      env.terminate("Missing class name");
    }
  }

  private void bootstrapIRType()
  {
    try
    {
      env.getTL().getType(NativeTypes.get_net_nosica_lang_None());
      addUsedType(NativeTypes.get_net_nosica_lang_None());

      env.getTL().getType(NativeTypes.get_net_nosica_lang_bool());
      addUsedType(NativeTypes.get_net_nosica_lang_bool());

      env.getTL().getType(NativeTypes.get_net_nosica_lang_word32());
      addUsedType(NativeTypes.get_net_nosica_lang_word32());

      env.getTL().getType(NativeTypes.get_net_nosica_lang_string_array());
      addUsedType(NativeTypes.get_net_nosica_lang_string_array());
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
    catch (GenericityException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
  }

  public void compil()
  {
    try
    {
      bootstrapIRType();
      bootstrap();
      compilTypes();

      // code analysis
      if (env.getMS().getNbError() == 0)
      {
        GlobalAnalysisDriver gaDriver = new GlobalAnalysisDriverImpl();
        gaDriver.globalAnalysis(mainType);
      }

      // code production
      if (env.getMS().getNbError() == 0)
      {
        CodeProduction cp = env.createCP(mainType);
        cp.produce();
      }
    }
    catch (CompilationAbortException e)
    {
      env.terminate("Compilation aborted");
    }

    if (env.getMS().getNbError() != 0)
    {
      env.terminate("Number of errors : " + env.getMS().getNbError());
    }
  }

  private void compilTypes()
  {
    // while there is a type to compil
    while (!typesToCompil.empty())
    {
      Type t = (Type)typesToCompil.pop();

      env.getMS().report(MessageLevel.DBG, "Compil001", t.getName().toString(), null);

      try
      {
        addIRType(t);
      }
      catch (SymbolException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }

      if (debugRecord)
      {
        IRRecordEnvironmentDisplayer displayer = new IRRecordEnvironmentDisplayer(
          new IndentedPrintWriter(System.out)
        );

        try
        {
          displayer.display(env.getIRRE().getRecord(t.getName()));
        }
        catch (SymbolException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }
      }
      else
      if (debugRecordType != null)
      {
        if (t.getName().toString().equals(debugRecordType))
        {
          IRRecordEnvironmentDisplayer displayer = new IRRecordEnvironmentDisplayer(
            new IndentedPrintWriter(System.out)
          );

          try
          {
            displayer.display(env.getIRRE().getRecord(t.getName()));
          }
          catch (SymbolException e)
          {
            env.terminate("Internal error : " + e.getMessage());
          }
        }
      }

      // we use the following vector as a temporary list of methods to compil
      ArrayList methods = new ArrayList();

      // initializer
      if (!t.getName().isArray() && !t.getName().equals(NativeTypes.get_net_nosica_lang_Object()))
        methods.add(t.getInitializer());

      // iterate cast operators if any
      Iterator i = t.iterateCastOperators();
      while (i.hasNext())
      {
        Method m = (Method)i.next();

        methods.add(m);
      }

      // iterate methods
      i = t.iterateOverloadables();
      while (i.hasNext())
      {
        Iterator j = ((ArrayList)i.next()).iterator();
        while (j.hasNext())
        {
          Method m = (Method)j.next();

          methods.add(m);
        }
      }

      // destructor
      methods.add(t.getDestructor());

      // deinitializer
      if (!t.getName().isArray() && !t.getName().equals(NativeTypes.get_net_nosica_lang_Object()))
        methods.add(t.getDeinitializer());

      // now compil those methods ...
      i = methods.iterator();
      while (i.hasNext())
      {
        Method m = (Method)i.next();

        compilOneMethod(m);
      }
    }
  }

  private void addIRType(Type type) throws SymbolException
  {
    if (env.getIRRE().recordExists(type.getName()))
      return;

    Iterator it = type.iterateExtends();
    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();
      Type extendsType = env.getTE().getType(extendsTypeName);
      addIRType(extendsType);
    }

  it = type.iterateImplements();
  while (it.hasNext())
  {
    TypeName extendsTypeName = (TypeName)it.next();
    Type implementsType = env.getTE().getType(extendsTypeName);
    addIRType(implementsType);
  }

    Record record = env.getIRRE().translateType(type);

    env.getIRRE().addRecord(record);
  }

  private void compilOneMethod(Method m)
  {
    env.getMS().report(MessageLevel.DBG, "Compil002", m.getName(), null);

    // load enclosing type if not loaded
    TypeName enclosingType = m.getType();

    Type type = null;

    try
    {
      type = env.getTE().getType(enclosingType);
    }
    catch (SymbolException e)
    {
    }

    AliasTable aliasTable = m.getAliasTable();

    // create a fake variable that will be used only to hold values on return statement
    net.nosica.compiler.intermediate.Variable returnValue = null;
    if (!m.isSub())
    {
      returnValue = new net.nosica.compiler.intermediate.VariableImpl(
        m.getResultType().getTypeName(),
        "___returnValue"
      );
    }
    Label endMethodLabel = null;

    // now create local environments for type checking of this method
    // 1 - a variable environment
    VariableEnvironment varEnv = new VariableEnvironmentImpl();
    addArgsToVariableEnvironment(varEnv, m.iterateArguments(), m.getModel().getNode());
    markArgsAsUsedType(m);
    // 2 - a label environment
    LabelRepository labelRepository = new LabelRepositoryImpl("method" + nbMethod++ + "_");
    endMethodLabel = labelRepository.generateLabel();
    // 3 - an exception environment
    ExceptionEnvironment exceptionEnvironment = new ExceptionEnvironmentImpl();
    // 4 - an IR translator
    IRTranslator IRT = new IRTranslatorImpl(m, varEnv);
    // 4 - typechecker
    TypeCheckVisitor visitor =
      new TypeCheckVisitor(
        type,
        m,
        aliasTable,
        varEnv,
        labelRepository,
        exceptionEnvironment,
        IRT,
        new UseTypeCallBackImpl()
       );
    // 5 - statement visitor
    StatementVisitor statementVisitor =
    new StatementVisitor(
      visitor,
      endMethodLabel,
      returnValue
    );

    // Type check method
    m.getModel().getNode().accept(statementVisitor);

    // check that all exceptions declared allowed to be thrown were really thrown
    Iterator it = m.iterateThrowList();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();

      if (!m.getThrownExceptionList().contains(typeName))
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck063", typeName.toString(), m.getModel().getNode());
      }
    }

    // now get translated statements
    net.nosica.compiler.intermediate.Statement statement = statementVisitor.getStatement();

    if (env.getMS().getNbError() != 0)
      return;

    if (statement != null)
    {
      try
      {
//      displayMethod("1", m, statement, "destructor");

        // add epilogue of method :
        // endMethodLabel :
        statement = visitor.getIRT().translateSequence(
          statement,
          visitor.getIRT().translateLabel(endMethodLabel)
        );

//      displayMethod("2", m, statement, "destructor");

        if (!m.isSub())
        {
          // RETURN(returnValue)
          statement = visitor.getIRT().translateSequence(
            statement,
            visitor.getIRT().translateReturn(returnValue)
          );
        }

//      displayMethod("3", m, statement, "main");

        // 1 : simplify translation by rewriting IR
        statement = rewrite(
          statement,
          new ExSeqRemoverVisitor()
        );

//      displayMethod("4", m, statement, "main");

/*      // 5 : decompose statements in Basic Block (wrong IR generation : disabled for now)
        statement = rewrite(
          statement,
          new BasicBlockVisitor()
        );

        System.out.println("after basic block");
        displayMethod(m, statement, "main");
*/
      }
      catch (VisitorException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }

    // add IR statement to IR repository
    try
    {
      Function f = new FunctionImpl(
        m,
        translateArgsInIRVar(m, IRT),
        statement,
        IRT.translateThis()
      );

      env.getIRFE().addFunction(f);

      if (debugIRMethods)
      {
        IRFunctionEnvironmentDisplayer displayer = new IRFunctionEnvironmentDisplayer(
          new IndentedPrintWriter(System.out)
        );

        try
        {
          displayer.display(f);
        }
        catch (SymbolException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }
      }
      else
      if (debugIRMethodsOfType != null)
      {
        if (f.getMethod().getType().toString().equals(debugIRMethodsOfType))
        {
          IRFunctionEnvironmentDisplayer displayer = new IRFunctionEnvironmentDisplayer(
            new IndentedPrintWriter(System.out)
          );

          try
          {
            displayer.display(f);
          }
          catch (SymbolException e)
          {
            env.terminate("Internal error : " + e.getMessage());
          }
        }
      }
    }
    catch (VisitorException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }
  }

  private void displayMethod(String message, Method m, Statement statement, String name)
  {
    if (name == null || (m.getName() != null && m.getName().equals(name)))
    {
      System.out.println(message);

      IRFunctionEnvironmentDisplayer displayer = new IRFunctionEnvironmentDisplayer(
        new IndentedPrintWriter(System.out)
      );

      try
      {
        statement.accept(displayer);
      }
      catch (VisitorException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }
  }

  private Statement rewrite(Statement statement, IRVisitorRewriter visitor) throws VisitorException
  {
    try
    {
      statement.accept(visitor);
    }
    catch (VisitorException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }

    statement = visitor.getRewrittenStatement();

    return statement;
  }

  private void verifyMainMethod(Type mainType)
  {
    Method m = env.getMR().findMainMethod(mainType.getName());

    if (m == null)
    {
      env.getMS().report(MessageLevel.ERR, "Symbol040", mainType.getName().toString(), null);
    }
  }

  private ArrayList translateArgsInIRVar(Method m, IRTranslator irt)
  {
    ArrayList IRVar = new ArrayList();
    Iterator it = m.getArguments().iterator();

    while (it.hasNext())
    {
      Variable var = (Variable)it.next();

      IRVar.add(irt.translateVariable(var));
    }

    return IRVar;
  }

  private void addArgsToVariableEnvironment(VariableEnvironment varEnv, Iterator it, Localizable node)
  {
    // add arguments of method to variable environment
    while (it.hasNext())
    {
      Argument arg = (Argument)it.next();

      try
      {
        varEnv.add(arg);

        env.getTL().getType(arg.getTypeName());
      }
      catch (TypeCheckerException e)
      {
        env.getMS().report(MessageLevel.ERR, "Symbol030", arg.getName(), node);
      }
      catch (SymbolException e)
      {
      }
      catch (GenericityException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }
  }

  private void markArgsAsUsedType(Method m)
  {
    Iterator it = m.iterateArguments();
    while (it.hasNext())
    {
      Argument arg = (Argument)it.next();
      addUsedType(arg.getTypeName());
    }

    if (!m.isSub())
    {
      addUsedType(m.getResultType().getTypeName());
    }
  }

  /**
   * \brief Converts a simple String representation of a typeName into a TypeName
   * By simple, we mean that there is no generic compound nor array compound<BR>
   * Be aware that these checks are not performed here, so conversion can silently occurs with errors
   *
   * \param className string representation of a TypeName to convert
   * \return converted TypeName
   */
  private TypeName string2TypeName(String className)
  {
    ArrayList components = new ArrayList();
    int dot     = 0;
    int lastDot = -1;

    while ((dot = className.indexOf('.', dot + 1)) != -1)
    {
      String strComponent = className.substring(lastDot + 1, dot);
      components.add(new TypeNameComponentImpl(strComponent, null));
      lastDot = dot;
    }

    String strLastComponent = className.substring(lastDot + 1);
    components.add(new TypeNameComponentImpl(strLastComponent, null));

    return new TypeNameImpl(components);
  }

  private void addUsedType(TypeName typeName)
  {
    if (typesCompiled.contains(typeName))
      return;

    try
    {
      Type type = env.getTE().getType(typeName);

      if (type.isInterface())
      {
        // an interface gets never compiled ... (it has no code)
        return;
      }

      if (type.hasExplicitExtends())
      {
        Iterator it = type.iterateExtends();
        while (it.hasNext())
        {
          TypeName extendsTypeName = (TypeName)it.next();

          addUsedType(extendsTypeName);
        }
      }
      else
      {
        // net.nosica.lang.Object is an implicit extends
        addUsedType(NativeTypes.get_net_nosica_lang_Object());
      }

      Iterator it = type.iterateFields();
      while (it.hasNext())
      {
        Field field = (Field)it.next();
        if (!field.getTypeName().equals(typeName))
        {
          // add typeName of field if different from type being added
          // (to avoid infinite recursion)
          addUsedType(field.getTypeName());
        }
      }

      typesToCompil.add(type);
      addIRType(type);
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }

    typesCompiled.add(typeName);
  }

  private class UseTypeCallBackImpl implements UseTypeCallBack
  {
    public void useType(TypeName typeName)
    {
      addUsedType(typeName);
    }
  }
}
