/*
  $Id: TypeEnvironmentDisplayer.java,v 1.12 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.util.IndentedPrintWriter;

import net.nosica.compiler.typechecker.Accessible;
import net.nosica.compiler.typechecker.Type;
import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.Field;
import net.nosica.compiler.typechecker.ResultType;
import net.nosica.compiler.typechecker.Argument;

import net.nosica.symbol.SymbolException;
import net.nosica.symbol.AliasTable;
import net.nosica.symbol.MethodKindEnum;

import java.util.Map;
import java.util.Iterator;
import java.util.ArrayList;

/**
 * \brief a class to display the contents of TypeEnvironment
 * Display all contents of TypeEnvironment. For degug purpose
 */

public class TypeEnvironmentDisplayer
{
  private IndentedPrintWriter writer;
  private Environment env = Environment.instance();

  /**
   * \brief creates a new displayer
   *
   * \param writer an IndentedPrintWriter
   */
  public TypeEnvironmentDisplayer(IndentedPrintWriter writer) { this.writer = writer; }

  /**
   * \brief Display all contents of TypeEnvironment
   * the writer is used as output channel
   */
  public void display()
  {
    Iterator it = env.getTE().iterateTypeName();

    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();

      try
      {
        Type type = env.getTE().getType(typeName);

        display(type);
      }
      catch (SymbolException e)
      {
        throw new Error("Internal error : " + e.getMessage());
      }
    }
  }

  public void display(Type type)
  {
    writer.println();

    AliasTable aliases = type.getAliasTable();

    // TypeName, kind of type, property of type
    display((Accessible)type);
    writer.print(type.getKind().toString() + " " + type.getName() + " [");
    if (type.isNative()) writer.print("N"); else writer.print("-");
    if (type.isPrimitive()) writer.print("P"); else writer.print("-");
    if (type.isAbstract()) writer.print("A"); else writer.print("-");
    if (type.isFinal()) writer.print("F"); else writer.print("-");
    if (type.isInner()) writer.print("I"); else writer.print("-");
    writer.println("]");

    // Extends
    writer.println("EXTENDS --");
    displayTypeNameList(type.iterateExtends());

    // Implements
    writer.println("IMPLEMENTS --");
    displayTypeNameList(type.iterateImplements());

    // Aliases
    writer.println("ALIASES --");
    display(aliases);

    // fields
    writer.println("FIELDS --");
    writer.addIndent();
    Iterator fieldIt = type.iterateFields();
    while (fieldIt.hasNext())
    {
      Field field = (Field)fieldIt.next();
      display(field);
    }
    writer.decIndent();

    // destructor
    writer.println("DESTRUCTOR--");
    writer.addIndent();
    display(type.getDestructor());
    writer.decIndent();

    // cast operators
    writer.println("CAST OPERATORS--");
    writer.addIndent();
    Iterator castIt = type.iterateCastOperators();
    while (castIt.hasNext())
    {
      Method method = (Method)castIt.next();
      display(method);
    }
    writer.decIndent();

    // overloadables
    writer.println("OVERLOADABLES--");
    writer.addIndent();
    Iterator methodsIt = type.iterateOverloadables();
    while (methodsIt.hasNext())
    {
      Iterator methodIt = ((ArrayList)methodsIt.next()).iterator();
      while (methodIt.hasNext())
      {
        Method method = (Method)methodIt.next();
        display(method);
      }
    }
    writer.decIndent();

    writer.println();
  }

  private void display(Accessible accessible)
  {
    if (accessible.isPublic())
      writer.print("public ");
    else
    if (accessible.isProtected())
      writer.print("protected ");
    else
    if (accessible.isPrivate())
      writer.print("private ");
  }

  private void display(AliasTable aliases)
  {
    if (aliases.hasEnclosing())
      display(aliases.getEnclosing());

    writer.addIndent();
    Iterator aliasesIt = aliases.iterateAliases();
    while (aliasesIt.hasNext())
    {
      Map.Entry entry = (Map.Entry)aliasesIt.next();
      TypeName alias = (TypeName)entry.getKey();
      TypeName aliased = (TypeName)entry.getValue();
      writer.println(alias + " -> " + aliased);
    }
    writer.decIndent();
  }

  private void display(Field field)
  {
    display((Accessible)field);
    if (field.isMutable()) writer.print("M"); else writer.print("X");
    if (field.isStatic()) writer.print("S"); else writer.print("X");
    if (field.isTransient()) writer.print("T"); else writer.print("X");
	writer.print((field.isConst() ? " const " : " ") + field.getTypeName());
	writer.print(" " + field.getName());
    writer.println();
  }

  private void display(Method method)
  {
    display((Accessible)method);

    if (method.getKind() == MethodKindEnum.castOperator)
    {
      writer.print("cast " + method.getResultType() + "() const");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.constructor)
    {
      writer.print("constructor(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.destructor)
    {
      writer.println("destructor()");
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.infix)
    {
      display(method.getResultType());
      writer.print(" infix(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.postfix)
    {
      display(method.getResultType());
      writer.print(" postfix(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.prefix)
    {
      display(method.getResultType());
      writer.print(" prefix " + method.getName() + "(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.property)
    {
      display(method.getResultType());
      writer.print(" property " + method.getName() + "(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.propertyArray)
    {
      display(method.getResultType());
      writer.print(" property []" + method.getName() + "(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
    if (method.getKind() == MethodKindEnum.method)
    {
      display(method.getResultType());
      writer.print(method.getName() + "(");
      displayArgs(method.iterateArguments());
      writer.print(")");
      displayThrowList(method.iterateThrowList());
      writer.println();
      writer.flush();
    }
    else
      throw new Error("Internal error : unrecognized method kind");
  }

  private void display(ResultType resultType)
  {
    if (resultType == null)
      writer.print("sub ");
    else
    {
      if (resultType.isConst())
        writer.print("const ");
      writer.print(resultType.getTypeName());
    }
  }

  private void display(Argument arg)
  {
    if (arg.isConst())
      writer.print("const ");
    writer.print(arg.getTypeName() + " " + arg.getName());
  }

  private void displayArgs(Iterator it)
  {
    while (it.hasNext())
    {
      display((Argument)it.next());
      if (it.hasNext())
        writer.print(", ");
    }
  }

  private void displayThrowList(Iterator it)
  {
    while (it.hasNext())
    {
      writer.print(((TypeName)it.next()).toString());
      if (it.hasNext())
        writer.print(", ");
    }
  }

  private void displayTypeNameList(Iterator it)
  {
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      writer.print(typeName);
      if (it.hasNext())
        writer.print(", ");
    }

    writer.println();
    writer.flush();
  }
}