/*
  $Id: TypeEnvironmentImpl.java,v 1.10 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.application;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.compiler.typechecker.Type;
import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.Argument;
import net.nosica.symbol.SymbolException;
import net.nosica.util.Pair;

import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Comparator;

public class TypeEnvironmentImpl implements TypeEnvironment
{
  private static TypeEnvironmentImpl theInstance = new TypeEnvironmentImpl();
  private TypeEnvironmentImpl() {}
  public static TypeEnvironmentImpl instance() { return theInstance; }
  public static void instanceReset() { theInstance = new TypeEnvironmentImpl(); }

  private Map types = new HashMap(); // Map<TypeName, Type>
  private Environment env = Environment.instance();

  public Type getType(TypeName typeName) throws SymbolException
  {
    if (types.containsKey(typeName))
      return (Type)types.get(typeName);

    throw new SymbolException("Type [" + typeName + "] does not exist");
  }

  public void addType(Type type) throws SymbolException
  {
    if (types.containsKey(type.getName()))
      throw new SymbolException("Type [" + type.getName() + "] already exists");
    types.put(type.getName(), type);
  }

  public boolean typeExists(TypeName typeName)
  {
    return types.containsKey(typeName);
  }

  public Iterator iterateTypeName() { return types.keySet().iterator(); }

  public boolean doesInherits(TypeName son, TypeName parent) throws SymbolException
  {
    if (doesExtends(son, parent))
      return true;

    return doesImplements(son, parent);
  }

  private boolean doesExtends(TypeName son, TypeName extend) throws SymbolException
  {
    // None inherits from anybody
    if (son.toString().equals("net.nosica.lang.None"))
      return true;

    Type sonType = getType(son);
    Iterator it = sonType.iterateExtends();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      if (typeName.equals(extend))
        return true;

      if (doesInherits(typeName, extend))
        return true;
    }

    return false;
  }

  private boolean doesImplements(TypeName son, TypeName implement) throws SymbolException
  {
    // None inherits from anybody
    if (son.toString().equals("net.nosica.lang.None"))
      return true;

    Type sonType = getType(son);
    Iterator it = sonType.iterateImplements();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      if (typeName.equals(implement))
        return true;

      if (doesInherits(typeName, implement))
        return true;
    }

    return false;
  }

  public SortedSet getParents(TypeName typeName) throws SymbolException, GenericityException
  {
    SortedSet result = new TreeSet(
                          new Comparator()
                          {
                            public int compare(Object obj1, Object obj2)
                            {
                              Pair pair1 = (Pair)obj1;
                              Pair pair2 = (Pair)obj2;

                              int degree1 = ((Integer)pair1.first).intValue();
                              int degree2 = ((Integer)pair2.first).intValue();

                              if (degree1 != degree2)
                              {
                                if (degree1 < degree2)
                                  return -1;
                                return 1;
                              }

                              TypeName typeName1 = (TypeName)pair1.second;
                              TypeName typeName2 = (TypeName)pair2.second;

                              return typeName1.toString().compareTo(typeName2.toString());
                            }
                          }
                       );

    getParentsRec(typeName, result, 0);

    return result;
  }

  private void getParentsRec(TypeName typeName, SortedSet set, int degree) throws SymbolException, GenericityException
  {
    Type type = env.getTL().getType(typeName);
    set.add(new Pair(new Integer(degree), typeName));

    Iterator it = type.iterateExtends();
    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();
      getParentsRec(extendsTypeName, set, degree + 1);
    }

    it = type.iterateImplements();
    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();
      getParentsRec(extendsTypeName, set, degree + 1);
    }
  }

  public TypeName getNearestParent(TypeName type1, TypeName type2) throws SymbolException, GenericityException
  {
    SortedSet type1Parents = env.getTE().getParents(type1);
    SortedSet type2Parents = env.getTE().getParents(type2);

    Iterator it = type1Parents.iterator();
    while (it.hasNext())
    {
      TypeName typeName1 = (TypeName)((Pair)it.next()).second;

      Iterator j = type2Parents.iterator();
      while (j.hasNext())
      {
        TypeName typeName2 = (TypeName)((Pair)j.next()).second;

        if (typeName1.equals(typeName2))
          return typeName1;
      }
    }

    return null;
  }

  public boolean doesImplements(Method m, Method signature, boolean compareName) throws SymbolException
  {
    if (m.isStatic() != signature.isStatic())
      return false;

    if (m.isConst() != signature.isConst())
      return false;

    if (compareName)
    {
      if (
           (m.getName() == null && signature.getName() != null) ||
           (m.getName() != null && signature.getName() == null)
         )
      {
        return false;
      }

      if (m.getName() != null &&
          signature.getName() != null &&
          !m.getName().equals(signature.getName())
         )
      {
        return false;
      }
    }

    if (m.getKind() != signature.getKind())
      return false;

    if (m.isSub() != signature.isSub())
      return false;

    if (!m.isSub())
    {
      if (m.getResultType().isConst() != signature.getResultType().isConst())
        return false;

      if (
          !m.getResultType().getTypeName().equals(signature.getResultType().getTypeName()) &&
          !doesInherits(m.getResultType().getTypeName(), signature.getResultType().getTypeName())
         )
        return false;
    }

    Iterator i = m.iterateArguments();
    Iterator j = signature.iterateArguments();

    while (i.hasNext() && j.hasNext())
    {
      Argument arg1 = (Argument)i.next();
      Argument arg2 = (Argument)j.next();

      if (arg1.isConst() != arg2.isConst())
        return false;
      if (
          !arg1.getTypeName().equals(arg2.getTypeName()) &&
          (
            !getType(arg1.getTypeName()).isPrimitive() ||
            !doesInherits(arg1.getTypeName(), arg2.getTypeName())
          )
         )
        return false;
    }

    return !i.hasNext() && !j.hasNext();
  }
}