/*
  $Id: AbstractIRVisitorRewriter.java,v 1.12 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.intermediate;

import net.nosica.compiler.application.Environment;

import net.nosica.framework.VisitorException;

import net.nosica.compiler.production.c.intermediate.*;
import net.nosica.compiler.typechecker.Type;

import net.nosica.parser.node.util.var.TypeName;

import java.util.Stack;
import java.util.ArrayList;
import java.util.Iterator;

public abstract class AbstractIRVisitorRewriter implements IRVisitorRewriter
{
  protected Environment env = Environment.instance();
  private Stack listStatement = new Stack(); // Stack<ArrayList<Statement> >

  protected Statement rewrittenStatement = null;

  private ArrayList pendingStatements = new ArrayList();
  private boolean removeThisStatement = false;

  /**
   * \brief insert a statement before currently analysed statement
   *
   * \param statement the statement to insert
   */
  public void insertBefore(Statement statement) throws VisitorException
  {
    try
    {
      ArrayList statements = (ArrayList)listStatement.peek();

      statements.add(statement);
    }
    catch (java.util.EmptyStackException e)
    {
      throw new VisitorException(e.getMessage());
    }
  }

  /**
   * \brief insert a statement after currently analysed statement
   *
   * \param statement the statement to add
   */
  public void insertAfter(Statement statement) throws VisitorException
  {
    pendingStatements.add(statement);
  }

  /**
   * \brief toggle an internal flag to remove currently analysed statement
   *
   * \warning this method is very dangerous, you should consider using another \
   * method before using this one. You should really understand what you're doing \
   * before using it.
   */
  public void removeThisStatement()
  {
    removeThisStatement = true;
  }

  public Statement getRewrittenStatement() throws VisitorException
  {
    return rewrittenStatement;
  }

  public boolean isPrimitive(TypeName typeName) throws VisitorException
  {
    Type type = env.getTE().getType(typeName);
    return type.isPrimitive();
  }

  public void visit(Call call) throws VisitorException
  {
    if (call.hasSubscript())
    {
      call.getSubscript().accept(this);
    }

    Iterator it = call.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
  }

  public void visit(MemberAccess memberAccess) throws VisitorException
  {
    if (memberAccess.hasSubscript())
    {
      memberAccess.getSubscript().accept(this);
    }
  }

  public void visit(Literal literal) throws VisitorException
  {
  }

  public void visit(Sequence sequence) throws VisitorException
  {
    sequence.getLeft().accept(this);
    sequence.getRight().accept(this);
  }

  public void visit(StatementExpression statementExpression) throws VisitorException
  {
    statementExpression.getExpression().accept(this);
  }

  public void visit(Variable variable) throws VisitorException
  {
  }

  public void visit(Move move) throws VisitorException
  {
    move.getVariable().accept(this);
    move.getRhs().accept(this);
  }

  public void visit(ExtendedSequence exSeq) throws VisitorException
  {
    exSeq.getStatement().accept(this);
    exSeq.getExpression().accept(this);
  }

  public void visit(Jump jump) throws VisitorException
  {
  }

  public void visit(LabelName labelName) throws VisitorException
  {
  }

  public void visit(ConditionalJump conditionalJump) throws VisitorException
  {
    conditionalJump.getCondition().accept(this);
  }

  public void visit(Cast castExp) throws VisitorException
  {
    castExp.getExpression().accept(this);
  }

  public void visit(InstanceOf instanceOfExp) throws VisitorException
  {
    instanceOfExp.getExpression().accept(this);
  }

  public void visit(RawAllocation rawAllocation) throws VisitorException
  {
  }

  public void visit(DecRef decRef) throws VisitorException
  {
  }

  public void visit(IncRef incRef) throws VisitorException
  {
  }

  public void visit(Init init) throws VisitorException
  {
    init.getVariable().accept(this);
    init.getRhs().accept(this);
  }

  public void visit(StatementList statementList) throws VisitorException
  {
    Iterator it = statementList.iterateStatementList();

    ArrayList statements = new ArrayList();

    listStatement.push(statements);

    while (it.hasNext())
    {
      Statement s = (Statement)it.next();
      s.accept(this);
      if (!removeThisStatement)
        statements.add(s);
      removeThisStatement = false;
      statements.addAll(pendingStatements);
      pendingStatements.clear();
    }

    statementList.setStatementList(statements);

    listStatement.pop();

    if (listStatement.isEmpty())
      rewrittenStatement = statementList;
  }

  public void visit(ArraySet arraySet) throws VisitorException
  {
    arraySet.getSubscript().accept(this);
    Iterator it = arraySet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
    arraySet.getRhs().accept(this);
  }

  public void visit(ArrayGet arrayGet) throws VisitorException
  {
    arrayGet.getSubscript().accept(this);
    Iterator it = arrayGet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
  }

  public void visit(CheckNull checkNull) throws VisitorException
  {
  }

  public void visit(ConstraintError constraintError) throws VisitorException
  {
  }

  public void visit(TryRegister tryRegister) throws VisitorException
  {
    if (tryRegister.hasFinallyStatement())
      tryRegister.getFinallyStatement().accept(this);
  }

  public void visit(TryUnregister tryUnregister) throws VisitorException
  {
  }

  public void visit(Throw throwStatement) throws VisitorException
  {
    throwStatement.getExpression().accept(this);
  }

  public void visit(VariableScopeLeave varScopeLeave) throws VisitorException
  {
  }

  public void visit(Return returnStatement) throws VisitorException
  {
    returnStatement.getResult().accept(this);
  }
}