/*
  $Id: ExSeqRemoverVisitor.java,v 1.11 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.intermediate;

import net.nosica.compiler.application.Environment;

import net.nosica.framework.VisitorException;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * \brief rewriter IR tree to get an easy to generate tree
 *
 * Nodes removed :
 * <ul>
 *   <li>ExtendedSequence
 *   <li>Statement
 * </ul>
 * These nodes are replaced by a StatementList<br>
 */
public class ExSeqRemoverVisitor implements IRVisitorRewriter
{
  protected ArrayList statementPart = new ArrayList(); // ArrayList<Statement>
  protected Expression expressionPart = null;
  protected Environment env = Environment.instance();

  public Statement getRewrittenStatement()
  {
    return new StatementListImpl(statementPart);
  }

  public void visit(Call call) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    if (call.hasSubscript())
    {
      call.getSubscript().accept(this);
      call.setSubscript(expressionPart);

      statements.addAll(statementPart);
    }

    ArrayList newArgs = new ArrayList();
    Iterator it = call.iterateArgs();

    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);

      newArgs.add(expressionPart);

      statements.addAll(statementPart);
    }

    call.setArgs(newArgs);

    expressionPart = call;
    statementPart = statements;
  }

  public void visit(ArrayGet arrayGet) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    // modify each arguments ...
    ArrayList args = new ArrayList();
    Iterator it = arrayGet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);

      statements.addAll(statementPart);
      args.add(expressionPart);
    }

    // now subscript
    arrayGet.getSubscript().accept(this);
    arrayGet.setSubscript(expressionPart);
    statements.addAll(statementPart);

    // return transformed ArraySet
    expressionPart = arrayGet;
    statementPart = statements;
  }

  public void visit(ArraySet arraySet) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    // modify each arguments ...
    ArrayList args = new ArrayList();
    Iterator it = arraySet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);

      statements.addAll(statementPart);
      args.add(expressionPart);
    }

    // now subscript
    arraySet.getSubscript().accept(this);
    arraySet.setSubscript(expressionPart);
    statements.addAll(statementPart);

    // now change rhs
    arraySet.getRhs().accept(this);
    arraySet.setRhs(expressionPart);
    statements.addAll(statementPart);

    // return transformed ArraySet
    expressionPart = null;

    statements.add(arraySet);

    statementPart = statements;
  }

  public void visit(MemberAccess memberAccess) throws VisitorException
  {
    if (memberAccess.hasSubscript())
    {
      memberAccess.getSubscript().accept(this);
      memberAccess.setSubscript(expressionPart);
    }
    else
      statementPart = new ArrayList();

    expressionPart = memberAccess;
  }

  public void visit(Literal literal) throws VisitorException
  {
    statementPart = new ArrayList();
    expressionPart = literal;
  }

  public void visit(Sequence sequence) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    sequence.getLeft().accept(this);
    statements.addAll(statementPart);

    sequence.getRight().accept(this);
    statements.addAll(statementPart);

    statementPart = statements;
    expressionPart = null;
  }

  public void visit(StatementExpression statementExpression) throws VisitorException
  {
    statementExpression.getExpression().accept(this);

    if (
        expressionPart instanceof Variable ||
        expressionPart instanceof MemberAccess ||
        expressionPart instanceof Literal
       )
    {
      expressionPart = null;

      return;
    }

    statementExpression.setExpression(expressionPart);

    expressionPart = null;
    statementPart.add(statementExpression);
  }

  public void visit(Variable variable) throws VisitorException
  {
    statementPart = new ArrayList();

    expressionPart = variable;
  }

  public void visit(Move move) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    move.getVariable().accept(this);
    move.setVariable((Variable)expressionPart);

    statements.addAll(statementPart);

    move.getRhs().accept(this);
    move.setRhs(expressionPart);

    statements.addAll(statementPart);

    statements.add(move);

    statementPart = statements;
    expressionPart = null;
  }

  public void visit(ExtendedSequence exSeq) throws VisitorException
  {
    ArrayList statements = new ArrayList();

  if (exSeq.getStatement() == null)
  {
    System.out.println("lknlknk");
  }

    exSeq.getStatement().accept(this);
    statements.addAll(statementPart);

    exSeq.getExpression().accept(this);
    statements.addAll(statementPart);

    statementPart = statements;
    // we let the expressionPart coming from expression.accept fall through
  }

  public void visit(Jump jump) throws VisitorException
  {
    statementPart = new ArrayList();
    statementPart.add(jump);
  }

  public void visit(LabelName labelName) throws VisitorException
  {
    statementPart = new ArrayList();
    statementPart.add(labelName);
  }

  public void visit(ConditionalJump conditionalJump) throws VisitorException
  {
    conditionalJump.getCondition().accept(this);
    conditionalJump.setCondition(expressionPart);

    statementPart.add(conditionalJump);
    expressionPart = null;
  }

  public void visit(Cast castExp) throws VisitorException
  {
    castExp.getExpression().accept(this);
    castExp.setExpression(expressionPart);

    expressionPart = castExp;
  }

  public void visit(InstanceOf instanceOfExp) throws VisitorException
  {
    instanceOfExp.getExpression().accept(this);
    instanceOfExp.setExpression(expressionPart);

    expressionPart = instanceOfExp;
  }

  public void visit(RawAllocation rawAllocation) throws VisitorException
  {
    expressionPart = rawAllocation;
    statementPart = new ArrayList();
  }

  public void visit(Init init) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    init.getVariable().accept(this);
    init.setVariable((Variable)expressionPart);

    statements.addAll(statementPart);

    init.getRhs().accept(this);
    init.setRhs(expressionPart);

    statements.addAll(statementPart);

    statements.add(init);

    statementPart = statements;
    expressionPart = null;
  }

  public void visit(StatementList statementList) throws VisitorException
  {
    ArrayList statements = new ArrayList();

    Iterator it = statementList.iterateStatementList();
    while (it.hasNext())
    {
      Statement statement = (Statement)it.next();
      statement.accept(this);

      statements.addAll(statementPart);
    }

    statementPart = statements;
    expressionPart = null;
  }

  public void visit(ConstraintError constraintError) throws VisitorException
  {
  }

  public void visit(TryRegister tryRegister) throws VisitorException
  {
    if (tryRegister.hasFinallyStatement())
    {
      statementPart = new ArrayList();
      expressionPart = null;
      tryRegister.getFinallyStatement().accept(this);
      tryRegister.setFinallyStatement(new StatementListImpl(statementPart));
    }

    expressionPart = null;
    statementPart = new ArrayList();
    statementPart.add(tryRegister);
  }

  public void visit(TryUnregister tryUnregister) throws VisitorException
  {
    expressionPart = null;
    statementPart = new ArrayList();
    statementPart.add(tryUnregister);
  }

  public void visit(Throw throwStatement) throws VisitorException
  {
    throwStatement.getExpression().accept(this);
    throwStatement.setExpression(expressionPart);

    expressionPart = null;
    statementPart.add(throwStatement);
  }

  public void visit(VariableScopeLeave varScopeLeave) throws VisitorException
  {
    expressionPart = null;
    statementPart = new ArrayList();
    statementPart.add(varScopeLeave);
  }

  public void visit(Return returnStatement) throws VisitorException
  {
    returnStatement.getResult().accept(this);
    returnStatement.setResult(expressionPart);

    expressionPart = null;
    statementPart.add(returnStatement);
  }
}
