/*
  $Id: CNameConverter.java,v 1.15 2004/04/01 21:27:55 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.production.c;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.Field;
import net.nosica.compiler.typechecker.Method;
import net.nosica.compiler.typechecker.Argument;
import net.nosica.compiler.typechecker.ResultType;
import net.nosica.framework.VisitorException;

/**
 * \brief interface to convert nosica typeName, function name, ... to c counterpart
 */
public interface CNameConverter
{
  /**
   * \brief convert method name to a c method name
   *
   * \param method method to convert
   * \return c counterpart
   */
  String cMethodName(Method method);

  /**
   * \brief convert polymorphic method name to a c polymorphic method name
   *
   * \param method method to convert
   * \param counter the polymorphic counter identifier
   * \return c counterpart
   */
  String cPolymorphicMethodName(Method method, int counter);

  /**
   * \brief convert function signature to a c function signature
   *
   * \param method method to convert
   * \return c counterpart
   */
  String cMethodSignature(Method method) throws VisitorException;

  /**
   * \brief convert method signature to a c method signature (only
   *        the C code line, with a provided/custom method name)
   *
   * \param method method to convert
   * \param methodName C name of the method in the signature
   * return c counterpart
   */
  String cMethodSignatureLine(Method method, String methodName) throws VisitorException;

  /**
   * \brief convert an array get to a c function signature
   *
   * Provide a comment of which method is instantiated + c signature line
   *
   * \param typeName type of array
   * \param forwardDeclaration is it a forward declaration ?
   * \return c signature
   */
  String cArrayGetSignature(TypeName typeName, boolean forwardDeclaration) throws VisitorException;

  /**
   * \brief convert an array get to a c function signature
   *
   * Only c signature
   *
   * \param typeName type of array
   * \param forwardDeclaration is it a forward declaration ?
   * \return c signature
   */
  String cArrayGetSignatureLine(TypeName typeName, boolean forwardDeclaration) throws VisitorException;

  /**
   * \brief convert an array get method call to a c function name
   *
   * \param typeName type of array
   * \return c signature
   */
  String cArrayGetName(TypeName typeName);

  /**
   * \brief convert an array set to a c function signature
   *
   * Provide a comment of which method is instantiated + c signature line
   *
   * \param typeName type of array
   * \param forwardDeclaration is it a forward declaration ?
   * \return c signature
   */
  String cArraySetSignature(TypeName typeName, boolean forwardDeclaration) throws VisitorException;

  /**
   * \brief convert an array set to a c function signature
   *
   * Only c signature
   *
   * \param typeName type of array
   * \param forwardDeclaration is it a forward declaration ?
   * \return c signature
   */
  String cArraySetSignatureLine(TypeName typeName, boolean forwardDeclaration) throws VisitorException;

  /**
   * \brief convert an array set method call to a c function name
   *
   * \param typeName type of array
   * \param c signature
   */
  String cArraySetName(TypeName typeName);

   /**
   * \brief convert typename to a c typename
   *
   * \remarks do take into account the nature of the typeName (interface, ...)
   * \remarks do not take into account argument modifier (pointer, by value)
   *
   * \param typeName to convert
   * \return c counterpart
   */
  String cTypeName(TypeName typeName) throws VisitorException;

  /**
   * \brief convert a typeName to a c typeName
   *
   * \remarks do not take into account the nature of the typeName (primitive, interface, ...)
   * \remarks do not take into account argument modifier (pointer, by value)
   *
   * \param typeName typeName to convert
   * \return c counterpart
   */
  String cRawTypeName(TypeName typeName);

  /**
   * \brief convert member name to a c member name
   *
   * \param field field to convert
   * \return c counterpart
   */
  String cMemberName(Field field);

  /**
   * \brief convert member typeName to a c member name
   *
   * \param field field to convert
   * \return c counterpart
   */
  String cMemberTypeName(Field field) throws VisitorException;

  /**
   * \brief convert a typeName to a full path
   *
   * \param typeName typeName to convert
   * \return c filename without extension
   */
  String cRadixFileName(TypeName typeName);

  /**
   * \brief convert a typeName to a full path
   *
   * \param typeName typeName to convert
   * \return c counterpart
   */
  String cIncludeTypeName(TypeName typeName);

  /**
   * \brief convert a source code character to a nosica UTF32 char literal
   *
   * \param ch : the character value in the source code
   * \return c counterpart (suitable as a word32 constant)
   */
  String cCharValue(char ch);

  /**
   * \brief indicates if the expression will be translated into a pointer
   *
   * \param exp the expression to analyse
   * \return true if exp will be translated into a pointer
   */
  boolean isExpTranslatedIntoPointer(Expression exp) throws VisitorException;

  /**
   * \brief indicates if the argument is translated into a pointer
   *
   * \remarks (the 'this' pointer is not handled here !)
   * \param arg the argument to analyse
   * \return true if argument will be translated into a pointer
   */
  boolean isArgTranslatedIntoPointer(Argument arg) throws VisitorException;

  /**
   * \brief indicates if the result type is translated into a pointer
   *
   * \param resultType the ResultType to analyse
   * \return true if resultType will be translated into a pointer
   */
  boolean isResultTypeTranslatedIntoPointer(ResultType resultType) throws VisitorException;

  /**
   * \brief return the c name of a given cleanUpPoint
   * \param cleanUpPoint the number of the cleanUpPoint
   * \return a string suitable for c label declaration
   */
  String cCleanUpPointLabelName(int cleanUpPoint);

  /**
   * \brief return the exception variable name
   *
   * \return exception variable name
   */
  String getExceptionVariableName();

  /**
   * \brief return c source file name
   *
   * \param main name to convert
   */
  String cSourceFileName(TypeName file);

  /**
   * \brief return c obj file name
   *
   * \param main name to convert
   */
  String cObjFileName(TypeName file);

  /**
   * \brief return c native library name
   *
   * \param packageName name of package to convert
   */
  String cNativeLibraryName(TypeName packageName);

  /**
   * \brief returns a string from a typeName suitable for protecting a c header file
   *
   * \param typeName the typeName from which to generate
   * \return c header macro define name
   */
  String cMacroDefineName(TypeName typeName);
}
