/*
  $Id: VariableProducerVisitor.java,v 1.18 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.production.c;

import net.nosica.framework.VisitorException;

import net.nosica.compiler.production.c.intermediate.CIRVisitor;

import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.Call;
import net.nosica.compiler.intermediate.MemberAccess;
import net.nosica.compiler.intermediate.Sequence;
import net.nosica.compiler.intermediate.StatementExpression;
import net.nosica.compiler.intermediate.Literal;
import net.nosica.compiler.intermediate.Variable;
import net.nosica.compiler.intermediate.Move;
import net.nosica.compiler.intermediate.ExtendedSequence;
import net.nosica.compiler.intermediate.Jump;
import net.nosica.compiler.intermediate.LabelName;
import net.nosica.compiler.intermediate.ConditionalJump;
import net.nosica.compiler.intermediate.Cast;
import net.nosica.compiler.intermediate.InstanceOf;
import net.nosica.compiler.intermediate.RawAllocation;
import net.nosica.compiler.intermediate.Init;
import net.nosica.compiler.intermediate.Statement;
import net.nosica.compiler.intermediate.StatementList;
import net.nosica.compiler.intermediate.ArrayGet;
import net.nosica.compiler.intermediate.ArraySet;
import net.nosica.compiler.intermediate.ConstraintError;
import net.nosica.compiler.intermediate.TryRegister;
import net.nosica.compiler.intermediate.TryUnregister;
import net.nosica.compiler.intermediate.Throw;
import net.nosica.compiler.intermediate.VariableScopeLeave;
import net.nosica.compiler.intermediate.Return;
import net.nosica.compiler.production.c.intermediate.CheckNull;
import net.nosica.compiler.production.c.intermediate.DecRef;
import net.nosica.compiler.production.c.intermediate.IncRef;

import net.nosica.compiler.application.Environment;

import net.nosica.util.IndentedPrintWriter;
import net.nosica.util.Pair;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;

public class VariableProducerVisitor implements CIRVisitor
{
  private IndentedPrintWriter printer;      // the printer to use
  private CNameConverter cConverter;        // the c converter
  private Environment env = Environment.instance();

  // variable handling
  private Set liveVariable = new HashSet(); // Set<Variable> list of live variable
                                            // to avoid printing several time the same variable
  private Set variableName = new HashSet(); // Set<String> list of unique variable name
  private int variableUniqueCounter = 0;    // a counter to safely rename variable
                                            // variable name will be unique in this method
  private ArrayList args;                      // ArrayList<String> arguments of method
                                            // so that we don't generate a variable
                                            // if it's in fact an argument

  /**
   * \brief construct a variable and label producer visitor
   *
   * This visitor has 2 goals :
   * - rename all variables having the same name so that the c compiler will accept them
   * - emit those variables at the beginning of the method (declare them in C style)
   */
  public VariableProducerVisitor(IndentedPrintWriter printer, CNameConverter cConverter, ArrayList args)
  {
    this.printer = printer;
    this.cConverter = cConverter;
    this.args = args;
  }

  public void visit(Call call) throws VisitorException
  {
    if (call.hasSubscript())
      call.getSubscript().accept(this);

    Iterator it = call.iterateArgs();
    while (it.hasNext())
      ((Expression)it.next()).accept(this);
  }

  public void visit(ArrayGet arrayGet) throws VisitorException
  {
    arrayGet.getSubscript().accept(this);
    Iterator it = arrayGet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
  }

  public void visit(ArraySet arraySet) throws VisitorException
  {
    arraySet.getSubscript().accept(this);
    Iterator it = arraySet.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
    arraySet.getRhs().accept(this);
  }

  public void visit(MemberAccess memberAccess) throws VisitorException
  {
    if (memberAccess.hasSubscript())
      memberAccess.getSubscript().accept(this);
  }

  public void visit(Literal literal) throws VisitorException
  {
    // nothing to do
  }

  public void visit(Sequence sequence) throws VisitorException
  {
    throw new UnsupportedOperationException();
  }

  public void visit(StatementExpression statementExpression) throws VisitorException
  {
    statementExpression.getExpression().accept(this);
  }

  public void visit(Variable variable) throws VisitorException
  {
    if (liveVariable.contains(variable))
      return;

    liveVariable.add(variable);
    rename(variable);

    if (!args.contains(variable.getName()))
      emitVariable(variable);
  }

  public void visit(Move move) throws VisitorException
  {
    move.getVariable().accept(this);
    move.getRhs().accept(this);
  }

  public void visit(ExtendedSequence exSeq) throws VisitorException
  {
    throw new UnsupportedOperationException();
  }

  public void visit(Jump jump) throws VisitorException
  {
  }

  public void visit(LabelName labelName) throws VisitorException
  {
  }

  public void visit(ConditionalJump conditionalJump) throws VisitorException
  {
    conditionalJump.getCondition().accept(this);
  }

  private void emitVariable(Variable variable) throws VisitorException
  {
    printer.print(cVariable(variable) + " " + variable.getName());
    if (!env.getTE().getType(variable.getTypeName()).isPrimitive())
      printer.print(" = null");
    printer.println(";");
  }

  public void visit(Cast castExp) throws VisitorException
  {
    castExp.getExpression().accept(this);
  }

  public void visit(InstanceOf instanceOfExp) throws VisitorException
  {
    instanceOfExp.getExpression().accept(this);
  }

  public void visit(RawAllocation rawAllocation) throws VisitorException
  {
  }

  public void visit(IncRef incRef) throws VisitorException
  {
    incRef.getVariable().accept(this);
  }

  public void visit(DecRef decRef) throws VisitorException
  {
  /* removed so that ___retval goes through.
    if (
        (!(decRef.getVariable() instanceof MemberAccess)) &&
        !liveVariable.contains(decRef.getVariable())
       )
      Environment.instance().terminate("Internal error : live variable should contain variable " + decRef.getVariable().getName());
  */
  }

  public void visit(Init init) throws VisitorException
  {
    visit((Move)init);
  }

  public void visit(StatementList statementList) throws VisitorException
  {
    Iterator it = statementList.iterateStatementList();
    while (it.hasNext())
    {
      Statement statement = (Statement)it.next();
      statement.accept(this);
    }
  }

  private void rename(Variable variable)
  {
    if (variable.getName().equals(""))
      variable.setName("___TEMP_");

    if (variableName.contains(variable.getName()))
      variable.setName(variable.getName() + ++variableUniqueCounter);

    variableName.add(variable.getName());
  }

  private String cVariable(Variable variable) throws VisitorException
  {
    StringBuffer result = new StringBuffer();
    result.append(cConverter.cTypeName(variable.getTypeName())).append(" ");
    if (cConverter.isExpTranslatedIntoPointer(variable))
      result.append("*");

    return result.toString();
  }

  public void visit(CheckNull checkNull) throws VisitorException
  {
    checkNull.getExpression().accept(this);
  }
  public void visit(ConstraintError constraintError) throws VisitorException
  {
  }
  public void visit(TryRegister tryRegister) throws VisitorException
  {
    if (tryRegister.hasFinallyStatement())
      tryRegister.getFinallyStatement().accept(this);

    Iterator it = tryRegister.getExceptionMap().values().iterator();
    while (it.hasNext())
    {
      Pair pair = (Pair)it.next();
      Variable var = (Variable)pair.second;

      if (liveVariable.contains(var))
        return;

      var.accept(this);
    }
  }
  public void visit(TryUnregister tryUnregister) throws VisitorException
  {
  }
  public void visit(Throw throwStatement) throws VisitorException
  {
    throwStatement.getExpression().accept(this);
  }
  public void visit(VariableScopeLeave varScopeLeave) throws VisitorException
  {
  }
  public void visit(Return returnStatement) throws VisitorException
  {
    returnStatement.getResult().accept(this);
  }
}