/*
  $Id: InsertRefCountVisitor.java,v 1.5 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.production.c.intermediate;

import net.nosica.compiler.intermediate.AbstractIRVisitorRewriter;
import net.nosica.compiler.intermediate.Function;
import net.nosica.compiler.intermediate.ArrayGetImpl;
import net.nosica.compiler.intermediate.ArraySet;
import net.nosica.compiler.intermediate.Call;
import net.nosica.compiler.intermediate.Cast;
import net.nosica.compiler.intermediate.Expression;
import net.nosica.compiler.intermediate.IRVisitorRewriter;
import net.nosica.compiler.intermediate.Init;
import net.nosica.compiler.intermediate.InitImpl;
import net.nosica.compiler.intermediate.InstanceOf;
import net.nosica.compiler.intermediate.MemberAccess;
import net.nosica.compiler.intermediate.Move;
import net.nosica.compiler.intermediate.Return;
import net.nosica.compiler.intermediate.Statement;
import net.nosica.compiler.intermediate.StatementExpression;
import net.nosica.compiler.intermediate.StatementList;
import net.nosica.compiler.intermediate.Variable;
import net.nosica.compiler.intermediate.VariableImpl;
import net.nosica.compiler.intermediate.VariableScopeLeave;
import net.nosica.compiler.intermediate.VariableScopeLeaveImpl;

import net.nosica.framework.VisitorException;

import java.util.Iterator;
import java.util.ArrayList;

/**
 * \brief insert IncRef / DecRef nodes to deal with reference counting
 *
 * Basically, it will :<br>
 * <ul>
 *   <li> Case 1 <br>
 *        <ul>
 *          <li>insert DECREF before each MOVE.
 *          <li>insert INCREF after each MOVE and INIT, except if rhs is a CALL.
 *          <li>insert INCREF/DECREF for ArraySet for the same reasons as MOVE
 *        </ul>
 *   <li> Case 2 <br>
 *        <ul>
 *          <li>insert INCREF for all reference arguments at the beginning of the method.
 *          <li>insert DECREF for all reference arguments before leaving method.
 *        </ul>
 *   <li> Case 3 <br>
 *        <ul>
 *          <li>add a temporary for each method that returns a reference
 *        </li>
 *   <li> Case 4 <br>
 *        <ul>
 *          <li>insert a DECREF for each method going out of scope
 *        </ul>
 *   <li> Case 5 <br>
 *        <ul>
 *          <li>insert a temporary for each expression of the type : var = var.field
 *        </ul>
 * </ul>
 * <br>
 * Rationale :<br>
 *   an INIT is a variable initialization : it means variable has never been used before,<br>
 *   therefore variable does not contains a valid pointer (and should contain null)<br>
 *   ==> only an INCREF has to be inserted after assignment<br>
 *   <br>
 *   a MOVE is a variable assignment. There is no assumption done about the content<br>
 *   of the variable before the MOVE : both an INCREF and a DECREF has to be done<br>
 *   <br>
 *   When a reference variable is returned from a CALL, reference count is equals to 1, not 0.<br>
 *   Therefore, we cannot insert an INCREF because in this case reference counting will never go<br>
 *   back to 0, and will never be freed.<br>
 *   <br>
 *   Each arguments must remain live till returning of the method.<br>
 *   If we weren't INCREF/DECREF them, the method would be able to delete them too early
 *   <br>
 *   Each method that returns a reference needs a temporary so that if the result of the
 *   method is not used it is properly destroyed ...
 *   <br>
 *   In case of a var = var.field pattern, the translation should be :
 *   <code>
 *   DECREF(var)
 *   var = var.field
 *   INCREF(var)
 *   </code>
 *   if var's refcount is 1, then DECREF(var) free the variable. Hence var = var.field<br>
 *   is a NullPointerException : we must add a temporary to do the following :
 *   <code>
 *   temp = var.field
 *   INCREF(temp)
 *   DECREF(var)
 *   var = temp
 *   INCREF(var)
 *   DECREF(temp)
 *   </code>
 */
public class InsertRefCountVisitor extends AbstractIRVisitorRewriter implements CIRVisitor, IRVisitorRewriter
{
  private Function function;

  public InsertRefCountVisitor(Function function)
  {
    this.function = function;
  }

  // CASE 2 ********************************************************************
  public Statement getRewrittenStatement() throws VisitorException
  {
    // before returning, for each reference argument,
    // add INCREF at the beginning
    // add DECREF at the end
    if (rewrittenStatement instanceof StatementList)
    {
      StatementList statementList = (StatementList)rewrittenStatement;
      ArrayList statements = statementList.getStatementList();

      Iterator it = function.iterateArgs();
      while (it.hasNext())
      {
        Variable arg = (Variable)it.next();
        if (!isPrimitive(arg.getTypeName()))
        {
          statements.add(0, new IncRefImpl(arg));
          statements.add(new DecRefImpl(arg));
        }
      }

      return rewrittenStatement;
    }

    env.terminate("Internal error : should be StatementList");
    return null;
  }

  public void visit(Return returnStatement) throws VisitorException
  {
    // add DECREF for each reference argument
    Iterator it = function.iterateArgs();
    while (it.hasNext())
    {
      Variable arg = (Variable)it.next();
      if (!isPrimitive(arg.getTypeName()))
        insertBefore(new DecRefImpl(arg));
    }
    returnStatement.getResult().accept(this);
  }

  // CASE 1 ********************************************************************
  public void visit(Move move) throws VisitorException
  {
    boolean variableIsPrimitive = isPrimitive(move.getVariable().getTypeName());
    Variable var = null; // a temporary variable for case 5

    // CASE 5
    if (!variableIsPrimitive)
    {
      if (move.getVariable() == move.getRhs())
      {
        // var = var
        removeThisStatement();
      }
      else
      {
        if (commonLeftExpression(move.getVariable(), move.getRhs()))
        {
          // insert a temporary
          var = new VariableImpl(
            move.getVariable().getTypeName(),
            env.getNG().generateTemporaryName()
          );
          insertBefore(new InitImpl(var, move.getRhs()));
          move.getRhs().accept(this);
          move.setRhs(var);
          insertBefore(new IncRefImpl(var));
        }
      }
    }

    if (!variableIsPrimitive)
    {
      insertBefore(
        new DecRefImpl(move.getVariable())
      );
    }

    move.getRhs().accept(this);

    if (!variableIsPrimitive && !(move.getRhs() instanceof Call))
    {
      insertAfter(
        new IncRefImpl(move.getVariable())
      );
    }

    // case 5
    if (var != null)
    {
      insertAfter(new DecRefImpl(var));
      insertAfter(new VariableScopeLeaveImpl(var));
    }
  }

  /**
   * \brief tries to look if two expressions share the same subscript
   *
   * In an expression like "left = right", analyse each expression to see if
   * right expression is in fact something like MemberAccess(left, field)
   *
   * \param left left expression
   * \param right right expression
   * \return true if left and right use same subscript
   */
  private boolean commonLeftExpression(Expression left, Expression right)
  {
    ArrayList leftSequencialExpression = new ArrayList();
    extractSequencialExpression(left, leftSequencialExpression);
    ArrayList rightSequencialExpression = new ArrayList();
    extractSequencialExpression(right, rightSequencialExpression);

    Iterator leftIt = leftSequencialExpression.iterator();
    Iterator rightIt = rightSequencialExpression.iterator();

    if (!leftIt.hasNext())
      return false;
    if (!rightIt.hasNext())
      return false;

    Variable leftVar = (Variable)leftIt.next();
    Variable rightVar = (Variable)rightIt.next();

    // incompatible expression
    if (leftVar != rightVar)
    {
      return false;
    }

    // now, it can only be MemberAccess
    while (leftIt.hasNext() && rightIt.hasNext())
    {
      MemberAccess leftMA = (MemberAccess)leftIt.next();
      MemberAccess rightMA = (MemberAccess)rightIt.next();

      if (leftMA.getField() != rightMA.getField())
        return false;
    }

    if (leftIt.hasNext())
      return false;

    if (!rightIt.hasNext())
      return false;

    return true;
  }

  private void extractSequencialExpression(Expression exp, ArrayList result)
  {
    if (exp instanceof MemberAccess)
    {
      MemberAccess m = (MemberAccess)exp;

      result.add(0, exp);
      extractSequencialExpression(m.getSubscript(), result);
    }
    else
    if (exp instanceof Variable)
      result.add(0, exp);
  }

  public void visit(Init init) throws VisitorException
  {
    init.getRhs().accept(this);

    if (!isPrimitive(init.getRhs().getTypeName()) && !(init.getRhs() instanceof Call))
    {
      insertAfter(
        new IncRefImpl(init.getVariable())
      );
    }
  }

  public void visit(ArraySet arraySet) throws VisitorException
  {
    boolean rhsIsPrimitive = isPrimitive(arraySet.getRhs().getTypeName());
    if (!rhsIsPrimitive)
    {
      // insert DECREF
      ArrayList args1 = new ArrayList(arraySet.getArgs());
      insertBefore(
        new DecRefImpl(new ArrayGetImpl(arraySet.getSubscript(), args1))
      );
    }

    // analyse
    arraySet.getSubscript().accept(this);
    Iterator it = arraySet.getArgs().iterator();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);
    }
    arraySet.getRhs().accept(this);

    // insert INCREF
    if (!rhsIsPrimitive && !(arraySet.getRhs() instanceof Call))
    {
      ArrayList args2 = new ArrayList(arraySet.getArgs());
      insertAfter(
        new IncRefImpl(new ArrayGetImpl(arraySet.getSubscript(), args2))
      );
    }
  }

  // CASE 4 ********************************************************************
  public void visit(VariableScopeLeave varScope) throws VisitorException
  {
    if (isPrimitive(varScope.getVariable().getTypeName()))
      return;

    insertBefore(
      new DecRefImpl(varScope.getVariable())
    );
  }

  // CASE 3 ********************************************************************
  public void visit(StatementExpression statementExpression) throws VisitorException
  {
    if (statementExpression.getExpression() instanceof Call)
    {
      Call call = (Call)statementExpression.getExpression();
      if (!call.getMethod().isSub() && !isPrimitive(call.getMethod().getResultType().getTypeName()))
      {
        // insert a temporary
        Variable var = new VariableImpl(
          call.getMethod().getResultType().getTypeName(),
          env.getNG().generateTemporaryName()
        );
        insertBefore(new InitImpl(var, call));
        call.accept(this);
        removeThisStatement();
        insertAfter(new DecRefImpl(var));
        insertAfter(new VariableScopeLeaveImpl(var));

        return;
      }
    }

    statementExpression.getExpression().accept(this);
  }

  public void visit(Cast cast) throws VisitorException
  {
    if (cast.getExpression() instanceof Call)
    {
      Call call = (Call)cast.getExpression();
      if (!call.getMethod().isSub() && !isPrimitive(call.getMethod().getResultType().getTypeName()))
      {
        // insert a temporary
        Variable var = new VariableImpl(
          call.getMethod().getResultType().getTypeName(),
          env.getNG().generateTemporaryName()
        );
        insertBefore(new InitImpl(var, call));
        call.accept(this);
        cast.setExpression(var);
        insertAfter(new DecRefImpl(var));
        insertAfter(new VariableScopeLeaveImpl(var));

        return;
      }
    }

    cast.getExpression().accept(this);
  }

  public void visit(InstanceOf instanceOf) throws VisitorException
  {
    if (instanceOf.getExpression() instanceof Call)
    {
      Call call = (Call)instanceOf.getExpression();
      if (!call.getMethod().isSub() && !isPrimitive(call.getMethod().getResultType().getTypeName()))
      {
        // insert a temporary
        Variable var = new VariableImpl(
          call.getMethod().getResultType().getTypeName(),
          env.getNG().generateTemporaryName()
        );
        insertBefore(new InitImpl(var, call));
        call.accept(this);
        instanceOf.setExpression(var);
        insertAfter(new DecRefImpl(var));
        insertAfter(new VariableScopeLeaveImpl(var));

        return;
      }
    }

    instanceOf.getExpression().accept(this);
  }

  public void visit(Call call) throws VisitorException
  {
    // subscript
    if (call.hasSubscript())
    {
      Expression exp = call.getSubscript();
      exp.accept(this);

      if (exp instanceof Call)
      {
        Call subscriptCall = (Call)exp;
        if (!subscriptCall.getMethod().isSub() && !isPrimitive(subscriptCall.getMethod().getResultType().getTypeName()))
        {
          // insert a temporary
          Variable var = new VariableImpl(
            subscriptCall.getMethod().getResultType().getTypeName(),
            env.getNG().generateTemporaryName()
          );
          insertBefore(new InitImpl(var, subscriptCall));
          subscriptCall.setSubscript(var);
          insertAfter(new DecRefImpl(var));
          insertAfter(new VariableScopeLeaveImpl(var));
        }
      }
    }

    // arguments
    ArrayList newArgs = new ArrayList();
    Iterator it = call.iterateArgs();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      exp.accept(this);

      if (exp instanceof Call)
      {
        Call argCall = (Call)exp;
        if (!argCall.getMethod().isSub() && !isPrimitive(argCall.getMethod().getResultType().getTypeName()))
        {
          // insert a temporary
          Variable var = new VariableImpl(
            argCall.getMethod().getResultType().getTypeName(),
            env.getNG().generateTemporaryName()
          );
          insertBefore(new InitImpl(var, argCall));
          newArgs.add(var);
          insertAfter(new DecRefImpl(var));
          insertAfter(new VariableScopeLeaveImpl(var));
        }
        else
        {
          newArgs.add(argCall);
        }
      }
      else
      {
        newArgs.add(exp);
      }
    }

    call.setArgs(newArgs);
  }
}