/*
  $Id: CompilerSymbolCallBackImpl.java,v 1.6 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.symbol;

import net.nosica.application.MessageLevel;
import net.nosica.application.MessageService;

import net.nosica.framework.Localizable;

import net.nosica.parser.node.Node;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameList;
import net.nosica.parser.node.util.var.FormalParameters;
import net.nosica.parser.node.util.var.FormalParameter;

import net.nosica.symbol.Accessible;
import net.nosica.symbol.SymbolCallBack;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.Field;
import net.nosica.symbol.MethodKindEnum;

import java.util.Stack;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;

/**
 * \brief implementation of SymbolCallBack for the Nosica compiler
 *
 * This implementation checks there is no duplicated alias, fields, methods, exceptions
 * That the declarations are authorized by the Nosica language ...
 */
public class CompilerSymbolCallBackImpl implements SymbolCallBack
{
  static private Set forbiddenOperators; // list of user forbidden operators
  static private Set prefixOperators;    // list of prefix operators
  static
  {
    forbiddenOperators = new HashSet();
    forbiddenOperators.add("=");
    forbiddenOperators.add(":=");
    forbiddenOperators.add("::=");

    prefixOperators = new HashSet();
    prefixOperators.add("++");
    prefixOperators.add("--");
    prefixOperators.add("!");
    prefixOperators.add("-");
  }

  private MessageService messageService;           // message service implementation to use for reporting
  private TypeName waitedTypeName;                 // the name to check
  private Stack savedWaitedTypeName = new Stack(); // Stack<TypeName>

  public CompilerSymbolCallBackImpl(MessageService messageService, TypeName waitedTypeName)
  {
    this.messageService = messageService;
    this.waitedTypeName = waitedTypeName;
  }

  public void onPackageDeclaration(TypeName packageName, Localizable localizable)
  {
    // sanity check
    if (!packageName.equals(waitedTypeName.getFirstComponents()))
    {
      messageService.report(MessageLevel.ERR, "Symbol001", waitedTypeName.getFirstComponents().toString(), localizable);
    }
  }

  public void onImportDeclaration(TypeName importName, Localizable localizable)
  {
    if (importName.isArray())
    {
      messageService.report(MessageLevel.ERR, "Symbol034", importName.toString(), localizable);
    }

    if (importName.isGeneric())
    {
      messageService.report(MessageLevel.ERR, "Symbol035", importName.toString(), localizable);
    }
  }

  public void onWrongImportDeclaration(TypeName importName, Localizable localizable)
  {
    messageService.report(MessageLevel.ERR, "Symbol041", importName.toString(), localizable);
  }

  public void onDuplicateAlias(String alias, Localizable localizable)
  {
    messageService.report(MessageLevel.ERR, "Symbol012", alias, localizable);
  }

  public void onClassDeclaration(GenericType type)
  {
    String strClassName = type.getName().getLastComponent().getPackageElement();
    // sanity check
    if (!strClassName.equals(waitedTypeName.getLastComponent().getPackageElement()))
    {
      String[] params = new String[2];
      params[0] = waitedTypeName.getLastComponent().getPackageElement();
      params[1] = strClassName;
      messageService.report(MessageLevel.ERR, "Symbol002", params, type.getNode());
    }

    sanityCheck(type);
  }

  public void onBeginNestedClassDeclaration(TypeName nestedTypeName)
  {
    savedWaitedTypeName.push(waitedTypeName);
    waitedTypeName = nestedTypeName;
  }

  public void onEndNestedClassDeclaration()
  {
    waitedTypeName = (TypeName)savedWaitedTypeName.pop();
  }

  public void onInterfaceDeclaration(GenericType type)
  {
    String strInterfaceName = type.getName().getLastComponent().getPackageElement();

    // sanity check
    if (!strInterfaceName.equals(waitedTypeName.getLastComponent().getPackageElement()))
    {
      messageService.report(MessageLevel.ERR, "Symbol003", type.getNode());
    }

    sanityCheck(type);
  }

  public void onBeginNestedInterfaceDeclaration(TypeName nestedTypeName)
  {
    savedWaitedTypeName.push(waitedTypeName);
    waitedTypeName = nestedTypeName;
  }

  public void onEndNestedInterfaceDeclaration()
  {
    waitedTypeName = (TypeName)savedWaitedTypeName.pop();
  }

  public void onFieldDeclaration(Field field, GenericType enclosing)
  {
    Iterator j = enclosing.iterateFields();
    while (j.hasNext())
    {
      Field field2 = (Field)j.next();
      if (field2.getName().equals(field.getName()))
      {
        messageService.report(MessageLevel.ERR, "Symbol014", field.getName(), enclosing.getNode());
        return;
      }
    }

    sanityCheck(field, enclosing);
  }

  public void onCheckFormalParameters(FormalParameters formalParameters)
  {
    Set paramNames = new HashSet(); // Set<String>

    Iterator it = formalParameters.iterateParameters();

    while (it.hasNext())
    {
      FormalParameter formalParam = (FormalParameter)it.next();

      if (formalParam.getVariable().hasInitializer())
      {
        messageService.report(MessageLevel.ERR, "Symbol015", formalParam.getVariable().getId(), formalParam);
      }
      if (!paramNames.add(formalParam.getVariable().getId()))
      {
        messageService.report(MessageLevel.ERR, "Symbol016", formalParam.getVariable().getId(), formalParam);
      }
    }
  }

  public void onCheckThrowList(TypeNameList typeNameList)
  {
    Set result = new HashSet();

    if (typeNameList == null)
      return;

    Iterator it = typeNameList.iterateTypeNames();

    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();

      if (!result.add(typeName))
      {
        messageService.report(MessageLevel.ERR, "Symbol017", typeName.toString(), typeName);
      }
    }
  }

  public void onConstructorDeclaration(GenericMethod method, GenericType enclosing)
  {
    sanityCheck(method, enclosing);
  }

  public void onDestructorDeclaration(GenericMethod method, GenericType enclosing)
  {
    if (enclosing.getDestructor() != null)
    {
      messageService.report(MessageLevel.ERR, "Symbol027", enclosing.getNode());
    }

    sanityCheck(method, enclosing);
  }

  public void onCastOperatorDeclaration(GenericMethod method, GenericType enclosing)
  {
    if (!enclosing.isPrimitive())
    {
      messageService.report(MessageLevel.ERR, "Symbol023", enclosing.getNode());
    }
    if (enclosing.getCastOperators().containsKey(method.getResultType().getTypeName()))
    {
      GenericMethod m = (GenericMethod)enclosing.getCastOperators().get(method.getResultType().getTypeName());
      messageService.report(MessageLevel.ERR, "Symbol025", new String[]{method.getResultType().getTypeName().toString(), m.getNode().localization()}, enclosing.getNode());
    }

    sanityCheck(method, enclosing);
  }

  public void onOperatorDeclaration(GenericMethod method, GenericType enclosing)
  {
    // if kind of method is infix, that is one of (=, :=, ::=) in a non primitive type, this is an error
    if (
       method.getKind() == MethodKindEnum.infix &&
       forbiddenOperators.contains(method.getName()) &&
       !enclosing.isPrimitive()
    )
    {
      messageService.report(MessageLevel.ERR, "Symbol023", method.getName(), enclosing.getNode());
    }

    sanityCheck(method, enclosing);
  }

  public void onPropertyDeclaration(GenericMethod method, GenericType enclosing)
  {
    sanityCheck(method, enclosing);
  }

  public void onMethodDeclaration(GenericMethod method, GenericType enclosing)
  {
    sanityCheck(method, enclosing);
  }

  public void onInitializer(GenericMethod method, GenericType enclosing)
  {
    if (enclosing.hasInitializer())
    {
      messageService.report(MessageLevel.ERR, "Symbol032", enclosing.getName().toString(), method.getNode());
      return;
    }

    sanityCheck(method, enclosing);
  }

  public void onDeinitializer(GenericMethod method, GenericType enclosing)
  {
    if (enclosing.hasDeinitializer())
    {
      messageService.report(MessageLevel.ERR, "Symbol037", enclosing.getName().toString(), method.getNode());
      return;
    }

    sanityCheck(method, enclosing);
  }

  private void sanityCheck(Accessible accessible, Node node)
  {
    if (
       accessible.isPublic() && (accessible.isPrivate() || accessible.isProtected()) ||
       accessible.isPrivate() && accessible.isProtected()
    )
    {
      messageService.report(MessageLevel.ERR, "Symbol011", node);
    }
  }

  private void sanityCheck(GenericType type)
  {
    sanityCheck((Accessible)type, type.getNode());

    if (type.isClass())
    {
      // sanity check
      if (type.isFinal() && type.isAbstract())
        messageService.report(MessageLevel.ERR, "Symbol018", type.getName().toString(), type.getNode());
      if (type.isNative() && !type.isPrimitive())
        messageService.report(MessageLevel.ERR, "Symbol019", type.getName().toString(), type.getNode());
    }
  }

  private void sanityCheck(GenericMethod method, GenericType enclosing)
  {
    sanityCheck((Accessible)method, method.getNode());

    if (method.isConst() && method.isStatic())
      messageService.report(MessageLevel.ERR, "Symbol004", method.getNode());
    if (method.isStatic() && method.isFinal())
      messageService.report(MessageLevel.ERR, "Symbol005", method.getNode());
    if (method.getKind() == MethodKindEnum.infix && method.getArguments().size() != 1)
      messageService.report(MessageLevel.ERR, "Symbol006", method.getNode());
    if (method.getKind() == MethodKindEnum.postfix && method.getArguments().size() != 1)
      messageService.report(MessageLevel.ERR, "Symbol007", method.getNode());
    if (method.getKind() == MethodKindEnum.prefix && method.getArguments().size() != 0)
      messageService.report(MessageLevel.ERR, "Symbol008", method.getNode());
    if (method.getKind() == MethodKindEnum.postfix && !method.getName().equals("<<") && !method.getName().equals(">>"))
      messageService.report(MessageLevel.ERR, "Symbol009", method.getNode());
    if (method.getKind() == MethodKindEnum.prefix && !prefixOperators.contains(method.getName()))
      messageService.report(MessageLevel.ERR, "Symbol010", method.getNode());

    if (
        !enclosing.isNative() &&
        !enclosing.isInterface() &&
        !enclosing.isAbstract() &&
        !method.isImplemented()
       )
      messageService.report(MessageLevel.ERR, "Symbol038", new String[]{enclosing.getName().toString(), method.getName()}, method.getNode());
  }

  private void sanityCheck(Field field, GenericType enclosing)
  {
    sanityCheck((Accessible)field, field.getNode());

    if (field.isConst() && field.isMutable())
      messageService.report(MessageLevel.ERR, "Symbol020", field.getName(), field.getNode());
    if (field.isStatic() && field.isTransient())
      messageService.report(MessageLevel.ERR, "Symbol022", field.getName(), field.getNode());
    if (enclosing.isPrimitive() && field.isMutable())
      messageService.report(MessageLevel.ERR, "Symbol039", new String[]{enclosing.getName().toString(), field.getName()}, field.getNode());
  }
}