/*
  $Id: Method.java,v 1.11 2004/03/18 23:15:20 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.parser.node.util.var.TypeName;

import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.AliasTable;
import net.nosica.symbol.GenericMethod;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;

/**
 * \brief representation of a fully instantiated method
 * That is, a method that is not generic anymore
 */

public interface Method extends Accessible
{
  /**
   * \brief is method const ?
   * requires : !isStatic()
   *
   * \return true if method is const
   */
  boolean isConst();

  /**
   * \brief is method final ?
   * requires : !isStatic()
   *
   * \return true id method is final
   */
  boolean isFinal();

  /**
   * \brief is method a Native Code callback ?
   *
   * \return true if method is a callback
   */
  boolean isCallback();

  /**
   * \brief is method implemented in the source code (non abstract?)
   *
   * \return true if the method is implemented in the source code
   */
  boolean isImplemented();

  /**
   * \brief is method sub ?
   *
   * \return true id method is sub
   */
  boolean isSub();

  /**
   * \brief return the resultType of this method
   * requires : !isSub()
   *
   * \return the resultType of this method
   */
  ResultType getResultType();

  /**
   * \brief return the name of this method
   *
   * \return name of the method
   */
  String getName();

  /**
   * \brief return the kind of this method
   *
   * \return kind of the method
   */
  MethodKindEnum getKind();

  /**
   * \brief return the AliasTable valid for this method
   * basically, this is the same table as the alias table of type, plus aliases for generic parameters
   *
   * \return aliasTable of method
   */
  AliasTable getAliasTable();

  /**
   * \brief set the alias table of this method
   *
   * \param aliasTable the alias table of the method
   */
  void setAliasTable(AliasTable aliasTable);

  /**
   * \brief returns the number of argument of this method
   *
   * \return number of argument
   */
  int nbArg();

  /**
   * \brief return an iterator over Arguments
   *
   * \return Iterator<Argument>
   */
  Iterator iterateArguments();

  /**
   * \brief return arguments
   *
   * \return ArrayList<Argument>
   */
  ArrayList getArguments();

  /**
   * \brief return an iterator over throw list
   *
   * \return Iterator<TypeName>
   */
  Iterator iterateThrowList();

  /**
   * \brief return thrown list
   *
   * \return Set<TypeName>
   */
  Set getThrowList();

  /**
   * \brief add an exception to the list of really thrown exception
   *
   * Used in typechecker to set which exceptions were really be thrown
   * At end of type checking, the method is OK if
   * - all exceptions from getThowList are either in getThrownExceptionList or in getInheritedException
   * - all methods of getThrownExceptionList are in getThrowList
   *
   * \param typeName exception really thrown (and allowed to go through)
   */
  void addThrownException(TypeName typeName);

  /**
   * \brief iterate list of really thrown exception
   *
   * Used in typechecker to set which exceptions were really be thrown
   * At end of type checking, the method is OK if
   * - all exceptions from getThowList are either in getThrownExceptionList or in getInheritedException
   * - all methods of getThrownExceptionList are in getThrowList
   *
   * \return Iterator<TypeName>
   */
  Iterator iterateThrownException();

  /**
   * \brief get the list of really thrown exception
   *
   * Used in typechecker to set which exceptions were really be thrown
   * At end of type checking, the method is OK if
   * - all exceptions from getThowList are either in getThrownExceptionList or in getInheritedException
   * - all methods of getThrownExceptionList are in getThrowList
   *
   * \return Set<TypeName>
   */
  Set getThrownExceptionList();

  /**
   * \brief set the inherited list of allowed to throw exception
   *
   * \param inheritedException Set<TypeName> list of inherited exception
   */
  void setInheritedException(Set inheritedException);

  /**
   * \brief return an iterator over inherited list of allowed to throw exception
   *
   * \return Iterator<TypeName>
   */
  Iterator iterateInheritedException();

  /**
   * \brief get the inherited list of allowed to throw exception
   *
   * \return Set<TypeName> list of inherited exception
   */
  Set getInheritedException();

  /**
   * \brief returns the type that defines this method
   *
   * \return name of type
   */
  TypeName getType();

  /**
   * \brief returns the model associated with this method
   *
   * \return model of method
   */
  GenericMethod getModel();

  /**
   * \brief compares method signature
   *
   * \param signature the signature to compare to
   * \return true if method has same signature as signature
   */
  boolean hasSameSignature(Method signature);
}