/*
  $Id: MethodImpl.java,v 1.17 2004/05/06 19:47:00 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.typedeclaration.MethodDeclarationImpl;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.GenericMethodImpl;
import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.AliasTable;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;

public class MethodImpl extends AccessibleImpl implements Method
{
  private GenericMethod method;
  private ResultType resultType;
  private ArrayList arguments;           // ArrayList<Argument>
  private Set throwList;
  private Set inheritedThrowList = java.util.Collections.EMPTY_SET;
  private Set reallyThrownException = new HashSet();
  private AliasTable aliasTable = null;
  private TypeName type;

  /**
   * \brief constructs a new method from a non generic one
   *
   * \param method generic method from which is issued this method
   * \param resultType result type of method, or null if sub
   * \param arguments arguments of method ArrayList<Argument>
   * \param throwList list of thrown exception
   * \param type typeName of type that contains this method
   * \param aliasTable attached aliasTable
   * \param uid unique identifier among methods that have same name in enclosing type
   */
  public MethodImpl(
    GenericMethod method,
    ResultType resultType,
    ArrayList arguments,
    Set throwList,
    TypeName type,
    AliasTable aliasTable
  )
  {
    super(method);

    this.method = method;
    this.resultType = resultType;
    this.arguments = arguments;
    this.throwList = throwList;
    this.type = type;
    this.aliasTable = aliasTable;
  }

  /**
   * \brief constructs a method from user parameters
   * This constructor is mostly used to construct a method signature for method search
   *
   * \param isStatic is method static
   * \param name name of method
   * \param kind kind of method being searched
   * \param arguments a ArrayList<Argument> of arguments
   * \param isConst constness of method being searched
   * \param node a localizable node for error reporting
   *
   * \note resultType is not provided as all resultType will match
   * \note access (public, protected, ...) is not provided as it is the task of the SecurityManager
   * to decide wether or not an access is ok
   * \note throwList is not a criteria
   */
  public MethodImpl(
    boolean isStatic,
    String name,
    MethodKindEnum kind,
    ArrayList arguments,
    boolean isConst,
    net.nosica.parser.node.typedeclaration.Method node
  )
  {
    super(new net.nosica.symbol.AccessibleImpl(true, false, false, isStatic));

    this.method = new GenericMethodImpl(
      node,
      true,
      false,
      false,
      new java.util.ArrayList(),
      null,
      isStatic,
      isConst,
      false,
      false,
      null,
      kind,
      name,
      arguments,
      new java.util.HashSet(),
      true
    );
    this.resultType = null;
    this.arguments = arguments;
    this.throwList = new java.util.HashSet();
    this.type = null;
  }

  /**
   * \brief constructs a method from user parameters
   * This constructor is mostly used to construct nosica defined method
   *
   * \param resultType result of method
   * \param name name of method
   * \param kind kind of method
   * \param arguments a ArrayList<Argument> of arguments
   * \param isConst constness of method
   * \param type enclosing typeName
   * \param aliasTable attached alias table
   * \param uid unique identifier among methods that have same name in enclosing type
   */
  public MethodImpl(
    ResultType resultType,
    String name,
    MethodKindEnum kind,
    ArrayList arguments,
    boolean isConst,
    TypeName type,
    AliasTable aliasTable
  )
  {
    super(new net.nosica.symbol.AccessibleImpl(true, false, false));

    this.method = new GenericMethodImpl(
      new MethodDeclarationImpl(
        null,
        null,
        isConst,
        null,
        null,
        true,
        false,
        false,
        false,
        true,
        false
      ),
      true,
      false,
      false,
      new java.util.ArrayList(),
      type,
      false,
      isConst,
      false,
      false,
      null,
      kind,
      name,
      arguments,
      new java.util.HashSet(),
      true
    );
    this.resultType = resultType;
    this.arguments = arguments;
    this.throwList = new java.util.HashSet();
    this.type = type;
    this.aliasTable = aliasTable;
  }

  public boolean isConst() { return method.isConst(); }
  public boolean isFinal() { return method.isFinal(); }
  public boolean isCallback() { return method.isCallback(); }
  public boolean isSub() { return resultType == null; }
  public boolean isImplemented() { return method.isImplemented(); }
  public ResultType getResultType() { return resultType; }
  public String getName() { return method.getName(); }
  public MethodKindEnum getKind() { return method.getKind(); }
  public AliasTable getAliasTable() { return aliasTable; }
  public void setAliasTable(AliasTable aliasTable) { this.aliasTable = aliasTable; }
  public int nbArg() { return arguments.size(); }
  public Iterator iterateArguments() { return arguments.iterator(); }
  public ArrayList getArguments() { return arguments; }
  public Iterator iterateThrowList() { return throwList.iterator(); }
  public Set getThrowList() { return throwList; }
  public void addThrownException(TypeName typeName) { reallyThrownException.add(typeName); }
  public Iterator iterateThrownException() { return reallyThrownException.iterator(); }
  public Set getThrownExceptionList() { return reallyThrownException; }
  public void setInheritedException(Set inheritedException) { this.inheritedThrowList = inheritedException; }
  public Iterator iterateInheritedException() { return inheritedThrowList.iterator(); }
  public Set getInheritedException() { return inheritedThrowList; }

  public TypeName getType() { return type; }
  public GenericMethod getModel() { return method; }

  public boolean hasSameSignature(Method signature)
  {
    if (isStatic() != signature.isStatic())
      return false;

    if (isConst() != signature.isConst())
      return false;

    if ((getName() == null) != (signature.getName() == null))
      return false;

    if (getName() != null && !getName().equals(signature.getName()))
      return false;

    if (getKind() != signature.getKind())
      return false;

    Iterator i = iterateArguments();
    Iterator j = signature.iterateArguments();

    while (i.hasNext() && j.hasNext())
    {
      Argument arg1 = (Argument)i.next();
      Argument arg2 = (Argument)j.next();

      if (!resultTypeHaveSameSignature(arg1, arg2))
        return false;
    }

    return !i.hasNext() && !j.hasNext();
  }

  private boolean resultTypeHaveSameSignature(ResultType rt1, ResultType rt2)
  {
    if (rt1.isConst() != rt2.isConst())
      return false;

    return rt1.getTypeName().equals(rt2.getTypeName());
  }

  public String toString()
  {
    StringBuffer result = new StringBuffer("[");

    if (isStatic())
      result.append("static ");

    if (isSub())
      result.append("sub ");
    else
    {
      if (getResultType().isConst())
        result.append("const ");
      result.append(getResultType().getTypeName().toString() + " ");
    }

    result.append(getType() + ".");

    if (getKind() == MethodKindEnum.property)
      result.append("property ");
    else
    if (getKind() == MethodKindEnum.propertyArray)
      result.append("property [] ");
    else
    if (getKind() == MethodKindEnum.infix)
      result.append("infix ");
    else
    if (getKind() == MethodKindEnum.postfix)
      result.append("postfix ");
    else
    if (getKind() == MethodKindEnum.prefix)
      result.append("prefix ");
    else
    if (getKind() == MethodKindEnum.castOperator)
      result.append("cast ");

    if (getKind() == MethodKindEnum.castOperator)
      result.append(getResultType().getTypeName().toString());
    else
    if (getName() != null)
      result.append(getName() + " ");

    result.append("(");

    Iterator it = iterateArguments();
    while (it.hasNext())
    {
      Argument arg = (Argument)it.next();
      if (arg.isConst())
        result.append("const ");
      result.append(arg.getTypeName().toString());
      if (it.hasNext())
        result.append(", ");
    }

    result.append(")");

    if (isConst())
      result.append(" const");

    result.append("]");

    return result.toString();
  }
}