/*
  $Id: MethodResolver.java,v 1.13 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.symbol.SymbolException;
import net.nosica.util.SettableInteger;

import java.util.ArrayList;

/**
 * \brief an interface to resolve method call
 */
public interface MethodResolver
{
  /**
   * \brief return the best matching method if it exists on a call context
   * The call context is determined by :
   * <ul>
   *   <li>the instance
   *   <li>the user supplied signature
   *   <li>the accessor
   * </ul>
   * In return, this method returns two set of methods :
   * <ul>
   *   <li>matchingMethods : sets of methods that match before weight cut
   *   <li>possibleMethods : sets of methods that have the same name
   * </ul>
   *
   * \param instance instance to look in
   * \param signature signature to look for
   * \param accessor name of type accessing method
   * \param matchingMethods ArrayList<MethodCallContext> returned by this method, list of matching methods
   * \param possibleMethods ArrayList<Method> returned by this method, list of methods having the same name
   *
   * \return a method call context if it exists allowing the call, null if no method match or if several methods match
   */
  MethodCallContext findMatchingMethod(ResultType instance, Method signature, TypeName accessor, ArrayList matchingMethods, ArrayList possibleMethods) throws SymbolException;

  /**
   * \brief returns all matching methods
   *
   * methods are first looked into typeName.
   * If no result, they are searched in all extends type recursively
   * if no result and nested class, they are searched into enclosing recursively
   *
   * \param instance instance to look in
   * \param signature signature to look for
   * \param accessor name of type accessing method
   * \param possibleMethods ArrayList<Method> list of methods that match the name
   * \return ArrayList<MethodCallContext> (can be empty if no method match)
   * \throw SymbolException if typeName does not exist
   */
  ArrayList findMatchingMethods(ResultType instance, Method signature, TypeName accessor, ArrayList possibleMethods) throws SymbolException;

  /**
   * \brief returns the best method in a set of possible methods
   * Methods have been tipically chosen by a call to findMatchingMethods
   * getBestMethod will then uses the MethodCallContext's weight to return the
   * best method in this call context
   *
   * \param possibleMethods ArrayList<MethodCallContext> a set of method call context
   * \return the best method if it exists, null otherwise
   */
  MethodCallContext getBestMethod(ArrayList possibleMethods);

  /**
   * \brief returns all property set methods named "name"
   *
   * methods are first looked into typeName.
   * If no result, they are searched in all extends type recursively
   * if no result and nested class, they are searched into enclosing recursively
   *
   * \param instance instance to look in
   * \param name name of the property set method
   * \param accessor name of type accessing method
   * \param possibleMethods ArrayList<Method> list of methods that match the name
   * \return ArrayList<Method> (can be empty if no method match)
   * \throw SymbolException if typeName does not exist
   */
  ArrayList findPropertySetMethods(ResultType instance, String name, TypeName accessor, ArrayList possibleMethods) throws SymbolException;

  /**
   * \brief returns all property set array methods named "name"
   *
   * methods are first looked into typeName.
   * If no result, they are searched in all extends type recursively
   * if no result and nested class, they are searched into enclosing recursively
   *
   * \param instance instance to look in
   * \param name name of the property set array method
   * \param nbArg number of argument this property array method should take
   * \param accessor name of type accessing method
   * \param possibleMethods ArrayList<Method> list of methods that match the name
   * \return ArrayList<Method> (can be empty if no method match)
   * \throw SymbolException if typeName does not exist
   */
  ArrayList findPropertySetArrayMethods(ResultType instance, String name, int nbArg, TypeName accessor, ArrayList possibleMethods) throws SymbolException;

  /**
   * \brief returns a cast operator if it exists between fromType to toType
   *
   * \param fromType type to search for
   * \param toType target type
   * \return the cast operator if it exists, null otherwise
   * \throw SymbolException if typeName does not exist
   */
  Method getCastOperator(TypeName fromType, TypeName toType) throws SymbolException;

  /**
   * \brief find a field
   *
   * field is first looked into typeName.
   * If no result, it is searched in all extends type recursively
   * if no result and nested class, it is searched into enclosing recursively
   *
   *
   * \param instance instance to look in
   * \param name name of field to search for
   * \param accessor name of the type accessing method
   * \param possibleFields ArrayList<Field> list of methods that match the name
   * \return field if it exists, null otherwise
   */
  FieldContext findField(ResultType instance, String name, TypeName accessor, ArrayList possibleFields) throws SymbolException;

  /**
   * \brief get default constructor of type
   *
   * \param typeName name of type to look in
   * \param accessor name of type accessing method
   * \param possibleConstructors ArrayList<Method> list of constructor that match the name
   * \return default constructor if it exists, null otherwise
   * \throw SymbolException if typeName does not exist
   */
  Method getDefaultConstructor(TypeName typeName, TypeName accessor, ArrayList possibleConstructors) throws SymbolException;

  /**
   * \brief match a resultType against a signature
   *
   * a resultType match if :
   *   if both are primitive :
   *     if typename are the same
   *       if constness are different a weight of CONSTNESS_WEIGHT is added
   *       MATCH
   *     else
   *     if a cast operator exists between signature to resultType
   *       weight is incremented by CAST_WEIGHT
   *       MATCH
   *     else
   *       NO MATCH
   *   else
   *   if both are references
   *     if signature is const and resultType is not
   *       NO MATCH
   *     if typename are the same
   *       MATCH
   *     if signature inherits from resultType
   *       weight is incremented by INHERITS_WEIGHT
   *     MATCH
   *   else
   *     NO MATCH
   *
   * \param resultType the model to match against (coming from a real signature)
   * \param signature the use supplied signature
   * \param weight an integer that is incremented to reflect the exactness of the match
   * \return the cast operator needed to make the matching, null if none was needed
   * \throw TypeCheckerException if NO MATH, don't throw if MATCH
   * \throw SymbolException if one of the needed types are not in TypeEnvironment
   */
  Method match(ResultType signature, ResultType resultType, SettableInteger weight) throws TypeMismatchException, SymbolException;

  /**
   * \brief return operator = of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorValueAssign(TypeName typeName);

  /**
   * \brief return operator ~ of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorInstanceAssign(TypeName typeName);


  /**
   * \brief return operator < of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorLt(TypeName typeName);

  /**
   * \brief return operator >* of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorMul(TypeName typeName);

  /**
   * \brief return operator + of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorAdd(TypeName typeName);

  /**
   * \brief return operator ++ of type
   *
   * \param typeName the name of the type to get operator from
   * \return asked operator
   * \remarks, method safely terminate program if operator is not found
   */
  Method getOperatorPlusPlus(TypeName typeName);

  /**
   * \brief return method length of array type
   *
   * \param typeName the name of the array type to get method from
   * \return asked operator
   * \remarks, method safely terminate program if method is not found
   */
  Method getMethodLength(TypeName typeName);

  /**
   * \brief find a particular method in a type
   *
   * \param typeName the type to look in
   * \param method the method to look for
   * \return the method if found
   * \throw SymbolException if not found
   */
  Method findMethod(TypeName typeName, Method method) throws SymbolException;

  /**
   * \brief find the main method in type
   *
   * \param typeName the type to look in
   * \return the method if found, null otherwise
   */
  Method findMainMethod(TypeName typeName);
}