/*
  $Id: MethodResolverImpl.java,v 1.24 2004/03/18 23:15:20 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.application.NativeTypes;
import net.nosica.compiler.application.Environment;
import net.nosica.compiler.application.GenericityException;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.symbol.SymbolException;
import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.GenericMethod;

import net.nosica.util.SettableInteger;

import java.util.Iterator;
import java.util.ArrayList;

public class MethodResolverImpl implements MethodResolver
{
  private static MethodResolverImpl theInstance = new MethodResolverImpl();
  private MethodResolverImpl() {}
  public static MethodResolverImpl instance() { return theInstance; }
  public static void instanceReset() { theInstance = new MethodResolverImpl(); }

  public static final int CONSTNESS_WEIGHT = 8;
  public static final int CAST_WEIGHT = 16;
  public static final int INHERITS_WEIGHT = 16;
  public static final int METHOD_CONSTNESS_WEIGHT = 160;  // should be very big ... how much ?
                                                          // set as 16 * 10 arguments for now

  private Environment env = Environment.instance();

  public MethodCallContext findMatchingMethod(
    ResultType instance,
    Method signature,   // the user supplied signature
    TypeName accessor,
    ArrayList matchingMethods,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    matchingMethods.addAll(findMatchingMethods(instance, signature, accessor, possibleMethods));

    return getBestMethod(matchingMethods);
  }

  public ArrayList findMatchingMethods(
    ResultType instance,
    Method signature,   // the user supplied signature
    TypeName accessor,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    Accessible mask = createMask(instance.getTypeName(), accessor);

    return findMatchingMethods(instance, signature, mask, possibleMethods);
  }

  private ArrayList findMatchingMethods(
    ResultType instance,
    Method signature,   // the user supplied signature
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    ArrayList result = recFindMatchingMethodsInExtends(instance, signature, mask, possibleMethods);

    if (signature.getName() != null && (signature.getName().equals("constructor") || signature.getName().equals("destructor")))
      return result;

    // enclosing type now ...
    if (result.isEmpty())
    {
      Type type = env.getTE().getType(instance.getTypeName());  // can throw if type does not exist
      if (type.isInner())
      {
        result = findMatchingMethods(new ResultTypeImpl(type.getEnclosing(), instance.isConst()), signature, mask, possibleMethods);
        Iterator it = result.iterator();
        while (it.hasNext())
        {
          MethodCallContext mcc = (MethodCallContext)it.next();
          mcc.insertEnclosing(type.getEnclosing());
        }
      }
    }

    return result;
  }

  public MethodCallContext getBestMethod(ArrayList possibleMethods)
  {
    MethodCallContext bestMcc = null;
    int bestWeight = -1;

    Iterator it = possibleMethods.iterator();
    while (it.hasNext())
    {
      MethodCallContext mcc = (MethodCallContext)it.next();

      int weight = getGlobalWeight(mcc);

      if (bestWeight == -1)
      {
        // special case, this is the first loop
        bestMcc = mcc;
        bestWeight = weight;
      }
      else
      {
        if (weight < bestWeight)
        {
          bestMcc = mcc;
          bestWeight = weight;
        }
        else
        if (weight == bestWeight)
        {
          // we have an ambiguity
          bestMcc = null;
        }
      }
    }

    return bestMcc;
  }

  private int getGlobalWeight(MethodCallContext mcc)
  {
    int weight = mcc.getWeight();

    Iterator it = mcc.iterateArgumentCasts();
    while (it.hasNext())
    {
      MethodCallContext argMcc = (MethodCallContext)it.next();

      if (argMcc != null)
      {
        weight += argMcc.getWeight();
      }
    }

    return weight;
  }

  /**
   * \brief search in extends recursively
   */
  private ArrayList recFindMatchingMethodsInExtends(
    ResultType instance,
    Method signature,   // the user supplied signature
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    ArrayList result = findMatchingMethodsRaw(instance, signature, mask, possibleMethods);

    if (signature.getName() != null && (signature.getName().equals("constructor") || signature.getName().equals("destructor")))
      return result;

    Type type = env.getTE().getType(instance.getTypeName());  // can throw if type does not exist

    // super types if needed
    if (result.isEmpty())
    {
      Iterator i = type.iterateExtends();
      while (i.hasNext() && result.isEmpty())
      {
        TypeName extendsTypeName = (TypeName)i.next();
        result = recFindMatchingMethodsInExtends(new ResultTypeImpl(extendsTypeName, instance.isConst()), signature, mask, possibleMethods);

        Iterator it = result.iterator();
        while (it.hasNext())
        {
          MethodCallContext mcc = (MethodCallContext)it.next();

          if (!mcc.getMethod().isStatic() && !env.getTE().getType(mcc.getMethod().getType()).isInterface())
            mcc.insertSuper(instance.getTypeName());
        }
      }
    }

    return result;
  }

  /**
   * \brief search only in wanted type
   */
  private ArrayList findMatchingMethodsRaw(
    ResultType instance,
    Method signature,   // the user supplied signature
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    Type type = env.getTE().getType(instance.getTypeName());

    ArrayList result = new ArrayList();

    try
    {
      // resolved method
      Iterator i = type.iterateMethodsByName(signature.getName());

      while (i.hasNext())
      {
        Method m = (Method)i.next();

        if (signature.getKind() == m.getKind() && !possibleMethods.contains(m))
          possibleMethods.add(m);

        // we cannot call a non const method on a const instance
        if (instance.isConst() && !m.isConst())
          continue;

        if (isAccessibleFrom(m, mask))
        {
          MethodCallContext methodCallContext = methodMatch(m, signature);

          if (methodCallContext != null)
          {
            result.add(methodCallContext);
          }
        }
      }
    }
    catch (SymbolException e)
    {
      // if method does not exist, iterateMethodsByName will throw
      // result will contain an empty vector, that's what we want
    }

    try
    {
      // generic method
      Iterator i = type.iterateGenericMethodsByName(signature.getName());

      while (i.hasNext())
      {
        GenericMethod gm = (GenericMethod)i.next();

        try
        {
          Method m = env.getGS().instantiateMethod(gm, signature, type.getName(), type.getAliasTable());
          type.addMethod(m);

          if (signature.getKind() == m.getKind() && !possibleMethods.contains(m))
            possibleMethods.add(m);

          // we cannot call a non const method on a const instance
          if (instance.isConst() && !m.isConst())
            continue;

          if (isAccessibleFrom(m, mask))
          {
            MethodCallContext methodCallContext = methodMatch(m, signature);

            if (methodCallContext != null)
            {
              methodCallContext.addWeight(1);
              result.add(methodCallContext);
            }
          }
        }
        catch (GenericityException e)
        {
          // method was not instantiable
        }
      }
    }
    catch (SymbolException e)
    {
      // if method does not exist, iterateMethodsByName will throw
      // result will contain an empty vector, that's what we want
    }

    return result;
  }

  public ArrayList findPropertySetMethods(
    ResultType instance,
    String name,
    TypeName accessor,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    return findPropertySetArrayMethods(instance, name, 1, accessor, possibleMethods);
  }

  public ArrayList findPropertySetArrayMethods(
    ResultType instance,
    String name,
    int nbArg,
    TypeName accessor,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    Accessible mask = createMask(instance.getTypeName(), accessor);

    return findPropertySetArrayMethods(instance, name, nbArg, mask, possibleMethods);
  }

  private ArrayList findPropertySetArrayMethods(
    ResultType instance,
    String name,
    int nbArg,
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    ArrayList result = recFindPropertySetArrayMethodsInExtends(instance, name, nbArg, mask, possibleMethods);

    // enclosing type now ...
    if (result.isEmpty())
    {
      Type type = env.getTE().getType(instance.getTypeName());
      if (type.isInner())
        result = findPropertySetArrayMethods(new ResultTypeImpl(type.getEnclosing(), instance.isConst()), name, nbArg, mask, possibleMethods);
    }

    return result;
  }

  private ArrayList recFindPropertySetArrayMethodsInExtends(
    ResultType instance,
    String name,
    int nbArg,
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    ArrayList result = findPropertySetArrayRaw(instance, name, nbArg, mask, possibleMethods);

    // super types if needed
    if (result.isEmpty())
    {
      Type type = env.getTE().getType(instance.getTypeName());

      Iterator i = type.iterateExtends();
      while (i.hasNext() && result.isEmpty())
      {
        TypeName extendsTypeName = (TypeName)i.next();
        result = recFindPropertySetArrayMethodsInExtends(new ResultTypeImpl(extendsTypeName, instance.isConst()), name, nbArg, mask, possibleMethods);
      }
    }

    return result;
  }

  private ArrayList findPropertySetArrayRaw(
    ResultType instance,
    String name,
    int nbArg,
    Accessible mask,
    ArrayList possibleMethods
  ) throws SymbolException
  {
    Type type = env.getTE().getType(instance.getTypeName());

    ArrayList result = new ArrayList();

    try
    {
      Iterator i = type.iterateMethodsByName(name);

      while (i.hasNext())
      {
        Method m = (Method)i.next();

        if (m.getKind() == MethodKindEnum.propertyArray && !possibleMethods.contains(m))
          possibleMethods.add(m);

        // we cannot call a non const method on a const instance
        if (instance.isConst() && !m.isConst())
          continue;

        if (nbArg == 1)
        {
          if (m.getKind() != MethodKindEnum.property)
            continue;
        }
        else
        {
          if (m.getKind() != MethodKindEnum.propertyArray)
            continue;
        }

        if (m.nbArg() != nbArg)
          continue;

        if (isAccessibleFrom(m, mask))
          result.add(m);
      }
    }
    catch (SymbolException e)
    {
      // if method does not exist, iterateMethodsByName will throw.
      // => result will contain an empty vector, that's what we want
    }

    return result;
  }

  public Method getCastOperator(TypeName fromType, TypeName toType) throws SymbolException
  {
    Type type = env.getTE().getType(fromType);

    try
    {
      return type.getCastOperator(toType);
    }
    catch (SymbolException e)
    {
      return null;
    }
  }

  public FieldContext findField(
    ResultType instance,
    String name,
    TypeName accessor,
    ArrayList possibleFields
  ) throws SymbolException
  {
    Accessible mask = createMask(instance.getTypeName(), accessor);
    return findField(instance, name, mask, possibleFields);
  }

  private FieldContext findField(
    ResultType instance,
    String name,
    Accessible mask,
    ArrayList possibleFields
  ) throws SymbolException
  {
    FieldContext field = recFindFieldInExtends(instance, name, mask, possibleFields);

    if (field == null)
    {
      Type type = env.getTE().getType(instance.getTypeName());

      if (type.isInner())
      {
        FieldContext result = findField(new ResultTypeImpl(type.getEnclosing(), instance.isConst()), name, mask, possibleFields);
        result.insertEnclosing(type.getEnclosing());
        return result;
      }
    }

    return field;
  }

  private FieldContext recFindFieldInExtends(
    ResultType instance,
    String name,
    Accessible mask,
    ArrayList possibleFields
  ) throws SymbolException
  {
    Type type = env.getTE().getType(instance.getTypeName());  // can throw if type does not exist

    try
    {
      Field field = type.getField(name);

      if (!possibleFields.contains(field))
        possibleFields.add(field);

      if (isAccessibleFrom(field, mask))
      {
        FieldContext fieldContext = new FieldContextImpl(field);

        // if instance is const, then accessed field is temporarily considedered as const
        if (instance.isConst())
          fieldContext.setConst(true);

        return fieldContext;
      }
    }
    catch (SymbolException e)
    {
      // throw if field does not exist in type
    }

    Iterator it = type.iterateExtends();
    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();

      try
      {
        // as a class has one and only one extends, we don't have to retrieve
        // result, compare it to null to see if we have to iterate or not ...
        FieldContext result = recFindFieldInExtends(new ResultTypeImpl(extendsTypeName, instance.isConst()), name, mask, possibleFields);
        if (result != null)
          result.insertSuper(extendsTypeName);

        return result;
      }
      catch (SymbolException ex)
      {
        // throw if field does not exist in type
      }
    }

    // not found
    return null;
  }

  public Method getDefaultConstructor(
    TypeName typeName,
    TypeName accessor,
    ArrayList possibleConstructors
  ) throws SymbolException
  {
    Accessible mask = createMask(typeName, accessor);

    Type type = env.getTE().getType(typeName);  // can throw if type does not exist

    Method defaultConstructor = null;

    Iterator it = type.iterateMethodsByName("constructor");
    while (it.hasNext())
    {
      Method m = (Method)it.next();

      if (!possibleConstructors.contains(m))
        possibleConstructors.add(m);

      if ((m.nbArg() == 0) && isAccessibleFrom(m, mask))
      {
        defaultConstructor = m;
        break;
      }
    }

    try
    {
      // generic method
      Iterator i = type.iterateGenericMethodsByName("constructor");
      Method signature = new MethodImpl(
                          null,
                          "constructor",
                          MethodKindEnum.constructor,
                          new ArrayList(),
                          false,
                          typeName,
                          null
      );

      while (i.hasNext())
      {
        GenericMethod gm = (GenericMethod)i.next();

        try
        {
          Method m = env.getGS().instantiateMethod(gm, signature, typeName, type.getAliasTable());

          if (!possibleConstructors.contains(m))
            possibleConstructors.add(m);

          if ((m.nbArg() == 0) && isAccessibleFrom(m, mask))
          {
            if (defaultConstructor == null)
              defaultConstructor = m;
            else
              defaultConstructor = null; // two constructors are matching !!
          }
        }
        catch (GenericityException e)
        {
          // method was not instantiable
        }
      }
    }
    catch (SymbolException e)
    {
      // if method does not exist, iterateMethodsByName will throw
      // result will contain an empty vector, that's what we want
    }

    return defaultConstructor;
  }

  /**
   * \brief verify protection is valid against a mask
   *
   * \param protection protection of member being accessed
   * \param mask mask of access
   * \return true if mask
   */
  private boolean isAccessibleFrom(Accessible protection, Accessible mask) throws SymbolException
  {
    return mask.isPublic() && protection.isPublic() ||
           mask.isProtected() && protection.isProtected() ||
           mask.isPrivate() && protection.isPrivate() ||
           mask.isPackage() && protection.isPackage();
  }

  private Accessible createMask(TypeName accessed, TypeName accessor) throws SymbolException
  {
    if (accessor.equals(accessed))
    {
      // full access
      return new AccessibleImpl(
        new net.nosica.symbol.AccessibleImpl(
          true,
          true,
          true,
          true,
          false
        )
      );
    }

    if (accessor.isArray())
    {
      // public access only
      return new AccessibleImpl(
        new net.nosica.symbol.AccessibleImpl(
          true,
          false,
          false,
          false,
          false
        )
      );
    }

    // look if accessor is an inner of accessed
    ArrayList components = accessor.getComponents();
    ArrayList partialComponents = new ArrayList();

    boolean isInner = false;

    Iterator i = components.iterator();
    while (i.hasNext())
    {
      partialComponents.add(i.next());
      if (new TypeNameImpl(partialComponents).equals(accessed))
      {
        isInner = true;
        break;
      }
    }

    if (isInner)
    {
      // full access
      return new AccessibleImpl(
        new net.nosica.symbol.AccessibleImpl(
          true,
          true,
          true,
          true,
          false
        )
      );
    }

    // inheritance ?
    boolean isSon = env.getTE().doesInherits(accessor, accessed);
    // compare package ...
    boolean isPackage = accessed.getFirstComponents().equals(accessor.getFirstComponents());


    return new AccessibleImpl(
      new net.nosica.symbol.AccessibleImpl(
        true,
        isSon,
        false,
        isPackage,
        false
      )
    );
  }

  /**
   * \brief returns MethodCallContext to use to call method
   *
   * \param m a real method of type to match against
   * \param signature the user supplied signature
   */
  private MethodCallContext methodMatch(Method m, Method signature) throws SymbolException
  {
    // constructs methodCallContext
    ArrayList argCasts = new ArrayList(); // vector will be filled later on this method
    MethodCallContext methodCallContext = new MethodCallContextImpl(signature.getModel().getNode(), m, argCasts);

    // staticness ?
    if (signature.isStatic() && !m.isStatic())
      return null;

    // constness ?
    if (signature.isConst() && !m.isConst())
      return null;

    // if method constness is different, add weight
    if (signature.isConst() != m.isConst())
      methodCallContext.addWeight(METHOD_CONSTNESS_WEIGHT);

    // name ?
    if (m.getName() != null)
    {
      if (!m.getName().equals(signature.getName()))
        return null;
    }
    else
    {
      if (signature.getName() != null)
        return null;
    }

    // kind ?
    if (m.getKind() != signature.getKind())
      return null;

    // args ?
    if (m.nbArg() != signature.nbArg())
      return null;

    Iterator i = m.iterateArguments();
    Iterator j = signature.iterateArguments();

    while (i.hasNext() && j.hasNext())
    {
      Argument arg1 = (Argument)i.next();
      Argument arg2 = (Argument)j.next();

      try
      {
        argCasts.add(argumentMatch(arg1, arg2));
      }
      catch (TypeMismatchException e)
      {
        return null;
      }
    }

    return methodCallContext;
  }


 /**
   * \brief performs an argument match
   *
   * \param arg the real argument of type to match against
   * \param signature the argument supplied by user signature
   * \return a cast operator branch or null if no cast was required
   */
  private MethodCallContext argumentMatch(
    Argument arg,
    Argument signature
  ) throws TypeMismatchException, SymbolException
  {
    // get types
    SettableInteger tempWeight = new SettableInteger(0);
    Method castOperator = match(arg, signature, tempWeight);

    if (castOperator != null)
    {
      MethodCallContext mcc = new MethodCallContextImpl(signature.getTypeName(), castOperator);
      mcc.addWeight(tempWeight.intValue());

      return mcc;
    }

    MethodCallContext mcc = new MethodCallContextImpl(signature.getTypeName(), null);
    mcc.addWeight(tempWeight.intValue());

    return mcc;
  }

  public Method match(
    ResultType resultType,
    ResultType signature,
    SettableInteger weight
  ) throws TypeMismatchException, SymbolException
  {
    // first check constness
    if (
        !signature.isConst() ||
        resultType.isConst()
       )
    {
      // if constness is different, add a weight
      if (resultType.isConst() != signature.isConst())
        weight.inc(CONSTNESS_WEIGHT);

      // get types
      Type argType = env.getTE().getType(resultType.getTypeName());
      Type signatureType = env.getTE().getType(signature.getTypeName());

      // primitive types ...
      if (argType.isPrimitive() && signatureType.isPrimitive())
      {
        // if same type
        if (signature.getTypeName().equals(resultType.getTypeName()))
          return null;

        // search for a cast operator
        Method castOperator = getCastOperator(signature.getTypeName(), resultType.getTypeName());

        if (castOperator != null)
        {
          weight.inc(CAST_WEIGHT);

          return castOperator;
        }

        // no match
      }
      else
      if (!argType.isPrimitive() && !signatureType.isPrimitive())
      {
        // if same type
        if (resultType.getTypeName().equals(signature.getTypeName()))
          return null;

        // check for inheritance
        if (env.getTE().doesInherits(signature.getTypeName(), resultType.getTypeName()))
        {
          weight.inc(INHERITS_WEIGHT);
          return null;
        }

        // no match
      }
    }

    throw new TypeMismatchException("argument do not match");
  }

  private Method getOperator(TypeName typeName, String operatorName, MethodKindEnum operatorType)
  {
    ArrayList opArgs = new ArrayList();
    if (operatorType != MethodKindEnum.prefix)
    {
      opArgs.add(new ArgumentImpl(true, typeName, ""));
    }

    Method operator =
      new MethodImpl(
        false,
        operatorName,
        operatorType,
        opArgs,
        false,
        null
    );

    ArrayList methods = null;

    try
    {
      ArrayList possibleMethods = new ArrayList();
      methods = env.getMR().findMatchingMethods(
        new ResultTypeImpl(typeName, false),
        operator,
        typeName,
        possibleMethods
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }

    if (methods.size() == 0)
    {
      env.terminate("Internal error : type " + typeName + " has no operator " + operatorName);
    }
    else
    if (methods.size() != 1)
    {
      env.terminate("Internal error : type " + typeName + " should have only one operator " + operatorName);
    }

    return ((MethodCallContext)methods.get(0)).getMethod();
  }

  public Method getOperatorValueAssign(TypeName typeName)
  {
    return getOperator(typeName, "=", MethodKindEnum.infix);
  }

  public Method getOperatorInstanceAssign(TypeName typeName)
  {
    return getOperator(typeName, "~", MethodKindEnum.infix);
  }

  public Method getOperatorLt(TypeName typeName)
  {
    return getOperator(typeName, "<", MethodKindEnum.infix);
  }

  public Method getOperatorMul(TypeName typeName)
  {
    return getOperator(typeName, "*", MethodKindEnum.infix);
  }

  public Method getOperatorAdd(TypeName typeName)
  {
    return getOperator(typeName, "+", MethodKindEnum.infix);
  }

  public Method getOperatorPlusPlus(TypeName typeName)
  {
    return getOperator(typeName, "++", MethodKindEnum.prefix);
  }

  public Method getMethodLength(TypeName typeName)
  {
    ArrayList opArgs = new ArrayList();
    opArgs.add(new ArgumentImpl(true, NativeTypes.get_net_nosica_lang_word32(), ""));

    Method methodLength =
      new MethodImpl(
        false,
        "length",
        MethodKindEnum.method,
        opArgs,
        true,
        null
    );

    ArrayList methods = null;

    try
    {
      ArrayList possibleMethods = new ArrayList();
      methods = env.getMR().findMatchingMethods(
        new ResultTypeImpl(typeName, false),
        methodLength,
        typeName,
        possibleMethods
      );
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error : " + e.getMessage());
    }

    if (methods.size() != 1)
      env.terminate("Internal error : type " + typeName + " should have only one method length");

    return ((MethodCallContext)methods.get(0)).getMethod();
  }

  public Method findMethod(TypeName typeName, Method method) throws SymbolException
  {
    Type type = env.getTE().getType(typeName);

    try
    {
      // @TODO@ take into account aliased methods ...
      if (method.getKind() == MethodKindEnum.castOperator)
      {
        // cast operator
        Method m = type.getCastOperator(method.getResultType().getTypeName());
        if (m.hasSameSignature(method))
          return m;
      }
      else
      {
        // overloadables
        Iterator it = type.iterateMethodsByName(method.getName());

        while (it.hasNext())
        {
          Method m = (Method)it.next();

          if (m.hasSameSignature(method))
            return m;
        }
      }
    }
    catch (SymbolException e)
    {
      // no method by that name in type
    }

    // method not found in type, let's try to find it in extend
    Iterator it = type.iterateExtends();
    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();
      return findMethod(extendsTypeName, method);
    }

    throw new SymbolException("Could not find method " + method.toString() + " in type [" + typeName.toString() + "]");
  }

  public Method findMainMethod(TypeName typeName)
  {
    try
    {
      Type mainType = env.getTE().getType(typeName);

      Iterator it = mainType.iterateOverloadables();
      while (it.hasNext())
      {
        ArrayList methods = (ArrayList)it.next();
        Iterator j = methods.iterator();

        while (j.hasNext())
        {
          Method m = (Method)j.next();

          if (!m.isStatic())
            continue;

          if (!m.isSub())
            continue;

          if (!m.getName().equals("main"))
            continue;

          if (m.getArguments().size() != 1)
            continue;

          net.nosica.compiler.typechecker.Argument arg =
            (net.nosica.compiler.typechecker.Argument)m.getArguments().get(0);

          if (!arg.isConst())
            continue;

          if (!arg.getTypeName().equals(NativeTypes.get_net_nosica_lang_string_array()))
            continue;

          return m;
        }
      }
    }
    catch (SymbolException e)
    {
    }

    return null;
  }
}