/*
  $Id: Type.java,v 1.9 2004/03/18 23:15:20 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.symbol.TypeKindEnum;
import net.nosica.symbol.SymbolException;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.AliasTable;

import java.util.Iterator;
import java.util.Map;

/**
 * \brief an instanciated Type
 * That is, a type that is not generic anymore.
 * One can make the difference between GenericType and Type this way :
 * <ul>
 *   <li>ArrayList<T, U> is a GenericType that can be instantiated or partially instantiated
 *   <li>ArrayList<T, string> is a GenericType that can be instantiated
 *   <li>ArrayList<int, string> is a fully instantiated type, this is a Type
 * </ul>
 */

public interface Type extends Accessible
{
  /**
   * \brief return the name of this type
   *
   * \return TypeName of this type
   */
  TypeName      getName();

  /**
   * \brief return kind of this type
   *
   * \return kind of type (class, interface, nestedclass, nestedinterface)
   */
  TypeKindEnum  getKind();

  /**
   * \brief Is this type a class ?
   *
   * \return true if this type is a class, false otherwise
   */
  boolean isClass();

  /**
   * \brief Is this type an interface ?
   *
   * \return true if this type is an interface, false otherwise
   */
  boolean isInterface();

  /**
   * \brief Is this type a nested type ?
   * ie, does this type has been defined inside another type ?
   *
   * \return true if type is nested, false otherwise
   */
  boolean isNested();

  /**
   * \brief Is this type an abstract type
   * ie, can it be instantiated ?<BR>
   * requires : (isClass() && isAbstract()) || isInterface()
   *
   * \return true if type is abstract, false otherwise
   */
  boolean       isAbstract();

  /**
   * \brief Is this type a final type
   * ie, can it be subclassed ?<BR>
   * requires : isClass() && !isAbstract()
   *
   * \return true if type is final, false otherwise
   */
  boolean       isFinal();

  /**
   * \brief Is this type a primitive type ?
   * ie, is it a reference type or a pass-by-value type ?<BR>
   * requires : isClass()
   *
   * \return true if type is primitive, false otherwise
   */
  boolean       isPrimitive();

  /**
   * \brief Is this type a native type
   * ie, does impelmentation of this type left to the programmer in C ort C++ ?<BR>
   * requires : isPrimitive()
   *
   * \return true if type is native, false otherwise
   */
  boolean       isNative();

  /**
   * \brief Is this type an inner type
   * ie, is this class inner of another class ?<BR>
   * requires : isNested() && isClass() && !isAbstract()
   *
   * \return true if type is inner, false otherwise
   */
  boolean       isInner();

  /**
   * \brief return an iterator over all extends
   *
   * \return Iterator<TypeName>
   */
  Iterator iterateExtends();

  /**
   * \brief return true if classe or interface has an explicit extends
   * That is if classes extends something else than Object, or interface extends something else
   * than General
   *
   * \return true if there is an explicit extends
   */
  boolean hasExplicitExtends();

  /**
   * \brief return an iterator over all implements
   *
   * \return Iterator<TypeName>
   */
  Iterator iterateImplements();

  /**
   * \brief return an iterator over all fields
   *
   * \return Iterator<Field>
   */
  Iterator iterateFields();

  /**
   * \brief returns a field by name
   *
   * \param name name of the field to get
   * \return field if found
   */
  Field getField(String name) throws SymbolException;

  /**
   * \brief return an iterator over all overloadable methods
   *
   * \return Iterator<ArrayList<Method>>
   */
  Iterator iterateOverloadables();

  /**
   * \brief return an iterator over all cast operators
   *
   * \return Iterator<Method>
   */
   Iterator iterateCastOperators();

   /**
    * \brief return the cast operator that transform this type into typeName
    *
    * \param typeName the targeted type
    * \return method if it exists
    */
   Method getCastOperator(TypeName typeName) throws SymbolException;

   /**
    * \brief return the initializer
    *
    * \return initializer
    */
   Method getInitializer();

    /**
    * \brief return the deinitializer
    *
    * \return deinitializer
    */
   Method getDeinitializer();

   /**
    * \brief return destructor
    *
    * \return destructor
    */
   Method getDestructor();

   /**
    * \brief returns an iterator over all methods having the same name
    *
    * \param name name of method
    * \return Iterator<Method>
    */
   Iterator iterateMethodsByName(String name) throws SymbolException;

   /**
    * \brief returns an iterator over all generic methods having the same name
    *
    * \param name name of method
    * \return Iterator<GenericMethod>
    */
   Iterator iterateGenericMethodsByName(String name) throws SymbolException;

   /**
    * \brief returns an iterator over all methods
    *
    * \return Iterator<Method>
    */
   Iterator iterateMethods() throws SymbolException;

   /**
    * \brief returns the model of type
    *
    * \return returns the type model from which is issued this type
    */
   GenericType getModel();

   /**
    * \brief return the alias table associated with this type
    *
    * \return aliastable associated with this type
    */
   AliasTable getAliasTable();

   /**
    * \brief return the method alias table for this type
    *
    * \return Map<String, TypeName>
    */
   Map getMethodAliasTable();

   /**
    * \brief returns enclosing of type (if any)
    * requires : isClass() && isNested()
    *
    * \return name of enclosing type
    */
   TypeName getEnclosing();

   /**
    * \brief adds a method to this type
    *
    * \param method the method to add
    */
   void addMethod(Method method) throws SymbolException;
}