/*
  $Id: TypeCheckVisitor.java,v 1.62 2004/03/04 22:16:16 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.framework.Localizable;
import net.nosica.framework.VisitorException;

import net.nosica.symbol.AliasTable;
import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.SymbolException;

import net.nosica.application.MessageLevel;
import net.nosica.application.NativeTypes;
import net.nosica.compiler.application.Environment;
import net.nosica.compiler.application.GenericityException;
import net.nosica.compiler.application.UseTypeCallBack;
import net.nosica.compiler.application.LabelRepository;
import net.nosica.compiler.application.IRTranslator;
import net.nosica.compiler.application.TypeCheckerErrorExpressionException;

import net.nosica.parser.node.Node;
import net.nosica.parser.node.NodeVisitor;
import net.nosica.parser.node.AbstractNodeVisitor;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponent;
import net.nosica.parser.node.util.var.ArrayDimsAndInits;
import net.nosica.parser.node.util.var.ArrayInitializer;
import net.nosica.parser.node.util.var.VariableInitializer;
import net.nosica.parser.node.util.var.Arguments;
import net.nosica.parser.node.util.operator.Operator;
import net.nosica.parser.node.util.operator.OperatorEnum;
import net.nosica.parser.node.util.literal.Literal;
import net.nosica.parser.node.util.literal.LiteralEnum;

import net.nosica.parser.node.typedeclaration.expression.AdditiveExpression;
import net.nosica.parser.node.typedeclaration.expression.AllocationExpression;
import net.nosica.parser.node.typedeclaration.expression.AssignExpression;
import net.nosica.parser.node.typedeclaration.expression.BinaryOperatorExpression;
import net.nosica.parser.node.typedeclaration.expression.CastExpression;
import net.nosica.parser.node.typedeclaration.expression.ConditionalAndExpression;
import net.nosica.parser.node.typedeclaration.expression.ConditionalExpression;
import net.nosica.parser.node.typedeclaration.expression.ConditionalOrExpression;
import net.nosica.parser.node.typedeclaration.expression.EqualityExpression;
import net.nosica.parser.node.typedeclaration.expression.Expression;
import net.nosica.parser.node.typedeclaration.expression.ImpliesExpression;
import net.nosica.parser.node.typedeclaration.expression.InstanceOfExpression;
import net.nosica.parser.node.typedeclaration.expression.MultiplicativeExpression;
import net.nosica.parser.node.typedeclaration.expression.PowerExpression;
import net.nosica.parser.node.typedeclaration.expression.PrefixExpression;
import net.nosica.parser.node.typedeclaration.expression.PrimaryExpression;
import net.nosica.parser.node.typedeclaration.expression.PrimaryComponent;
import net.nosica.parser.node.typedeclaration.expression.PrimarySuffixImpl;
import net.nosica.parser.node.typedeclaration.expression.RelationalExpression;
import net.nosica.parser.node.typedeclaration.expression.StreamExpression;
import net.nosica.parser.node.typedeclaration.expression.UnaryExpression;
import net.nosica.parser.node.typedeclaration.expression.UnaryExpressionNotPlusMinus;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.ListIterator;

/**
 * \brief typecheck an expression
 *
 * an expression is type checked by visiting all sub nodes of each expression.<br>
 * For example, in <br>
 * <pre>
 * lhs = rhs<br>
 * </pre>
 * lhs is an expression that will be visited.<br>
 * This will yields to the type of the expression lhs.<br>
 * Then, rhs is an expression that will be visited.<br>
 * This will yields to the type of the expression rhs.<br>
 * <br>
 * When a node has finished evaluating an expression, it places the result of the expression
 * (its type) into a <b>TypeCheckerResultType</b>.<br>
 * Instance <i>resultType</i> in <b>TypeCheckVisitor</b> serves this purpose.<br>
 * <br>
 * So the algo is :<br>
 * For a node :<br>
 * <ul>
 *   <li>visit all sub nodes, and collect all expression type using instance <i>resultType</i>
 *   <li>type check the current node,
 *   <ul>
 *     <li>if there is an error
 *     <ul>
 *       <li>throws <b>TypeCheckerErrorExpressionException</b>
 *     </ul>
 *     <li>else
 *     <ul>
 *       <li>places the result of the type check in instance <i>resultType</i> so that other super nodes can use it.
 *     </ul>
 *   </ul>
 * </ul>
 * <br>
 * As you can see, if we detect an error, we throw a <b>TypeCheckerErrorExpressionException</b>.<br>
 * This exception is caught by the <b>StatementVisitor</b> at the upper level.<br>
 * This way, we can interrupt the type checking of an expression that yields to error and try with the
 * following expressions ...<br>
 */
public class TypeCheckVisitor extends AbstractNodeVisitor implements NodeVisitor
{
  private Environment env = Environment.instance();
  private Type type;                          // current type being parsed
  private Method method;                      // the method being parsed
  private AliasTable aliasTable;              // alias table to use to resolve names

  private UseTypeCallBack typeCallBack;       // call back used to notify a type needs to be type-checked

  private VariableEnvironment variableEnv;    // variable environment
  private LabelRepository labelRepository;    // labels
  private ExceptionEnvironment exceptionEnvironment;  // exceptions
  private IRTranslator IRT;                   // IR Translator for the method being parsed

  private boolean staticContext;              // are we in a static context ?
  private boolean lhsContext = false;         // are we in a lhs context ?
  private boolean methodConstContext;         // is method being parsed a const method ?
  private TypeCheckerResultType resultType;   // type of expression being type checked.
  private TypeCheckState state;               // indicates a special state to drive TypeCheckVisitor
  private net.nosica.compiler.intermediate.Expression alternativeExp;          // in relation with state

  private net.nosica.compiler.intermediate.Expression translatedExp;           // translated expression

  public TypeCheckVisitor(
    Type type,
    Method method,
    AliasTable aliasTable,
    VariableEnvironment variableEnv,
    LabelRepository labelRepository,
    ExceptionEnvironment exceptionEnvironment,
    IRTranslator IRT,
    UseTypeCallBack typeCallBack
  )
  {
    this.type = type;
    this.method = method;
    this.aliasTable = aliasTable;

    this.variableEnv = variableEnv;
    this.labelRepository = labelRepository;
    this.exceptionEnvironment = exceptionEnvironment;
    this.typeCallBack = typeCallBack;
    this.state = TypeCheckState.NONE;
    this.alternativeExp = null;
    this.translatedExp = null;

    this.IRT = IRT;
    this.staticContext = method.isStatic();
    this.methodConstContext = method.isConst();
  }

  public Type getType() { return type; }
  public Method getMethod() { return method; }
  public AliasTable getAliasTable() { return aliasTable; }

  public ExceptionEnvironment getEE() { return exceptionEnvironment; }
  public VariableEnvironment getVE() { return variableEnv; }
  public LabelRepository getLabelRepository() { return labelRepository; }
  public UseTypeCallBack getTypeCallBack() { return typeCallBack; }

  public IRTranslator getIRT() { return IRT; }
  public TypeCheckerResultType getResultType() { return resultType; }
  public net.nosica.compiler.intermediate.Expression getTranslatedExp() { return translatedExp; }

  public void setState(TypeCheckState state) { this.state = state; }
  public TypeCheckState getState() { return state; }
  public void noState()
  {
    this.state = TypeCheckState.NONE;
    this.alternativeExp = null;
  }
  public void setAlternativeExp(net.nosica.compiler.intermediate.Expression exp) { this.alternativeExp = exp; }

  /*****************************************************************************
   * ASSIGN OPERATOR
   ****************************************************************************/
  public void visit(AssignExpression assignExp)
  {
    // type check left expression
    lhsContext = true;
    assignExp.getLeft().accept(this);
    TypeCheckerResultType left = resultType;
    net.nosica.compiler.intermediate.Expression tLeft = translatedExp;

    // type check right expression
    lhsContext = false;
    assignExp.getRight().accept(this);

    TypeCheckerResultType right = resultType;
    net.nosica.compiler.intermediate.Expression tRight = translatedExp;

    // type checking
    if ((left.isSub() && !left.isPropertySet()) || right.isSub())
    {
      if (left.isSub())
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", assignExp.getLeft());
      if (right.isSub())
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", assignExp.getRight());

      throw new TypeCheckerErrorExpressionException();
    }

    boolean isFieldInitialisation = false;
    if (
         method.getKind() == MethodKindEnum.constructor &&
         tLeft instanceof net.nosica.compiler.intermediate.MemberAccess
       )
    {
      net.nosica.compiler.intermediate.MemberAccess field = (net.nosica.compiler.intermediate.MemberAccess) tLeft;
      isFieldInitialisation = field.hasSubscript() && field.getSubscript().isThis();
    }

    // assignment resolving
    MethodCallContext mcc = TypeCheckerHelper.findAssignOperator(
      this.type.getName(),
      left,
      right,
      assignExp.getOperator(),
      isFieldInitialisation,
      assignExp
    );

    if (mcc == null)
      resultType = left;
    else
    {
      resultType = createResultType(mcc);
    }

    if (left.isPropertySet())
    {
      // property set call
      ArrayList args = new ArrayList(left.getTranslatedPropertySetArguments());
      args.add(0, tLeft);
      args.add(tRight);
      translatedExp = IRT.translateMethodCall(mcc, args);
    }
    else
    {
      // reference assignment / primitive assignment
      translatedExp = IRT.translateAssignment(mcc, tLeft, tRight, false);
    }
  }

  private TypeCheckerResultType createResultType(MethodCallContext m)
  {
    if (m.getMethod().isSub())
      return new TypeCheckerResultTypeIsSub();
    else
      return new TypeCheckerResultTypeImpl(
        m.getMethod().getResultType().getTypeName(),
        m.getMethod().getResultType().isConst()
      );
  }

  /*****************************************************************************
   * CONDITIONAL EXPRESSION
   ****************************************************************************/
  public void visit(ConditionalExpression condExp)
  {
    // condition
    lhsContext = false;
    ArrayList deadCondVar = null;
    variableEnv.save();
    try
    {
      condExp.getCondition().accept(this);
    }
    finally
    {
      deadCondVar = variableEnv.restore();
    }
    TypeCheckerResultType cond = resultType;
    net.nosica.compiler.intermediate.Expression tCond = translatedExp;

    // true expression
    lhsContext = false;
    ArrayList deadTrueVar = null;
    variableEnv.save();
    try
    {
      condExp.getTrueExpression().accept(this);
    }
    finally
    {
      deadTrueVar = variableEnv.restore();
    }
    TypeCheckerResultType trueExp = resultType;
    net.nosica.compiler.intermediate.Expression tTrue = translatedExp;

    // false expression
    lhsContext = false;
    ArrayList deadFalseVar = null;
    variableEnv.save();
    try
    {
      condExp.getFalseExpression().accept(this);
    }
    finally
    {
      deadFalseVar = variableEnv.restore();
    }
    TypeCheckerResultType falseExp = resultType;
    net.nosica.compiler.intermediate.Expression tFalse = translatedExp;

    // condition must not resolve to a sub
    if (cond.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", condExp.getCondition());
      throw new TypeCheckerErrorExpressionException();
    }

    // condition must resolve to a boolean
    if (!cond.getTypeName().equals(NativeTypes.get_net_nosica_lang_bool()))
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck020", cond.getTypeName().toString(), condExp.getCondition());
      throw new TypeCheckerErrorExpressionException();
    }

    // both expressions are sub, that is ok
    if (trueExp.isSub() && falseExp.isSub())
    {
      resultType = trueExp;
      translatedExp = TypeCheckerHelper.translateSubConditionalExpression(tCond, tTrue, tFalse, deadCondVar, deadTrueVar, deadFalseVar, this);

      return;
    }

    // if both are non sub, this should be ok
    if (!trueExp.isSub() && !falseExp.isSub())
    {
      try
      {
        // OK only if they have a common parent
        TypeName nearestParent = env.getTE().getNearestParent(trueExp.getTypeName(), falseExp.getTypeName());

        if (nearestParent == null)
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck022", new String[]{trueExp.getTypeName().toString(), falseExp.getTypeName().toString()}, condExp);

          throw new TypeCheckerErrorExpressionException();
        }

        resultType = new TypeCheckerResultTypeImpl(nearestParent, trueExp.isConst() || falseExp.isConst());
        translatedExp = TypeCheckerHelper.translateConditionalExpression(nearestParent, tCond, tTrue, tFalse, deadCondVar, deadTrueVar, deadFalseVar, this);

        return;
      }
      catch (SymbolException e)
      {
        throw new TypeCheckerErrorExpressionException();
      }
      catch (GenericityException e)
      {
        throw new TypeCheckerErrorExpressionException();
      }
    }

    // types not compatible
    env.getMS().report(MessageLevel.ERR, "TypeCheck022", new String[]{trueExp.isSub() ? "sub" : trueExp.getTypeName().toString(), falseExp.isSub() ? "sub" : falseExp.getTypeName().toString()}, condExp);

    throw new TypeCheckerErrorExpressionException();
  }

  public void visit(ConditionalAndExpression condAndExp)
  {
    ArrayList args = new ArrayList();
    ArrayList temporaries = new ArrayList();
    boolean foundError = false;

    Iterator it = condAndExp.iterateExpressions();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();

      lhsContext = false;
      ArrayList deadVariables = null;
      variableEnv.save();
      try
      {
        exp.accept(this);
      }
      finally
      {
        deadVariables = variableEnv.restore();
      }

      args.add(translatedExp);
      temporaries.add(deadVariables);

      // there must be a type
      if (resultType.isSub())
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp);
        foundError = true;
        continue;
      }

      // it must be bool
      if (!resultType.getTypeName().equals(NativeTypes.get_net_nosica_lang_bool()))
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck020", resultType.getTypeName().toString(), exp);
        foundError = false;
      }
    }

    if (foundError)
      throw new TypeCheckerErrorExpressionException();

    resultType = new TypeCheckerResultTypeImpl(
      NativeTypes.get_net_nosica_lang_bool(), true
    );
    translatedExp = TypeCheckerHelper.translateConditionalAndExpression(args, temporaries, this);
  }

  public void visit(ConditionalOrExpression condOrExp)
  {
    ArrayList args = new ArrayList();
    ArrayList temporaries = new ArrayList();
    boolean foundError = false;

    Iterator it = condOrExp.iterateExpressions();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();

      lhsContext = false;
      ArrayList deadVariables = null;
      variableEnv.save();
      try
      {
        exp.accept(this);
      }
      finally
      {
        deadVariables = variableEnv.restore();
      }

      args.add(translatedExp);
      temporaries.add(deadVariables);

      // there must be a type
      if (resultType.isSub())
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp);
        foundError = true;

        continue;
      }

      // it must be bool
      if (!resultType.getTypeName().equals(NativeTypes.get_net_nosica_lang_bool()))
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck020", resultType.getTypeName().toString(), exp);
        foundError = false;
      }
    }

    if (foundError)
      throw new TypeCheckerErrorExpressionException();

    resultType = new TypeCheckerResultTypeImpl(
      NativeTypes.get_net_nosica_lang_bool(), true
    );
    translatedExp = TypeCheckerHelper.translateConditionalOrExpression(args, temporaries, this);
  }

  public void visit(ImpliesExpression implyExp)
  {
    boolean foundError = false;

    lhsContext = false;
    ArrayList lhsTemporaries = null;
    variableEnv.save();
    try
    {
      implyExp.getLhs().accept(this);
    }
    finally
    {
      lhsTemporaries = variableEnv.restore();
    }

    TypeCheckerResultType lhsExp = resultType;
    net.nosica.compiler.intermediate.Expression tLhs = translatedExp;

    // there must be a type
    if (lhsExp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", implyExp.getLhs());
      throw new TypeCheckerErrorExpressionException();
    }

    // it must be bool
    if (!lhsExp.getTypeName().equals(NativeTypes.get_net_nosica_lang_bool()))
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck020", lhsExp.getTypeName().toString(), implyExp.getLhs());
      foundError = true;
    }

    lhsContext = false;
    ArrayList rhsTemporaries = null;
    variableEnv.save();
    try
    {
      implyExp.getRhs().accept(this);
    }
    finally
    {
      rhsTemporaries = variableEnv.restore();
    }

    TypeCheckerResultType rhsExp = resultType;
    net.nosica.compiler.intermediate.Expression tRhs = translatedExp;

    // there must be a type
    if (rhsExp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", implyExp.getRhs());
      throw new TypeCheckerErrorExpressionException();
    }

    // it must be bool
    if (!rhsExp.getTypeName().equals(NativeTypes.get_net_nosica_lang_bool()))
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck020", lhsExp.getTypeName().toString(), implyExp.getRhs());
      foundError = true;
    }

    if (foundError)
      throw new TypeCheckerErrorExpressionException();

    resultType = new TypeCheckerResultTypeImpl(
      NativeTypes.get_net_nosica_lang_bool(),
      true
    );
    translatedExp = TypeCheckerHelper.translateImplyExpression(tLhs, tRhs, lhsTemporaries, rhsTemporaries, this);
  }

  /*****************************************************************************
   * DYNAMIC RUNTIME TYPE IDENTIFICATION
   ****************************************************************************/
  public void visit(CastExpression castExp)
  {
    lhsContext = false;
    castExp.getExpression().accept(this);

    TypeCheckerResultType castedExpression = resultType;
    net.nosica.compiler.intermediate.Expression tExp = translatedExp;

    // expression must not be sub
    if (castedExpression.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", castExp.getExpression());
      throw new TypeCheckerErrorExpressionException();
    }

    try
    {
      Type expType = env.getTL().getType(castedExpression.getTypeName());

      TypeName castTypeName = env.getTR().resolve(castExp.getTypeName(), aliasTable, env.getFL());
      Type castType = env.getTL().getType(castTypeName);
      typeCallBack.useType(castTypeName);

      // if both types are primitive
      if (castType.isPrimitive() && expType.isPrimitive())
      {
        // if they are the same type
        if (castTypeName.equals(castedExpression.getTypeName()))
        {
          resultType = new TypeCheckerResultTypeImpl(castTypeName, castExp.castIsConst());
          translatedExp = tExp;

          return;
        }

        // let's look for a cast operator
        Method cast = env.getMR().getCastOperator(castedExpression.getTypeName(), castTypeName);
        typeCallBack.useType(cast.getType());

        if (cast != null)
        {
          ArrayList args = new ArrayList();
          args.add(tExp);

          resultType = new TypeCheckerResultTypeImpl(castTypeName, castExp.castIsConst());
          translatedExp = IRT.translateMethodCall(
            new MethodCallContextImpl(
              castExp,
              cast
            ),
            args
          );

          return;
        }
      }

      TypeName nearestParent = env.getTE().getNearestParent(castedExpression.getTypeName(), castTypeName);

      // types are not compatible (a reference type and a primitive type)
      if (nearestParent == null)
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck023", new String[]{castedExpression.getTypeName().toString(), castTypeName.toString()}, castExp);
        throw new TypeCheckerErrorExpressionException();
      }

      // nearest parent must either be the casted expression or the cast's typeName
      if (!nearestParent.equals(castTypeName) && !nearestParent.equals(castedExpression.getTypeName()))
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck023", new String[]{castedExpression.getTypeName().toString(), castTypeName.toString()}, castExp);
        throw new TypeCheckerErrorExpressionException();
      }

      // constness, same as AssignExpression with two references
      if (!castExp.castIsConst() && castedExpression.isConst())
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck024", castExp);
        throw new TypeCheckerErrorExpressionException();
      }

      resultType = new TypeCheckerResultTypeImpl(castTypeName, castExp.castIsConst());
      translatedExp = IRT.translateCastExpression(castTypeName, tExp);
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
    catch (GenericityException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  public void visit(InstanceOfExpression instanceOfExp)
  {
    lhsContext = false;
    instanceOfExp.getExpression().accept(this);

    TypeCheckerResultType exp = resultType;
    net.nosica.compiler.intermediate.Expression tExp = translatedExp;

    // expression must not be sub
    if (exp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", instanceOfExp.getExpression());
      throw new TypeCheckerErrorExpressionException();
    }

    try
    {
      TypeName instanceTypeName = env.getTR().resolve(instanceOfExp.getTypeName(), aliasTable, env.getFL());
      env.getTL().getType(instanceTypeName);
      typeCallBack.useType(instanceTypeName);

      TypeName nearestParent = env.getTE().getNearestParent(exp.getTypeName(), instanceTypeName);

      // types are not compatible (a reference type and a primitive type)
      if (nearestParent == null)
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck025", new String[]{exp.getTypeName().toString(), instanceTypeName.toString()}, instanceOfExp);
        throw new TypeCheckerErrorExpressionException();
      }

      // nearest parent must either be the casted expression or the cast's typeName
      if (!nearestParent.equals(instanceTypeName) && !nearestParent.equals(exp.getTypeName()))
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck025", new String[]{exp.getTypeName().toString(), instanceTypeName.toString()}, instanceOfExp);
        throw new TypeCheckerErrorExpressionException();
      }

      resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_bool(), true);
      translatedExp = IRT.translateInstanceOfExpression(instanceTypeName, tExp);
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
    catch (GenericityException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  /*****************************************************************************
   * UNARY EXPRESSION
   ****************************************************************************/
  public void visit(PrefixExpression prefixExp)
  {
    lhsContext = false;
    prefixExp.getExpression().accept(this);

    TypeCheckerResultType exp = resultType;
    net.nosica.compiler.intermediate.Expression tExp = translatedExp;

    // expression must not be sub
    if (exp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", prefixExp.getExpression());
      throw new TypeCheckerErrorExpressionException();
    }

    String name = (prefixExp.isIncrement() ? "++" : "--");

    Method signature = new MethodImpl(false, name, MethodKindEnum.prefix, new ArrayList(), exp.isConst(), null);

    ArrayList matchingNames = new ArrayList();
    ArrayList matchingMethods = new ArrayList();

    try
    {
      // search matching methods
      MethodCallContext match = env.getMR().findMatchingMethod(
        resultType,
        signature,
        this.type.getName(),
        matchingMethods,
        matchingNames
      );

      // report error
      if (match == null)
      {
        TypeCheckerHelper.reportMethodError(resultType, signature, matchingMethods, matchingNames, prefixExp);

        throw new TypeCheckerErrorExpressionException();
      }
      else
      {
        resultType = createResultType(match);

        ArrayList args = new ArrayList();
        args.add(tExp);
        translatedExp = IRT.translateMethodCall(match, args);
      }
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  public void visit(UnaryExpression unaryExp)
  {
    // if expression following minus is a literal, then handle it differently !
    Literal literal = null;

    if (unaryExp.getUnaryExpression() instanceof PrimaryExpression)
    {
      PrimaryExpression primary = (PrimaryExpression)unaryExp.getUnaryExpression();
      Iterator it = primary.iterateComponents();
      if (it.hasNext())
      {
        PrimaryComponent primaryComponent = (PrimaryComponent)it.next();
        if (primaryComponent.isLiteral())
          literal = primaryComponent.getLiteral();
      }
    }

    if (unaryExp.isMinus() && literal != null)
    {
      // we have someting like : "-2"
      // instead of interpreting it like 2.operator -()
      // we will just reverse the notation !
      if (
          literal.getLiteralType() == LiteralEnum.floatLiteral ||
          literal.getLiteralType() == LiteralEnum.integerLiteral
         )
        literal.invert();
    }

    lhsContext = false;
    unaryExp.getUnaryExpression().accept(this);

    TypeCheckerResultType exp = resultType;
    net.nosica.compiler.intermediate.Expression tExp = translatedExp;

    // expression must not be sub
    if (exp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", unaryExp.getUnaryExpression());
      throw new TypeCheckerErrorExpressionException();
    }

    if (unaryExp.isMinus())
    {
      Method signature = new MethodImpl(false, "-", MethodKindEnum.prefix, new ArrayList(), exp.isConst(), null);

      ArrayList matchingNames = new ArrayList();
      ArrayList matchingMethods = new ArrayList();

      try
      {
        // search matching methods
        MethodCallContext match = env.getMR().findMatchingMethod(
          resultType,
          signature,
          this.type.getName(),
          matchingMethods,
          matchingNames
        );

        // report error
        if (match == null)
        {
          TypeCheckerHelper.reportMethodError(resultType, signature, matchingMethods, matchingNames, unaryExp);

          throw new TypeCheckerErrorExpressionException();
        }
        else
        {
          resultType = createResultType(match);

          ArrayList args = new ArrayList();
          args.add(tExp);
          translatedExp = IRT.translateMethodCall(match, args);
        }
      }
      catch (SymbolException e)
      {
        throw new TypeCheckerErrorExpressionException();
      }
    }
  }

  public void visit(UnaryExpressionNotPlusMinus unaryExp)
  {
    lhsContext = false;
    unaryExp.getNotExpression().accept(this);

    TypeCheckerResultType exp = resultType;
    net.nosica.compiler.intermediate.Expression tExp = translatedExp;

    // expression must not be sub
    if (exp.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", unaryExp.getNotExpression());
      throw new TypeCheckerErrorExpressionException();
    }

    Method signature = new MethodImpl(false, "!", MethodKindEnum.prefix, new ArrayList(), exp.isConst(), null);

    ArrayList matchingNames = new ArrayList();
    ArrayList matchingMethods = new ArrayList();

    try
    {
      // search matching methods
      MethodCallContext match = env.getMR().findMatchingMethod(
        resultType,
        signature,
        this.type.getName(),
        matchingMethods,
        matchingNames
      );

      // report error
      if (match == null)
      {
        TypeCheckerHelper.reportMethodError(resultType, signature, matchingMethods, matchingNames, unaryExp);

        throw new TypeCheckerErrorExpressionException();
      }
      else
      {
        resultType = createResultType(match);

        ArrayList args = new ArrayList();
        args.add(tExp);
        translatedExp = IRT.translateMethodCall(match, args);
      }
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  /*****************************************************************************
   * PRIMARY EXPRESSION
   ****************************************************************************/
  public void visit(PrimaryExpression primaryExp)
  {
    ArrayList components = new ArrayList(primaryExp.getComponents());

    // in toto.f().g().h();
    // initial lhsContext can be either true or false, BUT
    // when typechecking 'toto', then '.f()', then '.g()', lhs context is
    // false ! That is we are in a rhs context.
    // for the last term, here '.h()', we restore lhsContext to its saved value
    boolean savedLhsContext = lhsContext;

    visitPrimaryExpression(components, savedLhsContext);
  }

  public void visitPrimaryExpression(ArrayList components, boolean savedLhsContext)
  {
    while (!components.isEmpty())
    {
      // act as a stack : get first element and remove it from stack
      PrimaryComponent component = (PrimaryComponent)components.get(0);
      components.remove(0);

      if (components.isEmpty())
      {
        // means we are analysing last component, restore lhsContext
        lhsContext = savedLhsContext;
      }
      else
      {
        lhsContext = false;
      }

      if (component.isLiteral()) // literal (*)
      {
        createLiteral(component.getLiteral());
      }
      else
      if (component.isThis())        // access from this (*)
      {
        createThis(component);
      }
      else
      if (component.isSuper())       // access to super (*)
      {
        createSuper(component);
      }
      else
      if (component.isExpression())  // expression (*)
      {
        component.getExpression().accept(this);
      }
      else
      if (component.isAllocation())  // allocation (*)
      {
        component.getAllocation().accept(this);
      }
      else
      if (component.isTypeName())    // typeName : a type, a variable, or this (*)
      {
        checkTypeName(component, components);
        // this one can add elements in primaryComponents !
        if (!components.isEmpty())
        {
          lhsContext = false;
        }
      }
      else
      if (component.isMember())      // a field, or a method
      {
        checkMember(component, components, savedLhsContext);
      }
      else
      if (component.isArrayAccess()) // anonymous property array
      {
        checkAnonymousArray(component);
      }
      else
      if (component.isArguments())
      {
        checkPrimitiveConstructor(component);
      }
      else
      {
        env.terminate("Internal error : cannot handle primaryExpression");
      }

      // require the use of the current type
      if (!resultType.isPropertySet() && !resultType.isSub())
      {
        typeCallBack.useType(resultType.getTypeName());
      }
    }
  }

  private void createLiteral(Literal literal)
  {
    try
    {
      LiteralEnum kind = literal.getLiteralType();

      if (literal == LiteralEnum.charLiteral)
      {
        env.getTL().getType(NativeTypes.get_net_nosica_lang_char());
        typeCallBack.useType(NativeTypes.get_net_nosica_lang_char());

        resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_char(), true);
        translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
        return;
      }
      else
      if (kind == LiteralEnum.floatLiteral)
      {
        Double d = new Double(literal.getString());

        if (Math.abs(d.doubleValue()) <= 3.40282347e+38)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_float32());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_float32());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_float32(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_float64());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_float64());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_float64(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
      }
      else
      if (kind == LiteralEnum.stringLiteral)
      {
        env.getTL().getType(NativeTypes.get_net_nosica_lang_string());
        typeCallBack.useType(NativeTypes.get_net_nosica_lang_string());

        resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_string(), false);
        translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
        return;
      }
      else
      if (kind == LiteralEnum.charLiteral)
      {
        env.getTL().getType(NativeTypes.get_net_nosica_lang_char());
        typeCallBack.useType(NativeTypes.get_net_nosica_lang_char());

        resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_char(), false);
        translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
        return;
      }
      else
      if (kind == LiteralEnum.integerLiteral)
      {
        long lValue = Long.parseLong(literal.getString());

        if (lValue >= 0 && lValue <= 255)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_word8());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_word8());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_word8(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= -128 && lValue <= 127)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_int8());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_int8());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_int8(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= 0 && lValue <= 65535)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_word16());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_word16());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_word16(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= -32768 && lValue <= 32767)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_int16());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_int16());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_int16(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= 0 && lValue <= 4294967295L)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_word32());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_word32());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_word32(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= -2147483648L && lValue <= 2147483647L)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_int32());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_int32());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_int32(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= 0)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_word64());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_word64());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_word64(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        if (lValue >= -9223372036854775808L && lValue <= 9223372036854775807L)
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_int64());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_int64());

          resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_int64(), true);
          translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
          return;
        }
        else
        {
          env.terminate("Internal error");
          return;
        }
      }
      else
      if (kind == LiteralEnum.booleanLiteral)
      {
        env.getTL().getType(NativeTypes.get_net_nosica_lang_bool());
        typeCallBack.useType(NativeTypes.get_net_nosica_lang_bool());

        resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_bool(), true);
        translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
        return;
      }
      else
      if (kind == LiteralEnum.nullLiteral)
      {
        env.getTL().getType(NativeTypes.get_net_nosica_lang_None());
        typeCallBack.useType(NativeTypes.get_net_nosica_lang_None());

        resultType = new TypeCheckerResultTypeImpl(NativeTypes.get_net_nosica_lang_None(), false);
        translatedExp = IRT.translateLiteral(resultType.getTypeName(), literal.getString());
        return;
      }

      env.terminate("Internal error : could not resolve literal type");
      return;
    }
    catch (GenericityException e)
    {
      env.terminate("Internal error");
      return;
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error");
      return;
    }
  }

  private void createThis(Node localizable)
  {
    if (staticContext)
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck004", "this", localizable);
      throw new TypeCheckerErrorExpressionException();
    }

    resultType = new TypeCheckerResultTypeImpl(this.type.getName(), methodConstContext);
    translatedExp = IRT.translateThis();
  }

  private void createSuper(Node localizable)
  {
    if (staticContext)
      env.getMS().report(MessageLevel.ERR, "TypeCheck004", "super", localizable);

    TypeName superName = null;

    Iterator j = type.iterateExtends();
    if (j.hasNext())
      superName = (TypeName)j.next();
    else
      env.terminate("Internal error : there should be an extend to this type");

    if (j.hasNext())
      env.terminate("Internal error : there should no more than one extend for this type");

    resultType = new TypeCheckerResultTypeImpl(superName, methodConstContext);
    translatedExp = IRT.translateSuper();
  }

  private void checkTypeName(PrimaryComponent component, ArrayList components)
  {
    // this can be a variable, the beginning of a type followed by one or more field, or a succession of field
    TypeName typeName = component.getTypeName();

    if (typeName.isArray())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck006", component);
      throw new TypeCheckerErrorExpressionException();
    }

    // here we have a non array type name !
    Iterator j = typeName.iterateComponents();

    // initialize component
    TypeNameComponent firstComponent = (TypeNameComponent)typeName.getComponents().get(0);
    String name = firstComponent.getPackageElement();

    // try first with a variable
    Variable var = variableEnv.get(name);

    if (var != null)
    {
      // It is a variable !

      // check there is no generic information
      if (firstComponent.hasGenericExtension())
        env.getMS().report(MessageLevel.ERR, "TypeCheck002", firstComponent);

      // construct result
      resultType = new TypeCheckerResultTypeImpl(var.getTypeName(), var.isConst());
      translatedExp = IRT.translateVariable(var);

      // eat component
      j.next();
    }
    else
    {
      // Not a variable : try with a type !
      ArrayList partialComponents = new ArrayList();
      TypeName lastKnownGoodType = null;
      ArrayList savedOriginalName = new ArrayList();

      while (j.hasNext())
      {
        TypeNameComponent typeNameComponent = (TypeNameComponent)j.next();

        partialComponents.add(typeNameComponent);
        TypeName partialName = new TypeNameImpl(partialComponents);

        try
        {
          partialName = env.getTR().resolve(partialName, aliasTable, env.getFL());
          env.getTL().getType(partialName);
          typeCallBack.useType(partialName);

          // type exist : save state !
          lastKnownGoodType = partialName;
          savedOriginalName.clear();
          savedOriginalName.addAll(partialComponents);
        }
        catch (SymbolException exx)
        {
        }
        catch (GenericityException exx)
        {
        }
      }

      // lastKnownGoodType denotes the last known type
      if (lastKnownGoodType != null)
      {
        // reset iterators to a known value
        j = typeName.iterateComponents();
        // we can't use lastKnownGoodType, because it has been resolved
        // (so number of component could have increased
        Iterator k = savedOriginalName.iterator();
        while (k.hasNext())
        {
          k.next();
          j.next();
        }

        // now there MUST be something following the type
        if (!j.hasNext())
        {
          if (components.isEmpty())
          {
            env.getMS().report(MessageLevel.ERR, "TypeCheck001", component);
            throw new TypeCheckerErrorExpressionException();
          }

          PrimaryComponent nextComponent = (PrimaryComponent)components.get(0);

          if (!nextComponent.isArguments())
          {
            env.getMS().report(MessageLevel.ERR, "TypeCheck001", component);
            throw new TypeCheckerErrorExpressionException();
          }

          // means we are calling a constructor ...
          // means type MUST be primitive because this is only legal on a primitive type
          try
          {
            Type primitiveType = env.getTE().getType(lastKnownGoodType);

            if (!primitiveType.isPrimitive())
            {
              env.getMS().report(MessageLevel.ERR, "TypeCheck003", lastKnownGoodType.toString(), component);
              throw new TypeCheckerErrorExpressionException();
            }
          }
          catch (SymbolException ex)
          {
            env.getMS().report(MessageLevel.ERR, "Symbol031", lastKnownGoodType.toString() + " (" + ex.getMessage() + ")", component);
            throw new TypeCheckerErrorExpressionException();
          }

          // everything's fine ...
          resultType = new TypeCheckerResultTypeStatic(lastKnownGoodType);
          typeCallBack.useType(lastKnownGoodType);
          translatedExp = null;
        }
        else
        {
          resultType = new TypeCheckerResultTypeStatic(lastKnownGoodType);
          typeCallBack.useType(lastKnownGoodType);
          translatedExp = null;
        }
      }
      else
      {
        // no type : there must be a succession of fields
        // reset iterator
        j = typeName.iterateComponents();

        // create resultType for member access
        if (staticContext)
        {
          resultType = new TypeCheckerResultTypeImpl(type.getName(), false);
          translatedExp = null;
        }
        else
          createThis(component);
      }
    }

    // finally transforms all remaining components in components
    // that can be examined by checkMember
    int pos = 0;
    while (j.hasNext())
    {
      TypeNameComponent typeNameComponent = (TypeNameComponent)j.next();

      if (typeNameComponent.hasGenericExtension())
        env.getMS().report(MessageLevel.ERR, "TypeCheck002", component);

      PrimaryComponent primarySuffix = new PrimarySuffixImpl(
        typeNameComponent.localization(),
        typeNameComponent.getPackageElement()
      );
      components.add(pos++, primarySuffix);
    }
  }

  private void checkMember(PrimaryComponent component, ArrayList components, boolean savedLhsContext)
  {
    // a member can be a field or a method call
    // performs a lookahead to take decision cleanly.
    boolean isField = false;
    boolean isMethod = false;

    // no following component : can only be a field
    if (components.isEmpty())
      isField = true;
    else
    {
      // there is a following component
      PrimaryComponent nextComponent = (PrimaryComponent)components.get(0);
      if (nextComponent.isArguments())
        isMethod = true;
      else
        isField = true;
    }

    // lookahead done : check appropriate member
    if (isField)
      checkField(component);
    else
    if (isMethod)
    {
      PrimaryComponent args = (PrimaryComponent)components.get(0);
      components.remove(0);

      if (components.isEmpty())
      {
        lhsContext = savedLhsContext;
      }

      checkMethod(component, args);
    }
    else
      env.terminate("Internal error : should never go here");
  }

  /**
   * \brief check a field
   * By field, it can be : a real field, a property set, or a property get
   */
  private void checkField(PrimaryComponent component)
  {
    // check field access, we also tries to find for property here
    String member = component.getMember();
    ArrayList possibleFields = new ArrayList();

    // result type must not be sub
    if (resultType.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", component);
      throw new TypeCheckerErrorExpressionException();
    }

    try
    {
      // try first with a field
      FieldContext fieldContext = env.getMR().findField(
        resultType,
        member,
        this.type.getName(),
        possibleFields
      );

      ArrayList matchingNames = new ArrayList();
      ArrayList matchingMethods = new ArrayList();

      // now with properties
      if (lhsContext)
      {
        // property set
        matchingMethods.addAll(
          env.getMR().findPropertySetMethods(
            resultType,
            member,
            this.type.getName(),
            matchingNames
          )
        );

        if (
            (fieldContext != null && matchingMethods.size() != 0) ||
            (fieldContext == null && matchingMethods.isEmpty())
           )
        {
          // several possible match or
          // no match at all
          // report error
          reportFieldError(member, fieldContext, matchingMethods, matchingNames, component);

          throw new TypeCheckerErrorExpressionException();
        }

        // a field match
        if (fieldContext != null)
        {
          resultType = new TypeCheckerResultTypeImpl(fieldContext.getTypeName(), fieldContext.isConst());
          translatedExp = IRT.translateMemberAccess(translatedExp, fieldContext);

          return;
        }
        else
        {
          // one or more set properties match
          resultType = new TypeCheckerResultTypeSetProperty(
            matchingMethods,
            new ArrayList(),
            new ArrayList()
          );
          // do not touch translatedExp, it will be modified in next AssignExpression

          return;
        }
      }
      else
      {
        // construct property get method
        Method signature = new MethodImpl(false, member, MethodKindEnum.property, new ArrayList(), resultType.isConst(), null);

        matchingMethods.addAll(env.getMR().findMatchingMethods(
                                        resultType,
                                        signature,
                                        this.type.getName(),
                                        matchingNames
        ));
      }

      MethodCallContext match = env.getMR().getBestMethod(matchingMethods);

      if (
          (fieldContext != null && match != null) ||
          (fieldContext == null && match == null)
         )
      {
        reportFieldError(member, fieldContext, matchingMethods, matchingNames, component);

        throw new TypeCheckerErrorExpressionException();
      }
      else
      {
        if (fieldContext != null)
        {
          resultType = new TypeCheckerResultTypeImpl(fieldContext.getField().getTypeName(), fieldContext.getField().isConst());
          translatedExp = IRT.translateMemberAccess(translatedExp, fieldContext);
        }
        else
        {
          resultType = createResultType(match);

          ArrayList args = new ArrayList();
          args.add(translatedExp);
          translatedExp = IRT.translateMethodCall(match, args);
        }
      }
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  private void reportFieldError(
    String member,
    FieldContext fieldContext,
    ArrayList matchingMethods,
    ArrayList matchingNames,
    Localizable localizable
  )
  {
    if (fieldContext == null && matchingMethods.isEmpty())
    {
      if (matchingNames.isEmpty())
        env.getMS().report(MessageLevel.ERR, "TypeCheck026", member, localizable);
      else
      {
        StringBuffer s = new StringBuffer();
        Iterator it = matchingNames.iterator();
        while (it.hasNext())
        {
          Method m = (Method)it.next();
          s.append(m.toString());
          if (it.hasNext())
            s.append(", ");
        }
        env.getMS().report(MessageLevel.ERR, "TypeCheck027", new String[]{member, s.toString()}, localizable);
      }
    }
    else
    if (fieldContext == null && !matchingMethods.isEmpty())
    {
      // several properties are matching
      StringBuffer s = new StringBuffer();
      Iterator it = matchingNames.iterator();
      while (it.hasNext())
      {
        Method m = (Method)it.next();
        s.append(m.toString());
        if (it.hasNext())
          s.append(", ");
      }
      env.getMS().report(MessageLevel.ERR, "TypeCheck028", new String[]{member, s.toString()}, localizable);
    }
    else
    if (fieldContext != null && !matchingMethods.isEmpty())
    {
      // several properties are matching, and a field is matching too
      StringBuffer s = new StringBuffer();
      Iterator it = matchingNames.iterator();
      while (it.hasNext())
      {
        Method m = (Method)it.next();
        s.append(m.toString());
        if (it.hasNext())
          s.append(", ");
      }
      env.getMS().report(MessageLevel.ERR, "TypeCheck029", new String[]{member, s.toString()}, localizable);
    }
  }

  private void checkMethod(PrimaryComponent component, PrimaryComponent arguments)
  {
    String member = component.getMember();

    ArrayList args = new ArrayList();           // ArrayList<TypeCheckerResultType>
    ArrayList argsTranslated = new ArrayList(); // ArrayList<Expression>
    visitArguments(arguments.getArguments(), args, argsTranslated);

    // resultType cannot be sub (check instance)
    if (resultType.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", component);
      throw new TypeCheckerErrorExpressionException();
    }

    Method signature = new MethodImpl(resultType.isStatic(), member, MethodKindEnum.method, args, resultType.isConst(), null);

    MethodCallContext match = findMethod(resultType, signature, component);

    resultType = createResultType(match);

    argsTranslated.add(0, translatedExp);
    translatedExp = IRT.translateMethodCall(match, argsTranslated);
  }

  /**
   * \brief check for an anonymous array
   */
  private void checkAnonymousArray(PrimaryComponent component)
  {
    // get arguments
    TypeCheckerResultType savedResultType = resultType;
    boolean savedLhsContext = lhsContext;
    net.nosica.compiler.intermediate.Expression savedTranslatedExp = translatedExp;

    ArrayList args = new ArrayList();            // ArrayList<Arguments>
    ArrayList argsTranslated = new ArrayList();  // ArrayList<Expression>
    ArrayList argsTemporaries = new ArrayList(); // ArrayList<Variable>
    Iterator it = component.iterateDimensions();
    while (it.hasNext())
    {
      Expression exp = (Expression)it.next();
      lhsContext = false;
      ArrayList deadVariables = null;  // ArrayList<Variable>
      variableEnv.save();
      try
      {
        exp.accept(this);
      }
      finally
      {
        deadVariables = variableEnv.restore();
      }

      args.add(new ArgumentImpl(resultType.isConst(), resultType.getTypeName(), ""));
      argsTranslated.add(translatedExp);
      argsTemporaries.add(deadVariables);
    }
    lhsContext = savedLhsContext;
    resultType = savedResultType;
    translatedExp = savedTranslatedExp;

    // resultType cannot be sub
    if (resultType.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", component);
      throw new TypeCheckerErrorExpressionException();
    }

    if (lhsContext)
    {
      try
      {
        ArrayList matchingNames = new ArrayList();
        ArrayList matchingMethods = env.getMR().findPropertySetArrayMethods(
          resultType,
          null,
          args.size() + 1,
          this.type.getName(),
          matchingNames
        );

        // a property must match
        if (matchingMethods.isEmpty())
        {
          // reportMethodError(signature, matchingMethods, matchingNames, node);
          throw new TypeCheckerErrorExpressionException();
        }

        resultType = new TypeCheckerResultTypeSetProperty(
          matchingMethods,
          args,
          argsTranslated
        );
        // translatedExp will be resolved by next call to AssignExpression
      }
      catch (SymbolException e)
      {
        throw new TypeCheckerErrorExpressionException();
      }
    }
    else
    {
      Method signature = new MethodImpl(false, null, MethodKindEnum.propertyArray, args, resultType.isConst(), null);

      MethodCallContext match = findMethod(resultType, signature, component);
      resultType = createResultType(match);

      argsTranslated.add(0, translatedExp);
      translatedExp = IRT.translateMethodCall(match, argsTranslated);
    }
  }

  private void checkPrimitiveConstructor(PrimaryComponent arguments)
  {
    // get arguments
    ArrayList args = new ArrayList();
    ArrayList argsTranslated = new ArrayList();
    visitArguments(arguments.getArguments(), args, argsTranslated);

    if (!resultType.isStatic())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck030", arguments);

      throw new TypeCheckerErrorExpressionException();
    }

    Method signature = new MethodImpl(false, "constructor", MethodKindEnum.constructor, args, false, null);

    MethodCallContext match = findMethod(resultType, signature, arguments);

    resultType = new TypeCheckerResultTypeImpl(resultType.getTypeName(), false);

    if (state == TypeCheckState.PARSING_VARIABLE_INITIALIZER)
    {
      translatedExp = IRT.translatePrimitiveConstruction(alternativeExp, match, argsTranslated);
      state = TypeCheckState.PARSED_PRIMITIVE_CONSTRUCTION;
    }
    else
      translatedExp = IRT.translatePrimitiveConstruction(match, argsTranslated);
  }

  public MethodCallContext findMethod(ResultType instance, Method signature, Node node)
  {
    try
    {
      ArrayList matchingNames = new ArrayList();
      ArrayList matchingMethods = new ArrayList();

      MethodCallContext match = env.getMR().findMatchingMethod(
        instance,
        signature,
        this.type.getName(),
        matchingMethods,
        matchingNames
      );

      if (match == null)
      {
        TypeCheckerHelper.reportMethodError(instance, signature, matchingMethods, matchingNames, node);
        throw new TypeCheckerErrorExpressionException();
      }

      Iterator it = match.getMethod().iterateThrowList();
      while (it.hasNext())
      {
        TypeName typeName = (TypeName)it.next();

        try
        {
          env.getTL().getType(typeName);
          typeCallBack.useType(typeName);
        }
        catch (GenericityException e)
        {
          env.terminate("Internal error " + e.getMessage());
        }

        checkException(typeName, match.getLocalizable());
      }

      return match;
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  /**
   */
  public void visitArguments(Arguments arguments, ArrayList argsResultType, ArrayList argsTranslatedExp)
  {
    // get arguments
    TypeCheckerResultType savedTypeCheckerResultType = resultType;
    boolean savedLhsContext = lhsContext;
    net.nosica.compiler.intermediate.Expression savedTranslatedExp = translatedExp;

    if (arguments.hasArgumentList())
    {
      Iterator it = arguments.getArgumentList().iterateArguments();
      while (it.hasNext())
      {
        Expression exp = (Expression)it.next();
        lhsContext = false;
        exp.accept(this);

        // check argument is not sub
        if (resultType.isSub())
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp);
          throw new TypeCheckerErrorExpressionException();
        }

        argsResultType.add(new ArgumentImpl(resultType.isConst(), resultType.getTypeName(), ""));
        argsTranslatedExp.add(translatedExp);
      }
    }

    resultType = savedTypeCheckerResultType;
    lhsContext = savedLhsContext;
    translatedExp = savedTranslatedExp;
  }

  public void visit(AllocationExpression allocExp)
  {
    TypeName typeName = allocExp.getType();

    try
    {
      typeName = env.getTR().resolve(typeName, aliasTable, env.getFL());
    }
    catch (SymbolException e)
    {
      env.getMS().report(MessageLevel.ERR, "Symbol031", typeName.toString() + " (" + e.getMessage() + ")", allocExp);
      throw new TypeCheckerErrorExpressionException();
    }

    Type allocation = null;

    try
    {
      allocation = env.getTL().getType(typeName);
    }
    catch (SymbolException e)
    {
      env.getMS().report(MessageLevel.ERR, "Symbol031", typeName.toString() + " (" + e.getMessage() + ")", allocExp);
      throw new TypeCheckerErrorExpressionException();
    }
    catch (GenericityException e)
    {
      env.getMS().report(MessageLevel.ERR, "Symbol031", typeName.toString() + " (" + e.getMessage() + ")", allocExp);
      throw new TypeCheckerErrorExpressionException();
    }

    try
    {
      typeCallBack.useType(typeName);

      // handle anonymous class if any
      if (allocation.isInterface() && allocExp.hasClassBody())
      {
        if (staticContext)
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck031", allocExp);
          throw new TypeCheckerErrorExpressionException();
        }

        allocation = env.getTL().getType(this.type.getName(), typeName, allocExp.getClassBody());
        typeName = allocation.getName();

        typeCallBack.useType(typeName);
      }

      if (allocExp.hasArguments())
      {
        if (allocation.isInterface() || (allocation.isClass() && allocation.isAbstract()))
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck032", typeName.toString(), allocExp);
          throw new TypeCheckerErrorExpressionException();
        }

        // get arguments
        ArrayList args = new ArrayList();
        ArrayList argsTranslated = new ArrayList();
        visitArguments(allocExp.getArguments(), args, argsTranslated);

        Method signature = new MethodImpl(
          false,
          "constructor",
          MethodKindEnum.constructor,
          args,
          false,
          null
        );

        // this will find and/or report error on constructor
        MethodCallContext match = findMethod(new ResultTypeImpl(typeName, false), signature, allocExp);
        resultType = new TypeCheckerResultTypeImpl(typeName, false);
        translatedExp = IRT.translateAllocation(match, argsTranslated);
      }
      else
      if (allocExp.hasArrayDims())
        checkArrayDims(allocExp.getArrayDims(), typeName);
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
    catch (GenericityException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  /**
   * \brief check array dims creation
   *
   * e.g : int32[] i = new int32[10];
   * ==> checkArrayDims(arrayDims, int32);
   * with arrayDims = [10];
   *
   * \param arrayDims instance of ArrayDims
   * \param typeName the typeName in the "new TypeName" expression
   */
  private void checkArrayDims(ArrayDimsAndInits arrayDims, TypeName typeName)
  {
    // ( ( "[" Expression ( "," Expression )* "]" )+ ( "[" ( "," )* "]" )* | ( "[" ( "," )* "]" )+ ArrayInitializer )
    if (arrayDims.hasInitializedDims())
    {
      // ( "[" Expression ( "," Expression )* "]" )+ ( "[" ( "," )* "]" )*
      // parser has read
      // TypeName[,][] like [,] then []
      // if we construct typeName this way, we will construct
      // 2 - typeName, then 1, 2 - typeName, in the exact reverse order !
      // so we need to construct type in the reverse order of the parsing ...
      ListIterator it = arrayDims.reverseIterateInitializedDims();

      ArrayList typeNames = new ArrayList();      // ArrayList<TypeName>
      ArrayList args = new ArrayList();           // ArrayList<ArrayList<Argument>>
      ArrayList argsTranslated = new ArrayList(); // ArrayList<ArrayList<Expression>>

      try
      {
        Type baseType = env.getTE().getType(typeName);
        if (baseType.isPrimitive())
        {
          typeNames.add(typeName);  // add primitive array element type to list of types
          args.add(new ArrayList());
          argsTranslated.add(new ArrayList());
        }
      }
      catch (SymbolException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }

      while (it.hasPrevious())
      {
        // ( "[" Expression ( "," Expression )* "]" )+
        ArrayList exps = (ArrayList)it.previous();
        Iterator j = exps.iterator();

        ArrayList arguments = new ArrayList();  // ArrayList<TypeCheckerResultType>
        ArrayList argumentsTranslated = new ArrayList(); // ArrayList<Expression>

        while (j.hasNext())
        {
          Expression exp = (Expression)j.next();

          exp.accept(this);

          // check argument is not sub
          if (resultType.isSub())
          {
            env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp);
            throw new TypeCheckerErrorExpressionException();
          }

          arguments.add(new ArgumentImpl(resultType.isConst(), resultType.getTypeName(), ""));
          argumentsTranslated.add(translatedExp);


          if (!NativeTypes.isPositiveInteger(resultType.getTypeName()))
          {
            env.getMS().report(MessageLevel.ERR, "TypeCheck033", resultType.getTypeName().toString(), arrayDims);
            throw new TypeCheckerErrorExpressionException();
          }
        }

        typeName = new TypeNameImpl(exps.size(), typeName);
        typeNames.add(0, typeName);
        args.add(0, arguments);
        argsTranslated.add(0, argumentsTranslated);
      }

      // we have added in typeNames TypeName like :
      // int[,], int[], int

      ArrayList constructors = new ArrayList();

      Iterator i = typeNames.iterator();
      Iterator j = args.iterator();
      while (i.hasNext() && j.hasNext())
      {
        TypeName tn = (TypeName)i.next();
        ArrayList arguments = (ArrayList)j.next();

        try
        {
          // now load type
          env.getTL().getType(tn);
          typeCallBack.useType(tn);

          Method signature = new MethodImpl(
            false,
            "constructor",
            MethodKindEnum.constructor,
            arguments,
            false,
            null
          );

          MethodCallContext mcc = findMethod(new ResultTypeImpl(tn, false), signature, arrayDims);

          constructors.add(mcc);
        }
        catch (SymbolException e)
        {
          env.getMS().report(MessageLevel.ERR, "Symbol031", typeName.toString() + " ("+e.getMessage()+")", arrayDims);
        }
        catch (GenericityException e)
        {
          env.getMS().report(MessageLevel.ERR, "Symbol031", typeName.toString() + " ("+e.getMessage()+")", arrayDims);
        }
      }

      resultType = new TypeCheckerResultTypeImpl(typeName, false);
      translatedExp = IRT.translateArrayAllocation(constructors, argsTranslated);

      it = arrayDims.reverseIterateUnitializedDims();

      while (it.hasPrevious())
      {
        // ( "[" ( "," )* "]" )*
        Integer dim = (Integer)it.previous();
        typeName = new TypeNameImpl(dim.intValue(), typeName);
      }
    }
    else
    {
      // ( "[" ( "," )* "]" )+ ArrayInitializer )
      ListIterator it = arrayDims.reverseIterateDimensions();
      while (it.hasPrevious())
      {
        // ( "[" ( "," )* "]" )+
        Integer dim = (Integer)it.previous();
        typeName = new TypeNameImpl(dim.intValue(), typeName);
      }

      arrayDims.getArrayInitializer().accept(this);
      // @TODO@ : needs to verify type returned is OK with typeName
      throw new UnsupportedOperationException();
    }
  }

  /*****************************************************************************
   * INITIALIZER
   ****************************************************************************/
  public void visit(ArrayInitializer arrayInit)
  {
    if (arrayInit.hasInitializers())
    {
      Iterator it = arrayInit.iterateInitializers();
      while (it.hasNext())
      {
        VariableInitializer varInit = (VariableInitializer)it.next();
        varInit.accept(this);
      }
    }
  }

  public void visit(VariableInitializer variableInit)
  {
    if (variableInit.isExpression())
      variableInit.getExpression().accept(this);
    else
      variableInit.getArrayInitializer().accept(this);
  }

  /*****************************************************************************
   * STREAM EXPRESSION
   ****************************************************************************/
  public void visit(StreamExpression streamExp)
  {
    if (streamExp.isStream())
    {
      lhsContext = true;
      streamExp.getStream().accept(this);

      // stream must not be sub
      if (resultType.isSub())
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", streamExp.getStream());
        throw new TypeCheckerErrorExpressionException();
      }

      lhsContext = false;
      streamExp.getNext().accept(this);
    }
    else
    if (streamExp.isRightOperand())
    {
      TypeCheckerResultType instance = resultType;
      net.nosica.compiler.intermediate.Expression tLeft = translatedExp;
      lhsContext = false;
      streamExp.getRightOperand().accept(this);

      TypeCheckerResultType data = resultType;
      net.nosica.compiler.intermediate.Expression tRight = translatedExp;

      // data must not be sub
      if (data.isSub())
      {
        env.getMS().report(MessageLevel.ERR, "TypeCheck021", streamExp.getRightOperand());
        throw new TypeCheckerErrorExpressionException();
      }

      MethodCallContext match = findBinaryOperator(instance, data, streamExp.getOp(), streamExp.getRightOperand());

      resultType = createResultType(match);
      ArrayList args = new ArrayList();
      args.add(tLeft); args.add(tRight);

      translatedExp = IRT.translateMethodCall(match, args);

      if (streamExp.hasNext())
        streamExp.getNext().accept(this);
    }
  }

  /*****************************************************************************
   * BINARY OPERATOR EXPRESSION
   ****************************************************************************/
  public void visit(AdditiveExpression addExp)
  {
    visitBinaryExpression(addExp);
  }
  public void visit(MultiplicativeExpression mulExp)
  {
    visitBinaryExpression(mulExp);
  }
  public void visit(EqualityExpression eqExpression)
  {
    visitBinaryExpression(eqExpression);
  }
  public void visit(RelationalExpression relationalExp)
  {
    visitBinaryExpression(relationalExp);
  }
  public void visit(PowerExpression powerExp)
  {
    visitBinaryExpression(powerExp);
  }

  private void visitBinaryExpression(BinaryOperatorExpression exp)
  {
    lhsContext = false;
    exp.getLeft().accept(this);

    TypeCheckerResultType left = resultType;
    net.nosica.compiler.intermediate.Expression tLeft = translatedExp;

    lhsContext = false;
    exp.getRight().accept(this);

    TypeCheckerResultType right = resultType;
    net.nosica.compiler.intermediate.Expression tRight = translatedExp;

    // left must not be sub
    if (left.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp.getLeft());
      throw new TypeCheckerErrorExpressionException();
    }

    // right must not be sub
    if (right.isSub())
    {
      env.getMS().report(MessageLevel.ERR, "TypeCheck021", exp.getRight());
      throw new TypeCheckerErrorExpressionException();
    }

    // special case for "==" and "!=" operators with primitive types
    OperatorEnum op = exp.getOperator().getOperator();
    if (op == OperatorEnum.eq || op == OperatorEnum.ne)
    {
      // operator "==" or "!=" detected
      try
      {
        Type leftType = env.getTE().getType(left.getTypeName());
        if (leftType.isPrimitive())
        {
          // change operator from value operators to instance operators
          if (op == OperatorEnum.eq)
            op = OperatorEnum.ieq;
          else
            op = OperatorEnum.ine;

          exp.getOperator().setOperator(op);
        }
      }
      catch (SymbolException e)
      {
        env.terminate("Internal error : " + e.getMessage());
      }
    }

    MethodCallContext match = findBinaryOperator(left, right, exp.getOperator(), exp);
    resultType = createResultType(match);

    translateBinaryOperator(tLeft, tRight, match);
  }

  public void translateBinaryOperator(
    net.nosica.compiler.intermediate.Expression tLeft,
    net.nosica.compiler.intermediate.Expression tRight,
    MethodCallContext match
  )
  {
    ArrayList args = new ArrayList();
    args.add(tLeft);  args.add(tRight);

    translatedExp = IRT.translateMethodCall(match, args);
  }

  public MethodCallContext findBinaryOperator(TypeCheckerResultType left, TypeCheckerResultType right, Operator op, Node exp)
  {
    // we can have two types of choice here.
    // either using the infix operator on left with right as argument
    // or using the postfix operator on right with left as argument

    try
    {
      // try with infix operator first
      ArrayList infixArguments = new ArrayList();
      infixArguments.add(new ArgumentImpl(right.isConst(), right.getTypeName(), ""));
      Method infixOp = new MethodImpl(false, op.getOperator().toString(), MethodKindEnum.infix, infixArguments, left.isConst(), null);

      ArrayList infixMatchingNames = new ArrayList();

      ArrayList infixMatchingMethods = env.getMR().findMatchingMethods(
        left,
        infixOp,
        this.type.getName(),
        infixMatchingNames
      );

      // and try with postfix operator
      ArrayList postfixArguments = new ArrayList();
      postfixArguments.add(new ArgumentImpl(left.isConst(), left.getTypeName(), ""));
      Method postfixOp = new MethodImpl(false, op.getOperator().toString(), MethodKindEnum.postfix, postfixArguments, right.isConst(), null);

      ArrayList postfixMatchingNames = new ArrayList();

      ArrayList postfixMatchingMethods = env.getMR().findMatchingMethods(
        right,
        postfixOp,
        this.type.getName(),
        postfixMatchingNames
      );

      // checks errors
      ArrayList matchingMethods = new ArrayList();
      matchingMethods.addAll(infixMatchingMethods);
      matchingMethods.addAll(postfixMatchingMethods);

      ArrayList matchingNames = new ArrayList();
      matchingNames.addAll(infixMatchingNames);
      matchingNames.addAll(postfixMatchingNames);

      MethodCallContext match = env.getMR().getBestMethod(matchingMethods);

      if (match == null)
      {
        TypeCheckerHelper.reportMethodError(resultType, infixOp, matchingMethods, matchingNames, exp);

        throw new TypeCheckerErrorExpressionException();
      }
      else
      {
        return match;
      }
    }
    catch (SymbolException e)
    {
      throw new TypeCheckerErrorExpressionException();
    }
  }

  public boolean checkException(TypeName typeName, Localizable localizable)
  {
    boolean checkedExceptionIsImmediatlyCaught = false;

    // check exception is caught or is allowed to go through method
    try
    {
      boolean isLegal = false;

      if (!exceptionEnvironment.caughtAndMarkExceptionAsLaunched(typeName))
      {
        // expression is not caught by an active site

        // check method allow exception to go through
        Iterator it = method.iterateThrowList();
        while (it.hasNext())
        {
          TypeName allowedTypeName = (TypeName)it.next();

          env.getTL().getType(allowedTypeName);
          typeCallBack.useType(allowedTypeName);

          if (
              typeName.equals(allowedTypeName) ||
              env.getTE().doesInherits(typeName, allowedTypeName)
             )
          {
            method.addThrownException(allowedTypeName);

            isLegal = true;
            break;
          }
        }
/*
        // legal ??

        // check method interface allow exception to go through
        if (!isLegal)
        {
          // try with inherited thrown exception
          it = method.iterateInheritedException();
          while (it.hasNext())
          {
            TypeName allowedTypeName = (TypeName)it.next();

            if (env.getTE().doesInherits(typeName, allowedTypeName))
            {
              method.addThrownException(allowedTypeName);

              isLegal = true;
              break;
            }
          }
        }
*/
      }
      else
      {
        isLegal = true;
        checkedExceptionIsImmediatlyCaught = true;
      }

      if (!isLegal)
      {
        // not allowed to be thrown ... only OK if it inherits from RuntimeException
        // load RuntimeException
        try
        {
          env.getTL().getType(NativeTypes.get_net_nosica_lang_RuntimeException());
          typeCallBack.useType(NativeTypes.get_net_nosica_lang_RuntimeException());
        }
        catch (GenericityException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }
        catch (VisitorException e)
        {
          env.terminate("Internal error : " + e.getMessage());
        }

        if (
            !typeName.equals(NativeTypes.get_net_nosica_lang_RuntimeException()) &&
            !env.getTE().doesInherits(typeName, NativeTypes.get_net_nosica_lang_RuntimeException())
           )
        {
          env.getMS().report(MessageLevel.ERR, "TypeCheck062", typeName.toString(), localizable);
        }
        else
        {
          method.addThrownException(typeName);
        }
      }
    }
    catch (GenericityException e)
    {
      env.terminate("Internal error " + e.getMessage());
    }
    catch (SymbolException e)
    {
      env.terminate("Internal error " + e.getMessage());
    }

    return checkedExceptionIsImmediatlyCaught;
  }
}