/*
  $Id: TypeCheckerHelper.java,v 1.18 2004/03/18 23:15:20 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.application.MessageLevel;
import net.nosica.application.NativeTypes;
import net.nosica.compiler.application.Environment;
import net.nosica.compiler.application.GenericityException;
import net.nosica.compiler.application.TypeCheckerErrorExpressionException;

import net.nosica.parser.node.Node;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.operator.Operator;
import net.nosica.parser.node.util.operator.OperatorEnum;

import net.nosica.framework.Localizable;

import net.nosica.symbol.SymbolException;
import net.nosica.symbol.MethodKindEnum;

import java.util.ArrayList;
import java.util.Iterator;

public class TypeCheckerHelper
{
  /**
   * \brief find an assign operator between left and right
   *
   * \param accessor the type that access operator
   * \param left the lhs
   * \param right the rhs
   * \param op the operator to search for (either "=", "~", "*=", "/=", "%=", "+=", "-=", "^=")
   * \param isFieldInitialisation is this a variable/field initialisation ?
   * \param node a localizable node for error reporting
   * \return a MethodCallContext
   */
  static public MethodCallContext findAssignOperator(
    TypeName accessor,
    TypeCheckerResultType left,
    TypeCheckerResultType right,
    Operator op,
    boolean isInitialisation,
    Node node
    )
  {
    try
    {
      if (left.isPropertySet())
      {
        // property set method

        // operator must be "="
        if (op.getOperator() != OperatorEnum.assign)
        {
          Environment.instance().getMS().report(MessageLevel.ERR, "TypeCheck037", op.toString(), op);
          throw new TypeCheckerErrorExpressionException();
        }

        ArrayList matchingNames = new ArrayList(); // ArrayList<Method>
        // ArrayList<MethodCallContext>
        ArrayList matchingMethods = left.resolvePropertySet(
          right,
          accessor,
          matchingNames,
          node
        );

        MethodCallContext match = Environment.instance().getMR().getBestMethod(matchingMethods);

        if (match == null)
        {
          Method signature = left.createAssignOperator(right, node);
          reportMethodError(left, signature, matchingMethods, matchingNames, node);

          throw new TypeCheckerErrorExpressionException();
        }
        else
        {
          return match;
        }
      }

      Type lhsType = Environment.instance().getTL().getType(left.getTypeName());
      if (lhsType.isPrimitive() && op.getOperator() == OperatorEnum.assign)
      {
        // for a primitive type, "=" and "~" are the same and are authorized
        // only when operator "~" is defined
        op.setOperator(OperatorEnum.iassign);
      }

      if (op.getOperator() == OperatorEnum.assign)
      {
        // reference assignment, very different, because handled directly by compiler
        // first, we have to test rhs is a reference type
        Type rhsType = Environment.instance().getTL().getType(right.getTypeName());

        if (rhsType.isPrimitive())
        {
          Environment.instance().getMS().report(MessageLevel.ERR, "TypeCheck014", right.getTypeName().toString(), node);
          throw new TypeCheckerErrorExpressionException();
        }

        // then we can assign rhs to lhs only
        // if :
        //   this is an initialisation
        //   left is not const and rhs is not const
        //   left is const and rhs is const
        //   left is const and rhs is not const
        if (!isInitialisation && !left.isConst() && right.isConst())
        {
          Environment.instance().getMS().report(MessageLevel.ERR, "TypeCheck015", node);
          throw new TypeCheckerErrorExpressionException();
        }

        // last, an affectation is possible only if
        // rhs inherits from lhs or lhs and rhs have same type
        if (!left.getTypeName().equals(right.getTypeName()))
        {
          if (!Environment.instance().getTE().doesInherits(right.getTypeName(), left.getTypeName()))
          {
            Environment.instance().getMS().report(MessageLevel.ERR, "TypeCheck016", new String[]{right.getTypeName().toString(), left.getTypeName().toString()}, node);
            throw new TypeCheckerErrorExpressionException();
          }
        }

        // special value : the reference assign operator
        return null;
      }
      else
      {
        // all other operators ("~", "*=", "/=", "%=", "+=", "-=", "^=")
        ArrayList arguments = new ArrayList();
        arguments.add(
          new ArgumentImpl(
            right.isConst(),
            right.getTypeName(), ""
          )
        );

        Method assignOperator = new MethodImpl(
          false,
          op.getOperator().toString(),
          MethodKindEnum.infix,
          arguments,
          !isInitialisation && left.isConst(),
          null
        );

        ArrayList matchingNames = new ArrayList();
        ArrayList matchingMethods = new ArrayList();

        MethodCallContext match = Environment.instance().getMR().findMatchingMethod(
          new ResultTypeImpl(left.getTypeName(), !isInitialisation && left.isConst()),
          assignOperator,
          accessor,
          matchingMethods,
          matchingNames
        );

        if (match == null)
        {
          reportMethodError(left, assignOperator, matchingMethods, matchingNames, node);

          throw new TypeCheckerErrorExpressionException();
        }
        else
        {
          return match;
        }
      }
    }
    catch (SymbolException e)
    {
      Environment.instance().terminate("Internal error : " + e.getMessage());

      return null;
    }
    catch (GenericityException e)
    {
      // genericity solver could not create type
      // errors have been already reported
      // try to continue with next expression
      throw new TypeCheckerErrorExpressionException();
    }
  }

  /**
   * \param brief report an error while searching for a method
   *
   * \brief instance the instance to search method from
   * \brief originalSearchedSignature the method searched
   * \brief mathingMethods the methods that match (if several match)
   * \brief matchingNames the methods that have the same name as searched method
   * \brief localizable a localizable instance for error reporting
   */
  static public void reportMethodError(
    ResultType instance,
    Method originalSearchedSignature,
    ArrayList matchingMethods,
    ArrayList matchingNames,
    Localizable localizable
  )
  {
    // only one method match : no error
    if (matchingMethods.size() == 1)
      return;

    // no methods found
    if (matchingMethods.isEmpty())
    {
      // no method by that name !
      if (matchingNames.isEmpty())
      {
        Environment.instance().getMS().report(
          MessageLevel.ERR,
          "TypeCheck017",
          originalSearchedSignature.toString(),
          localizable
        );
      }
      else
      {
        // found methods by that name
        StringBuffer methods = new StringBuffer();
        Iterator it = matchingNames.iterator();
        while (it.hasNext())
        {
          methods.append(((Method)it.next()).toString());
          if (it.hasNext())
            methods.append(", ");
        }

        String signature = buildSignature(instance, originalSearchedSignature);
        Environment.instance().getMS().report(
          MessageLevel.ERR,
          "TypeCheck018",
          new String[]{signature, methods.toString()},
          localizable
        );
      }
    }
    else
    {
      // several methods match
      StringBuffer methods = new StringBuffer();
      Iterator it = matchingMethods.iterator();
      while (it.hasNext())
      {
        methods.append(((MethodCallContext)it.next()).getMethod().toString());
        if (it.hasNext())
          methods.append(", ");
      }

      String signature = buildSignature(instance, originalSearchedSignature);
      Environment.instance().getMS().report(
        MessageLevel.ERR,
        "TypeCheck019",
        new String[]{signature, methods.toString()},
        localizable
      );
    }
  }

  /**
   * \brief construct a signature from an instance and a method signature
   *
   * \param instance the instance from which we want to search method
   * \param signature the signature to search for
   * \return complete signature
   */
  static public String buildSignature(ResultType instance, Method signature)
  {
    Method realSignature = new MethodImpl(
      signature.getResultType(),
      signature.getName(),
      signature.getKind(),
      signature.getArguments(),
      instance.isConst(),
      instance.getTypeName(),
      null
    );

    return realSignature.toString();
  }

  static public net.nosica.compiler.intermediate.Expression translateSubConditionalExpression
  (
    net.nosica.compiler.intermediate.Expression cond,
    net.nosica.compiler.intermediate.Expression exp1,
    net.nosica.compiler.intermediate.Expression exp2,
    ArrayList condTemp,
    ArrayList exp1Temp,
    ArrayList exp2Temp,
    TypeCheckVisitor visitor
  )
  {
    /*
      This is in fact the result of a if statement disguised in an expression
      Here's the structure :
      cond ? trueExpr : falseExp

      and how we transform it :
      ESEQ(
            temp = cond
            [condTemporaries destruction ]
            if (temp)
              goto trueLabel
            else
              goto falseLabel
            trueLabel :
              temp = trueExpr
              [ trueTemporaries destruction ]
              goto endLabel
            falseLabel :
              temp = falseExp
              [ falseTemporaries destruction ]
            endLabel
          ,
            temp
      )
    */

    Label trueLabel =visitor.getLabelRepository().generateLabel();
    Label falseLabel = visitor.getLabelRepository().generateLabel();
    Label endLabel = visitor.getLabelRepository().generateLabel();

    Variable tempVar = new VariableImpl(
      false,
      NativeTypes.get_net_nosica_lang_bool(),
      Environment.instance().getNG().generateTemporaryName(),
      true
    );

    net.nosica.compiler.intermediate.Variable temp = visitor.getIRT().translateVariable(tempVar);

    net.nosica.compiler.intermediate.Statement statement = null;

    // temp = cond
    statement = visitor.getIRT().translateInit(
      temp,
      cond
    );

    // conditional temporary variable destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, condTemp);

    // conditional jump
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateConditionalJump(
        temp,
        trueLabel,
        falseLabel
      )
    );

    // true label
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(trueLabel)
    );

    // true expression
    statement = visitor.getIRT().translateSequence(
      statement,
      exp1
    );

    // true expression temporary destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, exp1Temp);

    // jump end label
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateJump(endLabel)
    );

    // false label
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(falseLabel)
    );

    // false expression
    statement = visitor.getIRT().translateSequence(
      statement,
      exp2
    );

    // false expression temporary destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, exp2Temp);

    // endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(endLabel)
    );

    // return statements
    // in fact, we do not return anything
    // but, as we need an expression, we return a literal
    // that is not used, thus will be discarded later
    return visitor.getIRT().translateExtendedSequence(
      statement,
      visitor.getIRT().translateLiteral(
        NativeTypes.get_net_nosica_lang_None(),
        "null"
      )
    );
  }

  static public net.nosica.compiler.intermediate.Expression translateConditionalExpression
  (
    TypeName nearestParent,
    net.nosica.compiler.intermediate.Expression condExp,
    net.nosica.compiler.intermediate.Expression exp1,
    net.nosica.compiler.intermediate.Expression exp2,
    ArrayList condTemp,
    ArrayList exp1Temp,
    ArrayList exp2Temp,
    TypeCheckVisitor visitor
  )
  {
    /*
      This is in fact the result of a if statement disguised in an expression
      Here's the structure :
      cond ? trueExpr : falseExp

      and how we transform it :
      ESEQ(
            temp = cond
            [condTemporaries destruction ]
            if (temp)
              goto trueLabel
            else
              goto falseLabel
            trueLabel :
              temp = trueExpr
              [ trueTemporaries destruction ]
              goto endLabel
            falseLabel :
              temp = falseExp
              [ falseTemporaries destruction ]
            endLabel
          ,
            temp
      )
    */

    Label trueLabel = visitor.getLabelRepository().generateLabel();
    Label falseLabel = visitor.getLabelRepository().generateLabel();
    Label endLabel = visitor.getLabelRepository().generateLabel();

    // create variable that will hold the result
    Variable result = new VariableImpl(
      false,
      nearestParent,
      Environment.instance().getNG().generateTemporaryName(),
      false
    );

    net.nosica.compiler.intermediate.Variable IRResult = visitor.getIRT().translateVariable(result);
    boolean isResultPrimitive = false;
    try
    {
      isResultPrimitive = Environment.instance().getTE().getType(nearestParent).isPrimitive();
    }
    catch (SymbolException e)
    {
      Environment.instance().terminate("Internal error : " + e.getMessage());
    }

    try
    {
      if (!isResultPrimitive)
      {
        // because when the variable is primitive we don't want it to be 'destroyed'
        // as we use the c compiler's generated copy operator to handle the temporary
        visitor.getVE().add(result);
      }
    }
    catch (TypeCheckerException e)
    {
      Environment.instance().terminate("Internal error : " + e.getMessage());
    }

    // condition temporary creation
    Variable cond = new VariableImpl(
      false,
      NativeTypes.get_net_nosica_lang_bool(),
      Environment.instance().getNG().generateTemporaryName(),
      false
    );

    net.nosica.compiler.intermediate.Variable IRCond = visitor.getIRT().translateVariable(cond);

    net.nosica.compiler.intermediate.Statement statement = null;

    // temp = cond
    statement = visitor.getIRT().translateInit(IRCond, condExp);

    // condition temporary destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, condTemp);

    // conditional jump
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateConditionalJump(
        IRCond,
        trueLabel,
        falseLabel
      )
    );

    // trueLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(trueLabel)
    );

    // var = true expression
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(IRResult, exp1)
    );

    // true expression temporary destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, exp1Temp);

    // jump endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateJump(endLabel)
    );

    // false Label
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(falseLabel)
    );

    // var = false expression
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(IRResult, exp2)
    );

    // false expression temporary destruction
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, exp2Temp);

    // endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(endLabel)
    );

    // return var
    return visitor.getIRT().translateExtendedSequence(
      statement,
      IRResult
    );
  }

  static public net.nosica.compiler.intermediate.Expression translateConditionalAndExpression
  (
    ArrayList args,         // ArrayList<Expression>
    ArrayList temporaries,  // ArrayList<ArrayList<Variable>>
    TypeCheckVisitor visitor
  )
  {
    /*
      Here's the structure :
      cond1 && cond2 && .. && condn

      and here how we transform it :
      ESEQ(
            if (cond1 [temporary destruction] )
              goto cond2Label
            else
              goto falseLabel
            cond2Label :
            if (cond2 [temporary destruction] )
              goto cond3Label
            else
              goto falseLabel
            cond3Label :
              ...
            condnLabel :
            if (condn [temporary destruction] )
              goto trueLabel
            else
              goto falseLabel
            falseLabel :
              temp = false;
              goto endLabel
            trueLabel :
              temp = true;
            endLabel :
          ,
            temp
      )
     */

    net.nosica.compiler.intermediate.Statement statement = null;
    Label trueLabel = visitor.getLabelRepository().generateLabel();
    Label falseLabel = visitor.getLabelRepository().generateLabel();
    Label endLabel = visitor.getLabelRepository().generateLabel();
    Label nextLabel = null;

    Variable temp = new VariableImpl(
      false,
      NativeTypes.get_net_nosica_lang_bool(),
      Environment.instance().getNG().generateTemporaryName(),
      false
    );

    net.nosica.compiler.intermediate.Variable IRTemp = visitor.getIRT().translateVariable(temp);

    Iterator it = args.iterator();
    Iterator tempIt = temporaries.iterator();
    while (it.hasNext() && tempIt.hasNext())
    {
      net.nosica.compiler.intermediate.Expression branchExp = (net.nosica.compiler.intermediate.Expression)it.next();
      ArrayList temps = (ArrayList)tempIt.next(); // ArrayList<Variable>

      if (it.hasNext())
        nextLabel = visitor.getLabelRepository().generateLabel();
      else
        nextLabel = trueLabel;

      // INIT(temp, branchExp)
      statement = visitor.getIRT().translateSequence(
        statement,
        visitor.getIRT().translateInit(IRTemp, branchExp)
      );

      // temporaries deletion
      statement = visitor.getIRT().translateDeadVariableDestruction(statement, temps);

      // if (temp) goto nextLabel else goto falseLabel
      statement = visitor.getIRT().translateSequence(
        statement,
        visitor.getIRT().translateConditionalJump(
            IRTemp,
            nextLabel,
            falseLabel
        )
      );

      if (it.hasNext())
      {
        statement = visitor.getIRT().translateSequence(
          statement,
          visitor.getIRT().translateLabel(nextLabel)
        );
      }
    }

    // falseLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(falseLabel)
    );

    // temp = false
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(
        IRTemp,
        visitor.getIRT().translateLiteral(
          NativeTypes.get_net_nosica_lang_bool(),
          "false"
        )
      )
    );

    // goto endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateJump(endLabel)
    );

    // trueLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(trueLabel)
    );

    // temp = true
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(
        IRTemp,
        visitor.getIRT().translateLiteral(
          NativeTypes.get_net_nosica_lang_bool(),
          "true"
        )
      )
    );

    // endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(endLabel)
    );

    return visitor.getIRT().translateExtendedSequence(statement, IRTemp);
  }

  static public net.nosica.compiler.intermediate.Expression translateConditionalOrExpression
   (
     ArrayList args,
     ArrayList temporaries,
     TypeCheckVisitor visitor
   )
   {
     /*
      Here's the structure :
      cond1 || cond2 || .. || condn

      and here how we transform it :
      ESEQ(
            if (cond1 [temporary destruction] )
              goto trueLabel
            else
              goto cond2Label
            cond2Label :
            if (cond2 [temporary destruction] )
              goto trueLabel
            else
              goto cond2Label
            cond3Label :
              ...
            condnLabel :
            if (condn [temporary destruction] )
              goto trueLabel
            else
              goto falseLabel
            falseLabel :
              temp = false;
              goto endLabel
            trueLabel :
              temp = true;
            endLabel :
          ,
            temp
      )
     */

     net.nosica.compiler.intermediate.Statement statement = null;
     Label trueLabel = visitor.getLabelRepository().generateLabel();
     Label falseLabel = visitor.getLabelRepository().generateLabel();
     Label endLabel = visitor.getLabelRepository().generateLabel();
     Label nextLabel = null;

     Variable temp = new VariableImpl(
       false,
       NativeTypes.get_net_nosica_lang_bool(),
       Environment.instance().getNG().generateTemporaryName(),
       false
     );

     net.nosica.compiler.intermediate.Variable IRTemp = visitor.getIRT().translateVariable(temp);

     Iterator it = args.iterator();
     Iterator tempIt = temporaries.iterator();
     while (it.hasNext() && tempIt.hasNext())
     {
       net.nosica.compiler.intermediate.Expression branchExp = (net.nosica.compiler.intermediate.Expression)it.next();
       ArrayList temps = (ArrayList)tempIt.next(); // ArrayList<Variable>

       if (it.hasNext())
         nextLabel = visitor.getLabelRepository().generateLabel();
       else
         nextLabel = trueLabel;

       // INIT(temp, branchExp)
       statement = visitor.getIRT().translateSequence(
         statement,
         visitor.getIRT().translateInit(
           IRTemp,
           branchExp
         )
       );

       // temporaries deletion
       statement = visitor.getIRT().translateDeadVariableDestruction(statement, temps);

       // if (temp) goto trueLabel else goto nextLabel
       statement = visitor.getIRT().translateSequence(
         statement,
         visitor.getIRT().translateConditionalJump(
             IRTemp,
             trueLabel,
             nextLabel
         )
       );

       if (it.hasNext())
       {
         statement = visitor.getIRT().translateSequence(
           statement,
           visitor.getIRT().translateLabel(nextLabel)
         );
       }
     }

     // falseLabel
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateLabel(falseLabel)
     );

     // temp = false
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateInit(
         IRTemp,
         visitor.getIRT().translateLiteral(
           NativeTypes.get_net_nosica_lang_bool(),
           "false"
         )
       )
     );

     // goto endLabel
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateJump(endLabel)
    );

     // trueLabel
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateLabel(trueLabel)
     );

     // temp = true
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateInit(
         IRTemp,
         visitor.getIRT().translateLiteral(
           NativeTypes.get_net_nosica_lang_bool(),
           "true"
         )
       )
     );

     // endLabel
     statement = visitor.getIRT().translateSequence(
       statement,
       visitor.getIRT().translateLabel(endLabel)
     );

     return visitor.getIRT().translateExtendedSequence(statement, IRTemp);
   }

   static public net.nosica.compiler.intermediate.Expression translateImplyExpression
   (
     net.nosica.compiler.intermediate.Expression first,
     net.nosica.compiler.intermediate.Expression second,
     ArrayList firstTemporaries,
     ArrayList secondTemporaries,
     TypeCheckVisitor visitor
  )
  {
    /*
      Here's the structure
      "lhs => rhs" is equivalent to "!lhs || rhs"

      Here's how we transform it :
      ESEQ(
            if (lhs  [temporary destruction] )
              goto checkLabel
            else
              goto trueLabel
            checkLabel :
            if (rhs  [temporary destruction] )
              goto trueLabel
            else
              goto falseLabel
            falseLabel :
              temp = false
              goto endLabel
            trueLabel :
              temp = true;
            endLabel
          ,
            temp
      )
    */
    net.nosica.compiler.intermediate.Statement statement = null;
    Label checkLabel = visitor.getLabelRepository().generateLabel();
    Label trueLabel = visitor.getLabelRepository().generateLabel();
    Label falseLabel = visitor.getLabelRepository().generateLabel();
    Label endLabel = visitor.getLabelRepository().generateLabel();

    Variable temp = new VariableImpl(
      false,
      NativeTypes.get_net_nosica_lang_bool(),
      Environment.instance().getNG().generateTemporaryName(),
      false
    );

    net.nosica.compiler.intermediate.Variable IRTemp = visitor.getIRT().translateVariable(temp);

    // INIT(temp, lhs)
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(IRTemp, first)
    );

    // delete temporaries
    statement = visitor.getIRT().translateDeadVariableDestruction(statement, firstTemporaries);

    // if (temp) goto checkLabel else goto trueLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateConditionalJump(
        IRTemp,
        checkLabel,
        trueLabel
      )
    );

    // checkLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(checkLabel)
    );

    // INIT(temp, rhs)
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(IRTemp, second)
    );

    // delete temporaries
    statement = visitor.getIRT().translateDeadVariableDestruction(
      statement,
      secondTemporaries
    );

    // if (temp) goto trueLabel else goto falseLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateConditionalJump(
        IRTemp,
        trueLabel,
        falseLabel
      )
    );

    // falseLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(falseLabel)
    );

    // temp = false
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(
        IRTemp,
        visitor.getIRT().translateLiteral(
          NativeTypes.get_net_nosica_lang_bool(),
          "false"
        )
      )
    );

    // goto endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateJump(endLabel)
    );

    // trueLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(trueLabel)
    );

    // temp = true
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateInit(
        IRTemp,
        visitor.getIRT().translateLiteral(
          NativeTypes.get_net_nosica_lang_bool(),
          "true"
        )
      )
    );

    // endLabel
    statement = visitor.getIRT().translateSequence(
      statement,
      visitor.getIRT().translateLabel(endLabel)
    );

    return visitor.getIRT().translateExtendedSequence(statement, IRTemp);
  }
}
