/*
  $Id: TypeImpl.java,v 1.12 2004/04/01 21:27:55 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.compiler.typechecker;

import net.nosica.application.NativeTypes;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.symbol.TypeKindEnum;
import net.nosica.symbol.AliasTable;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.SymbolException;

import java.util.Set;
import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;

public class TypeImpl extends AccessibleImpl implements Type
{
  private GenericType genericType;  // generic type from which is contructed this type
  private TypeName typeName;        // full resolved name

  private Set extendsList;             // Set<TypeName>
  private Set implementsList;       // Set<TypeName>

  private ArrayList fields;                   // ArrayList<Field> because order of declaration is important !
  private Map castOperators;        // Map<TypeName, Method>
  private Map overloadables;        // Map<String, ArrayList<Method> >
  private Method destructor;
  private Method initializer;
  private Method deinitializer;
  private TypeName enclosing;       // may be null
  private AliasTable aliasTable;
  private Map methodAliasTable;     // Map<String, TypeName>

  /**
   * \brief constructs a new Type
   *
   * \param genericType type from which is contructed this type
   * \param typeName full resolved type name
   * \param extendsList kits of extends
   * \param implementsList list of implements
   * \param fields ArrayList<Field>
   * \param castOperators Map<TypeName, Method>
   * \param overloadables Map<String, ArrayList<Method>>
   * \param destructor destructor of type
   * \param initializer intializer of type
   * \param deinitializer de-initializer of type
   * \param enclosing name of enclosing, null if none
   * \param aliasTable alias table associated to type
   * \param methodAliasTable Map<String, TypeName>
   */
  public TypeImpl(
                   GenericType genericType,
                   TypeName typeName,
                   Set extendsList,
                   Set implementsList,
                   ArrayList fields,
                   Map castOperators,
                   Map overloadables,
                   Method destructor,
                   Method initializer,
                   Method deinitializer,
                   TypeName enclosing,
                   AliasTable aliasTable,
                   Map methodAliasTable
                 )
  {
    super(genericType);

    this.genericType = genericType;
    this.typeName = typeName;
    this.extendsList = extendsList;
    this.implementsList = implementsList;
    this.fields = fields;
    this.castOperators = castOperators;
    this.overloadables = overloadables;
    this.destructor = destructor;
    this.initializer = initializer;
    this.deinitializer = deinitializer;
    this.enclosing = enclosing;
    this.aliasTable = aliasTable;
    this.methodAliasTable = methodAliasTable;
  }

  public TypeName      getName() { return typeName; }
  public TypeKindEnum  getKind() { return genericType.getKind(); }
  public boolean       isClass() { return genericType.isClass(); }
  public boolean       isInterface() { return genericType.isInterface(); }
  public boolean       isNested() { return enclosing != null; }
  public boolean       isAbstract() { return genericType.isAbstract(); }
  public boolean       isFinal() { return genericType.isFinal(); }
  public boolean       isPrimitive() { return genericType.isPrimitive(); }
  public boolean       isNative() { return genericType.isNative(); }
  public boolean       isInner() { return genericType.isInner(); }
  public Iterator      iterateExtends() { return extendsList.iterator(); }
  public boolean       hasExplicitExtends()
  {
    return (isClass() && !extendsList.contains(NativeTypes.get_net_nosica_lang_Object())) ||
           (isInterface() && !extendsList.contains(NativeTypes.get_net_nosica_lang_General()));
  }
  public Iterator      iterateImplements() { return implementsList.iterator(); }
  public Iterator      iterateFields() { return fields.iterator(); }
  public Iterator      iterateOverloadables() { return overloadables.values().iterator(); }
  public Iterator      iterateCastOperators() { return castOperators.values().iterator(); }
  public Method        getDestructor() { return destructor; }
  public Method        getInitializer() { return initializer; }
  public Method        getDeinitializer() { return deinitializer; }
  public Iterator      iterateMethodsByName(String name) throws SymbolException
  {
    if (overloadables.containsKey(name))
      return ((ArrayList)overloadables.get(name)).iterator();

    throw new SymbolException("Method named [" + name + "] does not exist in type [" + typeName + "]");
  }
  public Iterator      iterateGenericMethodsByName(String name) throws SymbolException
  {
    if (genericType.getOverloadables().containsKey(name))
    {
      ArrayList result = new ArrayList();
      result.addAll((ArrayList)genericType.getOverloadables().get(name));
      Iterator it = result.iterator();
      while (it.hasNext())
      {
        GenericMethod m = (GenericMethod)it.next();
        if (!m.hasGenericParameter())
          it.remove();
      }

      return result.iterator();
    }

    throw new SymbolException("Method named [" + name + "] does not exist");
  }
  public Iterator      iterateMethods() { return new MethodIterator(); }
  public GenericType   getModel() { return genericType; }
  public AliasTable    getAliasTable() { return aliasTable; }
  public Map           getMethodAliasTable() { return methodAliasTable; }
  public Method        getCastOperator(TypeName typeName) throws SymbolException
  {
    if (castOperators.containsKey(typeName))
      return (Method)castOperators.get(typeName);

    throw new SymbolException("cast operator [" + typeName + "] not found in [" + this.typeName + "]");
  }
  public Field getField(String name) throws SymbolException
  {
    Iterator it = fields.iterator();
    while (it.hasNext())
    {
      Field field = (Field)it.next();
      if (field.getName().equals(name))
        return field;
    }
    throw new SymbolException("field [" + name + "] does not exist");
  }

  public TypeName getEnclosing() { return enclosing; }

  public void addMethod(Method method) throws SymbolException
  {
    String methodName = method.getName();
    ArrayList methods;
    if (overloadables.containsKey(methodName))
    {
      methods = (ArrayList)overloadables.get(methodName);
      //
      Iterator i=methods.iterator();
      while(i.hasNext())
      {
        Method m = (Method)i.next();
        if (m.hasSameSignature(method))
          throw new SymbolException("Method "+m.toString()+" already exists");
      }
    }
    else
    {
      methods = new ArrayList();
      overloadables.put(methodName, methods);
    }
    //
    methods.add(method);
  }

  private class MethodIterator implements Iterator
  {
    private ArrayList methods = new ArrayList();
    private Iterator j;

    public MethodIterator()
    {
      methods.add(getInitializer());
      methods.add(getDeinitializer());
      methods.add(getDestructor());

      Iterator it = overloadables.values().iterator();
      while (it.hasNext())
      {
        ArrayList m = (ArrayList)it.next();
        methods.addAll(m);
      }

      it = castOperators.values().iterator();
      while (it.hasNext())
        methods.add(it.next());

      j = methods.iterator();
    }

    public boolean hasNext() { return j.hasNext(); }
    public Object next() { return j.next(); }
    public void remove() { throw new UnsupportedOperationException(); }
  }
}