/*
  $Id: Main.java,v 1.16 2004/04/05 19:34:48 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.doc;

import net.nosica.application.ApplicationContext;
import net.nosica.application.ApplicationContextImpl;
import net.nosica.application.PropertyException;
import net.nosica.application.NosicaPropertySetup;
import net.nosica.application.SymbolEnvironment;
import net.nosica.application.SymbolEnvironmentImpl;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.SymbolException;

import net.nosica.util.CommandLineAnalysis;
import net.nosica.util.CommandLineFlag;
import net.nosica.util.CommandLineOption;
import net.nosica.util.LightResolver;
import net.nosica.util.LightParser;

import net.nosica.compiler.Config;

import net.nosica.doc.application.ReverseEnvironment;
import net.nosica.doc.application.ReverseEnvironmentImpl;
import net.nosica.doc.production.xml.XMLGenerator;
import net.nosica.doc.production.xml.XMLGeneratorImpl;
import net.nosica.doc.production.html.HTMLGenerator;
import net.nosica.doc.production.html.JavaLikeHTMLGenerator;

import java.util.Iterator;

import java.io.File;

public class Main
{
  private ApplicationContext context = ApplicationContextImpl.instance();

  // the parsed types
  private SymbolEnvironment symbolEnv = SymbolEnvironmentImpl.instance();
  // reversed mapping
  private ReverseEnvironment reverseEnv = ReverseEnvironmentImpl.instance();

  // configuration
  private String systemPropertyFile = Config.SYSTEM_PROPERTY_FILE;
  private String userPropertyFile     = "";
  private boolean generateXML  = true;
  private boolean generateHTML = true;
  private String outputDirectory = ".";
  private String[] files;

  static public void main(String[] args)
  {
    try
    {
      Main main = new Main(args);
      main.run();
    }
    catch (Throwable e)
    {
      e.printStackTrace();
      System.exit(2);
    }
  }

  public Main(String[] args)
  {
    try
    {
      NosicaPropertySetup.setup(context);
    }
    catch (PropertyException e)
    {
      context.terminate("Internal error : " + e.getMessage());
    }

    commandLineAnalysis(args);

    try
    {
      File systemProperties = new File(systemPropertyFile);
      if (systemProperties.exists())
      {
        System.out.println("A system property file has been found, loading from [" + systemProperties.getAbsolutePath() + "]");
        context.loadPropertiesFromFile(systemProperties);
      }
      else
      {
        context.terminate("Could not load system property file from " + systemPropertyFile);
      }

      File userProperties = new File(userPropertyFile);
      if (userProperties.exists())
      {
        System.out.println("A user property file has been found, loading from [" + userProperties.getAbsolutePath() + "]");
        context.loadPropertiesFromFile(userProperties);
      }
    }
    catch (PropertyException e)
    {
      context.terminate(e.getMessage());
    }
    catch (java.io.IOException e)
    {
      context.terminate(e.getMessage());
    }
  }

  private void commandLineAnalysis(String[] args)
  {
    CommandLineAnalysis cmd = new CommandLineAnalysis(
      "nosicadoc",
      "[options] files",
      "",
      Config.VERSION
    );
    CommandLineFlag   cl_m = new CommandLineFlag(cmd,   'm', "nohtml",  "Do not produce html", false);
    CommandLineFlag   cl_x = new CommandLineFlag(cmd,   'x', "noxml",   "Do not produce xml", false);
    CommandLineFlag   cl_h = new CommandLineFlag(cmd,   'h', "help",    "Display this help message", false);
    CommandLineFlag   cl_v = new CommandLineFlag(cmd,   'v', "version", "Display application version", false);
    CommandLineOption cl_d = new CommandLineOption(cmd, 'o', "output",  "Set output directory", ".");
    CommandLineOption cl_s = new CommandLineOption(cmd, 's', "system",  "Set system property", Config.SYSTEM_PROPERTY_FILE);
    CommandLineOption cl_p = new CommandLineOption(cmd, 'p', "user",    "Set user property", "");

    files = cmd.process(args);

    if (cl_v.getValue())
    {
      context.terminate(cmd.version());
    }
    if (cl_h.getValue())
    {
      context.terminate(cmd.help());
    }

    systemPropertyFile = cl_s.getValue();
    userPropertyFile   = cl_p.getValue();
    outputDirectory    = cl_d.getValue();
    generateXML        = !cl_x.getValue();
    generateHTML       = !cl_m.getValue();

    if (files.length == 0)
    {
      StringBuffer message = new StringBuffer("No files were given on command line\n");
      message.append(cmd.help());
      context.terminate(message.toString());
    }
  }

  private String removeTrailingSlash(String path)
  {
    // remove trailing '/' (or '\' on windows)
    int index = path.lastIndexOf('/');
    if (index == -1)
    {
      index = path.lastIndexOf('\\');
    }
    if (index == path.length() - 1)
    {
      path = path.substring(0, path.length() - 1);
    }

    return path;
  }

  public void run()
  {
    for (int i = 0; i < files.length; ++i)
    {
      File file = new File(files[i]);

      if (!file.exists())
      {
        context.terminate("Path " + files[i] + " does not exist");
      }

      LightParser parser = new LightParser();

      // parse files given on command line
      try
      {
        if (file.isDirectory())
        {
          parser.handleDirectory(file);
        }
        else
        {
          parser.handleFile(file);
        }
      }
      catch (java.io.IOException e)
      {
        context.terminate("I/O error : " + e.getMessage());
      }
    }

    // resolve typename
    LightResolver resolver = new LightResolver();
    resolver.resolveTypes();

    computeReverseEnvironment();

    // now generate documentation
    generateDocumentation();
  }

  public void computeReverseEnvironment()
  {
    Iterator it = symbolEnv.iterateTypeName();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      try
      {
        GenericType type = symbolEnv.getGenericType(typeName);

        // compute reverse information now that the type was resolved
        Iterator j = type.iterateExtends();
        while (j.hasNext())
        {
          TypeName inherited = (TypeName)j.next();
          reverseEnv.addReverse(inherited, typeName);
        }
        j = type.iterateImplements();
        while (j.hasNext())
        {
          TypeName inherited = (TypeName)j.next();
          reverseEnv.addReverse(inherited, typeName);
        }
      }
      catch (SymbolException e)
      {
        context.terminate("Internal error : " + e.getMessage());
      }
    }
  }

  public void generateDocumentation()
  {
    if (generateXML)
    {
      System.out.println("Generating XML...");
      XMLGenerator generator = new XMLGeneratorImpl(outputDirectory + "/xml");

      Iterator it = symbolEnv.iterateTypeName();
      while (it.hasNext())
      {
        TypeName typeName = (TypeName)it.next();
        try
        {
          GenericType type = symbolEnv.getGenericType(typeName);
          generator.generate(type);
        }
        catch (SymbolException e)
        {
          context.terminate("Internal error : " + e.getMessage());
        }
      }
    }

    if (generateHTML)
    {
      System.out.println("Generating HTML...");
      HTMLGenerator generator = new JavaLikeHTMLGenerator(outputDirectory + "/html");
      generator.generate();
    }
  }
}
