/*
  $Id: XMLGeneratorImpl.java,v 1.12 2004/03/27 22:37:25 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.doc.production.xml;

import net.nosica.application.ApplicationContext;
import net.nosica.application.ApplicationContextImpl;
import net.nosica.doc.application.CommentHelper;

import net.nosica.symbol.Accessible;
import net.nosica.symbol.Symbol;
import net.nosica.symbol.GenericSymbol;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.MethodKindEnum;
import net.nosica.symbol.Field;
import net.nosica.symbol.ResultType;
import net.nosica.symbol.Argument;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.generic.GenericParameter;

import net.nosica.util.IndentedPrintWriter;

import net.nosica.util.FileHelper;

import java.io.File;
import java.io.FileOutputStream;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.Map;

public class XMLGeneratorImpl implements XMLGenerator
{
  private ApplicationContext context = ApplicationContextImpl.instance();
  private String outputDirectory;

  private IndentedPrintWriter writer;
  private GenericType type;

  public XMLGeneratorImpl(String outputDirectory)
  {
    this.outputDirectory = outputDirectory;
  }

  public void generate(GenericType type)
  {
    this.type = type;

    File file = FileHelper.getFile(outputDirectory + "/" + type.getName().toString() + ".xml");

    FileOutputStream out = null;

    try
    {
      out = new FileOutputStream(file);
      this.writer = new IndentedPrintWriter(out);

      String XMLEntity = type.isClass() ? "CLASS" : "INTERFACE";

      writer.println("<" + XMLEntity + ">");
      writer.addIndent();

      generateDoc(type);
      generatePackage();
      generateName();
      generateProtection(type);
      generateGeneric(type);
      generateExtends();
      generateImplements();
      generateAttributes();
      generateFields();
      generateInitializer();
      generateConstructors();
      generateMethods();
      generateCastOperators();
      generateDestructor();
      generateDeinitializer();

      writer.decIndent();
      writer.println("</" + XMLEntity + ">");

      writer.flush();
      out.close();
    }
    catch (java.io.FileNotFoundException e)
    {
      context.terminate(e.getMessage());
    }
    catch (java.io.IOException e)
    {
      context.terminate(e.getMessage());
    }
    finally
    {
      try
      {
        out.close();
      }
      catch (java.io.IOException e)
      {
        context.terminate(e.getMessage());
      }
    }
  }

  private void generateDoc(Symbol symbol)
  {
    ArrayList comments = symbol.getNode().getComments();

    if (comments == null)
      return;

    comments = CommentHelper.filterComments(comments);
    if (comments.size() == 0)
      return;

    writer.println("<DOC>");
    writer.addIndent();
    Iterator it = comments.iterator();
    while (it.hasNext())
    {
      String comment = (String)it.next();
      writer.println(encodeString(comment));
    }
    writer.decIndent();
    writer.println("</DOC>");
  }

  private void generatePackage()
  {
    if (type.getName().getComponents().size() == 1)
      return;

    writer.println("<PACKAGE>");
    writer.addIndent();
    writer.println(type.getName().getFirstComponents().toString());
    writer.decIndent();
    writer.println("</PACKAGE>");
  }

  private void generateName()
  {
    writer.println("<NAME>");
    writer.addIndent();
    writer.println(type.getName().getLastComponent().getPackageElement());
    writer.decIndent();
    writer.println("</NAME>");
  }

  private void generateGeneric(GenericSymbol symbol)
  {
    if (!symbol.hasGenericParameter())
      return;

    writer.println("<GENERICS>");
    writer.addIndent();

    Iterator it = symbol.iterateGenericParameters();
    while (it.hasNext())
    {
      GenericParameter genericParameter = (GenericParameter)it.next();

      writer.println("<GENERIC>");
      writer.addIndent();
      writer.println("<TYPENAME>" + encodeTypeName(genericParameter.getName()) + "</TYPENAME>");
      if (genericParameter.hasImplements())
      {
        writer.println("<IMPLEMENTS>" + encodeTypeName(genericParameter.getImplements()) + "</IMPLEMENTS>");
      }
      if (genericParameter.hasDefaultType())
      {
        writer.println("<DEFAULT>" + encodeTypeName(genericParameter.getDefaultType()) + "</DEFAULT>");
      }
      writer.decIndent();
      writer.println("</GENERIC>");
    }

    writer.decIndent();
    writer.println("</GENERICS>");
  }

  private void generateExtends()
  {
    Iterator it = type.iterateExtends();
    boolean hasExtends = false;

    while (it.hasNext())
    {
      TypeName extendsTypeName = (TypeName)it.next();

      if (!hasExtends)
      {
        hasExtends = true;
        writer.println("<EXTENDS>");
        writer.addIndent();
      }

      writer.print("<TYPENAME>");
      writer.print(encodeTypeName(extendsTypeName));
      writer.println("</TYPENAME>");
    }

    if (hasExtends)
    {
      writer.decIndent();
      writer.println("</EXTENDS>");
    }
  }

  private void generateImplements()
  {
    Iterator it = type.iterateImplements();
    boolean hasImplements = false;

    while (it.hasNext())
    {
      TypeName implementsTypeName = (TypeName)it.next();

      if (!hasImplements)
      {
        hasImplements = true;
        writer.println("<IMPLEMENTS>");
        writer.addIndent();
      }

      writer.print("<TYPENAME>");
      writer.print(encodeTypeName(implementsTypeName));
      writer.println("</TYPENAME>");
    }

    if (hasImplements)
    {
      writer.decIndent();
      writer.println("</IMPLEMENTS>");
    }
  }

  private void generateProtection(Accessible accessible)
  {
    writer.println("<PROTECTION>");
    writer.addIndent();
    if (accessible.isPublic())
      writer.println("PUBLIC");
    else
    if (accessible.isProtected())
      writer.println("PROTECTED");
    else
    if (accessible.isPrivate())
      writer.println("PRIVATE");
    else
      writer.println("PACKAGE");
    writer.decIndent();
    writer.println("</PROTECTION>");
  }

  private void generateAttributes()
  {
    if (type.isNative())
    {
      writer.println("<ATTRIBUTE>NATIVE</ATTRIBUTE>");
    }
    if (type.isPrimitive())
    {
      writer.println("<ATTRIBUTE>PRIMITIVE</ATTRIBUTE>");
    }
    if (type.isAbstract())
    {
      writer.println("<ATTRIBUTE>ABSTRACT</ATTRIBUTE>");
    }
    if (type.isFinal())
    {
      writer.println("<ATTRIBUTE>FINAL</ATTRIBUTE>");
    }
    if (type.isInner())
    {
      writer.println("<ATTRIBUTE>INNER</ATTRIBUTE>");
    }
    if (type.isNested())
    {
      writer.println("<ATTRIBUTE>NESTED</ATTRIBUTE>");
    }
  }

  private void generateFields()
  {
    Iterator it = type.iterateFields();
    while (it.hasNext())
    {
      Field field = (Field)it.next();

      writer.println("<FIELD>");
      writer.addIndent();
      generateProtection(field);
      generateDoc(field);
      writer.println("<TYPENAME>" + encodeTypeName(field.getTypeName()) + "</TYPENAME>");
      writer.println("<NAME>" + field.getName() + "</NAME>");
      if (field.isConst())
      {
        writer.println("<ATTRIBUTE>CONST</ATTRIBUTE>");
      }
      if (field.isMutable())
      {
        writer.println("<ATTRIBUTE>MUTABLE</ATTRIBUTE>");
      }
      if (field.isStatic())
      {
        writer.println("<ATTRIBUTE>STATIC</ATTRIBUTE>");
      }
      if (field.isTransient())
      {
        writer.println("<ATTRIBUTE>TRANSIENT</ATTRIBUTE>");
      }
      writer.decIndent();
      writer.println("</FIELD>");
    }
  }

  private void generateResultType(ResultType result)
  {
    if (result == null)
      return;

    writer.println("<RESULT>");
    writer.addIndent();
    writer.println("<TYPENAME>" + encodeTypeName(result.getTypeName()) + "</TYPENAME>");
    if (result.isConst())
      writer.println("<ATTRIBUTE>CONST</ATTRIBUTE>");
    writer.decIndent();
    writer.println("</RESULT>");
  }

  private void generateArguments(ArrayList arguments)
  {
    if (arguments.size() == 0)
      return;

    writer.println("<ARGUMENTS>");
    writer.addIndent();

    Iterator it = arguments.iterator();
    while (it.hasNext())
    {
      Argument arg = (Argument)it.next();

      writer.println("<ARGUMENT>");
      writer.addIndent();
      writer.println("<NAME>" + arg.getName() + "</NAME>");
      writer.println("<TYPENAME>" + encodeTypeName(arg.getTypeName()) + "</TYPENAME>");
      if (arg.isConst())
        writer.println("<ATTRIBUTE>CONST</ATTRIBUTE>");
      writer.decIndent();
      writer.println("</ARGUMENT>");
    }

    writer.decIndent();
    writer.println("</ARGUMENTS>");
  }

  private void generateThrowList(Set throwList)
  {
    if (throwList.size() == 0)
      return;

    writer.println("<THROWS>");
    writer.addIndent();
    Iterator it = throwList.iterator();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();

      writer.println("<TYPENAME>" + encodeTypeName(typeName) + "</TYPENAME>");
    }
    writer.decIndent();
    writer.println("</THROWS>");
  }

  private void generateMethod(GenericMethod method)
  {
    String XMLEntity;

    if (method.getKind() == MethodKindEnum.constructor)
      XMLEntity = "CONSTRUCTOR";
    else
    if (method.getKind() == MethodKindEnum.destructor)
      XMLEntity = "DESTRUCTOR";
    else
    if (method.getKind() == MethodKindEnum.initializer)
      XMLEntity = "INITIALIZER";
    else
    if (method.getKind() == MethodKindEnum.deinitializer)
      XMLEntity = "DEINITIALIZER";
    else
    if (method.getKind() == MethodKindEnum.method)
      XMLEntity = "METHOD";
    else
    if (method.getKind() == MethodKindEnum.property)
      XMLEntity = "PROPERTY";
    else
    if (method.getKind() == MethodKindEnum.propertyArray)
      XMLEntity = "ARRAY_PROPERTY";
    else
    if (method.getKind() == MethodKindEnum.castOperator)
      XMLEntity = "CAST";
    else
      XMLEntity = "OPERATOR";

    writer.println("<" + XMLEntity + ">");
    writer.addIndent();
    generateProtection(method);
    generateDoc(method);
    generateGeneric(method);
    generateResultType(method.getResultType());
    if (method.getName() != null)
    {
      writer.println("<NAME>" + encodeString(method.getName()) + "</NAME>");
    }
    generateArguments(method.getArguments());
    generateThrowList(method.getThrowList());
    writer.decIndent();
    writer.println("</" + XMLEntity + ">");
  }

  private void generateInitializer()
  {
    if (type.hasInitializer())
    {
      generateMethod(type.getInitializer());
    }
  }

  private void generateConstructors()
  {
    ArrayList constructors = (ArrayList)type.getOverloadables().get("constructor");
    if (constructors == null)
      return;

    Iterator it = constructors.iterator();
    while (it.hasNext())
    {
      GenericMethod method = (GenericMethod)it.next();
      generateMethod(method);
    }
  }

  private void generateMethods()
  {
    Iterator it = type.getOverloadables().entrySet().iterator();
    while (it.hasNext())
    {
      Map.Entry entry = (Map.Entry)it.next();

      if (entry.getKey() != null && entry.getKey().equals("constructor"))
        continue;

      ArrayList methods = (ArrayList)entry.getValue();
      Iterator j = methods.iterator();
      while (j.hasNext())
      {
        GenericMethod method = (GenericMethod)j.next();
        generateMethod(method);
      }
    }
  }

  private void generateCastOperators()
  {
    Iterator it = type.getCastOperators().values().iterator();
    while (it.hasNext())
    {
      GenericMethod method = (GenericMethod)it.next();
      generateMethod(method);
    }
  }

  private void generateDestructor()
  {
    if (type.hasDestructor())
    {
      generateMethod(type.getDestructor());
    }
  }

  private void generateDeinitializer()
  {
    if (type.hasDeinitializer())
    {
      generateMethod(type.getDeinitializer());
    }
  }

  private String replaceAll(String source, String toSearch, String toReplace)
  {
    int lastIndex = 0;

    for (;;)
    {
      int index = source.indexOf(toSearch, lastIndex);
      if (index == -1)
        break;

      String temp = source.substring(0, index);
      temp += toReplace;
      temp += source.substring(index + toSearch.length());
      source = temp;
      lastIndex = index + toReplace.length();
    }

    return source;
  }

  private String encodeString(String string)
  {
    String result = replaceAll(string, "<", "&lt;");
    result = replaceAll(result, ">", "&gt;");
    return result;
  }

  private String encodeTypeName(TypeName typeName)
  {
    return encodeString(typeName.toString());
  }
}
