/*
  $Id: ConstructorDeclaration.java,v 1.5 2004/03/25 22:54:39 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.parser.node.typedeclaration;

import  net.nosica.parser.node.util.generic.GenericDeclaration;
import  net.nosica.parser.node.util.var.FormalParameters;
import  java.util.Iterator;

/**
 * \brief node that represents a constructor declaration
 *
 * e.g : in source code :
 * <pre>
 * class Test
 * {
 *   public constructor() {}
 * }
 * </pre>
 * will define a ConstructorDeclaration<br>
 * <br>
 * grammar :<br>
 *  ConstructorDeclaration ::=<br>
 *   ( "public" | "protected" | "private" )?
 *   "constructor"
 *   ( GenericDeclaration )?
 *   FormalParameters
 *   ( "throws" TypeNameList )?
 *   (
 *       "{"
 *         ( ExplicitConstructorInvocation )?
 *         ( BlockStatement )*
 *       "}"
 *     |
 *       ";"
 *   )
 */
public interface ConstructorDeclaration extends Method, ClassBodyDeclaration
{
  /**
   * \brief is constructor public ?
   *
   * \return true if constructor is public
   */
  boolean isPublic();

  /**
   * \brief is constructor protected ?
   *
   * \return true if constructor is protected
   */
  boolean isProtected();

  /**
   * \brief is constructor private ?
   *
   * \return true if constructor is private
   */
  boolean isPrivate();

  /**
   * \brief is constructor generic ?
   *
   * \return true if constructor is generic
   */
  boolean hasGenericDeclaration();

  /**
   * \brief does constructor uses an explicit constructor invocation ?
   *
   * \return true if constructor uses a constructor invocation
   */
  boolean hasExplicitConstructorInvocation();

  /**
   * \brief has constructor an implementation ?
   *
   * \return true if constructor has an implementation
   */
  boolean hasBody();

  /**
   * \brief return the list of generic parameters introduced by constructor
   *
   * \remarks need hasGenericDeclaration() to return true
   * \return generic declarations ...
   */
  GenericDeclaration getGenericDeclaration();

  /**
   * \brief return formal parameters of constructor
   *
   * \return formal parameters
   */
  FormalParameters getFormalParameters();

  /**
   * \brief return the explicit constructor invocation
   *
   * An explicit constructor invocation if it exists is always the first expression of the implementation of a constructor
   * \return explicit constructor invocation
   */
  ExplicitConstructorInvocation getExplicitConstructorInvocation();

  /**
   * \brief return list of block
   *
   * \return Iterator<BlockStatement>
   */
  Iterator iterateBlockStatements();
}
