/*
  $Id: OperatorDeclaration.java,v 1.4 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.parser.node.typedeclaration;

import  net.nosica.parser.node.util.var.ResultType;
import  net.nosica.parser.node.util.operator.Operator;
import  net.nosica.parser.node.util.generic.GenericDeclaration;
import  net.nosica.parser.node.util.var.FormalParameters;

/**
 * \brief a node that represents an operator declaration
 *
 * e.g : in source code :
 * <pre>
 * class Test
 * {
 *   Test infix<<(Toto toto);
 *   Test prefix -();
 *   Test postfix<<(Toto toto);
 * }
 * </pre>
 * Declares three new operators with different meanings<br>
 * First one is an infix operator, that is, code<br>
 * <pre>
 * Test test = new Test();
 * Toto toto = new Toto();
 * test << toto;
 * </pre>
 * will call Test.infix<<(Toto)<br>
 * Second one is a prefix operator, that is, code<br>
 * <pre>
 * Test test = new Test();
 * -test;
 * </pre>
 * will call Test.prefix-()<br>
 * Third one is a postfix operator, that is, code<br>
 * <pre>
 * Test test = new Test();
 * Toto toto = new Toto();
 * toto << test;
 * </pre>
 * will call Test.postfix(Toto)<br>
 * <br>
 * grammar :
 *    OperatorDeclaration ::=
 *      ( "public" | "protected" | "private" | "final" )*
 *      ResultType
 *      ( "infix" | "prefix" | "postfix" )
 *      OperatorList
 *      ( GenericDeclaration )?
 *      FormalParameters
 *      ( "const" )?
 *      ( "throws" TypeNameList )?
 *      ( Block | ";" )
 *      OptionalBlock
 */

public interface OperatorDeclaration extends Method, ClassBodyDeclaration, InterfaceMemberDeclaration
{
  /**
   * is operator public ?
   *
   * \brief return true if operator is public
   */
  boolean isPublic();

  /**
   * is operator protected ?
   *
   * \brief return true if operator is protected
   */
  boolean isProtected();

  /**
   * is operator private ?
   *
   * \brief return true if operator is private
   */
  boolean isPrivate();

  /**
   * is operator final ?
   *
   * \brief return true if operator is final
   */
  boolean isFinal();

  /**
   * is operator infix ?
   *
   * \brief return true if operator is infix
   */
  boolean isInfix();

  /**
   * is operator prefix ?
   *
   * \brief return true if operator is prefix
   */
  boolean isPrefix();

  /**
   * is operator postfix ?
   *
   * \brief return true if operator is postfix
   */
  boolean isPostfix();

  /**
   * \brief return the underlying operator
   *
   * \return operator
   */
  Operator getOperator();

  /**
   * \brief is this method const ?
   *
   * \brief return true if method is const
   */
  boolean isMethodConst();

  /**
   * \brief has this method generic modifier ?
   *
   * That is, is method a generic one ?
   *
   * \breif return true if method is generic
   */
  boolean hasGenericDeclaration();

  /**
   * \brief return resultType
   *
   * \return ResultType
   */
  ResultType getType();

  /**
   * \brief return generic declaration
   *
   * \remarks needs hasGenericDeclaration() to return true
   * \return generic declaration
   */
  GenericDeclaration getGenericDeclaration();

  /**
   * \brief return formal parameters of method
   *
   * \return parameters of method
   */
  FormalParameters getFormalParameters();
}
