/*
  $Id: PropertyDeclaration.java,v 1.4 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.parser.node.typedeclaration;

import net.nosica.parser.node.util.generic.GenericDeclaration;
import net.nosica.parser.node.util.var.ResultType;
import net.nosica.parser.node.util.var.FormalParameters;

/**
 * \brief a node that represents a property declaration
 *
 * e.g : in source code :
 * <pre>
 * class Test
 * {
 *   private int i;
 *
 *   public int property i() const { return i; }
 * }
 * </pre>
 * will create a method that acts as a readonly member<br>
 * ie, the following code is valid :<br>
 * <pre>
 * Test test = new Test();
 * int i = test.i; // this will call in fact test.property i();
 * </pre>
 * <br>
 * a property has no defined prototype<br>
 * You can use the different signatures :<br>
 * <ul>
 *   <li> int property i() const;       // called a property get
 *   <li> sub property i(int i);  // called a property set
 *   <li> int property i(int i);  // called a property getset
 * </ul>
 * But properties can also be defined to act as an array operator<br>
 * e.g :
 * <pre>
 * class Test
 * {
 *   private int[] i;
 *
 *   public int property[](word index) const { return i[index]; }
 * }
 * </pre>
 * will create a method that act as a readonly operator []<br>
 * the following code<br>
 * <pre>
 * Test test = new Test();
 * int i = test[0]; // will call test.property[](0);
 * </pre>
 * A property array has no defined prototype<br>
 * You can use the different signatures :<br>
 * <ul>
 *   <li> int property [](word dim0) const; // called an anonymous property array get
 *   <li> sub property [](int i, word dim0); // called an anonymous property array set
 *   <li> int property [](int i, word dim0); // called an anonymous property array getset
 *   <li> int property [] name(word dim0) const // called a property array get
 *   <li> sub property [] name(int i, word dim0); // called a property array set
 *   <li> int property [] name(int i, word dim0); // called a property array getset
 * </ul>
 * <br>
 * Property array can also be multidimensional !<br>
 * example :
 * <ul>
 *   <li> int property [](word dim0, word dim1) // multi dimension anonymous property array get
 *   <li> sub property [](int i, word dim0, word dim1) // multi dimension anonymous property array set
 *   <li> int property [](int i, word dim0, word dim1) // multi dimension anonymous property array getset
 * </ul>
 * and so on with named multi dimensional property array ...<br>
 * <br>
 * <br>
 * grammar :<br>
 *   PropertyDeclaration ::=
 *      ( "public" | "protected" | "private" | "final" | "static")*
 *      ResultType
 *      "property"
 *      ( "[" "]" )?
 *      ( <IDENTIFIER> )?
 *      ( GenericDeclaration )?
 *      FormalParameters
 *      ( "const" )?
 *      ( "throws" TypeNameList )?
 *      ( Block | ";" )
 *      OptionalBlock
 */

public interface PropertyDeclaration extends Method, ClassBodyDeclaration, InterfaceMemberDeclaration
{
  /**
   * \brief return ResultType of PropertyDeclaration
   *
   * \return the ResultType
   */
  ResultType getResultType();

  /**
   * \brief return the name of the PropertyDeclaration (if any)
   *
   * \remarks only a property array can be anonymous
   * \return name of property (null if anonymous)
   */
  String getName();

  /**
   * \brief is this PropertyDeclaration a property array or a simple property ?
   *
   * \return true if this is a property array
   */
  boolean isPropertyArray();

  /**
   * \brief does this method defines generic parameters ?
   * That is is this method a generic method
   *
   * \return true if this is a generic method
   */
  boolean hasGenericDeclaration();

  /**
   * \brief return the generic declaration of this method
   *
   * \remarks need isPropertyArray() to return true
   * \return underlying GenericDeclaration
   */
  GenericDeclaration getGenericDeclaration();

  /**
   * \brief returns the formal parameters of this PropertyDeclaration
   *
   * \return formal parameters of property
   */
  FormalParameters getFormalParameters();

  /**
   * \brief returns constness of method
   *
   * This is because of the constness of methods (and also because of the number of parameters)
   * that one can say if this is a property get or a property set
   *
   * \return constness of method
   */
  boolean isMethodConst();

  /**
   * \brief is method public ?
   *
   * \return true if method is public
   */
  boolean isPublic();

  /**
   * \brief is method protected ?
   *
   * \return true if method is protected
   */
  boolean isProtected();

  /**
   * \brief is method private ?
   *
   * \return true if method is private
   */
  boolean isPrivate();

  /**
   * \brief is method final ?
   *
   * \return true if method is final
   */
  boolean isFinal();

  /**
   * \brief is property static ?
   *
   * \return true if method is static
   */
  boolean isStatic();
}
