/*
  $Id: FeedSymbolVisitor.java,v 1.12 2004/04/18 00:52:04 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.symbol;

import net.nosica.parser.node.AbstractNodeVisitor;
import net.nosica.parser.node.NodeVisitor;
import net.nosica.parser.node.Node;
import net.nosica.parser.node.filestructure.CompilationUnit;
import net.nosica.parser.node.filestructure.PackageDeclaration;
import net.nosica.parser.node.filestructure.PackageDeclarationImpl;
import net.nosica.parser.node.filestructure.ImportDeclaration;
import net.nosica.parser.node.filestructure.ImportDeclarationImpl;
import net.nosica.parser.node.typedeclaration.TypeDeclaration;
import net.nosica.parser.node.typedeclaration.ClassDeclaration;
import net.nosica.parser.node.typedeclaration.ClassBody;
import net.nosica.parser.node.typedeclaration.NestedClassDeclaration;
import net.nosica.parser.node.typedeclaration.InterfaceDeclaration;
import net.nosica.parser.node.typedeclaration.NestedInterfaceDeclaration;
import net.nosica.parser.node.typedeclaration.AliasDeclaration;
import net.nosica.parser.node.typedeclaration.FieldDeclaration;
import net.nosica.parser.node.typedeclaration.ConstructorDeclaration;
import net.nosica.parser.node.typedeclaration.ConstructorDeclarationImpl;
import net.nosica.parser.node.typedeclaration.DestructorDeclaration;
import net.nosica.parser.node.typedeclaration.CastOperatorDeclaration;
import net.nosica.parser.node.typedeclaration.OperatorDeclaration;
import net.nosica.parser.node.typedeclaration.PropertyDeclaration;
import net.nosica.parser.node.typedeclaration.MethodDeclaration;
import net.nosica.parser.node.typedeclaration.Initializer;
import net.nosica.parser.node.typedeclaration.Deinitializer;
import net.nosica.parser.node.util.var.VariableDeclarator;
import net.nosica.parser.node.util.var.FormalParametersImpl;
import net.nosica.parser.node.util.var.TypeNameListImpl;
import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponent;
import net.nosica.parser.node.util.var.TypeNameComponentImpl;
import net.nosica.parser.node.util.generic.GenericParameter;
import net.nosica.parser.node.util.generic.GenericDeclaration;

import net.nosica.application.NativeTypes;
import net.nosica.application.ApplicationContext;
import net.nosica.application.ApplicationContextImpl;
import net.nosica.application.FileLoader;

import net.nosica.util.NosicaFileFilter;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.io.File;

/**
 * \brief A visitor that will extract symbol information from a Node
 */

public class FeedSymbolVisitor extends AbstractNodeVisitor implements NodeVisitor
{
  private static AliasTable net_nosica_lang_imports = null;  // table of imports of net.nosica.lang imports

  private ApplicationContext context = ApplicationContextImpl.instance();
  private FileLoader fileLoader = null;
  private SymbolCallBack callBack = null;

  private TypeName packageName = null;
  private AliasTable packageAliasTable = null;
  private GenericType currentGenericType = null;
  private AliasTable currentAliasTable = null;
  private Map currentMethodAliasTable = null;
  private ArrayList types = new ArrayList();

  /**
   * \brief Constructs a new FeedSymbolVisitor that will extract symbol information
   * from waitedTypeName referenced file
   *
   * \param waitedTypeName TypeName of type we are about to parse
   */
  public FeedSymbolVisitor(FileLoader fileLoader, SymbolCallBack callBack)
  {
    this.fileLoader = fileLoader;
    this.callBack = callBack;

    load_net_nosica_lang_package();

    // create a new alias table for this CompilationUnit
    packageAliasTable = new AliasTableImpl(net_nosica_lang_imports);
  }

  public FeedSymbolVisitor(FileLoader fileLoader, SymbolCallBack callBack, GenericType enclosingGenericType)
  {
    this(fileLoader, callBack);

    this.currentGenericType = enclosingGenericType;
    this.currentAliasTable = enclosingGenericType.getAliasTable();
    this.currentMethodAliasTable = enclosingGenericType.getMethodAliasTable();
  }

  private void load_net_nosica_lang_package()
  {
    // if net_nosica_lang_imports was not initialized, load it
    if (net_nosica_lang_imports == null)
    {
      ImportDeclaration importDecl = new ImportDeclarationImpl(NativeTypes.get_net_nosica_lang());
      packageAliasTable = new AliasTableImpl();
      importDecl.accept(this);
      try
      {
        packageAliasTable.addAlias(NativeTypes.get_int(), NativeTypes.get_net_nosica_lang_int32());
        packageAliasTable.addAlias(NativeTypes.get_word(), NativeTypes.get_net_nosica_lang_word32());
        packageAliasTable.addAlias(NativeTypes.get_float(), NativeTypes.get_net_nosica_lang_float32());
      }
      catch (SymbolException e)
      {
        context.terminate("Internal error : [" + e.getMessage() + "]");
      }
      net_nosica_lang_imports = packageAliasTable;
    }
  }

  /**
   * \brief returns all types extracted from this visitor
   *
   * \return ArrayList<GenericType>
   */
  public ArrayList getTypes() { return types; }

  /**
   * \brief parse a CompilationUnit, as stated in grammar, a CompilationUnit contains
   * information like package, imports and types<BR>
   * CompilationUnit imports all nosica file of the same package ot the file being parsed
   *
   * All imports of net.nosica.lang are also imported.
   */
  public void visit(CompilationUnit compilationUnit)
  {
    PackageDeclaration packageDecl;

    // check package
    if (compilationUnit.hasPackageDeclaration())
      packageDecl = compilationUnit.getPackageDeclaration();
    else
    {
      packageDecl = new PackageDeclarationImpl(new TypeNameImpl());
      packageDecl.setLocalization(compilationUnit.localization());
    }
    packageDecl.accept(this);

    if (!packageDecl.getTypeName().equals(NativeTypes.get_net_nosica_lang()))
    {
      // add all files of current directory as import
      ImportDeclaration currentDir = new ImportDeclarationImpl(packageDecl.getTypeName());
      currentDir.accept(this);
    }

    // add all imports of CompilationUnit to alias table
    Iterator it = compilationUnit.iterateImportDeclaration();
    while (it.hasNext())
    {
      ImportDeclaration importDecl = (ImportDeclaration)it.next();
      importDecl.accept(this);
    }

    // go on with types
    currentAliasTable = packageAliasTable;
    it = compilationUnit.iterateTypesDeclaration();
    while (it.hasNext())
    {
      TypeDeclaration typeDecl = (TypeDeclaration)it.next();
      typeDecl.accept(this);
    }
  }

  /**
   * \brief Verify that package declaration is OK according to waited typeName
   * requires packageDecl.getTypeName() == waitedTypeName.getFirstComponents()
   */
  public void visit(PackageDeclaration packageDecl)
  {
    callBack.onPackageDeclaration(packageDecl.getTypeName(), packageDecl);
    packageName = packageDecl.getTypeName();
  }

  /**
   * \brief Make one or more alias(es) from an ImportDeclaration
   * An ImportDeclaration can refer to several files on disk<BR>
   * (because of partial generic types)<BR>
   * For example, let's say user has defined :
   * <ul>
   *   <li>/home/toto/nosica/source1
   *   <li>/home/toto/nosica/source2
   *   <li>/usr/nosica/
   * </ul>
   * If we have an ImportDeclaration of net.nosica.Test<BR>
   * we have 3 path to search for this file.<BR>
   * It is perfectly legal that we find several files refered by net.nosica.Test<BR>
   * Only one file will contains the complete generic version of the type, the other one<br>
   * are allowed to contain a partial generic version of the type
   * Therefore, it is illegal to use a generic type in an import.
   */
  public void visit(ImportDeclaration importDecl)
  {
    callBack.onImportDeclaration(importDecl.getTypeName(), importDecl);

    // imports all files that match this typeName
    Iterator it = fileLoader.findFiles(importDecl.getTypeName()).iterator();

    if (!it.hasNext())
    {
      callBack.onWrongImportDeclaration(importDecl.getTypeName(), importDecl);
    }

    // for each files
    while (it.hasNext())
    {
      File file = (File)it.next();
      if (file.isDirectory())
      {
        // get all .nos files of directory
        File[] files = file.listFiles(new NosicaFileFilter(false));

        // add an alias for each files in directory
        for (int i = 0; i < files.length; ++i)
        {
          // extract alias
          String name = files[i].getName();
          int dot = name.lastIndexOf('.');
          String alias = name.substring(0, dot);

          // build TypeName
          ArrayList tmp = new ArrayList();
          tmp.addAll(importDecl.getTypeName().getComponents());
          tmp.add(new TypeNameComponentImpl(alias, null));

          try
          {
            // add alias
            ArrayList components = new ArrayList();
            components.add(new TypeNameComponentImpl(alias, null));
            packageAliasTable.addAlias(new TypeNameImpl(components), new TypeNameImpl(tmp));
          }
          catch (SymbolException e)
          {
            callBack.onDuplicateAlias(alias, importDecl);
          }
        }
      }
      else
      {
        // this is a file
        String name = file.getName();
        int dot = name.lastIndexOf('.');
        String alias = name.substring(0, dot);

        try
        {
          // add alias
          ArrayList components = new ArrayList();
          components.add(new TypeNameComponentImpl(alias, null));
          packageAliasTable.addAlias(new TypeNameImpl(components), importDecl.getTypeName());
        }
        catch (SymbolException e)
        {
          callBack.onDuplicateAlias(alias, importDecl);
        }
      }
    }
  }

  /**
   * \brief Just pass control to ClassDeclaration of InterfaceDeclaration
   */
  public void visit(TypeDeclaration typeDecl)
  {
    // create alias table for this type
    AliasTable savedAliasTable = currentAliasTable;
    Map savedMethodAliasTable = currentMethodAliasTable;

    currentAliasTable = new AliasTableImpl(currentAliasTable);
    currentMethodAliasTable = new HashMap();

    if (typeDecl.isClassDeclaration())
      typeDecl.getClassDeclaration().accept(this);
    else
      typeDecl.getInterfaceDeclaration().accept(this);

    currentMethodAliasTable = savedMethodAliasTable;
    currentAliasTable = savedAliasTable;
  }

  /**
   * \brief Create a GenericType upon a ClassDeclaration
   * class name must be the class name of the waited type<BR>
   * A generic type is then created, as well as its associated AliasTable<BR>
   * Add an alias to itself in the alias table<BR>
   * Then control is passed to the underlying ClassBody<BR>
   */
  public void visit(ClassDeclaration classDecl)
  {
    // extract class name
    String strClassName = classDecl.getIdentifier();

    // extract extends set
    Set extendsSet = new HashSet();
    if (classDecl.hasExtend())
    {
      extendsSet.add(classDecl.getExtend());
    }

    // extract implements set
    Set implementsSet = new HashSet();
    if (classDecl.hasImplementList())
    {
      Iterator it = classDecl.getImplementList().iterateTypeNames();
      while (it.hasNext())
      {
        implementsSet.add(it.next());
      }
    }

    ArrayList genericDecl;
    if (classDecl.hasGenericDeclaration())
    {
      genericDecl = classDecl.getGenericDeclaration().getGenericParameters();
    }
    else
    {
      genericDecl = new ArrayList();
    }

    TypeName typeName = createTypeName(packageName, strClassName, classDecl.getGenericDeclaration());
    TypeName enclosing = currentGenericType == null ? null : currentGenericType.getName();

    // create generic type and sanity check during construction
    currentGenericType = new GenericTypeImpl(
      classDecl,
      classDecl.isPublic(),
      classDecl.isProtected(),
      classDecl.isPrivate(),
      genericDecl,
      null, // will be set later
      classDecl.isAbstract(),
      classDecl.isFinal(),
      classDecl.isPrimitive(),
      classDecl.isNative(),
      typeName,
      extendsSet,
      implementsSet,
      currentAliasTable,
      currentMethodAliasTable
     );

    callBack.onClassDeclaration(currentGenericType);

    types.add(currentGenericType);

    classDecl.getClassBody().accept(this);
  }

  static private TypeName createTypeName(TypeName packageName, String strTypeName, GenericDeclaration genericDecl)
  {
    ArrayList typeNameComponents = new ArrayList();
    typeNameComponents.addAll(packageName.getComponents());

    ArrayList genericExtension = null;
    if (genericDecl != null)
    {
      genericExtension = new ArrayList();
      Iterator it = genericDecl.iterateGenericParameters();
      while (it.hasNext())
      {
        GenericParameter genericParm = (GenericParameter)it.next();
        genericExtension.add(genericParm.getName());
      }
    }

    TypeNameComponent genericComponent = new TypeNameComponentImpl(
      strTypeName,
      genericExtension
    );

    typeNameComponents.add(genericComponent);

    return new TypeNameImpl(typeNameComponents);
  }

  /**
   * \brief Iterates through all members of ClassBody
   * Pass control to all members of type
   * <ul>
   *   <li>NestedClassDeclaration
   *   <li>NestedInterfaceDeclaration
   *   <li>AliasDeclaration
   * </ul>
   */
  public void visit(ClassBody classBody)
  {
    if (classBody.hasClassBodyDeclaration())
    {
      // first import symbols for nested class and nested interfaces
      Iterator it = classBody.iterateClassBodyDeclaration();
      while (it.hasNext())
      {
        Node node = (Node)it.next();
        node.accept(this);
      }
    }
  }

  /**
   * \brief Pass control to ClassDeclaration then adds Nested information to created genericType
   * Alias of nested class is added to enclosing type.
   */
  public void visit(NestedClassDeclaration nestedClassDecl)
  {
    // save global parameters
    TypeName savedPackageName = packageName;
    GenericType savedCurrentGenericType = currentGenericType;
    AliasTable savedCurrentAliasTable = currentAliasTable;
    Map savedMethodAliasTable = currentMethodAliasTable;

    // create a new packageName so that ClassDeclaration will interpret everything the right way
    ArrayList packageComponents = new ArrayList();
    packageComponents.addAll(currentGenericType.getName().getComponents());
    packageName = new TypeNameImpl(packageComponents);

    TypeNameComponent classNameComponent = new TypeNameComponentImpl(
      nestedClassDecl.getClassDeclaration().getIdentifier(),
      null
    );

    ArrayList fullNameComponents = new ArrayList();
    fullNameComponents.addAll(currentGenericType.getName().getComponents());
    fullNameComponents.add(classNameComponent);
    TypeName nestedClassName = new TypeNameImpl(fullNameComponents);

    currentAliasTable = new AliasTableImpl(savedCurrentAliasTable);
    currentMethodAliasTable = new HashMap();

    // accept ClassDeclaration
    callBack.onBeginNestedClassDeclaration(nestedClassName);
    nestedClassDecl.getClassDeclaration().accept(this);
    callBack.onEndNestedClassDeclaration();

    // create short name
    ArrayList shortNameComponents = new ArrayList();
    shortNameComponents.add(classNameComponent);
    TypeName shortName = new TypeNameImpl(shortNameComponents);
    // as we add an alias from shortName to nestedTypeName on enclosing class, we don't have
    // to add an alias to self in our own alias table

    // now add nested information to class
    currentGenericType.setEnclosing(savedCurrentGenericType.getName());
    currentGenericType.setNested();
    if (nestedClassDecl.getClassDeclaration().isInner())
    {
      currentGenericType.setInner();
    }

    // restore global parameters
    packageName = savedPackageName;
    currentGenericType = savedCurrentGenericType;
    currentMethodAliasTable = savedMethodAliasTable;
    currentAliasTable = savedCurrentAliasTable;

    // and add an alias of this nested class to enclosing
    try
    {
      // it is possible that a nested class with the same name has already added an alias to enclosing
      if (!currentAliasTable.aliasExists(shortName))
      {
        currentAliasTable.addAlias(shortName, nestedClassName);
      }
    }
    catch (SymbolException e)
    {
      context.terminate("Internal error : " + e.getMessage());
    }
  }

  /**
   * \brief Create a GenericType upon an InterfaceDeclaration
   * interface name must be the name of the waited type<BR>
   * A generic type is then created, as well as its associated AliasTable<BR>
   * Add an alias to itself in the alias table<BR>
   * Then iterates through all members<BR>
   * Pass control to all members of type
   * <ul>
   *   <li>NestedClassDeclaration
   *   <li>NestedInterfaceDeclaration
   *   <li>AliasDeclaration
   * </ul>
   */
  public void visit(InterfaceDeclaration interfaceDecl)
  {
    // extract class name
    String strInterfaceName = interfaceDecl.getName();

    // extract extends set
    Set extendsSet = new HashSet();
    if (interfaceDecl.hasExtendList())
    {
      Iterator it = interfaceDecl.getExtendList().iterateTypeNames();
      while (it.hasNext())
        extendsSet.add(it.next());
    }

    // extract generic decl
    ArrayList genericDecl;
    if (interfaceDecl.hasGenericDeclaration())
      genericDecl = interfaceDecl.getGenericDeclaration().getGenericParameters();
    else
      genericDecl = new ArrayList();

    TypeName typeName = createTypeName(packageName, strInterfaceName, interfaceDecl.getGenericDeclaration());
    TypeName enclosing = currentGenericType == null ? null : currentGenericType.getName();

    // create generic type and sanity check during construction
    currentGenericType = new GenericTypeImpl(
       interfaceDecl,
       interfaceDecl.isPublic(),
       interfaceDecl.isProtected(),
       interfaceDecl.isPrivate(),
       genericDecl,
       null, // will be set later
       typeName,
       extendsSet,
       currentAliasTable,
       currentMethodAliasTable
    );

    callBack.onInterfaceDeclaration(currentGenericType);

    types.add(currentGenericType);

    Iterator it = interfaceDecl.iterateInterfaceMemberDeclaration();
    while (it.hasNext())
    {
      Node node = (Node)it.next();
      node.accept(this);
    }
  }

  /**
   * \brief Pass control to IntefaceDeclaration then adds Nested information to created genericType
   * Alias of nested interface is added to enclosing type.
   */
  public void visit(NestedInterfaceDeclaration nestedInterfaceDecl)
  {
    // save global parameters
    TypeName savedPackageName = packageName;
    GenericType savedCurrentGenericType = currentGenericType;
    AliasTable savedCurrentAliasTable = currentAliasTable;
    Map savedMethodAliasTable = currentMethodAliasTable;

    // create a new packageName so that InterfaceDeclaration will interpret everything the right way
    ArrayList tmp = new ArrayList();
    tmp.addAll(currentGenericType.getName().getComponents());
    packageName = new TypeNameImpl(tmp);

    currentAliasTable = new AliasTableImpl(savedCurrentAliasTable);
    currentMethodAliasTable = new HashMap();

    TypeNameComponent interfaceNameComponent = new TypeNameComponentImpl(
      nestedInterfaceDecl.getInterfaceDeclaration().getName(),
      null
    );

    ArrayList fullNameComponents = new ArrayList();
    fullNameComponents.addAll(currentGenericType.getName().getComponents());
    fullNameComponents.add(interfaceNameComponent);
    TypeName nestedInterfaceName = new TypeNameImpl(fullNameComponents);

    // accept InterfaceDeclaration
    callBack.onBeginNestedInterfaceDeclaration(nestedInterfaceName);
    nestedInterfaceDecl.getInterfaceDeclaration().accept(this);
    callBack.onEndNestedInterfaceDeclaration();

    // create short name
    ArrayList shortNameComponents = new ArrayList();
    shortNameComponents.add(interfaceNameComponent);
    TypeName shortName = new TypeNameImpl(shortNameComponents);
    // as we add an alias from shortName to nestedTypeName on enclosing class, we don't have
    // to add an alias to self in our own alias table

    // now add nested information to interface
    currentGenericType.setEnclosing(savedCurrentGenericType.getName());
    currentGenericType.setNested();

    // restore global parameters
    packageName = savedPackageName;
    currentGenericType = savedCurrentGenericType;
    currentMethodAliasTable = savedMethodAliasTable;
    currentAliasTable = savedCurrentAliasTable;

    // and add an alias of this nested interface to enclosing
    try
    {
      // an interface cannot partially implements another generic interface : alias must be unique
      currentAliasTable.addAlias(shortName, nestedInterfaceName);
    }
    catch (SymbolException e)
    {
      callBack.onDuplicateAlias(shortName.toString(), nestedInterfaceDecl);
    }
  }

  /**
   * \brief Add an alias to the active alias table
   */
  public void visit(AliasDeclaration aliasDecl)
  {
    try
    {
      ArrayList components = new ArrayList();
      components.add(new TypeNameComponentImpl(aliasDecl.getAlias(), null));

      if (aliasDecl.isImport())
        currentAliasTable.addAlias(new TypeNameImpl(components), aliasDecl.getType());
      else
      {
        if (currentMethodAliasTable.containsKey(aliasDecl.getAlias()))
        {
          callBack.onDuplicateAlias(aliasDecl.getAlias(), aliasDecl);
        }

        currentMethodAliasTable.put(aliasDecl.getAlias(), aliasDecl.getType());
      }
    }
    catch (SymbolException e)
    {
      callBack.onDuplicateAlias(aliasDecl.getAlias(), aliasDecl);
    }
  }

  /**
   * \brief Add this FieldDeclaration to active GenericType
   */
  public void visit(FieldDeclaration fieldDecl)
  {
    Iterator it = fieldDecl.iterateVariableDeclarator();
    while (it.hasNext())
    {
      VariableDeclarator varDecl = (VariableDeclarator)it.next();

      Field field = new FieldImpl(
        fieldDecl,
        fieldDecl.isPublic(),
        fieldDecl.isProtected(),
        fieldDecl.isPrivate(),
        fieldDecl.isConst(),
        fieldDecl.isMutable(),
        fieldDecl.isTransient(),
        fieldDecl.isStatic(),
        fieldDecl.getTypeName(),
        varDecl.getId(),
        varDecl.getVariableInitializer(),
        currentGenericType.getName()
      );

      callBack.onFieldDeclaration(field, currentGenericType);

      currentGenericType.addField(field);
    }
  }

  /**
   * \brief Imports a constructor
   */
  public void visit(ConstructorDeclaration constructorDecl)
  {
    ArrayList genericDecl;
    if (constructorDecl.hasGenericDeclaration())
      genericDecl = constructorDecl.getGenericDeclaration().getGenericParameters();
    else
      genericDecl = new ArrayList();

    callBack.onCheckFormalParameters(constructorDecl.getFormalParameters());
    callBack.onCheckThrowList(constructorDecl.getThrowList());

    GenericMethod method = new GenericMethodImpl(
      constructorDecl,
      constructorDecl.isPublic(),
      constructorDecl.isProtected(),
      constructorDecl.isPrivate(),
      genericDecl,
      currentGenericType.getName(),
      false,
      false,
      false,
      false,
      null,
      MethodKindEnum.constructor,
      "constructor",
      FeedSymbolVisitorHelper.formalParameters2Arguments(constructorDecl.getFormalParameters()),
      FeedSymbolVisitorHelper.typeNameList2ThrowList(constructorDecl.getThrowList()),
      constructorDecl.hasBody()
    );

    callBack.onConstructorDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  /**
   * \brief Imports a destructor
   */
  public void visit(DestructorDeclaration destructorDecl)
  {
    GenericMethod method = new GenericMethodImpl(
      destructorDecl,
      true,
      false,
      false,
      new ArrayList(),
      currentGenericType.getName(),
      false,
      false,
      false,
      false,
      null,
      MethodKindEnum.destructor,
      "destructor",
      new ArrayList(),
      new HashSet(),
      destructorDecl.hasBlock()
    );

    callBack.onDestructorDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  /**
   * \brief Imports a cast operator
   */
  public void visit(CastOperatorDeclaration castOperator)
  {
    callBack.onCheckThrowList(castOperator.getThrowList());

    GenericMethod method  = new GenericMethodImpl(
      castOperator,
      true,
      false,
      false,
      new ArrayList(),
      currentGenericType.getName(),
      false,
      true,
      true,
      false,
      new ResultTypeImpl(true, castOperator.getCastedType()),
      MethodKindEnum.castOperator,
      castOperator.getCastedType().toString(),
      new ArrayList(),
      FeedSymbolVisitorHelper.typeNameList2ThrowList(castOperator.getThrowList()),
      castOperator.hasBlock()
    );

    callBack.onCastOperatorDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  /**
   * \brief Imports an operator
   */
  public void visit(OperatorDeclaration operatorDecl)
  {
    ArrayList genericDecl;
    if (operatorDecl.hasGenericDeclaration())
      genericDecl = operatorDecl.getGenericDeclaration().getGenericParameters();
    else
      genericDecl = new ArrayList();

    callBack.onCheckFormalParameters(operatorDecl.getFormalParameters());
    callBack.onCheckThrowList(operatorDecl.getThrowList());

    GenericMethod method = new GenericMethodImpl(
      operatorDecl,
      operatorDecl.isPublic(),
      operatorDecl.isProtected(),
      operatorDecl.isPrivate(),
      genericDecl,
      currentGenericType.getName(),
      false,
      operatorDecl.isMethodConst(),
      operatorDecl.isFinal(),
      false,
      FeedSymbolVisitorHelper.resultType2ResultType(operatorDecl.getType()),
      operatorDecl.isInfix() ? MethodKindEnum.infix : operatorDecl.isPrefix() ? MethodKindEnum.prefix : MethodKindEnum.postfix,
      operatorDecl.getOperator().getString(),
      FeedSymbolVisitorHelper.formalParameters2Arguments(operatorDecl.getFormalParameters()),
      FeedSymbolVisitorHelper.typeNameList2ThrowList(operatorDecl.getThrowList()),
      operatorDecl.hasBlock()
    );

    callBack.onOperatorDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  /**
   * \brief Imports a property
   */
  public void visit(PropertyDeclaration propertyDecl)
  {
    ArrayList genericDecl;
    if (propertyDecl.hasGenericDeclaration())
      genericDecl = propertyDecl.getGenericDeclaration().getGenericParameters();
    else
      genericDecl = new ArrayList();

    MethodKindEnum methodKind;
    if (propertyDecl.isPropertyArray())
      methodKind = MethodKindEnum.propertyArray;
    else
      methodKind = MethodKindEnum.property;

    callBack.onCheckFormalParameters(propertyDecl.getFormalParameters());
    callBack.onCheckThrowList(propertyDecl.getThrowList());

    GenericMethod method = new GenericMethodImpl(
      propertyDecl,
      propertyDecl.isPublic(),
      propertyDecl.isProtected(),
      propertyDecl.isPrivate(),
      genericDecl,
      currentGenericType.getName(),
      propertyDecl.isStatic(),
      propertyDecl.isMethodConst(),
      propertyDecl.isFinal(),
      false,
      FeedSymbolVisitorHelper.resultType2ResultType(propertyDecl.getResultType()),
      methodKind,
      propertyDecl.getName(),
      FeedSymbolVisitorHelper.formalParameters2Arguments(propertyDecl.getFormalParameters()),
      FeedSymbolVisitorHelper.typeNameList2ThrowList(propertyDecl.getThrowList()),
      propertyDecl.hasBlock()
    );

    callBack.onPropertyDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  /**
   * \brief Imports a method
   */
  public void visit(MethodDeclaration methodDecl)
  {
    ArrayList genericDecl;
    if (methodDecl.getMethodDeclarator().hasGenericDeclaration())
      genericDecl = methodDecl.getMethodDeclarator().getGenericDeclaration().getGenericParameters();
    else
      genericDecl = new ArrayList();

    callBack.onCheckFormalParameters(methodDecl.getMethodDeclarator().getFormalParameters());
    callBack.onCheckThrowList(methodDecl.getThrowList());

    GenericMethod method = new GenericMethodImpl(
      methodDecl,
      methodDecl.isPublic(),
      methodDecl.isProtected(),
      methodDecl.isPrivate(),
      genericDecl,
      currentGenericType.getName(),
      methodDecl.isStatic(),
      methodDecl.isMethodConst(),
      methodDecl.isFinal(),
      methodDecl.isCallback(),
      FeedSymbolVisitorHelper.resultType2ResultType(methodDecl.getResultType()),
      MethodKindEnum.method,
      methodDecl.getMethodDeclarator().getName(),
      FeedSymbolVisitorHelper.formalParameters2Arguments(methodDecl.getMethodDeclarator().getFormalParameters()),
      FeedSymbolVisitorHelper.typeNameList2ThrowList(methodDecl.getThrowList()),
      methodDecl.hasBlock()
    );

    callBack.onMethodDeclaration(method, currentGenericType);

    currentGenericType.addMethod(method);
  }

  public void visit(Initializer initializer)
  {
    GenericMethod method = new GenericMethodImpl(
      initializer,
      false,
      false,
      false,
      new ArrayList(),
      currentGenericType.getName(),
      true,
      false,
      false,
      false,
      null,
      MethodKindEnum.initializer,
      "initializer",
      new ArrayList(),
      java.util.Collections.EMPTY_SET,
      initializer.hasBlock()
    );

    callBack.onInitializer(method, currentGenericType);

    currentGenericType.setInitializer(method);
  }

  public void visit(Deinitializer deinitializer)
  {
    GenericMethod method = new GenericMethodImpl(
      deinitializer,
      false,
      false,
      false,
      new ArrayList(),
      currentGenericType.getName(),
      true,
      false,
      false,
      false,
      null,
      MethodKindEnum.deinitializer,
      "deinitializer",
      new ArrayList(),
      java.util.Collections.EMPTY_SET,
      deinitializer.hasBlock()
    );

    callBack.onDeinitializer(method, currentGenericType);

    currentGenericType.setDeinitializer(method);
  }
}
