/*
  $Id: GenericType.java,v 1.9 2004/04/01 21:27:55 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.symbol;

import net.nosica.parser.node.util.var.TypeName;
import java.util.Map;
import java.util.Set;
import java.util.Iterator;

/**
 * \brief Represents a Generic Type
 * that is, a type template that is able to generate other Types once
 * it is given its generic parameters
 */

public interface GenericType extends GenericSymbol
{
  /**
   * \brief return the name of this type
   *
   * \return TypeName of this type
   */
  TypeName      getName();

  /**
   * \brief set the type name
   *
   * \param typeName new TypeName of this type
   */
  void      setName(TypeName typeName);

  /**
   * \brief return kind of this type
   *
   * \return kind of type (class, interface, nestedclass, nestedinterface)
   */
  TypeKindEnum  getKind();

  /**
   * \brief Add a method to this type
   * There is only one destructor<BR>
   * A cast operator is unique : no cast operators to the same type can be defined<BR>
   * If !isPrimitive(), operators [=, :=, ::=, ==, !=] cannot be overloaded<BR>
   * Two methods having the same signature cannot be added (same arguments, same result type, same constness)
   *
   * \param method the method to add
   */
  void          addMethod(GenericMethod method);

  /**
   * \brief Add a field to this type
   * Two fields cannot have the same signature
   *
   * \param field the field to add
   */
  void          addField(Field field);

  /**
   * \brief Iterate through list of extends
   *
   * \return the list of extends as a Iterator<TypeName>
   */
  Iterator      iterateExtends();

  /**
   * \brief set extends set
   *
   * \param extendsSet set the list of extends
   */
  void          setExtends(Set extendsSet);

  /**
   * \brief return extends set
   *
   * \return Set<TypeName>
   */
  Set           getExtends();

  /**
   * \brief Iterate through list of implements
   *
   * \return the list of implements as a Iterator<TypeName>
   */
  Iterator      iterateImplements();

  /**
   * \brief set implements set
   *
   * \param implementsSet set the list of implements
   */
  void          setImplements(Set implementsSet);

  /**
   * \brief return implements
   *
   * \return Set<TypeName>
   */
  Set          getImplements();

  /**
   * \brief Iterate over fields of this type
   *
   * \return Iterator<Field>
   */
  Iterator      iterateFields();

  /**
   * \brief Return overloadables of this type
   * That is all methods that can be overloaded<BR>
   * Normal methods, properties, constructors can be overloaded
   *
   * \return Map<String, ArrayList<GenericMethod> >
   */
  Map           getOverloadables();

  /**
   * \brief Return list of cast operators
   * A cast operator is unique among a type
   *
   * \return Map<TypeName, GenericMethod>
   */
  Map           getCastOperators();

  /**
   * \brief Return destructor associated with this type
   *
   * \return destructor
   */
  GenericMethod getDestructor();

  /**
   * \brief Is this type a class ?
   *
   * \return true if this type is a class, false otherwise
   */
  boolean isClass();

  /**
   * \brief Is this type an interface ?
   *
   * \return true if this type is an interface, false otherwise
   */
  boolean isInterface();

  /**
   * \brief Is this type a nested type ?
   * ie, does this type has been defined inside another type ?
   *
   * \return true if type is nested, false otherwise
   */
  boolean isNested();

  /**
   * \brief set this type to be a nested type
   */
  void setNested();

  /**
   * \brief Is this type an abstract type
   * ie, can it be instantiated ?<BR>
   * requires : (isClass() && isAbstract()) || isInterface()
   *
   * \return true if type is abstract, false otherwise
   */
  boolean       isAbstract();

  /**
   * \brief Is this type a final type
   * ie, can it be subclassed ?<BR>
   * requires : isClass() && !isAbstract()
   *
   * \return true if type is final, false otherwise
   */
  boolean       isFinal();

  /**
   * \brief Is this type a primitive type ?
   * ie, is it a reference type or a pass-by-value type ?<BR>
   * requires : isClass()
   *
   * \return true if type is primitive, false otherwise
   */
  boolean       isPrimitive();

  /**
   * \brief Is this type a native type
   * ie, does impelmentation of this type left to the programmer in C ort C++ ?<BR>
   * requires : isPrimitive()
   *
   * \return true if type is native, false otherwise
   */
  boolean       isNative();

  /**
   * \brief Is this type an inner type
   * ie, is this class inner of another class ?<BR>
   * requires : isNested() && isClass() && !isAbstract()
   *
   * \return true if type is inner, false otherwise
   */
  boolean       isInner();

  /**
   * \brief Set this type an inner type
   */
  void          setInner();

  /**
   * \brief has this type an initializer ?
   *
   * \return true if type has an initializer
   */
  boolean hasInitializer();

  /**
   * \brief return initializer of type (if any)
   * requires hasInitializer
   *
   * \return initializer
   */
  GenericMethod getInitializer();

  /**
   * \brief set the initializer for this type
   *
   * \param initializer
   */
  void setInitializer(GenericMethod initializer);

  /**
   * \brief has this type a deinitializer ?
   *
   * \return true if type has a deinitializer
   */
  boolean hasDeinitializer();

  /**
   * \brief return deinitializer of type (if any)
   * requires hasDeinitializer
   *
   * \return deinitializer
   */
  GenericMethod getDeinitializer();

  /**
   * \brief set the deinitializer for this type
   *
   * \param deinitializer
   */
  void setDeinitializer(GenericMethod deinitializer);

  /**
   * \brief indicates if a destructor was defined in this type
   *
   * \return true if a a destructor exists
   */
  boolean hasDestructor();

  /**
   * \brief returns the associated alias table
   *
   * \return associated alias table
   */
  AliasTable getAliasTable();

  /**
   * \brief returns the alias table for methods
   *
   * \return Map<String, TypeName>
   */
  Map getMethodAliasTable();

  /**
   *  \brief iterates over all methods of this generic type
   *
   * \return Iterator<GenericMethod>
   */
  Iterator iterateMethods();
}