/*
  $Id: GenericTypeImpl.java,v 1.11 2004/04/01 21:27:55 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.symbol;

import net.nosica.parser.node.util.var.TypeName;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import java.util.Iterator;

public class GenericTypeImpl extends AccessibleGenericSymbolImpl implements GenericType
{
  // fields ********************************************************************
  private TypeName typeName;                            // name of type
  private TypeKindEnum typeKind;                        // kind (interface, class, nested, ...)
  private boolean isNested;                             // is this class nested in another class ?
  private boolean isAbstract;                           // is this class an abstract class ?
  private boolean isFinal;                              // is this class final ?
  private boolean isPrimitive;                          // is this class primitive ?
  private boolean isNative;                             // is this class native ?
  private boolean isInner;                              // is this class an inner class ?
  private Set extendsSet              = new HashSet();  // Set<TypeName>
  private Set implementsSet           = new HashSet();  // Set<TypeName>
  private ArrayList fields                     = new ArrayList();  // ArrayList<Field> because order of declaration is important
  private GenericMethod theDestructor = null;
  private GenericMethod initializer   = null;
  private GenericMethod deinitializer = null;
  private Map castOperators           = new HashMap();  // Map<TypeName, GenericMethod>
  private Map overloadables           = new HashMap();  // Map<String, ArrayList<GenericMethod> >
  private AliasTable aliasTable       = null;           // associated alias table
  private Map methodAliasTable        = null;           // Map<String, TypeName>

  // constructors **************************************************************

  /**
   * \brief Constructs a class
   * a class cannot be final and abstract at the same time<br>
   * if a class isNative, then it must also be primitive
   *
   * \param node node that define type
   * \param isPublic is class public ?
   * \param isProtected is class protected ?
   * \param isPrivate is class Private ?
   * \param genericParameterList list of generic params introduced by this type
   * \param isAbstract is class abstract ?
   * \param isFinal is class final ?
   * \param isPrimitive is class primitive ?
   * \param isNative is class native ?
   * \param typeName complete name of type
   * \param extendsSet list of extends
   * \param implementsSet list of implements
   * \param aliasTable the associated alias table
   * \param methodAliasTable the associated method alias table Map<String, TypeName>
   */
  public GenericTypeImpl(
    net.nosica.parser.node.typedeclaration.Symbol node,
    boolean isPublic,
    boolean isProtected,
    boolean isPrivate,
    ArrayList genericParameterList,
    TypeName enclosing,
    boolean isAbstract,
    boolean isFinal,
    boolean isPrimitive,
    boolean isNative,
    TypeName typeName,
    Set extendsSet,
    Set implementsSet,
    AliasTable aliasTable,
    Map methodAliasTable
  )
  {
    super(node, isPublic, isProtected, isPrivate, genericParameterList, enclosing);

    this.typeKind         = TypeKindEnum.topClass;
    this.isNested         = false;
    this.isAbstract       = isAbstract;
    this.isFinal          = isFinal;
    this.isPrimitive      = isPrimitive;
    this.isNative         = isNative;
    this.typeName         = typeName;
    this.isInner          = false;
    this.extendsSet       = extendsSet;
    this.implementsSet    = implementsSet;
    this.aliasTable       = aliasTable;
    this.methodAliasTable = methodAliasTable;

    setStatic(true);
  }

  /**
   * \brief Constructs an interface
   *
   * \param node node that defines interface
   * \param isPublic is class public ?
   * \param isProtected is class protected ?
   * \param isPrivate is class Private ?
   * \param genericParameterList list of generic params introduced by this type
   * \param TypeName enclosing
   * \param typeName complete name of type
   * \param extendsSet list of extends Set<TypeName>
   * \param aliasTable the associated alias table
   * \param methodAliasTable Map<String, TypeName>
   */
  public GenericTypeImpl(
    net.nosica.parser.node.typedeclaration.Symbol node,
    boolean isPublic,
    boolean isProtected,
    boolean isPrivate,
    ArrayList genericParameterList,
    TypeName enclosing,
    TypeName typeName,
    Set extendsSet,
    AliasTable aliasTable,
    Map methodAliasTable
  )
  {
    super(node, isPublic, isProtected, isPrivate, genericParameterList, enclosing);

    this.typeName         = typeName;
    this.typeKind         = TypeKindEnum.topInterface;
    this.isNested         = false;
    this.isAbstract       = false;
    this.isFinal          = false;
    this.isPrimitive      = false;
    this.isNative         = false;
    this.isInner          = false;
    this.extendsSet       = extendsSet;
    this.aliasTable       = aliasTable;
    this.methodAliasTable = methodAliasTable;

    setStatic(true);
  }

  // methods *******************************************************************
  public TypeName getName() { return typeName; }
  public void setName(TypeName typeName) { this.typeName = typeName; }
  public TypeKindEnum getKind() { return typeKind; }

  public void addMethod(GenericMethod method)
  {
    if (method.getKind() == MethodKindEnum.destructor)
    {
      // register destructor
      theDestructor = method;
    }
    else
    if (method.getKind() == MethodKindEnum.castOperator)
    {
      // register cast operators
      castOperators.put(method.getResultType().getTypeName(), method);
    }
    else
    {
      // register other kinds of methods : constructors, operators, properties, method

      // add method
      ArrayList methods;

      if (overloadables.containsKey(method.getName()))
        methods = (ArrayList)overloadables.get(method.getName());
      else
      {
        methods = new ArrayList();
        overloadables.put(method.getName(), methods);
      }

      // everything's fine : add method !
      methods.add(method);
    }
  }

  public void addField(Field field)
  {
    fields.add(field);
  }

  public Iterator iterateExtends() { return extendsSet.iterator(); }
  public void setExtends(Set extendsSet) { this.extendsSet = extendsSet; }
  public Set getExtends() { return extendsSet; }
  public Iterator iterateImplements() { return implementsSet.iterator(); }
  public void setImplements(Set implementsSet) { this.implementsSet = implementsSet; }
  public Set getImplements() { return implementsSet; }
  public Iterator iterateFields() { return fields.iterator(); }
  public Map getOverloadables() { return overloadables; }
  public Map getCastOperators() { return castOperators; }
  public GenericMethod getDestructor() { return theDestructor; }
  public boolean isClass() { return typeKind == TypeKindEnum.topClass || typeKind == TypeKindEnum.nestedClass; }
  public boolean isInterface() { return typeKind == TypeKindEnum.topInterface || typeKind == TypeKindEnum.nestedInterface; }
  public boolean isNested() { return isNested; }
  public void setNested() { isNested = true; }
  public boolean isAbstract() { return isAbstract; }
  public boolean isFinal() { return isFinal; }
  public boolean isPrimitive() { return isPrimitive; }
  public boolean isNative() { return isNative; }
  public boolean isInner() { return isInner; }
  public void setInner() { isInner = true; setStatic(false); }
  public boolean hasDestructor() { return theDestructor != null; }
  public boolean hasInitializer() { return initializer != null; }
  public GenericMethod getInitializer() { return initializer; }
  public void setInitializer(GenericMethod initializer) { this.initializer = initializer; }
  public boolean hasDeinitializer() { return deinitializer != null; }
  public GenericMethod getDeinitializer() { return deinitializer; }
  public void setDeinitializer(GenericMethod deinitializer) { this.deinitializer = deinitializer; }
  public AliasTable getAliasTable() { return aliasTable; }
  public Map getMethodAliasTable() { return methodAliasTable; }
  public Iterator iterateMethods() { return new MethodIterator(); }

  private class MethodIterator implements Iterator
  {
    private ArrayList methods = new ArrayList();
    private Iterator j;

    public MethodIterator()
    {
      if (hasInitializer())
        methods.add(getInitializer());
      if (hasDeinitializer())
        methods.add(getDeinitializer());
      if (hasDestructor())
        methods.add(getDestructor());

      Iterator it = overloadables.values().iterator();
      while (it.hasNext())
      {
        ArrayList m = (ArrayList)it.next();
        methods.addAll(m);
      }

      it = castOperators.values().iterator();
      while (it.hasNext())
        methods.add(it.next());

      j = methods.iterator();
    }

    public boolean hasNext() { return j.hasNext(); }
    public Object next() { return j.next(); }
    public void remove() { throw new UnsupportedOperationException(); }
  }
}