/*
  $Id: SymbolCallBack.java,v 1.5 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.symbol;

import net.nosica.framework.Localizable;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameList;
import net.nosica.parser.node.util.var.FormalParameters;

import net.nosica.symbol.GenericType;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.Field;

/**
 * \brief a callback used by FeedSymbolVisitor
 *
 * As the purpose of FeedSymbolVisitor is really to extract symbol
 * (GenericType, GenericMethod, Field), remaining tasks (such as reporting errors)
 * are left to this call back
 */
public interface SymbolCallBack
{
  /**
   * \brief called on package declaration
   *
   * \param packageName name of the package extracted
   * \param localizable localizable information
   */
  void onPackageDeclaration(TypeName packageName, Localizable localizable);

  /**
   * \brief called on new import declaration
   *
   * \param importName name of import extracted
   * \param localizable localizable information
   */
  void onImportDeclaration(TypeName importName, Localizable localizable);

  /**
   * \brief called when an import declaration points to no file
   *
   * \param importName name of "wrong" import
   * \param localizable localizable information
   */
  void onWrongImportDeclaration(TypeName importName, Localizable localizable);

  /**
   * \brief called whenever the new alias added was detected as a duplicate
   *
   * \param alias the duplicated alias
   * \param localizable localizable information
   */
  void onDuplicateAlias(String alias, Localizable localizable);

  /**
   * \brief called whenever a new class was extracted
   *
   * \param type the extracted type
   */
  void onClassDeclaration(GenericType type);

  /**
   * \brief called whenever a nested class declaration is about to be extracted
   *
   * \param nestedTypeName name of the nested class declaration
   */
  void onBeginNestedClassDeclaration(TypeName nestedTypeName);

  /**
   * \brief called at end of parsing a nested class declaration
   */
  void onEndNestedClassDeclaration();

  /**
   * \brief called whenever a new interface was extracted
   *
   * \param type the extracted type
   */
  void onInterfaceDeclaration(GenericType type);

  /**
   * \brief called whenever a nested interface declaration is about to be extracted
   *
   * \param nestedTypeName name of the nested interface declaration
   */
  void onBeginNestedInterfaceDeclaration(TypeName nestedTypeName);

  /**
   * \brief called at end of parsing a nested interface declaration
   */
  void onEndNestedInterfaceDeclaration();

  /**
   * \brief called whenever a new field declaration was extracted
   *
   * \param field the extracted field
   * \param enclosing the enclosing type
   */
  void onFieldDeclaration(Field field, GenericType enclosing);

  /**
   * \brief called when a new list of formal parameters need to be checked
   *
   * \param parameters the parameters to check
   */
  void onCheckFormalParameters(FormalParameters parameters);

  /**
   * \brief called whenever a new list of thrown exception need to be checked
   *
   * \param typeNameList the list of exceptions to check
   */
  void onCheckThrowList(TypeNameList typeNameList);

  /**
   * \brief called whenever a new constructor declaration was extracted
   *
   * \param method the extracted constructor
   * \param enclosing the enclosing type
   */
  void onConstructorDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new destructor declaration was extracted
   *
   * \param method the extracted destructor
   * \param enclosing the enclosing type
   */
  void onDestructorDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new cast operator declaration was extracted
   *
   * \param method the extracted cast operator
   * \param enclosing the enclosing type
   */
  void onCastOperatorDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new operator declaration was extracted
   *
   * \param method the extracted operator
   * \param enclosing the enclosing type
   */
  void onOperatorDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new property declaration was extracted
   *
   * \param method the extracted property
   * \param enclosing the enclosing type
   */
  void onPropertyDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new method declaration was extracted
   *
   * \param method the extracted method
   * \param enclosing the enclosing type
   */
  void onMethodDeclaration(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new initializer declaration was extracted
   *
   * \param method the extracted initializer
   * \param enclosing the enclosing type
   */
  void onInitializer(GenericMethod method, GenericType enclosing);

  /**
   * \brief called whenever a new deinitializer declaration was extracted
   *
   * \param method the extracted deinitializer
   * \param enclosing the enclosing type
   */
  void onDeinitializer(GenericMethod method, GenericType enclosing);
}