/*
  $Id: Helper.java,v 1.5 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.util;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponent;
import net.nosica.parser.node.util.var.TypeNameComponentImpl;

import net.nosica.compiler.application.Environment;

import java.util.Iterator;
import java.util.ArrayList;
import java.util.Stack;

/**
 * \brief contains various helper methods
 */

public class Helper
{
  /**
   * \brief Create a non generic type name from a (possibly) generic
   * parameter must not be an array, this is verified and interrupts the program if this is the case
   *
   * \param typeName a non array type name
   * \return the same typename without genericity information
   */
  public static TypeName createNonGenericTypeName(TypeName typeName)
  {
    Environment.check(!typeName.isArray(), "Cannot handle an array type name");

    ArrayList result = new ArrayList();
    Iterator it = typeName.getComponents().iterator();
    while (it.hasNext())
    {
      TypeNameComponent component = (TypeNameComponent)it.next();
      result.add(new TypeNameComponentImpl(component.getPackageElement(), null));
    }

    return new TypeNameImpl(result);
  }

  /**
   * \brief create a non array type name from a typename (possibly scalar and/or generic)
   *
   * \param typeName the typeName to convert
   * \return a new non array typeName
   */
  public static TypeName createNonArrayTypeName(TypeName typeName)
  {
    TypeName result = typeName;

    while (result.isArray())
      result = result.getArrayElementType();

    return result;
  }

  /**
   * \brief Create a radix type name
   * a radix type name is a type name that is scalar and generic
   *
   * \param typeName typeName to convert
   * \return new typeName such that (!typeName.isGeneric() && !typeName.isArray()) is true
   * \note dotted structure is conserved, only genericity and array information is given up
   */
  public static TypeName createRadixTypeName(TypeName typeName)
  {
    return createNonGenericTypeName(createNonArrayTypeName(typeName));
  }

  /**
   * \brief Reconstructs a typename from a model to add generic information
   * That is, apply genericity information of model to radix to create a new TypeName<BR>
   * CAUTION : this is not the inverse of method createNonGenericTypeName<BR>
   * createNonGenericTypeName removes genericity information on each component, whereas
   * this method add genericity information only on last component
   *
   * \param radix a non array non generic typename
   * \param model a type name with (possibly) some genericity information on last component
   * \return a new TypeName
   */
  public static TypeName reconstructGenericTypeNameFromRadix(TypeName radix, TypeName model)
  {
    if (!model.isGeneric())
      return radix;

    Environment.check(!radix.isArray(), "Cannot handle an array type name");

    ArrayList modelComponents = model.getComponents();

    ArrayList radixComponents = new ArrayList();
    radixComponents.addAll(radix.getComponents());

    for (int i = 0; i < modelComponents.size(); ++i)
    {
      TypeNameComponent curModelComponent =
        (TypeNameComponent)modelComponents.get(modelComponents.size() - i - 1);
      TypeNameComponent curRadixComponent =
        (TypeNameComponent)radixComponents.get(radixComponents.size() - i - 1);

      ArrayList genericDecl = null;
      if (curModelComponent.hasGenericExtension())
      {
        genericDecl = new ArrayList();
        genericDecl.addAll(curModelComponent.getGenericExtension());
      }

      radixComponents.set(
	    radixComponents.size() - i - 1,
        new TypeNameComponentImpl(
          curRadixComponent.getPackageElement(),
          genericDecl
        )
      );
    }

    return new TypeNameImpl(radixComponents);
  }

  /**
   * \brief Reconstructs a type name from a model to add array information
   * Apply array information of model to radix<BR>
   * This is the inverse method of createNonArrayTypeName<BR>
   *
   * \param radix the type to reconstruct (possibly generic, possibly array)
   * \param model the model that contains array information to apply
   * \return a new type name
   */

  public static TypeName reconstructArrayTypeNameFromRadix(TypeName radix, TypeName model)
  {
    if (!model.isArray())
      return radix;

    Stack arrayInfo = new Stack();
    TypeName tmp = model;
    TypeName result = radix;

    while (tmp.isArray())
    {
      arrayInfo.push(new Integer(tmp.getDim()));
      tmp = tmp.getArrayElementType();
    }

    while (!arrayInfo.isEmpty())
    {
      Integer dim = (Integer)arrayInfo.pop();
      result = new TypeNameImpl(dim.intValue(), result);
    }

    return result;
  }

  /**
   * \brief Reconstructs a typeName from a model to add generic and array information
   * For the same reason given in reconstructGenericTypeNameFromRadix, this method is not the
   * inverse of method createRadixTypeName
   *
   * \param radix the typename to reconstruct
   * \param model the model that contains genericity an array information to apply
   * \return a new typeName
   */
  public static TypeName reconstructTypeNameFromRadix(TypeName radix, TypeName model)
  {
    return reconstructArrayTypeNameFromRadix(
              reconstructGenericTypeNameFromRadix(radix, createNonArrayTypeName(model)),
              model);
  }
}
