/*
  $Id: IndentedPrintWriter.java,v 1.4 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.util;

import java.io.PrintWriter;
import java.io.Writer;
import java.io.OutputStream;

/**
 * \brief a PrintWriter with indenting-able capacities
 * Used widely to gives classes debugger a simple way to display data nicely.
 */

public class IndentedPrintWriter extends PrintWriter
{
  private int     indent = 0;
  private int     indentSize = 2;
  private String  indentPattern = " ";
  private char[]  indentString = new char[0];
  private boolean needIndent = false;

  /**
   * Create a PrintWriter with an OutputStream
   *
   * \param out the OutputStream to use
   */
  public IndentedPrintWriter(OutputStream out) { super(out); }

  /**
   * Create a PrintWriter with a Writer
   *
   * \param out the Writer to use
   */
  public IndentedPrintWriter(Writer out) { super(out); }

  /**
   * \brief print a boolean value
   */
  public void print(boolean v) { beforePrint(); super.print(v); }

  /**
   * \brief print a char
   */
  public void print(char    v) { beforePrint(); super.print(v); }

  /**
   * \brief print an int
   */
  public void print(int     v) { beforePrint(); super.print(v); }

  /**
   * \brief print a string
   */
  public void print(String  v) { beforePrint(); super.print(v); }

  /**
   * \brief print a boolean value followed by a newline
   */
  public void println(boolean v) { beforePrint(); super.println(v);  needIndent=true; }

  /**
   * \brief print a char followed by a newline
   */
  public void println(char    v) { beforePrint(); super.println(v);  needIndent=true; }

  /**
   * \brief print an int followed by a newline
   */
  public void println(int     v) { beforePrint(); super.println(v);  needIndent=true; }

  /**
   * \brief print a string followed by a newline
   */
  public void println(String  v) { beforePrint(); super.println(v);  needIndent=true; }

  /**
   * \brief print a newline
   */
  public void println()          { beforePrint(); super.println();   needIndent=true; }

  private void beforePrint()
  {
    if (needIndent)
    {
      super.flush();
      super.print(indentString);
      needIndent = false;
    }
  }

  private void cacheIndent()
  {
    StringBuffer tmp = new StringBuffer();

    for (int i = 0; i < indent * indentSize; ++i)
      tmp.append(indentPattern);

    indentString = tmp.toString().toCharArray();
  }

  /**
   * \brief Add an indent level to the output stream
   */
  public void addIndent()
  {
    ++indent;
    cacheIndent();
  }

  /**
   * \brief Remove an indent level to the output stream
   */
  public void decIndent()
  {
    --indent;
    if (indent < 0)
      throw new RuntimeException("indentation cannot be negative");
    cacheIndent();
  }

  /**
   * \brief Sets the size of the indent to use, it is 2 by default
   *
   * \param indentSize the new indentSize to use
   */
  public void setIndentSize(int indentSize) { this.indentSize = indentSize; }

  /**
   * \brief return the currently used indent size
   *
   * \return indent size
   */
  public int getIndentSize() { return indentSize; }

  /**
   * \brief Set the new pattern to use for indentation
   *
   * \param pattern the pattern to use
   */
  public void setIndentPattern(String pattern) { this.indentPattern = pattern; }

  /**
   * \brief return the current pattern used to indent stream
   *
   * \return current pattern
   */
  public String getIndentPattern() { return indentPattern; }
}
