/*
  $Id: LightResolver.java,v 1.2 2004/04/01 21:27:55 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.util;

import net.nosica.parser.node.util.var.TypeName;
import net.nosica.parser.node.util.var.TypeNameImpl;
import net.nosica.parser.node.util.var.TypeNameComponentImpl;
import net.nosica.parser.node.util.generic.GenericParameter;

import net.nosica.symbol.AliasTable;
import net.nosica.symbol.Argument;
import net.nosica.symbol.Field;
import net.nosica.symbol.GenericMethod;
import net.nosica.symbol.GenericType;
import net.nosica.symbol.SymbolException;

import net.nosica.application.ApplicationContext;
import net.nosica.application.ApplicationContextImpl;
import net.nosica.application.FileLoaderImpl;
import net.nosica.application.SymbolEnvironment;
import net.nosica.application.SymbolEnvironmentImpl;
import net.nosica.application.TypeNameResolverImpl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;

public class LightResolver
{
  private SymbolEnvironment symbolEnv = SymbolEnvironmentImpl.instance();
  private ApplicationContext context  = ApplicationContextImpl.instance();

  public void resolveTypes()
  {
    Iterator it = symbolEnv.iterateTypeName();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      try
      {
        GenericType type = symbolEnv.getGenericType(typeName);

        // add missing aliases for proper resolving
        addAliases(type);

        // resolve type
        resolveType(type);
      }
      catch (SymbolException e)
      {
        context.terminate("Internal error : " + e.getMessage());
      }
    }
  }

  public TypeName resolve(TypeName typeName, AliasTable aliasTable)
  {
    try
    {
      typeName = TypeNameResolverImpl.instance().resolveExceptGeneric(typeName, aliasTable, FileLoaderImpl.instance());
    }
    catch (SymbolException e)
    {
      System.err.println("Warning : Could not resolve type " + typeName.toString() + " from " + typeName.localization());
    }

    return typeName;
  }

  public Set resolveSet(Iterator it, AliasTable aliasTable)
  {
    Set extendsSet = new HashSet();
    while (it.hasNext())
    {
      TypeName typeName = (TypeName)it.next();
      typeName = resolve(typeName, aliasTable);
      extendsSet.add(typeName);
    }

    return extendsSet;
  }

  public void resolveMethod(GenericMethod m, AliasTable aliasTable)
  {
    // resolve result type
    if (!m.isSub())
    {
      m.getResultType().setTypeName(resolve(m.getResultType().getTypeName(), aliasTable));
    }

    // resolve parameters
    Iterator it = m.iterateArguments();
    while (it.hasNext())
    {
      Argument arg = (Argument)it.next();
      arg.setTypeName(resolve(arg.getTypeName(), aliasTable));
    }

    // resolve throw list
    m.setThrowList(resolveSet(m.iterateThrowList(), aliasTable));
  }

  private void addAliases(GenericType type)
  {
    // if this type has an enclosing type
    // then first add aliases to this enclosing type
    TypeName enclosingTypeName = type.getEnclosing();
    if (enclosingTypeName != null)
    {
      try
      {
        GenericType enclosing = symbolEnv.getGenericType(enclosingTypeName);
        addAliases(enclosing);
      }
      catch (SymbolException e)
      {
        context.terminate("Internal error : " + e.getMessage());
      }
    }

    // add some alias for type name resolving
    Iterator it = type.iterateGenericParameters();
    while (it.hasNext())
    {
      GenericParameter param = (GenericParameter)it.next();

      try
      {
        type.getAliasTable().addAlias(param.getName(), param.getName());
      }
      catch (SymbolException e)
      {
        // can happen if we already add aliases for this type via a nested type
      }
    }

    ArrayList thisComponents = new ArrayList();
    thisComponents.add(new TypeNameComponentImpl("This", new ArrayList()));
    TypeName thisTypeName = new TypeNameImpl(thisComponents);

    try
    {
      type.getAliasTable().addAlias(thisTypeName, type.getName());
    }
    catch (SymbolException e)
    {
      // can happen because it is already defined in enclosing AliasTable
    }
  }

  public void resolveType(GenericType type)
  {
    // resolve extends
    Set extendsSet = resolveSet(type.iterateExtends(), type.getAliasTable());
    type.setExtends(extendsSet);

    // resolve implements
    Set implementsSet = resolveSet(type.iterateImplements(), type.getAliasTable());
    type.setImplements(implementsSet);

    // resolve generic information (implements and defaultType)
    {
      Iterator it = type.iterateGenericParameters();
      while (it.hasNext())
      {
        GenericParameter parm = (GenericParameter)it.next();
        if (parm.hasImplements())
          parm.setImplements(resolve(parm.getImplements(), type.getAliasTable()));
        if (parm.hasDefaultType())
          parm.setDefaultType(resolve(parm.getDefaultType(), type.getAliasTable()));
      }
    }

    // resolve each fields
    {
      Iterator it = type.iterateFields();
      while (it.hasNext())
      {
        Field field = (Field)it.next();
        field.setTypeName(resolve(field.getTypeName(), type.getAliasTable()));
      }
    }

    // resolve each methods
    Iterator it = type.iterateMethods();
    while (it.hasNext())
    {
      GenericMethod m = (GenericMethod)it.next();
      resolveMethod(m, type.getAliasTable());
    }
  }
}