/*
  $Id: XHTMLWriter.java,v 1.4 2004/03/04 22:16:17 djobet Exp $
  Copyright 2001-2002 David Jobet(david.jobet@free.fr) and Eric Nicolas(enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package net.nosica.util;


import java.util.Stack;

/**
 * \brief a simple helper class to write xhtml document without the need to worry to close tag
 *
 * this class helps writing xhtml tag by remembering which tags are opened. This way, you don't have to remember
 * which tags are opened. You can start as many tags as wanted and later close all of them by asking
 * the writer to close the first one (all intermediate tags will be closed as well)
 */
public class XHTMLWriter
{
  private static final XHTMLEntity ADD_INDENT = new XHTMLEntity("");

  private IndentedPrintWriter writer;
  private Stack pendingTag = new Stack();

  public XHTMLWriter(IndentedPrintWriter writer)
  {
    this.writer = writer;
  }

  /**
   * \brief will print xml version
   */
  public void printXMLVersion()
  {
    writer.println("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
  }

  /**
   * \brief will print used DTD
   */
  public void printDTD()
  {
    writer.println("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">");
  }

  /**
   * \brief start a tag, storing it for later handling, and returns it
   *
   * \param xmlEntity the entity to store in created tag
   * \param created tag
   */
  public XHTMLEntity beginTag(String xmlEntity)
  {
    XHTMLEntity entity = new XHTMLEntity(xmlEntity);
    pendingTag.push(entity);
    writer.print("<" + xmlEntity + ">");

    return entity;
  }

  /**
   * \brief start a tag, storing it for later handling, and returns it
   *
   * \param xmlEntity the entity to store in created tag
   * \param option is stored inside starting tag
   * \param created tag
   */
  public XHTMLEntity beginTag(String xmlEntity, String option)
  {
    XHTMLEntity entity = new XHTMLEntity(xmlEntity);
    pendingTag.push(entity);
    writer.print("<" + xmlEntity + " " + option + ">");

    return entity;
  }

  /**
   * \brief start a tag, storing it for later handling, and returns it
   *
   * this version adds line feed
   *
   * \param xmlEntity the entity to store in created tag
   * \param created tag
   */
  public XHTMLEntity beginTagLn(String xmlEntity)
  {
    XHTMLEntity entity = beginTag(xmlEntity);
    writer.println();

    return entity;
  }

  /**
   * \brief start a tag, storing it for later handling, and returns it
   *
   * this version add line feed
   *
   * \param xmlEntity the entity to store in created tag
   * \param option is stored inside starting tag
   * \param created tag
   */
  public XHTMLEntity beginTagLn(String xmlEntity, String option)
  {
    XHTMLEntity entity = beginTag(xmlEntity, option);
    writer.println();

    return entity;
  }

  /**
   * \brief end a previously strated tag. All other intermediate tag are closed
   *
   * \param xmlEntity the entity to close
   */
  public void endTag(XHTMLEntity entity)
  {
    XHTMLEntity tag = null;

    do
    {
      tag = (XHTMLEntity)pendingTag.pop();
      if (tag == ADD_INDENT)
      {
        writer.decIndent();
      }
      else
      {
        writer.print("</" + tag.getTag() + ">");
      }
    }
    while (entity != tag);
  }

  /**
   * \brief end a previously strated tag. All other intermediate tag are closed
   *
   * This version add line feed
   *
   * \param xmlEntity the entity to close
   */
  public void endTagLn(XHTMLEntity entity)
  {
    endTag(entity);
    writer.println();
  }

  /**
   * \brief print an empty tag
   *
   * \param xmlEntity the empty entity to print
   */
  public void printTag(String xmlEntity)
  {
    writer.print("<" + xmlEntity + "/>");
  }

  /**
   * \brief print an empty tag with an option
   *
   * \param xmlEntity the empty entity to print
   * \param option option to enclose in empty entity
   */
  public void printTag(String xmlEntity, String option)
  {
    writer.print("<" + xmlEntity + " " + option + "/>");
  }

  /**
   * \brief print an empty tag
   *
   * This version adds line feed
   *
   * \param xmlEntity the empty entity to print
   */
  public void printTagLn(String xmlEntity)
  {
    printTag(xmlEntity);
    writer.println();
  }

  /**
   * \brief print an empty tag with an option
   *
   * This version add line feed
   *
   * \param xmlEntity the empty entity to print
   * \param option option to enclose in empty entity
   */
  public void printTagLn(String xmlEntity, String option)
  {
    printTag(xmlEntity, option);
    writer.println();
  }

  /**
   * \brief print a comment
   *
   * \param comment the comment to print
   */
  public void printComment(String comment)
  {
    writer.print("<!--" + comment + "-->");
  }

  /**
   * \brief print a comment
   *
   * This version adds line feed
   *
   * \param comment the comment to print
   */
  public void printCommentLn(String comment)
  {
    printComment(comment);
    writer.println();
  }

  /**
   * \brief print some text
   *
   * \param text the text to print
   */
  public void print(String text)
  {
    writer.print(text);
  }

  /**
   * \brief print some text
   *
   * This version adds line feed
   *
   * \param text the text to print
   */
  public void println(String text)
  {
    writer.println(text);
  }

  /**
   * \brief prints line feed
   */
  public void println()
  {
    writer.println();
  }

  /**
   * \brief add indentation to the stream
   */
  public void addIndent()
  {
    writer.addIndent();
    pendingTag.push(ADD_INDENT);
  }
}