#!/usr/bin/perl -w
use strict;

#
# command line :
# runtest.pl [startDir [nosFile] ]
#

my $nbTest = 0;
my @errors = ();
my @internalerrors = ();

my $JAVA="java";
my $MAKE="make";
my $DIFF="diff";
my $NOSICA_JAR="../nosica.jar";
my $NOSICACC="";
my $ALTERNATE_NOSICACC="../nosicacc";
my $NOSICA_MAIN="net.nosica.compiler.Main";

if (-f $NOSICA_JAR)
{
  $NOSICACC = "$JAVA -cp $NOSICA_JAR $NOSICA_MAIN"
}

if (-f $ALTERNATE_NOSICACC)
{
  $NOSICACC = $ALTERNATE_NOSICACC;
}

# *****************************
# Handles a test that passed OK
#
# Parameters :
#   $exitValue return value of program that exited
#   $dir the complete path to the file
#   $file the name of the file compiled
#
# if exitValue == 0, then this test succeeded
# if exitValue != 0, then this test failed
#
# This method prints "OK" if the test succeeded
# it prints "Error !" if the test failed and records the name of the file
# that failed in the globale array @errors
# *****************************
sub testOK
{
  my $exitValue = shift;
  my $dir = shift;
  my $file = shift;
  my $logFile = shift;

  # everything went fine
  if ($exitValue == 0)
  {
    print "OK";
    unlink $logFile;
  }
  else
  {
    print "Error !";
    # This program should have thrown errors
    # note that it failed
    push @errors, $dir.$file;
  }
}

# ***************************
# Search a file for a text
#
# Parameters :
#  $file name of the file to look in
#  $text text to search in file
#
#  returns 0 if $text was not found in $file, 1 otherwise
# ***************************

sub fileContainsText
{
  my $file = shift;
  my $text = shift;
  my $foundText = 0;
  open(HDL, "<$file") || die "Cannot open $file";
  while (<HDL>)
  {
    if (/$text/)
    {
      $foundText = 1;
    }
  }
  close(HDL);

  return $foundText;
}

# ********************************************************
# Trim all spaces around the sting (including end of line)
#
# Parameters:
#   $text text to trim
#
#   returns the trimmed text
# ********************************************************

sub trim
{
  my $text = shift;

  $/ = "\n";
  chomp($text);
  $/ = "\r";      # ms-windows hack: also remove the second EOL marker
  chomp($text);

  $text =~ s/^\s*(.+)\s*$/$1/;
  return $text;
}

# **********************************
# Handles a test that passed bad
#
# Parameters
#   $exitValue return value of program that exited
#   $dir the complete path to the file
#   $file the name of the file compiled
#
# if exitValue == 0, then this test failed
# if exitValue != 0, then this test perhaps succeeded
#
# if the test failed, it prints "Error !" and records the name of the file in
# global variable @errors
# if the test perhaps succeeded.
#   if reference file exists, then the error code contained in it must be the one
#     that was outputed by test
#     If it is the case, the test succeeds : "OK" is printed
#     else this is an error, "Error !" is printed and the name of the file is recorded
#     in global variable @errors
#   else
#     the output is parsed to find an error code
#     if the error code is found, the reference file is creted with it
#     else this is an error
# **********************************

sub testError
{
  my $exitValue = shift;
  my $dir = shift;
  my $file = shift;
  my $nosFile = shift;
  my $logFile = shift;

  # There were errors ...
  if ($exitValue == 0)
  {
    print "Error !";
    # This program should have run OK
    push @errors, $dir.$file;
  }
  else
  {
    # source was designed to fail and it failed
    my $refFile = $dir.$nosFile.".ref";

    if (-f $refFile)
    {
      # ref file exists
      my $waitedErrorCode = trim(`cat $refFile`);
      my $foundWaitedErrorCode = fileContainsText($logFile, $waitedErrorCode);

      if ($foundWaitedErrorCode == 0)
      {
        print "Error !";
        push @errors, $dir.$file;
      }
      else
      {
        print "OK";
        unlink $logFile;
      }
    }
    else
    {
      print "refFile : $refFile\n";
      print "Creating reference file ...";
      # ref file was not found : creating it

      my $errorCode = "";
      open(REF_HDL, "<$logFile") || die "Cannot open $logFile";
      while (<REF_HDL>)
      {
        if (/^ERR:(.*)$/)
        {
          $errorCode = $1;
        }
      }
      close(REF_HDL);

      if ($errorCode ne "")
      {
        `echo $errorCode>$refFile`;
        print $errorCode;
        unlink $logFile;
      }
      else
      {
        print "Could not create reference file";
      }
    }
  }
}

# ****************************************************************************************
# Handles a test that passed with warning
#
# Parameters
#   $exitValue value of program that exited
#   $dir the complete path to the file
#   $file the name of the file compiled (e.g : TestWarning.nos)
#   $nosFile the nosica class name (e.g : TestWarning)
#
# if the exit value was 0, then this test succeeded
# if the exit value was not 0, then this test failed
#
# if the test failed, it prints "Error !" and records the name of the file in
# global variable @errors
# if the test succeeded.
#   if reference file exists, then the error code contained in it must be the one
#     that was outputed by test
#     If it is the case, the test succeeds : "OK" is printed
#     else this is an error, "Error !" is printed and the name of the file is recorded
#     in global variable @errors
#   else
#     the output is parsed to find an error code
#     if the error code is found, the reference file is creted with it
#     else this is an error
# ****************************************************************************************

sub testWarning
{
  my $exitValue = shift;
  my $dir = shift;
  my $file = shift;
  my $nosFile = shift;
  my $logFile = shift;

  if ($exitValue == 0)
  {
    # source was designed to run with warning
    my $refFile = $dir.$nosFile.".ref";

    if (-f $refFile)
    {
      # ref file exists
      my $waitedWarningCode = trim(`cat $refFile`);

      my $foundWaitedWarningCode = fileContainsText("$logFile", $waitedWarningCode);

      if ($foundWaitedWarningCode == 0)
      {
        print "Error !";
        push @errors, $dir.$file;
      }
      else
      {
        print "OK";
        unlink $logFile;
      }
    }
    else
    {
      print "refFile : $refFile\n";
      print "Creating reference file ...";
      # ref file was not found : creating it

      my $warningCode = "";
      open(REF_HDL, "<$logFile") || die "Cannot open $logFile";
      while (<REF_HDL>)
      {
        if (/^WRN:(.*)$/)
        {
          $warningCode = $1;
        }
      }
      close(REF_HDL);

      if ($warningCode ne "")
      {
        `echo $warningCode>$refFile`;
        print $warningCode;
        unlink $logFile;
      }
      else
      {
        print "Could not create reference file";
      }
    }
  }
  else
  {
    print "Error !";
    # This program should have run OK
    push @errors, $dir.$file;
  }
}

# **************************************************************************************************
# Handles a test that should have passed OK, then compil the resulting source and analyse its output
#
#   $exitValue value of program that exited
#   $dir the complete path to the file
#   $file the name of the file compiled (e.g : TestWarning.nos)
#   $nosFile the nosica class name (e.g : TestWarning)
#
# if the exit value was 0, then compilation of file succeeded, it is now time to compil the resulting
#   source, execute it, and compare it with a previously grabbed output
# if the exit value was not 0, then this test failed
#
# if the test failed, it prints "Error !" and records the name of the file in
# global variable @errors
# ****************************************************************************************************
sub testExec
{
  my $exitValue = shift;
  my $dir = shift;
  my $file = shift;
  my $nosFile = shift;
  my $logFile = shift;

  if ($exitValue == 0)
  {
    # Test ran OK, now cd into generation directory
    # launch a make here
    # handle the result of the make
    # run the resulting test program
    # monitor it's output

    my $command = $dir."generated/".$nosFile." 1>$logFile 2>&1";
    my $testExitValue = system($command) >> 8;

    if ($testExitValue == 0)
    {
      my $refFile = $dir.$nosFile.".ref";

      if (-f $refFile)
      {
        $command = "$DIFF $logFile ".$dir.$nosFile.".ref";
        my $diffExitValue = system($command) >> 8;

        if ($diffExitValue == 0)
        {
          print "OK";
          unlink $logFile;
        }
        else
        {
          print "Error !";
          push @errors, $dir.$file;
        }
      }
      else
      {
        print "refFile : $refFile\n";
        print "Creating reference file ...";

        # ref file was not found : creating it
        system("mv -f $logFile $refFile");
      }
    }
    else
    {
      print "Error !";
      #this program should have run OK
      push @errors, $dir.$file;
    }
  }
  else
  {
    print "Error !";
    # this program should have run OK
    push @errors, $dir.$file;
  }
}

#
# test($dir, $file)
#
sub test
{
  my $dir = shift;
  my $file = shift;

  if ($file =~ /^(.*)(OK|Error|Warning|Exec)\.nos$/)
  {
    my $base = $1;
    my $extension = $2;
    my $nosFile = $base.$extension;

    print "compiling nosica file [".$nosFile."] ... ";
    ++$nbTest;

    my $logFile = $dir.$nosFile;
    if ($logFile =~ /^\.\//)
    {
      # detect a "./" in front of logFile and remove it
      $logFile =~ s/^\.\/(.*)/$1/g;
    }
    $logFile =~ s/\//_/g;         # replace '/' by '_'
    $logFile = $logFile.".log";

    # generate a local nosicacc.properties to set 
    # code.output
    # sourcepaths
    my $reverseDir = $dir;
    $reverseDir =~ s/(\w+)\//\.\.\//g;

    my $tempNosicaProperties = $dir."generated/nosicacc.properties";
    open(TMP_HDL, ">$tempNosicaProperties") || die "cannot generate $tempNosicaProperties";
    print TMP_HDL "code.output = ".$dir."generated\n";
    print TMP_HDL "sourcepaths= ../../stdlib/, $dir\n";
    print TMP_HDL "NOSICA_NATIVESTDLIB_INCLUDE_PATH = ".$reverseDir."../../../nativestdlib/c\n";
    print TMP_HDL "NOSICA_NATIVESTDLIB_LIB_PATH = ".$reverseDir."../../../nativestdlib/c/lib\n";
    close(TMP_HDL);

    my $command="$NOSICACC -s nosicacc.properties -p $tempNosicaProperties $nosFile 1>$logFile 2>&1";

    my $exitValue = system($command) >> 8;

    if ($exitValue == 2)
    {
      print "Internal error !";
      push @internalerrors, $dir.$file;
    }
    elsif ($extension eq "OK")
    {
      testOK($exitValue, $dir, $file, $logFile);
    }
    elsif ($extension eq "Error")
    {
      testError($exitValue, $dir, $file, $nosFile, $logFile);
    }
    elsif ($extension eq "Warning")
    {
      testWarning($exitValue, $dir, $file, $nosFile, $logFile);
    }
    elsif ($extension eq "Exec")
    {
      testExec($exitValue, $dir, $file, $nosFile, $logFile);
    }
    else
    {
      print "Could not recognize file type ... giving up !";
    }

    print "\n";
  }
}

#
# recBrowsePath($dir)
#
sub recBrowsePath
{
  my $dir = shift;
  local *DIR_HDL;

  print "entering directory [$dir]\n";

  # first, let's find directory and recurse into it
  opendir DIR_HDL, $dir || die "cannot open '$dir'";
  while (my $file = readdir DIR_HDL)
  {
    if (-d $dir.$file)
    {
      if ($file eq "." || $file eq ".." || $file eq "CVS")
      {
        next;
      }

      # recursively
      recBrowsePath($dir.$file."/");
    }
  }
  closedir(DIR_HDL);

  # now, only find nosica files
  opendir DIR_HDL, $dir || die "cannot open '$dir'";
  while (my $file = readdir DIR_HDL)
  {
    if (-f $dir.$file)
    {
      test($dir, $file);
    }
  }

  closedir(DIR_HDL);

  print "leaving directory [$dir]\n";
}

#
# main
#

my $startDir = "./";
my $nosFile = "";
my $timestamp1 = time();

# command line analysis
if ($#ARGV > -1)
{
  $startDir = $ARGV[0];

  # if startDir does not contain an ending /, let's add it
  if (!($startDir =~ /^(.*)\/$/))
  {
    $startDir .= "/";
  }

  if ($#ARGV > 0)
  {
    $nosFile = $ARGV[1];
  }
}

if (-d $startDir)
{
}
else
{
  print "[".$startDir."] is not a directory !\n";
  exit(1);
}

if ($NOSICACC eq "")
{
  print "No compiler available : I did not found nosica.jar or nosicacc.bin\n";
  print "Either launch [make] or [make srcbin] or [make jarbin] in upper directory\n";
  exit(1);
}
print "Command used to launch nosica compiler : ".$NOSICACC."\n";

# we have specified a nos file on command line
if ($nosFile eq "")
{
  recBrowsePath($startDir);
}
else
{
  # rec test on path $startDir
  test($startDir, $nosFile.".nos");
}

my $timestamp2 = time();

# error reporting
print "Ran $nbTest tests\n";

if ($#errors != -1 || $#internalerrors != -1)
{
  if ($#errors != -1)
  {
    print "Files in error :\n";

    for (@errors)
    {
      print "\t".$_."\n";
    }
  }

  if ($#internalerrors != -1)
  {
    print "Files in internal error :\n";

    for (@internalerrors)
    {
      print "\t".$_."\n";
    }
  }
}
else
{
  print "All right dude!\n";
}

# time information
print "Elapsed ".($timestamp2 - $timestamp1)." seconds\n";
