/*
  $Id: alloc.c,v 1.1 2004/04/18 00:52:04 djobet Exp $
  Copyright 2001-2002 David Jobet (david.jobet@free.fr) and Eric Nicolas (enicolas@dvdfr.com).
  Copyright 2003-2004 David Jobet(david.jobet@free.fr).
  This file is part of Nosica, a compiler for the Nosica language.

  Nosica is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Nosica is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Nosica; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// Nosica standard library ----------------------------------------------------

// Implementation of the memory checker
// This code monitors all malloc() and free() in a Nosica program and detects:
// - Memory leaks upon program termination
// - Memory corruption just before or just after the block
// And *may help* detecting;
// - Using a freed block
// - Multiple free() on the same block

#include "types0.h"
#include <stdlib.h>
#include <stdio.h>

// --- Declarations

#define PADDING_BEGIN 0xdead1234
#define PADDING_END   0xdead5678
#define PADDING_DEAD  0x66666666
#define PADDING_DEAD1 0x66

typedef struct _Chunk Chunk;
struct _Chunk
{
  Chunk* next;          // link to the next allocated memory chunk
  char*  file;          // source file for the allocation location
  int    line;          // source line for the allocation location
  word32 size;          // size of the usefull part of the chunk, in word32
  word32 length;        // size of the usefull part of the chunk, in bytes
  word32 data[0];       // data (actual size = 'size+2')
                        // data[0] and data[size+1] are security padding
};

static Chunk *chunks;
void  memDisplay();

// --- Implementation

void memMallocInit()
{
  chunks = 0;
}

void memMallocDone()
{
  Chunk* chunk;

  // Report memory leaks and memory corruption
  chunk=chunks;
  while(chunk)
  {
    printf("MEM:LEAK   : %d-byte block allocated at %s:%d\n", chunk->length, chunk->file, chunk->line);
    if (chunk->data[0] != PADDING_BEGIN)
    {
      printf("MEM:CORRUPT: %d-byte block allocated at %s:%d damaged before begining\n", chunk->length, chunk->file, chunk->line);
      abort();
    }
    if (chunk->data[chunk->size+1] != PADDING_END)
    {
      printf("MEM:CORRUPT: %d-byte block allocated at %s:%d damaged after the end\n", chunk->length, chunk->file, chunk->line);
      abort();
    }
    //
    chunk = chunk->next;
  }
}

void* memMalloc(word32 length, char *file, int line)
{
  Chunk *chunk;
  word32 size;

  if (length==0) return 0;

  // Allocate chunk
  size  = (length + 3)/4;
  chunk = (Chunk*)malloc(sizeof(Chunk)+4*size+8);

  // Fill in chunk information
  chunk->file=file;
  chunk->line=line;
  chunk->size=size;
  chunk->length=length;
  chunk->data[0] = PADDING_BEGIN;
  chunk->data[size+1] = PADDING_END;

  // Link into the list of chunks
  chunk->next=chunks;
  chunks=chunk;

  // Return allocated block
  return &(chunk->data[1]);
}

void memFree(void* block)
{
  Chunk *chunk;
  if (block==0) return;

  // Retrieve address of chunk
  chunk = (Chunk*)((word8*)block-sizeof(Chunk)-4);

  // Report memory corruption
  if (chunk->file   == (char*)PADDING_DEAD ||
      chunk->line   == PADDING_DEAD ||
      chunk->size   == PADDING_DEAD ||
      chunk->length == PADDING_DEAD)
  {
    printf("MEM:CORRUPT: block is invalid and probably is being freed twice");
    abort();
  }
  if (chunk->data[0] != PADDING_BEGIN)
  {
    printf("MEM:CORRUPT: %d-byte block allocated at %s:%d damaged before begining\n", chunk->length, chunk->file, chunk->line);
    abort();
  }
  if (chunk->data[chunk->size+1] != PADDING_END)
  {
    printf("MEM:CORRUPT: %d-byte block allocated at %s:%d damaged after the end\n", chunk->length, chunk->file, chunk->line);
    abort();
  }

  // Unlink from the list of chunks
  if (chunks == chunk)
  {
    // it was the top of the list, change the top
    chunks=chunk->next;
  }
  else
  {
    // look for link just before in linked list
    Chunk *prev;
    for(prev=chunks;;prev=prev->next)
      if (prev->next == chunk) break;
    // remove this link
    prev->next=chunk->next;
  }

  // Make sure the block is market with dirt, and free it
  memset(chunk, PADDING_DEAD1, sizeof(Chunk)+4*chunk->size+8);
  free(chunk);
}

void memDisplay()
{
  Chunk *chunk;
  printf("---memory status------------------------------------------------\n");
  chunk=chunks;
  while(chunk)
  {
    printf("%p: %d-byte block (next=%p) allocated at %s:%d\n", chunk, chunk->length, chunk->next, chunk->file, chunk->line);
    chunk=chunk->next;
  }
}
